@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -80,7 +81,7 @@ nonzerotrace_F2xq(GEN T)
     return pol1_F2x(vs);
   do
   {
-    avma = av;
+    set_avma(av);
     a = random_F2x(n, vs);
   } while (F2xq_trace(a, T)==0);
   return a;
@@ -233,7 +234,7 @@ _F2xqE_mul(void *E, GEN P, GEN n)
   if (!s || ell_is_inf(P)) return ellinf();
   if (s<0) P = F2xqE_neg(P, e->a2, e->T);
   if (is_pm1(n)) return s>0? gcopy(P): P;
-  return gerepileupto(av, gen_pow(P, n, e, &_F2xqE_dbl, &_F2xqE_add));
+  return gerepilecopy(av, gen_pow_i(P, n, e, &_F2xqE_dbl, &_F2xqE_add));
 }
 
 GEN
@@ -244,7 +245,7 @@ F2xqE_mul(GEN P, GEN n, GEN a2, GEN T)
   return _F2xqE_mul(&E, P, n);
 }
 
-/* Finds a random non-singular point on E */
+/* Finds a random nonsingular point on E */
 GEN
 random_F2xqE(GEN a, GEN a6, GEN T)
 {
@@ -252,13 +253,13 @@ random_F2xqE(GEN a, GEN a6, GEN T)
   GEN x, y, rhs, u;
   do
   {
-    avma= ltop;
+    set_avma(ltop);
     x   = random_F2x(F2x_degree(T),T[1]);
     if (typ(a) == t_VECSMALL)
     {
       GEN a2 = a, x2;
       if (!lgpol(x))
-        { avma=ltop; retmkvec2(pol0_Flx(T[1]), F2xq_sqrt(a6,T)); }
+        { set_avma(ltop); retmkvec2(pol0_Flx(T[1]), F2xq_sqrt(a6,T)); }
       u = x; x2  = F2xq_sqr(x, T);
       rhs = F2x_add(F2xq_mul(x2,F2x_add(x,a2),T),a6);
       rhs = F2xq_div(rhs,x2,T);
@@ -428,31 +429,21 @@ F2xqE_chord_update(GEN R, GEN P, GEN Q, GEN a2, GEN T, GEN *pt_R)
   }
 }
 
-/* Returns the Miller function f_{m, Q} evaluated at the point P using
-   the standard Miller algorithm.
- */
-
-struct _F2xqE_miller
-{
-  GEN T, a2, P;
-};
+struct _F2xqE_miller { GEN T, a2, P; };
 
 static GEN
 F2xqE_Miller_dbl(void* E, GEN d)
 {
   struct _F2xqE_miller *m = (struct _F2xqE_miller *)E;
   GEN T = m->T, a2 = m->a2, P = m->P;
-  GEN v, line;
-  GEN num = F2xq_sqr(gel(d,1), T);
-  GEN denom = F2xq_sqr(gel(d,2), T);
-  GEN point = gel(d,3);
+  GEN v, line, point = gel(d,3);
+  GEN N = F2xq_sqr(gel(d,1), T);
+  GEN D = F2xq_sqr(gel(d,2), T);
   line = F2xqE_tangent_update(point, P, a2, T, &point);
-  num  = F2xq_mul(num, line, T);
+  N = F2xq_mul(N, line, T);
   v = F2xqE_vert(point, P, a2, T);
-  denom = F2xq_mul(denom, v, T);
-  return mkvec3(num, denom, point);
+  D = F2xq_mul(D, v, T); return mkvec3(N, D, point);
 }
-
 static GEN
 F2xqE_Miller_add(void* E, GEN va, GEN vb)
 {
@@ -461,41 +452,40 @@ F2xqE_Miller_add(void* E, GEN va, GEN vb)
   GEN v, line, point;
   GEN na = gel(va,1), da = gel(va,2), pa = gel(va,3);
   GEN nb = gel(vb,1), db = gel(vb,2), pb = gel(vb,3);
-  GEN num   = F2xq_mul(na, nb, T);
-  GEN denom = F2xq_mul(da, db, T);
+  GEN N = F2xq_mul(na, nb, T);
+  GEN D = F2xq_mul(da, db, T);
   line = F2xqE_chord_update(pa, pb, P, a2, T, &point);
-  num  = F2xq_mul(num, line, T);
+  N  = F2xq_mul(N, line, T);
   v = F2xqE_vert(point, P, a2, T);
-  denom = F2xq_mul(denom, v, T);
-  return mkvec3(num, denom, point);
+  D = F2xq_mul(D, v, T); return mkvec3(N, D, point);
 }
-
+/* Returns the Miller function f_{m, Q} evaluated at the point P using
+ * the standard Miller algorithm. */
 static GEN
 F2xqE_Miller(GEN Q, GEN P, GEN m, GEN a2, GEN T)
 {
-  pari_sp ltop = avma;
+  pari_sp av = avma;
   struct _F2xqE_miller d;
-  GEN v, num, denom, g1;
+  GEN v, N, D, g1;
 
   d.a2 = a2; d.T = T; d.P = P;
   g1 = pol1_F2x(T[1]);
-  v = gen_pow(mkvec3(g1,g1,Q), m, (void*)&d, F2xqE_Miller_dbl, F2xqE_Miller_add);
-  num = gel(v,1); denom = gel(v,2);
-  return gerepileupto(ltop, F2xq_div(num, denom, T));
+  v = gen_pow_i(mkvec3(g1,g1,Q), m, (void*)&d, F2xqE_Miller_dbl, F2xqE_Miller_add);
+  N = gel(v,1); D = gel(v,2);
+  return gerepileupto(av, F2xq_div(N, D, T));
 }
 
 GEN
 F2xqE_weilpairing(GEN P, GEN Q, GEN m, GEN a2, GEN T)
 {
-  pari_sp ltop = avma;
-  GEN num, denom, result;
+  pari_sp av = avma;
+  GEN N, D;
   if (ell_is_inf(P) || ell_is_inf(Q)
     || (F2x_equal(gel(P,1),gel(Q,1)) && F2x_equal(gel(P,2),gel(Q,2))))
     return pol1_F2x(T[1]);
-  num    = F2xqE_Miller(P, Q, m, a2, T);
-  denom  = F2xqE_Miller(Q, P, m, a2, T);
-  result = F2xq_div(num, denom, T);
-  return gerepileupto(ltop, result);
+  N    = F2xqE_Miller(P, Q, m, a2, T);
+  D  = F2xqE_Miller(Q, P, m, a2, T);
+  return gerepileupto(av, F2xq_div(N, D, T));
 }
 
 GEN
@@ -623,10 +613,10 @@ _can_invd(void *E, GEN V, GEN v, GEN q, long M)
 }
 
 /* Lift P to Q such that Q(x^2)=Q(x)*Q(-x) mod 2^n
-   if Q = Q0(X^2)+X*Q1(X^2), solve Q(x^2) = Q0^2-X*Q1^2
+   if Q = Q0(X^2)+X*Q1(X^2), solve Q(X^2) = Q0^2-X*Q1^2
 */
-static GEN
-F2x_canonlift(GEN P, long n)
+GEN
+F2x_Teichmuller(GEN P, long n)
 { return gen_ZpX_Newton(F2x_to_ZX(P),gen_2, n, NULL, _can_iter, _can_invd); }
 
 static GEN
@@ -775,7 +765,7 @@ F2xq_elltrace_Harley(GEN a6, GEN T2)
   sqx = mkvec2(F2xq_sqrt(polx_F2x(T2[1]),T2), T2);
   if (DEBUGLEVEL>1) timer_printf(&ti,"Sqrtx");
   ispcyc = zx_is_pcyc(F2x_to_Flx(T2));
-  T = ispcyc? F2x_to_ZX(T2): F2x_canonlift(T2, N-2);
+  T = ispcyc? F2x_to_ZX(T2): F2x_Teichmuller(T2, N-2);
   if (DEBUGLEVEL>1) timer_printf(&ti,"Teich");
   T = FpX_get_red(T, int2n(N));
   if (DEBUGLEVEL>1) timer_printf(&ti,"Barrett");
