@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -68,16 +69,6 @@ FlxqE_changepointinv(GEN x, GEN ch, GEN T, ulong p)
   return z;
 }
 
-static ulong
-nonsquare_Fl(ulong p)
-{
-  ulong a;
-  do
-    a = random_Fl(p);
-  while (krouu(a, p) >= 0);
-  return a;
-}
-
 static GEN
 nonsquare_Flxq(GEN T, ulong p)
 {
@@ -88,7 +79,7 @@ nonsquare_Flxq(GEN T, ulong p)
     return mkvecsmall2(vs, nonsquare_Fl(p));
   do
   {
-    avma = av;
+    set_avma(av);
     a = random_Flx(n, vs, p);
   } while (Flxq_issquare(a, T, p));
   return a;
@@ -223,7 +214,7 @@ _FlxqE_mul(void *E, GEN P, GEN n)
   if (!s || ell_is_inf(P)) return ellinf();
   if (s<0) P = FlxqE_neg(P, e->T, e->p);
   if (is_pm1(n)) return s>0? gcopy(P): P;
-  return gerepileupto(av, gen_pow(P, n, e, &_FlxqE_dbl, &_FlxqE_add));
+  return gerepilecopy(av, gen_pow_i(P, n, e, &_FlxqE_dbl, &_FlxqE_add));
 }
 
 GEN
@@ -236,16 +227,16 @@ FlxqE_mul(GEN P, GEN n, GEN a4, GEN T, ulong p)
 
 /* 3*x^2+2*a2*x = -a2*x, and a2!=0 */
 
-/* Finds a random non-singular point on E */
+/* Finds a random nonsingular point on E */
 static GEN
 random_F3xqE(GEN a2, GEN a6, GEN T)
 {
   pari_sp ltop = avma;
   GEN x, y, rhs;
-  const ulong p=3;
+  const ulong p = 3;
   do
   {
-    avma= ltop;
+    set_avma(ltop);
     x   = random_Flx(get_Flx_degree(T),get_Flx_var(T),p);
     rhs = Flx_add(Flxq_mul(Flxq_sqr(x, T, p), Flx_add(x, a2, p), T, p), a6, p);
   } while ((!lgpol(rhs) && !lgpol(x)) || !Flxq_issquare(rhs, T, p));
@@ -254,7 +245,7 @@ random_F3xqE(GEN a2, GEN a6, GEN T)
   return gerepilecopy(ltop, mkvec2(x, y));
 }
 
-/* Finds a random non-singular point on E */
+/* Finds a random nonsingular point on E */
 GEN
 random_FlxqE(GEN a4, GEN a6, GEN T, ulong p)
 {
@@ -264,7 +255,7 @@ random_FlxqE(GEN a4, GEN a6, GEN T, ulong p)
     return random_F3xqE(gel(a4,1), a6, T);
   do
   {
-    avma= ltop;
+    set_avma(ltop);
     x   = random_Flx(get_Flx_degree(T),get_Flx_var(T),p);
     x2  = Flxq_sqr(x, T, p); /*  x^3+a4*x+a6 = x*(x^2+a4)+a6  */
     rhs = Flx_add(Flxq_mul(x, Flx_add(x2, a4, p), T, p), a6, p);
@@ -363,9 +354,7 @@ FlxqE_Miller_line(GEN R, GEN Q, GEN slope, GEN a4, GEN T, ulong p)
 }
 
 /* Computes the equation of the line tangent to R and returns its
-   evaluation at the point Q. Also doubles the point R.
- */
-
+ * evaluation at the point Q. Also doubles the point R. */
 static GEN
 FlxqE_tangent_update(GEN R, GEN Q, GEN a4, GEN T, ulong p, GEN *pt_R)
 {
@@ -386,9 +375,7 @@ FlxqE_tangent_update(GEN R, GEN Q, GEN a4, GEN T, ulong p, GEN *pt_R)
 }
 
 /* Computes the equation of the line through R and P, and returns its
-   evaluation at the point Q. Also adds P to the point R.
- */
-
+ * evaluation at the point Q. Also adds P to the point R. */
 static GEN
 FlxqE_chord_update(GEN R, GEN P, GEN Q, GEN a4, GEN T, ulong p, GEN *pt_R)
 {
@@ -418,10 +405,6 @@ FlxqE_chord_update(GEN R, GEN P, GEN Q, GEN a4, GEN T, ulong p, GEN *pt_R)
   }
 }
 
-/* Returns the Miller function f_{m, Q} evaluated at the point P using
-   the standard Miller algorithm.
- */
-
 struct _FlxqE_miller
 {
   ulong p;
@@ -434,15 +417,13 @@ FlxqE_Miller_dbl(void* E, GEN d)
   struct _FlxqE_miller *m = (struct _FlxqE_miller *)E;
   ulong p  = m->p;
   GEN T = m->T, a4 = m->a4, P = m->P;
-  GEN v, line;
-  GEN num = Flxq_sqr(gel(d,1), T, p);
-  GEN denom = Flxq_sqr(gel(d,2), T, p);
-  GEN point = gel(d,3);
+  GEN v, line, point = gel(d,3);
+  GEN N = Flxq_sqr(gel(d,1), T, p);
+  GEN D = Flxq_sqr(gel(d,2), T, p);
   line = FlxqE_tangent_update(point, P, a4, T, p, &point);
-  num  = Flxq_mul(num, line, T, p);
+  N  = Flxq_mul(N, line, T, p);
   v = FlxqE_vert(point, P, a4, T, p);
-  denom = Flxq_mul(denom, v, T, p);
-  return mkvec3(num, denom, point);
+  D = Flxq_mul(D, v, T, p); return mkvec3(N, D, point);
 }
 
 static GEN
@@ -454,50 +435,50 @@ FlxqE_Miller_add(void* E, GEN va, GEN vb)
   GEN v, line, point;
   GEN na = gel(va,1), da = gel(va,2), pa = gel(va,3);
   GEN nb = gel(vb,1), db = gel(vb,2), pb = gel(vb,3);
-  GEN num   = Flxq_mul(na, nb, T, p);
-  GEN denom = Flxq_mul(da, db, T, p);
+  GEN N = Flxq_mul(na, nb, T, p);
+  GEN D = Flxq_mul(da, db, T, p);
   line = FlxqE_chord_update(pa, pb, P, a4, T, p, &point);
-  num  = Flxq_mul(num, line, T, p);
+  N  = Flxq_mul(N, line, T, p);
   v = FlxqE_vert(point, P, a4, T, p);
-  denom = Flxq_mul(denom, v, T, p);
-  return mkvec3(num, denom, point);
+  D = Flxq_mul(D, v, T, p); return mkvec3(N, D, point);
 }
 
+/* Returns the Miller function f_{m, Q} evaluated at the point P using
+ * the standard Miller algorithm. */
 static GEN
 FlxqE_Miller(GEN Q, GEN P, GEN m, GEN a4, GEN T, ulong p)
 {
-  pari_sp ltop = avma;
+  pari_sp av = avma;
   struct _FlxqE_miller d;
-  GEN v, num, denom, g1;
+  GEN v, N, D, g1;
 
   d.a4 = a4; d.T = T; d.p = p; d.P = P;
   g1 = pol1_Flx(get_Flx_var(T));
-  v = gen_pow(mkvec3(g1,g1,Q), m, (void*)&d, FlxqE_Miller_dbl, FlxqE_Miller_add);
-  num = gel(v,1); denom = gel(v,2);
-  return gerepileupto(ltop, Flxq_div(num, denom, T, p));
+  v = gen_pow_i(mkvec3(g1,g1,Q), m, (void*)&d,
+                FlxqE_Miller_dbl, FlxqE_Miller_add);
+  N = gel(v,1); D = gel(v,2);
+  return gerepileupto(av, Flxq_div(N, D, T, p));
 }
 
 GEN
 FlxqE_weilpairing(GEN P, GEN Q, GEN m, GEN a4, GEN T, ulong p)
 {
-  pari_sp ltop = avma;
-  GEN num, denom, result;
+  pari_sp av = avma;
+  GEN N, D, result;
   if (ell_is_inf(P) || ell_is_inf(Q)
     || (Flx_equal(gel(P,1),gel(Q,1)) && Flx_equal(gel(P,2),gel(Q,2))))
     return pol1_Flx(get_Flx_var(T));
-  num    = FlxqE_Miller(P, Q, m, a4, T, p);
-  denom  = FlxqE_Miller(Q, P, m, a4, T, p);
-  result = Flxq_div(num, denom, T, p);
-  if (mpodd(m))
-    result  = Flx_neg(result, p);
-  return gerepileupto(ltop, result);
+  N = FlxqE_Miller(P, Q, m, a4, T, p);
+  D = FlxqE_Miller(Q, P, m, a4, T, p);
+  result = Flxq_div(N, D, T, p);
+  if (mpodd(m)) result = Flx_neg(result, p);
+  return gerepileupto(av, result);
 }
 
 GEN
 FlxqE_tatepairing(GEN P, GEN Q, GEN m, GEN a4, GEN T, ulong p)
 {
-  if (ell_is_inf(P) || ell_is_inf(Q))
-    return pol1_Flx(get_Flx_var(T));
+  if (ell_is_inf(P) || ell_is_inf(Q)) return pol1_Flx(get_Flx_var(T));
   return FlxqE_Miller(P, Q, m, a4, T, p);
 }
 
@@ -546,134 +527,6 @@ Flxq_ellgens(GEN a4, GEN a6, GEN ch, GEN D, GEN m, GEN T, ulong p)
 /**                                                                   **/
 /***********************************************************************/
 
-static GEN _can_invl(void *E, GEN V) {(void) E; return V; }
-
-static GEN _can_lin(void *E, GEN F, GEN V, GEN q)
-{
-  GEN v = RgX_splitting(V, 3);
-  (void) E;
-  return FpX_sub(V,ZXV_dotproduct(v, F), q);
-}
-
-static GEN
-_can_iter(void *E, GEN f, GEN q)
-{
-  GEN h = RgX_splitting(f,3);
-  GEN h1s = ZX_sqr(gel(h,1)), h2s = ZX_sqr(gel(h,2)), h3s = ZX_sqr(gel(h,3));
-  GEN h12 = ZX_mul(gel(h,1), gel(h,2));
-  GEN h13 = ZX_mul(gel(h,1), gel(h,3));
-  GEN h23 = ZX_mul(gel(h,2), gel(h,3));
-  GEN h1c = ZX_mul(gel(h,1), h1s);
-  GEN h3c = ZX_mul(gel(h,3), h3s);
-  GEN th = ZX_mul(ZX_sub(h2s,ZX_mulu(h13,3)),gel(h,2));
-  GEN y = FpX_sub(f,ZX_add(RgX_shift_shallow(h3c,2),ZX_add(RgX_shift_shallow(th,1),h1c)),q);
-  (void) E;
-  return mkvecn(7,y,h1s,h2s,h3s,h12,h13,h23);
-}
-
-static GEN
-_can_invd(void *E, GEN V, GEN v, GEN qM, long M)
-{
-  GEN h1s=gel(v,2), h2s=gel(v,3), h3s=gel(v,4);
-  GEN h12=gel(v,5), h13=gel(v,6), h23=gel(v,7);
-  GEN F = mkvec3(ZX_sub(h1s,RgX_shift_shallow(h23,1)),RgX_shift_shallow(ZX_sub(h2s,h13),1),
-                 ZX_sub(RgX_shift_shallow(h3s,2),RgX_shift_shallow(h12,1)));
-  (void)E;
-  return gen_ZpX_Dixon(ZXV_Z_mul(F, utoi(3)), V, qM, utoi(3), M, NULL,
-                                                 _can_lin, _can_invl);
-}
-
-static GEN
-F3x_canonlift(GEN P, long n)
-{ return gen_ZpX_Newton(Flx_to_ZX(P),utoi(3), n, NULL, _can_iter, _can_invd); }
-
-static GEN _can5_invl(void *E, GEN V) {(void) E; return V; }
-
-static GEN _can5_lin(void *E, GEN F, GEN V, GEN q)
-{
-  ulong p = *(ulong*)E;
-  GEN v = RgX_splitting(V, p);
-  return FpX_sub(V,ZXV_dotproduct(v, F), q);
-}
-
-/* P(X,t) -> P(X*t^n,t) mod (t^p-1) */
-static GEN
-_shift(GEN P, long n, ulong p, long v)
-{
-  long i, l=lg(P);
-  GEN r = cgetg(l,t_POL); r[1] = P[1];
-  for(i=2;i<l;i++)
-  {
-    long s = n*(i-2)%p;
-    GEN ci = gel(P,i);
-    if (typ(ci)==t_INT)
-      gel(r,i) = monomial(ci, s, v);
-    else
-      gel(r,i) = RgX_rotate_shallow(ci, s, p);
-  }
-  return FpXX_renormalize(r, l);
-}
-
-struct _can_mul
-{
-  GEN T, q;
-  ulong p;
-};
-
-static GEN
-_can5_mul(void *E, GEN A, GEN B)
-{
-  struct _can_mul *d = (struct _can_mul *)E;
-  GEN a = gel(A,1), b = gel(B,1);
-  long n = itos(gel(A,2));
-  GEN bn = _shift(b, n, d->p, get_FpX_var(d->T));
-  GEN c = FpXQX_mul(a, bn, d->T, d->q);
-  return mkvec2(c, addii(gel(A,2), gel(B,2)));
-}
-
-static GEN
-_can5_sqr(void *E, GEN A)
-{
-  return _can5_mul(E,A,A);
-}
-
-static GEN
-_can5_iter(void *E, GEN f, GEN q)
-{
-  pari_sp av = avma;
-  struct _can_mul D;
-  ulong p = *(ulong*)E;
-  long i, vT = fetch_var();
-  GEN N, P, d, V, fs;
-  D.q = q; D.T = ZX_Z_sub(pol_xn(p,vT),gen_1);
-  D.p = p;
-  fs = mkvec2(_shift(f, 1, p, vT), gen_1);
-  N = gel(gen_powu(fs,p-1,(void*)&D,_can5_sqr,_can5_mul),1);
-  N = simplify_shallow(FpXQX_red(N,polcyclo(p,vT),q));
-  P = FpX_mul(N,f,q);
-  P = RgX_deflate(P, p);
-  d = RgX_splitting(N, p);
-  V = cgetg(p+1,t_VEC);
-  gel(V,1) = ZX_mulu(gel(d,1), p);
-  for(i=2; i<= (long)p; i++)
-    gel(V,i) = ZX_mulu(RgX_shift_shallow(gel(d,p+2-i), 1), p);
-  (void)delete_var(); return gerepilecopy(av, mkvec2(ZX_sub(f,P),V));
-}
-
-static GEN
-_can5_invd(void *E, GEN H, GEN v, GEN qM, long M)
-{
-  ulong p = *(long*)E;
-  return gen_ZpX_Dixon(gel(v,2), H, qM, utoi(p), M, E, _can5_lin, _can5_invl);
-}
-
-static GEN
-Flx_canonlift(GEN P, long n, ulong p)
-{
-  return p==3 ? F3x_canonlift(P,n):
-         gen_ZpX_Newton(Flx_to_ZX(P),utoi(p), n, &p, _can5_iter, _can5_invd);
-}
-
 /* assume a and n  are coprime */
 static GEN
 RgX_circular_shallow(GEN P, long a, long n)
@@ -857,7 +710,7 @@ ZpXQ_norm_pcyc(GEN x, GEN T, GEN q, GEN p)
   D.T = T; D.p = q; D.n = d+1;
   if (d==1) return ZX_copy(x);
   z = mkvec2(x,mkvecsmall(p[2]));
-  z = gen_powu(z,d,(void*)&D,ZpXQ_norm_sqr,ZpXQ_norm_mul);
+  z = gen_powu_i(z,d,(void*)&D,ZpXQ_norm_sqr,ZpXQ_norm_mul);
   return gmael(z,1,2);
 }
 
@@ -1051,7 +904,7 @@ Flxq_ellcard_Kohel(GEN a4, GEN a6, GEN T, ulong p)
   timer_start(&ti);
   if (!ispcyc)
   {
-    T2 = Flx_canonlift(get_Flx_mod(T),N,p);
+    T2 = Flx_Teichmuller(get_Flx_mod(T),p,N);
     if (DEBUGLEVEL) timer_printf(&ti,"Teich");
   } else
     T2 = Flx_to_ZX(get_Flx_mod(T));
@@ -1189,7 +1042,7 @@ Flxq_ellcard_Harley(GEN a4, GEN a6, GEN T, ulong p)
   timer_start(&ti);
   if (!ispcyc)
   {
-    T2 = Flx_canonlift(get_Flx_mod(T),N,p);
+    T2 = Flx_Teichmuller(get_Flx_mod(T),p,N);
     if (DEBUGLEVEL) timer_printf(&ti,"Teich");
   } else
     T2 = Flx_to_ZX(get_Flx_mod(T));
@@ -1368,7 +1221,7 @@ F3xq_ellcard_naive(GEN a2, GEN a6, GEN T)
     if (!lgpol(rhs)) a++; else if (Flxq_issquare(rhs, T, 3)) a+=2;
     Flx_next(x, 3);
   }
-  avma = av;
+  set_avma(av);
   return a;
 }
 
@@ -1388,20 +1241,19 @@ Flxq_ellcard_naive(GEN a4, GEN a6, GEN T, ulong p)
     if (!lgpol(rhs)) a++; else if (Flxq_issquare(rhs,T,p)) a+=2;
     Flx_next(x,p);
   }
-  avma = av;
+  set_avma(av);
   return a;
 }
 
 /* assume T irreducible mod p, m = (q-1)/(p-1) */
-static int
+static long
 Flxq_kronecker(GEN x, GEN m, GEN T, ulong p)
 {
   pari_sp av;
   ulong z;
   if (lgpol(x) == 0) return 0;
-  av = avma;
-  z = Flxq_pow(x, m, T, p)[2];
-  avma = av; return krouu(z, p);
+  av = avma; z = Flxq_pow(x, m, T, p)[2];
+  return gc_long(av, krouu(z, p));
 }
 
 /* Find x such that kronecker(u = x^3+a4x+a6, p) is KRO.
