@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -53,7 +54,7 @@ FpXX_red(GEN z, GEN p)
       pari_sp av = avma;
       c = FpX_red(zi,p);
       switch(lg(c)) {
-        case 2: avma = av; c = gen_0; break;
+        case 2: set_avma(av); c = gen_0; break;
         case 3: c = gerepilecopy(av, gel(c,2)); break;
       }
     }
@@ -103,17 +104,17 @@ FpXX_subspec(GEN x, GEN y, GEN p, long nx, long ny)
   GEN z;
   if (ny <= nx)
   {
-    lz = nx+2; z = cgetg(lz, t_POL)+2;
-    for (i=0; i<ny; i++) gel(z,i) = Fq_sub(gel(x,i), gel(y,i), NULL, p);
-    for (   ; i<nx; i++) gel(z,i) = gcopy(gel(x,i));
+    lz = nx+2; z = cgetg(lz, t_POL);
+    for (i=0; i<ny; i++) gel(z,i+2) = Fq_sub(gel(x,i), gel(y,i), NULL, p);
+    for (   ; i<nx; i++) gel(z,i+2) = gcopy(gel(x,i));
   }
   else
   {
-    lz = ny+2; z = cgetg(lz, t_POL)+2;
-    for (i=0; i<nx; i++) gel(z,i) = Fq_sub(gel(x,i), gel(y,i), NULL, p);
-    for (   ; i<ny; i++) gel(z,i) = Fq_neg(gel(y,i), NULL, p);
+    lz = ny+2; z = cgetg(lz, t_POL);
+    for (i=0; i<nx; i++) gel(z,i+2) = Fq_sub(gel(x,i), gel(y,i), NULL, p);
+    for (   ; i<ny; i++) gel(z,i+2) = Fq_neg(gel(y,i), NULL, p);
   }
-  return FpXX_renormalize(z-2, lz);
+  z[1] = 0; return FpXX_renormalize(z, lz);
 }
 
 GEN
@@ -195,8 +196,12 @@ FpXX_integ(GEN P, GEN p)
   for (i=3; i<=l ; i++)
   {
     GEN x = gel(P,i-1);
-    GEN i1 = Fp_inv(utoi(i-2), p);
-    gel(res,i) = typ(x)==t_INT? Fp_mul(x,i1,p): FpX_Fp_mul(x,i1,p);
+    if (signe(x))
+    {
+      GEN i1 = Fp_inv(utoi(i-2), p);
+      gel(res,i) = typ(x)==t_INT? Fp_mul(x,i1,p): FpX_Fp_mul(x,i1,p);
+    } else
+      gel(res,i) = gen_0;
   }
   return FpXX_renormalize(res, l+1);
 }
@@ -246,37 +251,6 @@ Kronecker_to_FpXQX(GEN Z, GEN T, GEN p)
   return FpXQX_renormalize(x, i+1);
 }
 
-/* shallow, n = deg(T) */
-GEN
-Kronecker_to_ZXX(GEN z, long n, long v)
-{
-  long i,j,lx,l, N = (n<<1)+1;
-  GEN x, t;
-  l = lg(z); lx = (l-2) / (N-2);
-  x = cgetg(lx+3,t_POL);
-  x[1] = z[1];
-  for (i=2; i<lx+2; i++)
-  {
-    t = cgetg(N,t_POL); t[1] = evalvarn(v);
-    for (j=2; j<N; j++) gel(t,j) = gel(z,j);
-    z += (N-2);
-    gel(x,i) = ZX_renormalize(t,N);
-  }
-  N = (l-2) % (N-2) + 2;
-  t = cgetg(N,t_POL); t[1] = evalvarn(v);
-  for (j=2; j<N; j++) gel(t,j) = gel(z,j);
-  gel(x,i) = ZX_renormalize(t,N);
-  return ZXX_renormalize(x, i+1);
-}
-/* shallow */
-GEN
-ZXX_mul_Kronecker(GEN x, GEN y, long n)
-{ return ZX_mul(ZXX_to_Kronecker(x,n), ZXX_to_Kronecker(y,n)); }
-
-GEN
-ZXX_sqr_Kronecker(GEN x, long n)
-{ return ZX_sqr(ZXX_to_Kronecker(x,n)); }
-
 GEN
 FpXQX_red(GEN z, GEN T, GEN p)
 {
@@ -293,21 +267,6 @@ FpXQX_red(GEN z, GEN T, GEN p)
 static GEN
 to_intmod(GEN x, GEN p) { retmkintmod(modii(x, p), p); }
 
-static GEN
-FpX_to_mod_raw(GEN z, GEN p)
-{
-  long i,l = lg(z);
-  GEN x;
-  if (l == 2)
-  {
-    x = cgetg(3,t_POL); x[1] = z[1];
-    gel(x,2) = mkintmod(gen_0,p); return x;
-  }
-  x = cgetg(l,t_POL);
-  for (i=2; i<l; i++) gel(x,i) = to_intmod(gel(z,i), p);
-  x[1] = z[1]; return normalizepol_lg(x,l);
-}
-
 GEN
 FpXQX_to_mod(GEN z, GEN T, GEN p)
 {
@@ -326,6 +285,60 @@ FpXQX_to_mod(GEN z, GEN T, GEN p)
   return normalizepol_lg(x,l);
 }
 
+static GEN
+FpXQX_to_mod_raw(GEN z, GEN T, GEN p)
+{
+  long i,l = lg(z);
+  GEN x = cgetg(l, t_POL);
+  x[1] = z[1];
+  if (l == 2) return x;
+  for (i=2; i<l; i++)
+  {
+    GEN zi = gel(z,i);
+    gel(x,i) = typ(zi) == t_POL? mkpolmod(FpX_to_mod_raw(zi, p), T)
+                               : to_intmod(zi, p);
+  }
+  return normalizepol_lg(x,l);
+}
+
+INLINE GEN
+FqX_to_mod_raw(GEN f, GEN T, GEN p)
+{ return T?FpXQX_to_mod_raw(f, T, p): FpX_to_mod_raw(f, p); }
+
+static GEN
+FqXC_to_mod_raw(GEN x, GEN T, GEN p)
+{ pari_APPLY_type(t_COL, FqX_to_mod_raw(gel(x,i), T, p)) }
+
+GEN
+FqXC_to_mod(GEN z, GEN T, GEN p)
+{
+  GEN x;
+  long i,l = lg(z);
+  if (!T) return FpXC_to_mod(z, p);
+  x = cgetg(l, t_COL);
+  if (l == 1) return x;
+  p = icopy(p);
+  T = FpX_to_mod_raw(T, p);
+  for (i=1; i<l; i++)
+    gel(x,i) = FqX_to_mod_raw(gel(z, i), T, p);
+  return x;
+}
+
+GEN
+FqXM_to_mod(GEN z, GEN T, GEN p)
+{
+  GEN x;
+  long i,l = lg(z);
+  if (!T) return FpXM_to_mod(z, p);
+  x = cgetg(l, t_MAT);
+  if (l == 1) return x;
+  p = icopy(p);
+  T = FpX_to_mod_raw(T, p);
+  for (i=1; i<l; i++)
+    gel(x,i) = FqXC_to_mod_raw(gel(z, i), T, p);
+  return x;
+}
+
 static int
 ZXX_is_ZX_spec(GEN a,long na)
 {
@@ -435,7 +448,7 @@ FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
     if (pr)
     {
       av0 = avma; x = FpXQX_red(x, T, p);
-      if (pr == ONLY_DIVIDES) { avma=av0; return signe(x)? NULL: pol_0(vx); }
+      if (pr == ONLY_DIVIDES) { set_avma(av0); return signe(x)? NULL: pol_0(vx); }
       if (pr == ONLY_REM) return x;
       *pr = x;
     }
@@ -455,7 +468,7 @@ FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
   }
   av0 = avma; dz = dx-dy;
   lead = gequal1(lead)? NULL: gclone(Fq_inv(lead,T,p));
-  avma = av0;
+  set_avma(av0);
   z = cgetg(dz+3,t_POL); z[1] = x[1];
   x += 2; y += 2; z += 2;
   for (dy1=dy-1; dy1>=0 && !signe(gel(y, dy1)); dy1--);
@@ -470,7 +483,7 @@ FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
     if (lead) p1 = Fq_mul(p1, lead, NULL,p);
     tetpil=avma; gel(z,i-dy) = gerepile(av,tetpil,Fq_red(p1,T,p));
   }
-  if (!pr) { if (lead) gunclone(lead); return z-2; }
+  if (!pr) { guncloneNULL(lead); return z-2; }
 
   rem = (GEN)avma; av = (pari_sp)new_chunk(dx+3);
   for (sx=0; ; i--)
@@ -480,13 +493,13 @@ FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
       p1 = Fq_sub(p1, Fq_mul(gel(z,j),gel(y,i-j),NULL,p),NULL,p);
     tetpil=avma; p1 = Fq_red(p1, T, p); if (signe(p1)) { sx = 1; break; }
     if (!i) break;
-    avma=av;
+    set_avma(av);
   }
   if (pr == ONLY_DIVIDES)
   {
-    if (lead) gunclone(lead);
-    if (sx) { avma=av0; return NULL; }
-    avma = (pari_sp)rem; return z-2;
+    guncloneNULL(lead);
+    if (sx) return gc_NULL(av0);
+    return gc_const((pari_sp)rem, z-2);
   }
   lr=i+3; rem -= lr;
   rem[0] = evaltyp(t_POL) | evallg(lr);
@@ -501,7 +514,7 @@ FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
     tetpil=avma; gel(rem,i) = gerepile(av,tetpil, Fq_red(p1, T, p));
   }
   rem -= 2;
-  if (lead) gunclone(lead);
+  guncloneNULL(lead);
   if (!sx) (void)FpXQX_renormalize(rem, lr);
   if (pr == ONLY_REM) return gerepileupto(av0,rem);
   *pr = rem; return z-2;
@@ -581,6 +594,9 @@ matid2_FpXQXM(long v)
             mkcol2(pol_0(v),pol_1(v)));
 }
 
+static GEN
+FpXX_shift(GEN a, long n) { return RgX_shift_shallow(a, n); }
+
 static GEN
 FpXQX_halfgcd_split(GEN x, GEN y, GEN T, GEN p)
 {
@@ -589,12 +605,12 @@ FpXQX_halfgcd_split(GEN x, GEN y, GEN T, GEN p)
   GEN y1, r, q;
   long l = lgpol(x), n = l>>1, k;
   if (lgpol(y)<=n) return matid2_FpXQXM(varn(x));
-  R = FpXQX_halfgcd(RgX_shift_shallow(x,-n),RgX_shift_shallow(y,-n), T, p);
+  R = FpXQX_halfgcd(FpXX_shift(x,-n),FpXX_shift(y,-n), T, p);
   V = FpXQXM_FpXQX_mul2(R,x,y, T, p); y1 = gel(V,2);
   if (lgpol(y1)<=n) return gerepilecopy(av, R);
   q = FpXQX_divrem(gel(V,1), y1, T, p, &r);
   k = 2*n-degpol(y1);
-  S = FpXQX_halfgcd(RgX_shift_shallow(y1,-k), RgX_shift_shallow(r,-k), T, p);
+  S = FpXQX_halfgcd(FpXX_shift(y1,-k), FpXX_shift(r,-k), T, p);
   return gerepileupto(av, FpXQXM_mul2(S,FpXQX_FpXQXM_qmul(q,R, T, p), T, p));
 }
 
@@ -650,7 +666,7 @@ FpXQX_gcd_basecase(GEN a, GEN b, GEN T, GEN p)
     }
     av = avma; c = FpXQX_rem(a, b, T, p); a=b; b=c;
   }
-  avma = av; return a;
+  return gc_const(av, a);
 }
 
 GEN
@@ -773,6 +789,71 @@ FpXQX_dotproduct(GEN x, GEN y, GEN T, GEN p)
   return gerepileupto(av, Fq_red(c,T,p));
 }
 
+/* Res(A,B) = Res(B,R) * lc(B)^(a-r) * (-1)^(ab), with R=A%B, a=deg(A) ...*/
+GEN
+FpXQX_resultant(GEN a, GEN b, GEN T, GEN p)
+{
+  long da,db,dc;
+  pari_sp av;
+  long vT = get_FpX_var(T);
+  GEN c,lb, res = pol_1(vT);
+
+  if (!signe(a) || !signe(b)) return pol_0(vT);
+  if (lgefint(p) == 3)
+  {
+    pari_sp av = avma;
+    GEN Pl, Ql, Tl, R;
+    ulong pp = to_FlxqX(a, b, T, p, &Pl, &Ql, &Tl);
+    R = FlxqX_resultant(Pl, Ql, Tl, pp);
+    return gerepileupto(av, Flx_to_ZX(R));
+  }
+
+  da = degpol(a);
+  db = degpol(b);
+  if (db > da)
+  {
+    swapspec(a,b, da,db);
+    if (both_odd(da,db)) res = FpX_neg(res, p);
+  }
+  if (!da) return pol_1(vT); /* = res * a[2] ^ db, since 0 <= db <= da = 0 */
+  av = avma;
+  while (db)
+  {
+    lb = gel(b,db+2);
+    c = FpXQX_rem(a,b, T,p);
+    a = b; b = c; dc = degpol(c);
+    if (dc < 0) { set_avma(av); return pol_0(vT); }
+
+    if (both_odd(da,db)) res = FpX_neg(res, p);
+    if (!equali1(lb)) res = FpXQ_mul(res, FpXQ_powu(lb, da - dc, T, p), T, p);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpXQX_resultant (da = %ld)",da);
+      gerepileall(av,3, &a,&b,&res);
+    }
+    da = db; /* = degpol(a) */
+    db = dc; /* = degpol(b) */
+  }
+  res = FpXQ_mul(res, FpXQ_powu(gel(b,2), da, T, p), T, p);
+  return gerepileupto(av, res);
+}
+
+/* disc P = (-1)^(n(n-1)/2) lc(P)^(n - deg P' - 2) Res(P,P'), n = deg P */
+GEN
+FpXQX_disc(GEN P, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  GEN L, dP = FpXX_deriv(P, p), D = FpXQX_resultant(P, dP, T, p);
+  long dd;
+  if (!signe(D)) return pol_0(get_FpX_var(T));
+  dd = degpol(P) - 2 - degpol(dP); /* >= -1; > -1 iff p | deg(P) */
+  L = leading_coeff(P);
+  if (dd && !gequal1(L))
+    D = (dd == -1)? FpXQ_div(D,L,T,p): FpXQ_mul(D, FpXQ_powu(L, dd, T, p), T, p);
+  if (degpol(P) & 2) D = FpX_neg(D, p);
+  return gerepileupto(av, D);
+}
+
 /***********************************************************************/
 /**                                                                   **/
 /**                       Barrett reduction                           **/
@@ -866,7 +947,6 @@ FpXQX_invBarrett_Newton(GEN S, GEN T, GEN p)
   return gerepilecopy(av, x);
 }
 
-/* 1/polrecip(S)+O(x^(deg(S)-1)) */
 GEN
 FpXQX_invBarrett(GEN S, GEN T, GEN p)
 {
@@ -922,26 +1002,25 @@ FpXQX_divrem_Barrettspec(GEN x, long l, GEN mg, GEN S, GEN T, GEN p, GEN *pr)
 }
 
 static GEN
-FpXQX_divrem_Barrett_noGC(GEN x, GEN mg, GEN S, GEN T, GEN p, GEN *pr)
+FpXQX_divrem_Barrett(GEN x, GEN mg, GEN S, GEN T, GEN p, GEN *pr)
 {
-  long l = lgpol(x), lt = degpol(S), lm = 2*lt-1;
-  GEN q = NULL, r;
+  GEN q = NULL, r = FpXQX_red(x, T, p);
+  long l = lgpol(r), lt = degpol(S), lm = 2*lt-1, v = varn(S);
   long i;
   if (l <= lt)
   {
-    if (pr == ONLY_REM) return ZXX_copy(x);
-    if (pr == ONLY_DIVIDES) return signe(x)? NULL: pol_0(varn(x));
-    if (pr) *pr =  ZXX_copy(x);
-    return pol_0(varn(x));
+    if (pr == ONLY_REM) return r;
+    if (pr == ONLY_DIVIDES) return signe(r)? NULL: pol_0(v);
+    if (pr) *pr = r;
+    return pol_0(v);
   }
   if (lt <= 1)
-    return FpXQX_divrem_basecase(x,S,T,p,pr);
+    return FpXQX_divrem_basecase(r,S,T,p,pr);
   if (pr != ONLY_REM && l>lm)
   {
-    q = cgetg(l-lt+2, t_POL);
+    q = cgetg(l-lt+2, t_POL); q[1] = S[1];
     for (i=0;i<l-lt;i++) gel(q+2,i) = gen_0;
   }
-  r = l>lm ? shallowcopy(x): x;
   while (l>lm)
   {
     GEN zr, zq = FpXQX_divrem_Barrettspec(r+2+l-lm,lm,mg,S,T,p,&zr);
@@ -954,49 +1033,47 @@ FpXQX_divrem_Barrett_noGC(GEN x, GEN mg, GEN S, GEN T, GEN p, GEN *pr)
     for(i=0; i<lz; i++) gel(r+2+l-lm,i) = gel(zr,2+i);
     l = l-lm+lz;
   }
-  if (pr != ONLY_REM)
+  if (pr == ONLY_REM)
   {
     if (l > lt)
+      r = FpXQX_divrem_Barrettspec(r+2,l,mg,S,T,p,ONLY_REM);
+    else
+      r = FpXQX_renormalize(r, l+2);
+    setvarn(r, v); return r;
+  }
+  if (l > lt)
+  {
+    GEN zq = FpXQX_divrem_Barrettspec(r+2,l,mg,S,T,p,pr ? &r: NULL);
+    if (!q) q = zq;
+    else
     {
-      GEN zq = FpXQX_divrem_Barrettspec(r+2,l,mg,S,T,p,&r);
-      if (!q) q = zq;
-      else
-      {
-        long lq = lgpol(zq);
-        for(i=0; i<lq; i++) gel(q+2,i) = gel(zq,2+i);
-      }
+      long lq = lgpol(zq);
+      for(i=0; i<lq; i++) gel(q+2,i) = gel(zq,2+i);
     }
-    else
-    { setlg(r, l+2); r = ZXX_copy(r); }
   }
-  else
-  {
-    if (l > lt)
-      (void) FpXQX_divrem_Barrettspec(r+2,l,mg,S,T,p,&r);
-    else
-    { setlg(r, l+2); r = ZXX_copy(r); }
-    r[1] = x[1]; return FpXQX_renormalize(r, lg(r));
-  }
-  if (pr) { r[1] = x[1]; r = FpXQX_renormalize(r, lg(r)); }
-  q[1] = x[1]; q = FpXQX_renormalize(q, lg(q));
+  else if (pr)
+    r = FpX_renormalize(r, l+2);
+  setvarn(q, v); q = FpXQX_renormalize(q, lg(q));
   if (pr == ONLY_DIVIDES) return signe(r)? NULL: q;
-  if (pr) *pr = r;
+  if (pr) { setvarn(r, v); *pr = r; }
   return q;
 }
 
 GEN
 FpXQX_divrem(GEN x, GEN S, GEN T, GEN p, GEN *pr)
 {
-  GEN B, y = get_FpXQX_red(S, &B);
-  long dy = degpol(y), dx = degpol(x), d = dx-dy;
-  if (pr==ONLY_REM) return FpXQX_rem(x, y, T, p);
+  GEN B, y;
+  long dy, dx, d;
+  if (pr==ONLY_REM) return FpXQX_rem(x, S, T, p);
+  y = get_FpXQX_red(S, &B);
+  dy = degpol(y); dx = degpol(x); d = dx-dy;
   if (lgefint(p) == 3)
   {
     GEN a, b, t, z;
     pari_sp tetpil, av = avma;
     ulong pp = to_FlxqX(x, y, T, p, &a, &b, &t);
     z = FlxqX_divrem(a, b, t, pp, pr);
-    if (pr == ONLY_DIVIDES && !z) { avma = av; return NULL; }
+    if (pr == ONLY_DIVIDES && !z) return gc_NULL(av);
     tetpil=avma;
     z = FlxX_to_ZXX(z);
     if (pr && pr != ONLY_DIVIDES && pr != ONLY_REM)
@@ -1011,8 +1088,8 @@ FpXQX_divrem(GEN x, GEN S, GEN T, GEN p, GEN *pr)
   {
     pari_sp av=avma;
     GEN mg = B? B: FpXQX_invBarrett(y, T, p);
-    GEN q = FpXQX_divrem_Barrett_noGC(x,mg,y,T,p,pr);
-    if (!q) {avma=av; return NULL;}
+    GEN q = FpXQX_divrem_Barrett(x,mg,y,T,p,pr);
+    if (!q) return gc_NULL(av);
     if (!pr || pr==ONLY_DIVIDES) return gerepilecopy(av, q);
     gerepileall(av,2,&q,pr);
     return q;
@@ -1040,7 +1117,7 @@ FpXQX_rem(GEN x, GEN S, GEN T, GEN p)
   {
     pari_sp av=avma;
     GEN mg = B? B: FpXQX_invBarrett(y, T, p);
-    GEN r = FpXQX_divrem_Barrett_noGC(x, mg, y, T, p, ONLY_REM);
+    GEN r = FpXQX_divrem_Barrett(x, mg, y, T, p, ONLY_REM);
     return gerepileupto(av, r);
   }
 }
@@ -1059,9 +1136,14 @@ Fq_addmul(GEN x, GEN y, GEN z, GEN T, GEN p)
 GEN
 FpXQX_div_by_X_x(GEN a, GEN x, GEN T, GEN p, GEN *r)
 {
-  long l = lg(a)-1, i;
-  GEN z = cgetg(l, t_POL);
-  z[1] = evalsigne(1) | evalvarn(0);
+  long l = lg(a), i;
+  GEN z;
+  if (l <= 3)
+  {
+    if (r) *r = l == 2? gen_0: gcopy(gel(a,2));
+    return pol_0(0);
+  }
+  l--; z = cgetg(l, t_POL); z[1] = evalsigne(1) | evalvarn(0);
   gel(z, l-1) = gel(a,l);
   for (i=l-2; i>1; i--) /* z[i] = a[i+1] + x*z[i+1] */
     gel(z, i) = Fq_addmul(gel(a,i+1), x, gel(z,i+1), T, p);
@@ -1270,6 +1352,7 @@ FpXYQQ_pow(GEN x, GEN n, GEN S, GEN T, GEN p)
     ulong pp = to_FlxqX(x, NULL, T, p, &x, NULL, &T);
     S = ZX_to_Flx(S, pp);
     y = FlxX_to_ZXX( FlxYqq_pow(x, n, S, T, pp) );
+    y = gerepileupto(av, y);
   }
   else
   {
@@ -1278,7 +1361,7 @@ FpXYQQ_pow(GEN x, GEN n, GEN S, GEN T, GEN p)
     D.p = p;
     y = gen_pow(x, n, (void*)&D, &FpXYQQ_sqr, &FpXYQQ_mul);
   }
-  return gerepileupto(av, y);
+  return y;
 }
 
 GEN
@@ -1399,6 +1482,7 @@ FpXQXQ_pow(GEN x, GEN n, GEN S, GEN T, GEN p)
     ulong pp = to_FlxqX(x, S, T, p, &x, &S, &T);
     GEN z = FlxqXQ_pow(x, n, S, T, pp);
     y = FlxX_to_ZXX(z);
+    return gerepileupto(ltop, y);
   }
   else
   {
@@ -1406,9 +1490,9 @@ FpXQXQ_pow(GEN x, GEN n, GEN S, GEN T, GEN p)
     S = FpXQX_get_red(S, T, p);
     D.S = S; D.T = T; D.p = p;
     if (s < 0) x = FpXQXQ_inv(x,S,T,p);
-    y = gen_pow(x, n, (void*)&D,&_FpXQXQ_sqr,&_FpXQXQ_mul);
+    y = gen_pow_i(x, n, (void*)&D,&_FpXQXQ_sqr,&_FpXQXQ_mul);
+    return gerepilecopy(ltop, y);
   }
-  return gerepileupto(ltop, y);
 }
 
 /* generates the list of powers of x of degree 0,1,2,...,l*/
@@ -1445,7 +1529,7 @@ FpXQXQ_transmul_init(GEN tau, GEN S, GEN T, GEN p)
     bht = FpXQXn_mul(bt, h, n-1, T, p);
   else
   {
-    GEN bh = FpXQX_div(RgX_shift_shallow(tau, n-1), S, T, p);
+    GEN bh = FpXQX_div(FpXX_shift(tau, n-1), S, T, p);
     bht = FpXQX_recipspec(bh+2, lgpol(bh), n-1);
     setvarn(bht, vT);
   }
@@ -1459,11 +1543,11 @@ FpXQXQ_transmul(GEN tau, GEN a, long n, GEN T, GEN p)
   GEN t1, t2, t3, vec;
   GEN bt = gel(tau, 1), bht = gel(tau, 2), ft = gel(tau, 3);
   if (signe(a)==0) return pol_0(varn(a));
-  t2 = RgX_shift_shallow(FpXQX_mul(bt, a, T, p),1-n);
+  t2 = FpXX_shift(FpXQX_mul(bt, a, T, p),1-n);
   if (signe(bht)==0) return gerepilecopy(ltop, t2);
-  t1 = RgX_shift_shallow(FpXQX_mul(ft, a, T, p),-n);
+  t1 = FpXX_shift(FpXQX_mul(ft, a, T, p),-n);
   t3 = FpXQXn_mul(t1, bht, n-1, T, p);
-  vec = FpXX_sub(t2, RgX_shift_shallow(t3, 1), p);
+  vec = FpXX_sub(t2, FpXX_shift(t3, 1), p);
   return gerepileupto(ltop, vec);
 }
 
@@ -1584,11 +1668,13 @@ FpXQXQ_autpow_mul(void * E, GEN x, GEN y)
 GEN
 FpXQXQ_autpow(GEN aut, long n, GEN S, GEN T, GEN p)
 {
+  pari_sp av = avma;
   struct _FpXQXQ D;
   T = FpX_get_red(T, p);
   S = FpXQX_get_red(S, T, p);
   D.S=S; D.T=T; D.p=p;
-  return gen_powu(aut,n,&D,FpXQXQ_autpow_sqr,FpXQXQ_autpow_mul);
+  aut = gen_powu_i(aut,n,&D,FpXQXQ_autpow_sqr,FpXQXQ_autpow_mul);
+  return gerepilecopy(av, aut);
 }
 
 static GEN
@@ -1614,11 +1700,13 @@ FpXQXQ_auttrace_sqr(void *E, GEN x)
 GEN
 FpXQXQ_auttrace(GEN aut, long n, GEN S, GEN T, GEN p)
 {
+  pari_sp av = avma;
   struct _FpXQXQ D;
   T = FpX_get_red(T, p);
   S = FpXQX_get_red(S, T, p);
   D.S=S; D.T=T; D.p=p;
-  return gen_powu(aut,n,&D,FpXQXQ_auttrace_sqr,FpXQXQ_auttrace_mul);
+  aut = gen_powu_i(aut,n,&D,FpXQXQ_auttrace_sqr,FpXQXQ_auttrace_mul);
+  return gerepilecopy(av, aut);
 }
 
 static GEN
@@ -1648,11 +1736,13 @@ FpXQXQ_autsum_sqr(void * T, GEN x)
 GEN
 FpXQXQ_autsum(GEN aut, long n, GEN S, GEN T, GEN p)
 {
+  pari_sp av = avma;
   struct _FpXQXQ D;
   T = FpX_get_red(T, p);
   S = FpXQX_get_red(S, T, p);
   D.S=S; D.T=T; D.p=p;
-  return gen_powu(aut,n,&D,FpXQXQ_autsum_sqr,FpXQXQ_autsum_mul);
+  aut = gen_powu_i(aut,n,&D,FpXQXQ_autsum_sqr,FpXQXQ_autsum_mul);
+  return gerepilecopy(av, aut);
 }
 
 GEN
@@ -1687,26 +1777,98 @@ INLINE GEN
 FpXXn_red(GEN a, long n)
 { return RgXn_red_shallow(a, n); }
 
+/* (f*g) \/ x^n */
+static GEN
+FpXQX_mulhigh_i(GEN f, GEN g, long n, GEN T, GEN p)
+{
+  return FpXX_shift(FpXQX_mul(f,g,T, p),-n);
+}
+
+static GEN
+FpXQXn_mulhigh(GEN f, GEN g, long n2, long n, GEN T, GEN p)
+{
+  GEN F = RgX_blocks(f, n2, 2), fl = gel(F,1), fh = gel(F,2);
+  return FpXX_add(FpXQX_mulhigh_i(fl, g, n2, T, p), FpXQXn_mul(fh, g, n - n2, T, p), p);
+}
+
+/* Compute intformal(x^n*S)/x^(n+1) */
+static GEN
+FpXX_integXn(GEN x, long n, GEN p)
+{
+  long i, lx = lg(x);
+  GEN y;
+  if (lx == 2) return ZXX_copy(x);
+  y = cgetg(lx, t_POL); y[1] = x[1];
+  for (i=2; i<lx; i++)
+  {
+    ulong j = n+i-1;
+    GEN xi = gel(x,i);
+    if (!signe(xi))
+      gel(y,i) = gen_0;
+    else
+      gel(y,i) = typ(xi)==t_INT ? Fp_divu(xi, j, p)
+                                : FpX_divu(xi, j, p);
+  }
+  return ZXX_renormalize(y, lx);;
+}
+
+/* Compute intformal(x^n*S)/x^(n+1) */
+static GEN
+ZlXX_integXn(GEN x, long n, GEN p, ulong pp)
+{
+  long i, lx = lg(x);
+  GEN y;
+  if (lx == 2) return ZXX_copy(x);
+  if (!pp) return FpXX_integXn(x, n, p);
+  y = cgetg(lx, t_POL); y[1] = x[1];
+  for (i=2; i<lx; i++)
+  {
+    GEN xi = gel(x,i);
+    if (!signe(xi))
+      gel(y,i) = gen_0;
+    else
+    {
+      ulong j;
+      long v = u_lvalrem(n+i-1, pp, &j);
+      if (typ(xi)==t_INT)
+      {
+        if (v==0)
+          gel(y,i) = Fp_divu(xi, j, p);
+        else
+          gel(y,i) = Fp_divu(diviuexact(xi, upowuu(pp, v)), j, p);
+      } else
+      {
+        if (v==0)
+          gel(y,i) = FpX_divu(xi, j, p);
+        else
+          gel(y,i) = FpX_divu(ZX_divuexact(xi, upowuu(pp, v)), j, p);
+      }
+    }
+  }
+  return ZXX_renormalize(y, lx);;
+}
+
 GEN
-FpXQXn_exp(GEN h, long e, GEN T, GEN p)
+ZlXQXn_expint(GEN h, long e, GEN T, GEN p, ulong pp)
 {
   pari_sp av = avma, av2;
   long v = varn(h), n=1;
   GEN f = pol_1(v), g = pol_1(v);
   ulong mask = quadratic_prec_mask(e);
   av2 = avma;
-  if (signe(h)==0 || degpol(h)<1 || !gequal0(gel(h,2)))
-    pari_err_DOMAIN("FpXQXn_exp","valuation", "<", gen_1, h);
   for (;mask>1;)
   {
-    GEN q, w;
+    GEN u, w;
     long n2 = n;
     n<<=1; if (mask & 1) n--;
     mask >>= 1;
-    g = FpXX_sub(FpXX_mulu(g,2,p), FpXQXn_mul(f, FpXQXn_sqr(g, n2, T, p), n2, T, p), p);
-    q = FpXX_deriv(FpXXn_red(h,n2), p);
-    w = FpXX_add(q, FpXQXn_mul(g, FpXX_sub(FpXX_deriv(f, p), FpXQXn_mul(f,q,n-1, T, p), p),n-1, T, p), p);
-    f = FpXX_add(f, FpXQXn_mul(f, FpXX_sub(FpXXn_red(h, n), FpXX_integ(w, p), p), n, T, p), p);
+    u = FpXQXn_mul(g, FpXQX_mulhigh_i(f, FpXXn_red(h, n2-1), n2-1, T, p), n-n2, T, p);
+    u = FpXX_add(u, FpXX_shift(FpXXn_red(h, n-1), 1-n2), p);
+    w = FpXQXn_mul(f, ZlXX_integXn(u, n2-1, p, pp), n-n2, T, p);
+    f = FpXX_add(f, FpXX_shift(w, n2), p);
+    if (mask<=1) break;
+    u = FpXQXn_mul(g, FpXQXn_mulhigh(f, g, n2, n, T, p), n-n2, T, p);
+    g = FpXX_sub(g, FpXX_shift(u, n2), p);
     if (gc_needed(av2,2))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"FpXQXn_exp, e = %ld", n);
@@ -1716,18 +1878,16 @@ FpXQXn_exp(GEN h, long e, GEN T, GEN p)
   return gerepileupto(av, f);
 }
 
-/* (f*g) \/ x^n */
-static GEN
-FpXQX_mulhigh_i(GEN f, GEN g, long n, GEN T, GEN p)
-{
-  return RgX_shift_shallow(FpXQX_mul(f,g,T, p),-n);
-}
+GEN
+FpXQXn_expint(GEN h, long e, GEN T, GEN p)
+{ return ZlXQXn_expint(h, e, T, p, 0); }
 
-static GEN
-FpXQXn_mulhigh(GEN f, GEN g, long n2, long n, GEN T, GEN p)
+GEN
+FpXQXn_exp(GEN h, long e, GEN T, GEN p)
 {
-  GEN F = RgX_blocks(f, n2, 2), fl = gel(F,1), fh = gel(F,2);
-  return FpXX_add(FpXQX_mulhigh_i(fl, g, n2, T, p), FpXQXn_mul(fh, g, n - n2, T, p), p);
+  if (signe(h)==0 || degpol(h)<1 || !gequal0(gel(h,2)))
+    pari_err_DOMAIN("FpXQXn_exp","valuation", "<", gen_1, h);
+  return FpXQXn_expint(FpXX_deriv(h, p), e, T, p);
 }
 
 GEN
@@ -1762,7 +1922,7 @@ FpXQXn_inv(GEN f, long e, GEN T, GEN p)
     mask >>= 1;
     fr = FpXXn_red(f, n);
     u = FpXQXn_mul(W, FpXQXn_mulhigh(fr, W, n2, n, T, p), n-n2, T, p);
-    W = FpXX_sub(W, RgX_shift_shallow(u, n2), p);
+    W = FpXX_sub(W, FpXX_shift(u, n2), p);
     if (gc_needed(av2,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"FpXQXn_inv, e = %ld", n);
