@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -98,18 +99,6 @@ Flx_Xn1(long sv, long n, ulong p)
   t[i] = 1; return t;
 }
 
-static ulong
-Fl_nonsquare(ulong p)
-{
-  long k = 2;
-  for (;; k++)
-  {
-    long i = krouu(k, p);
-    if (!i) pari_err_PRIME("Fl_nonsquare",utoipos(p));
-    if (i < 0) return k;
-  }
-}
-
 static GEN
 Flx_root_mod_2(GEN f)
 {
@@ -228,7 +217,6 @@ static ulong
 Flx_otherroot(GEN x, ulong r, ulong p)
 { return Fl_neg(Fl_add(x[3], r, p), p); }
 
-
 /* 'todo' contains the list of factors to be split.
  * 'done' the list of finished factors, no longer touched */
 struct split_t { GEN todo, done; };
@@ -392,7 +380,7 @@ Flx_cubic_root(GEN ff, ulong p)
       if (Fa!=P)
         Sa = Fl_mul(Sa, Fl_div(Fa, P, p),p);
     }
-    avma = av;
+    set_avma(av);
     return Fl_sub(Fl_double(Sa,p),t,p);
   }
 }
@@ -512,11 +500,11 @@ Flx_oneroot(GEN f, ulong p)
   switch(lg(f))
   {
     case 2: return 0;
-    case 3: avma = av; return p;
+    case 3: return p;
   }
   if (p == 2) return Flx_oneroot_mod_2(f);
   r = Flx_oneroot_i(Flx_normalize(f, p), p, 0);
-  avma = av; return r;
+  return gc_ulong(av,r);
 }
 
 ulong
@@ -527,11 +515,11 @@ Flx_oneroot_split(GEN f, ulong p)
   switch(lg(f))
   {
     case 2: return 0;
-    case 3: avma = av; return p;
+    case 3: return p;
   }
   if (p == 2) return Flx_oneroot_mod_2(f);
   r = Flx_oneroot_i(Flx_normalize(f, p), p, 1);
-  avma = av; return r;
+  return gc_ulong(av,r);
 }
 
 /* assume that p is prime */
@@ -541,8 +529,8 @@ FpX_oneroot(GEN f, GEN pp) {
   ZX_rootmod_init(&f, pp);
   switch(lg(f))
   {
-    case 2: avma = av; return gen_0;
-    case 3: avma = av; return NULL;
+    case 2: set_avma(av); return gen_0;
+    case 3: return gc_NULL(av);
   }
   if (typ(f) == t_VECSMALL)
   {
@@ -551,11 +539,11 @@ FpX_oneroot(GEN f, GEN pp) {
       r = Flx_oneroot_mod_2(f);
     else
       r = Flx_oneroot_i(f, p, 0);
-    avma = av;
+    set_avma(av);
     return (r == p)? NULL: utoi(r);
   }
   f = FpX_oneroot_i(f, pp);
-  if (!f) { avma = av; return NULL; }
+  if (!f) return gc_NULL(av);
   return gerepileuptoint(av, f);
 }
 
@@ -713,18 +701,41 @@ FpX_nbroots(GEN f, GEN p)
 {
   pari_sp av = avma;
   GEN z = FpX_split_part(f, p);
-  avma = av; return degpol(z);
+  return gc_long(av, degpol(z));
 }
 
+/* 1 < deg(f) <= p */
+static int
+Flx_is_totally_split_i(GEN f, ulong p)
+{
+  GEN F = Flx_Frobenius(f, p);
+  return degpol(F)==1 && uel(F,2)==0UL && uel(F,3)==1UL;
+}
+int
+Flx_is_totally_split(GEN f, ulong p)
+{
+  pari_sp av = avma;
+  ulong n = degpol(f);
+  if (n <= 1) return 1;
+  if (n > p) return 0; /* includes n < 0 */
+  return gc_bool(av, Flx_is_totally_split_i(f,p));
+}
 int
 FpX_is_totally_split(GEN f, GEN p)
 {
-  long n=degpol(f);
   pari_sp av = avma;
+  ulong n = degpol(f);
+  int u;
   if (n <= 1) return 1;
-  if (abscmpui(n, p) > 0) return 0;
-  f = FpX_red(f, p);
-  avma = av; return gequalX(FpX_Frobenius(f, p));
+  if (abscmpui(n, p) > 0) return 0; /* includes n < 0 */
+  if (lgefint(p) != 3)
+    u = gequalX(FpX_Frobenius(FpX_red(f,p), p));
+  else
+  {
+    ulong pp = (ulong)p[2];
+    u = Flx_is_totally_split_i(ZX_to_Flx(f,pp), pp);
+  }
+  return gc_bool(av, u);
 }
 
 long
@@ -743,7 +754,7 @@ Flx_nbroots(GEN f, ulong p)
   }
   z = Flx_sub(Flx_Frobenius(f, p), polx_Flx(f[1]), p);
   z = Flx_gcd(z, f, p);
-  avma = av; return degpol(z);
+  return gc_long(av, degpol(z));
 }
 
 long
@@ -772,7 +783,7 @@ FpX_ddf_degree(GEN T, GEN XP, GEN p)
   for (i = 3; i <= l+1; i++)
   {
     b = FpX_FpXQV_eval(b, xq, T, p);
-    if (gequalX(b)) { avma = av; return i-1; }
+    if (gequalX(b)) return gc_long(av,i-1);
     hash_insert_long(&h, b, i-1);
   }
   if (DEBUGLEVEL>=7) timer_printf(&ti,"FpX_ddf_degree: baby");
@@ -782,9 +793,9 @@ FpX_ddf_degree(GEN T, GEN XP, GEN p)
   for(i = 2; i <= m+1; i++)
   {
     g = FpX_FpXQV_eval(g, xq, T, p);
-    if (hash_haskey_long(&h, g, &j)) { avma=av; return l*i-j; }
+    if (hash_haskey_long(&h, g, &j)) return gc_long(av, l*i-j);
   }
-  avma = av; return n;
+  return gc_long(av,n);
 }
 
 /* See <http://www.shoup.net/papers/factorimpl.pdf> */
@@ -882,11 +893,11 @@ FpX_edf_simple(GEN Tp, GEN XP, long d, GEN p, GEN V, long idx)
       GEN R = FpXQ_pow(FpX_Fp_add(t, randomi(p), p), p2, T, p);
       f = FpX_gcd(FpX_Fp_sub(R, gen_1, p), Tp, p);
       if (degpol(f) > 0 && degpol(f) < n) break;
-      avma = btop2;
+      set_avma(btop2);
     }
     if (degpol(f) > 0 && degpol(f) < n) break;
     if (++ct == 10 && !BPSW_psp(p)) pari_err_PRIME("FpX_edf_simple",p);
-    avma = btop;
+    set_avma(btop);
   }
   f = FpX_normalize(f, p);
   ff = FpX_div(Tp, f ,p);
@@ -906,7 +917,7 @@ FpX_edf_rec(GEN T, GEN hp, GEN t, long d, GEN p2, GEN p, GEN V, long idx)
   av = avma;
   do
   {
-    avma = av;
+    set_avma(av);
     R = FpXQ_pow(deg1pol(gen_1, randomi(p), vT), p2, h, p);
     u1 = FpX_gcd(FpX_Fp_sub(R, gen_1, p), hp, p);
     if (++ct == 10 && !BPSW_psp(p)) pari_err_PRIME("FpX_edf_rec",p);
@@ -1033,15 +1044,15 @@ FpX_ispower(GEN f, ulong k, GEN p, GEN *pt_r)
   {
     ulong pp = p[2];
     GEN fp = ZX_to_Flx(f, pp);
-    if (!Flx_ispower(fp, k, pp, pt_r)) { avma = av; return 0; }
-    if (pt_r) *pt_r = gerepileupto(av, Flx_to_ZX(*pt_r)); else avma = av;
+    if (!Flx_ispower(fp, k, pp, pt_r)) return gc_long(av,0);
+    if (pt_r) *pt_r = gerepileupto(av, Flx_to_ZX(*pt_r)); else set_avma(av);
     return 1;
   }
   lc = Fp_sqrtn(leading_coeff(f), stoi(k), p, NULL);
   if (!lc) { av = avma; return 0; }
   F = FpX_factor_Yun(f, p); l = lg(F)-1;
   for(i=1; i <= l; i++)
-    if (i%k && degpol(gel(F,i))) { avma = av; return 0; }
+    if (i%k && degpol(gel(F,i))) return gc_long(av,0);
   if (pt_r)
   {
     GEN r = scalarpol(lc, v), s = pol_1(v);
@@ -1103,7 +1114,7 @@ static GEN Flx_simplefact_Cantor(GEN T, ulong p);
 static GEN
 FpX_simplefact_Cantor(GEN T, GEN p)
 {
-  GEN V, XP;
+  GEN V;
   long i, l;
   if (lgefint(p) == 3)
   {
@@ -1111,9 +1122,9 @@ FpX_simplefact_Cantor(GEN T, GEN p)
     return Flx_simplefact_Cantor(ZX_to_Flx(T,pp), pp);
   }
   T = FpX_get_red(T, p);
-  XP = FpX_Frobenius(T, p);
   V = FpX_factor_Yun(get_FpX_mod(T), p); l = lg(V);
-  for (i=1; i < l; i++) gel(V,i) = FpX_ddf_Shoup(gel(V,i), XP, p);
+  for (i=1; i < l; i++)
+    gel(V,i) = FpX_ddf_Shoup(gel(V,i), FpX_Frobenius(gel(V,i), p), p);
   return vddf_to_simplefact(V, get_FpX_degree(T));
 }
 
@@ -1122,11 +1133,11 @@ FpX_isirred_Cantor(GEN Tp, GEN p)
 {
   pari_sp av = avma;
   pari_timer ti;
-  long n, d;
+  long n;
   GEN T = get_FpX_mod(Tp);
   GEN dT = FpX_deriv(T, p);
   GEN XP, D;
-  if (degpol(FpX_gcd(T, dT, p)) != 0) { avma = av; return 0; }
+  if (degpol(FpX_gcd(T, dT, p)) != 0) return gc_bool(av,0);
   n = get_FpX_degree(T);
   T = FpX_get_red(Tp, p);
   if (DEBUGLEVEL>=6) timer_start(&ti);
@@ -1134,8 +1145,7 @@ FpX_isirred_Cantor(GEN Tp, GEN p)
   if (DEBUGLEVEL>=6) timer_printf(&ti,"FpX_Frobenius");
   D = FpX_ddf_Shoup(T, XP, p);
   if (DEBUGLEVEL>=6) timer_printf(&ti,"FpX_ddf_Shoup");
-  d = degpol(gel(D, n));
-  avma = av; return d==n;
+  return gc_bool(av, degpol(gel(D,n)) == n);
 }
 
 static GEN FpX_factor_deg2(GEN f, GEN p, long d, long flag);
@@ -1159,7 +1169,7 @@ FpX_nbfact_Frobenius(GEN T, GEN XP, GEN p)
 {
   pari_sp av = avma;
   long s = ddf_to_nbfact(FpX_ddf_Shoup(T, XP, p));
-  avma = av; return s;
+  return gc_long(av,s);
 }
 
 long
@@ -1168,7 +1178,7 @@ FpX_nbfact(GEN T, GEN p)
   pari_sp av = avma;
   GEN XP = FpX_Frobenius(T, p);
   long n = FpX_nbfact_Frobenius(T, XP, p);
-  avma = av; return n;
+  return gc_long(av,n);
 }
 
 /* p > 2 */
@@ -1313,7 +1323,7 @@ split_nonsquares(struct split_t *S, GEN g, ulong p, GEN xt)
   long d = degpol(a);
   if (d < 0)
   {
-    ulong i, z = Fl_nonsquare(p);
+    ulong i, z = nonsquare_Fl(p);
     split_add_done(S, (GEN)z);
     for (i = 2; i <= q; i++) split_add_done(S, (GEN)Fl_mul(z, Fl_sqr(i,p), p));
   } else {
@@ -1427,7 +1437,7 @@ Flx_roots(GEN f, ulong p)
   switch(lg(f))
   {
     case 2: pari_err_ROOTS0("Flx_roots");
-    case 3: avma = av; return cgetg(1, t_VECSMALL);
+    case 3: set_avma(av); return cgetg(1, t_VECSMALL);
   }
   if (p == 2) return Flx_root_mod_2(f);
   return gerepileuptoleaf(av, Flx_roots_i(Flx_normalize(f, p), p));
@@ -1488,25 +1498,6 @@ Flx_factor_deg2(GEN f, ulong p, long d, long flag)
   }
 }
 
-void
-F2xV_to_FlxV_inplace(GEN v)
-{
-  long i;
-  for(i=1;i<lg(v);i++) gel(v,i)= F2x_to_Flx(gel(v,i));
-}
-void
-FlxV_to_ZXV_inplace(GEN v)
-{
-  long i;
-  for(i=1;i<lg(v);i++) gel(v,i)= Flx_to_ZX(gel(v,i));
-}
-void
-F2xV_to_ZXV_inplace(GEN v)
-{
-  long i;
-  for(i=1;i<lg(v);i++) gel(v,i)= F2x_to_ZX(gel(v,i));
-}
-
 static GEN
 F2x_Berlekamp_ker(GEN u)
 {
@@ -1579,14 +1570,14 @@ F2x_split_Berlekamp(GEN *t)
         pari_sp av = avma;
         long lb;
         b = F2x_rem(pol, a);
-        if (F2x_degree(b) <= 0) { avma=av; continue; }
+        if (F2x_degree(b) <= 0) { set_avma(av); continue; }
         b = F2x_gcd(a,b); lb = F2x_degree(b);
         if (lb && lb < la)
         {
           t[L] = F2x_div(a,b);
           t[i]= b; L++;
         }
-        else avma = av;
+        else set_avma(av);
       }
     }
   }
@@ -1737,7 +1728,7 @@ F2x_edf_simple(GEN Tp, GEN XP, long d, GEN V, long idx)
     if (lgpol(t) == 0) continue;
     f = F2x_gcd(t, Tp); df = F2x_degree(f);
     if (df > 0 && df < n) break;
-    avma = btop;
+    set_avma(btop);
   }
   ff = F2x_div(Tp, f);
   F2x_edf_simple(f, XP, d, V, idx);
@@ -1837,18 +1828,17 @@ F2x_isirred_Cantor(GEN T)
 {
   pari_sp av = avma;
   pari_timer ti;
-  long n, d;
+  long n;
   GEN dT = F2x_deriv(T);
   GEN XP, D;
-  if (F2x_degree(F2x_gcd(T, dT)) != 0) { avma = av; return 0; }
+  if (F2x_degree(F2x_gcd(T, dT)) != 0) return gc_bool(av,0);
   n = F2x_degree(T);
   if (DEBUGLEVEL>=6) timer_start(&ti);
   XP = F2x_Frobenius(T);
   if (DEBUGLEVEL>=6) timer_printf(&ti,"F2x_Frobenius");
   D = F2x_ddf_simple(T, XP);
   if (DEBUGLEVEL>=6) timer_printf(&ti,"F2x_ddf_simple");
-  d = F2x_degree(gel(D, n));
-  avma = av; return d==n;
+  return gc_bool(av, F2x_degree(gel(D,n)) == n);
 }
 #endif
 
@@ -1931,7 +1921,7 @@ Flx_ispower(GEN f, ulong k, ulong p, GEN *pt_r)
   if (lc == ULONG_MAX) { av = avma; return 0; }
   F = Flx_factor_squarefree(f, p); l = lg(F)-1;
   for (i = 1; i <= l; i++)
-    if (i%k && degpol(gel(F,i))) { avma = av; return 0; }
+    if (i%k && degpol(gel(F,i))) return gc_long(av,0);
   if (pt_r)
   {
     GEN r = Fl_to_Flx(lc, v), s = pol1_Flx(v);
@@ -1942,7 +1932,7 @@ Flx_ispower(GEN f, ulong k, ulong p, GEN *pt_r)
       r = Flx_mul(r, s, p);
     }
     *pt_r = gerepileuptoleaf(av, r);
-  } else av = avma;
+  } else set_avma(av);
   return 1;
 }
 
@@ -2053,10 +2043,10 @@ Flx_edf_simple(GEN Tp, GEN XP, long d, ulong p, GEN V, long idx)
       GEN R = Flxq_powu(Flx_Fl_add(t, random_Fl(p), p), p2, T, p);
       f = Flx_gcd(Flx_Fl_add(R, p-1, p), Tp, p);
       if (degpol(f) > 0 && degpol(f) < n) break;
-      avma = btop2;
+      set_avma(btop2);
     }
     if (degpol(f) > 0 && degpol(f) < n) break;
-    avma = btop;
+    set_avma(btop);
   }
   f = Flx_normalize(f, p);
   ff = Flx_div(Tp, f ,p);
@@ -2080,7 +2070,7 @@ Flx_edf_rec(GEN T, GEN XP, GEN hp, GEN t, long d, ulong p, GEN V, long idx)
   av = avma;
   do
   {
-    avma = av;
+    set_avma(av);
     R = Flxq_powu(mkvecsmall3(vT, random_Fl(p), 1), p2, h, p);
     u1 = Flx_gcd(Flx_Fl_add(R, p-1, p), hp, p);
   } while (degpol(u1)==0 || degpol(u1)==n);
@@ -2193,12 +2183,12 @@ Flx_ddf(GEN T, ulong p)
 static GEN
 Flx_simplefact_Cantor(GEN T, ulong p)
 {
-  GEN XP, V;
+  GEN V;
   long i, l;
   T = Flx_get_red(T, p);
-  XP = Flx_Frobenius(T, p);
   V = Flx_factor_squarefree(get_Flx_mod(T), p); l = lg(V);
-  for (i=1; i < l; i++) gel(V,i) = Flx_ddf_Shoup(gel(V,i), XP, p);
+  for (i=1; i < l; i++)
+    gel(V,i) = Flx_ddf_Shoup(gel(V,i), Flx_Frobenius(gel(V,i), p), p);
   return vddf_to_simplefact(V, get_Flx_degree(T));
 }
 
@@ -2207,11 +2197,9 @@ Flx_isirred_Cantor(GEN Tp, ulong p)
 {
   pari_sp av = avma;
   pari_timer ti;
-  long n, d;
-  GEN T = get_Flx_mod(Tp);
-  GEN dT = Flx_deriv(T, p);
-  GEN XP, D;
-  if (degpol(Flx_gcd(T, dT, p)) != 0) { avma = av; return 0; }
+  long n;
+  GEN T = get_Flx_mod(Tp), dT = Flx_deriv(T, p), XP, D;
+  if (degpol(Flx_gcd(T, dT, p)) != 0) return gc_bool(av,0);
   n = get_Flx_degree(T);
   T = Flx_get_red(Tp, p);
   if (DEBUGLEVEL>=6) timer_start(&ti);
@@ -2219,8 +2207,7 @@ Flx_isirred_Cantor(GEN Tp, ulong p)
   if (DEBUGLEVEL>=6) timer_printf(&ti,"Flx_Frobenius");
   D = Flx_ddf_Shoup(T, XP, p);
   if (DEBUGLEVEL>=6) timer_printf(&ti,"Flx_ddf_Shoup");
-  d = degpol(gel(D, n));
-  avma = av; return d==n;
+  return gc_bool(av, degpol(gel(D,n)) == n);
 }
 
 /* f monic */
@@ -2272,7 +2259,7 @@ Flx_nbfact_by_degree(GEN T, long *nb, ulong p)
     s += V[i];
   }
   *nb = s;
-  avma = av; return V;
+  set_avma(av); return V;
 }
 
 long
@@ -2280,7 +2267,7 @@ Flx_nbfact_Frobenius(GEN T, GEN XP, ulong p)
 {
   pari_sp av = avma;
   long s = ddf_to_nbfact(Flx_ddf_Shoup(T, XP, p));
-  avma = av; return s;
+  return gc_long(av,s);
 }
 
 /* T must be squarefree mod p*/
@@ -2290,15 +2277,15 @@ Flx_nbfact(GEN T, ulong p)
   pari_sp av = avma;
   GEN XP = Flx_Frobenius(T, p);
   long n = Flx_nbfact_Frobenius(T, XP, p);
-  avma = av; return n;
+  return gc_long(av,n);
 }
 
 int
 Flx_is_irred(GEN f, ulong p)
 {
   pari_sp av = avma;
-  int z = !!Flx_factor_i(Flx_normalize(f,p),p,2);
-  avma = av; return z;
+  f = Flx_normalize(f,p);
+  return gc_bool(av, !!Flx_factor_i(f,p,2));
 }
 
 /* Use this function when you think f is reducible, and that there are lots of
@@ -2314,7 +2301,7 @@ FpX_is_irred(GEN f, GEN p)
     case 1:  z = !!Flx_factor_i(f,p[2],2); break;
     default: z = !!FpX_factor_i(f,p,2); break;
   }
-  avma = av; return z;
+  return gc_bool(av,z);
 }
 GEN
 FpX_degfact(GEN f, GEN p) {
