@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -60,7 +61,7 @@ ZX_divides_i(GEN x, GEN y, GEN B)
     for (j=0; j<=i && j<=dz; j++)
       p1 = subii(p1, mulii(gel(z,j),gel(y,i-j)));
     if (signe(p1)) return NULL;
-    avma = av;
+    set_avma(av);
   }
   return z - 2;
 }
@@ -141,7 +142,7 @@ Beauzamy_bound(GEN S)
     s = addrr(s, divri(itor(sqri(c), prec), gel(bin,i+1)));
   }
   /* s = [S]_2^2 */
-  C = powruhalf(stor(3,prec), 3 + 2*d); /* 3^{3/2 + d} */
+  C = powruhalf(utor(3,prec), 3 + 2*d); /* 3^{3/2 + d} */
   C = divrr(mulrr(C, s), mulur(4*d, mppi(prec)));
   lS = absi_shallow(leading_coeff(S));
   return mulir(lS, sqrtr(C));
@@ -270,7 +271,7 @@ nextK:
       if (!signe(y) || !dvdii(constant_coeff(lcpol), y))
       {
         if (DEBUGLEVEL>3) err_printf("T");
-        avma = av; goto NEXT;
+        set_avma(av); goto NEXT;
       }
       y = lc; /* full computation */
       for (i=1; i<=K; i++)
@@ -284,7 +285,7 @@ nextK:
       if (! (q = ZX_divides_i(lcpol,y,bound)) )
       {
         if (DEBUGLEVEL>3) err_printf("*");
-        avma = av; goto NEXT;
+        set_avma(av); goto NEXT;
       }
       /* found a factor */
       list = cgetg(K+1, t_VEC);
@@ -378,7 +379,7 @@ root_bound(GEN P0)
     pari_sp av = avma;
     /* y = 2^k; Q(y) >= lP y^d ? */
     if (cmpii(shifteval(Q,k), shifti(lP, d*k)) >= 0) break;
-    avma = av;
+    set_avma(av);
   }
   if (k < 0) k = 0;
   x = int2n(k);
@@ -607,7 +608,7 @@ AGAIN:
     if (i == r && ZM_equal(CM_L, oldCM_L))
     {
       CM_L = oldCM_L;
-      avma = av2; continue;
+      set_avma(av2); continue;
     }
 
     CM_Lp = FpM_image(CM_L, utoipos(27449)); /* inexpensive test */
@@ -718,7 +719,7 @@ pick_prime(GEN a, long fl, pari_timer *T)
   if (equali1(lead)) lead = NULL;
   u_forprime_init(&S, 2, ULONG_MAX);
   av1 = avma;
-  for (np = 0; np < MAXNP; avma = av1)
+  for (np = 0; np < MAXNP; set_avma(av1))
   {
     ulong p = u_forprime_next(&S);
     long nfacp;
@@ -749,7 +750,7 @@ pick_prime(GEN a, long fl, pari_timer *T)
     }
     np++;
   }
-  avma = av; return chosenp;
+  return gc_ulong(av, chosenp);
 }
 
 /* Assume A a squarefree ZX; return the vector of its rational roots */
@@ -872,34 +873,34 @@ GEN
 ZX_squff(GEN f, GEN *ex)
 {
   GEN T, V, P, e;
-  long i, k, n, val;
+  long i, k, val = ZX_valrem(f, &f), n = 2 + degpol(f);
 
-  if (signe(leading_coeff(f)) < 0) f = gneg_i(f);
-  val = ZX_valrem(f, &f);
-  n = 1 + degpol(f); if (val) n++;
+  if (signe(leading_coeff(f)) < 0) f = ZX_neg(f);
   e = cgetg(n,t_VECSMALL);
   P = cgetg(n,t_COL);
-
   T = ZX_gcd_all(f, ZX_deriv(f), &V);
-  for (k=i=1;; k++)
+  for (k = i = 1;; k++)
   {
-    pari_sp av = avma;
-    GEN W = ZX_gcd_all(T,V, &T);
-    long dW = degpol(W);
-    /* W = prod P^e, e > k; V = prod P^e, e >= k */
-    if (dW == degpol(V)) /* V | T */
+    GEN W = ZX_gcd_all(T,V, &T); /* V and W are squarefree */
+    long dW = degpol(W), dV = degpol(V);
+    /* f = prod_i T_i^{e_i}
+     * W = prod_{i: e_i > k} T_i,
+     * V = prod_{i: e_i >= k} T_i,
+     * T = prod_{i: e_i > k} T_i^{e_i - k} */
+    if (!dW)
+    {
+      if (dV) { gel(P,i) = Q_primpart(V); e[i] = k; i++; }
+      break;
+    }
+    if (dW == dV)
     {
       GEN U;
-      if (!dW) { avma = av; break; }
       while ( (U = ZX_divides(T, V)) ) { k++; T = U; }
-      T = gerepilecopy(av, T);
     }
     else
     {
       gel(P,i) = Q_primpart(RgX_div(V,W));
-      e[i] = k; i++;
-      if (!dW) break;
-      V = W;
+      e[i] = k; i++; V = W;
     }
   }
   if (val) { gel(P,i) = pol_x(varn(f)); e[i] = val; i++;}
@@ -969,7 +970,7 @@ ZX_is_irred(GEN x)
   if (l == 4) return 1; /* degree 1 */
   if (ZX_val(x)) return 0;
   if (!ZX_is_squarefree(x)) return 0;
-  y = ZX_DDF(x); avma = av;
+  y = ZX_DDF(x); set_avma(av);
   return (lg(y) == 2);
 }
 
@@ -999,7 +1000,6 @@ ZX_is_squarefree(GEN x)
   pari_sp av = avma;
   GEN d;
   long m;
-  int r;
   if (lg(x) == 2) return 0;
   m = ZX_deflate_order(x);
   if (m > 1)
@@ -1008,7 +1008,161 @@ ZX_is_squarefree(GEN x)
     x = RgX_deflate(x, m);
   }
   d = ZX_gcd(x,ZX_deriv(x));
-  r = (lg(d) == 3); avma = av; return r;
+  return gc_bool(av, lg(d) == 3);
+}
+
+static int
+ZX_gcd_filter(GEN *pt_A, GEN *pt_P)
+{
+  GEN A = *pt_A, P = *pt_P;
+  long i, j, l = lg(A), n = 1, d = degpol(gel(A,1));
+  GEN B, Q;
+  for (i=2; i<l; i++)
+  {
+    long di = degpol(gel(A,i));
+    if (di==d) n++;
+    else if (d > di)
+    { n=1; d = di; }
+  }
+  if (n == l-1)
+    return 0;
+  B = cgetg(n+1, t_VEC);
+  Q = cgetg(n+1, typ(P));
+  for (i=1, j=1; i<l; i++)
+  {
+    if (degpol(gel(A,i))==d)
+    {
+      gel(B,j) = gel(A,i);
+      Q[j] = P[i];
+      j++;
+    }
+  }
+  *pt_A = B; *pt_P = Q; return 1;
+}
+
+static GEN
+ZX_gcd_Flx(GEN a, GEN b, ulong g, ulong p)
+{
+  GEN H = Flx_gcd(a, b, p);
+  if (!g)
+    return Flx_normalize(H, p);
+  else
+  {
+    ulong t = Fl_mul(g, Fl_inv(Flx_lead(H), p), p);
+    return Flx_Fl_mul(H, t, p);
+  }
+}
+
+static GEN
+ZX_gcd_slice(GEN A, GEN B, GEN g, GEN P, GEN *mod)
+{
+  pari_sp av = avma;
+  long i, n = lg(P)-1;
+  GEN H, T;
+  if (n == 1)
+  {
+    ulong p = uel(P,1), gp = g ? umodiu(g, p): 0;
+    GEN a = ZX_to_Flx(A, p), b = ZX_to_Flx(B, p);
+    GEN Hp = ZX_gcd_Flx(a, b, gp, p);
+    H = gerepileupto(av, Flx_to_ZX(Hp));
+    *mod = utoi(p);
+    return H;
+  }
+  T = ZV_producttree(P);
+  A = ZX_nv_mod_tree(A, P, T);
+  B = ZX_nv_mod_tree(B, P, T);
+  g = g ?  Z_ZV_mod_tree(g, P, T): NULL;
+  H = cgetg(n+1, t_VEC);
+  for(i=1; i <= n; i++)
+  {
+    ulong p = P[i];
+    GEN a = gel(A,i), b = gel(B,i);
+    gel(H,i) = ZX_gcd_Flx(a, b, g? g[i]: 0, p);
+  }
+  if (ZX_gcd_filter(&H, &P))
+    T = ZV_producttree(P);
+  H = nxV_chinese_center_tree(H, P, T, ZV_chinesetree(P, T));
+  *mod = gmael(T, lg(T)-1, 1);
+  gerepileall(av, 2, &H, mod);
+  return H;
+}
+
+GEN
+ZX_gcd_worker(GEN P, GEN A, GEN B, GEN g)
+{
+  GEN V = cgetg(3, t_VEC);
+  gel(V,1) = ZX_gcd_slice(A, B, equali1(g)? NULL: g , P, &gel(V,2));
+  return V;
+}
+
+static GEN
+ZX_gcd_chinese(GEN A, GEN P, GEN *mod)
+{
+  ZX_gcd_filter(&A, &P);
+  return nxV_chinese_center(A, P, mod);
+}
+
+GEN
+ZX_gcd_all(GEN A, GEN B, GEN *Anew)
+{
+  pari_sp av = avma;
+  long k, valH, valA, valB, vA = varn(A), dA = degpol(A), dB = degpol(B);
+  GEN worker, c, cA, cB, g, Ag, Bg, H = NULL, mod = gen_1, R;
+  GEN Ap, Bp, Hp;
+  forprime_t S;
+  ulong pp;
+  if (dA < 0) { if (Anew) *Anew = pol_0(vA); return ZX_copy(B); }
+  if (dB < 0) { if (Anew) *Anew = pol_1(vA); return ZX_copy(A); }
+  A = Q_primitive_part(A, &cA);
+  B = Q_primitive_part(B, &cB);
+  valA = ZX_valrem(A, &A); dA -= valA;
+  valB = ZX_valrem(B, &B); dB -= valB;
+  valH = minss(valA, valB);
+  valA -= valH; /* valuation(Anew) */
+  c = (cA && cB)? gcdii(cA, cB): NULL; /* content(gcd) */
+  if (!dA || !dB)
+  {
+    if (Anew) *Anew = RgX_shift_shallow(A, valA);
+    return monomial(c? c: gen_1, valH, vA);
+  }
+  g = gcdii(leading_coeff(A), leading_coeff(B)); /* multiple of lead(gcd) */
+  if (is_pm1(g)) {
+    g = NULL;
+    Ag = A;
+    Bg = B;
+  } else {
+    Ag = ZX_Z_mul(A,g);
+    Bg = ZX_Z_mul(B,g);
+  }
+  init_modular_big(&S);
+  do {
+    pp = u_forprime_next(&S);
+    Ap = ZX_to_Flx(Ag, pp);
+    Bp = ZX_to_Flx(Bg, pp);
+  } while (degpol(Ap) != dA || degpol(Bp) != dB);
+  if (degpol(Flx_gcd(Ap, Bp, pp)) == 0)
+  {
+    if (Anew) *Anew = RgX_shift_shallow(A, valA);
+    return monomial(c? c: gen_1, valH, vA);
+  }
+  worker = snm_closure(is_entry("_ZX_gcd_worker"), mkvec3(A, B, g? g: gen_1));
+  av = avma;
+  for (k = 1; ;k *= 2)
+  {
+    gen_inccrt_i("ZX_gcd", worker, g, (k+1)>>1, 0, &S, &H, &mod, ZX_gcd_chinese, NULL);
+    gerepileall(av, 2, &H, &mod);
+    Hp = ZX_to_Flx(H, pp);
+    if (lgpol(Flx_rem(Ap, Hp, pp)) || lgpol(Flx_rem(Bp, Hp, pp))) continue;
+    if (!ZX_divides(Bg, H)) continue;
+    R = ZX_divides(Ag, H);
+    if (R) break;
+  }
+  /* lead(H) = g */
+  if (g) H = Q_primpart(H);
+  if (c) H = ZX_Z_mul(H,c);
+  if (DEBUGLEVEL>5) err_printf("done\n");
+  if (Anew) *Anew = RgX_shift_shallow(R, valA);
+  return valH? RgX_shift_shallow(H, valH): H;
 }
 
 #if 0
@@ -1030,95 +1184,13 @@ maxnorm(GEN p)
 }
 #endif
 
-/* A, B in Z[X] */
 GEN
-ZX_gcd_all(GEN A, GEN B, GEN *Anew)
+ZX_gcd(GEN A, GEN B)
 {
-  GEN R, a, b, q, H, Hp, g, Ag, Bg;
-  long m, n, valX, valA, vA = varn(A);
-  ulong p;
-  int small;
-  pari_sp ltop, av;
-  forprime_t S;
-
-  if (!signe(A)) { if (Anew) *Anew = pol_0(vA); return ZX_copy(B); }
-  if (!signe(B)) { if (Anew) *Anew = pol_1(vA); return ZX_copy(A); }
-  valA = ZX_valrem(A, &A);
-  valX = minss(valA, ZX_valrem(B, &B));
-  ltop = avma;
-
-  n = 1 + minss(degpol(A), degpol(B)); /* > degree(gcd) */
-  g = gcdii(leading_coeff(A), leading_coeff(B)); /* multiple of lead(gcd) */
-  if (is_pm1(g)) {
-    g = NULL;
-    Ag = A;
-    Bg = B;
-  } else {
-    Ag = ZX_Z_mul(A,g);
-    Bg = ZX_Z_mul(B,g);
-  }
-  small = (ZX_max_lg(A) == 3 && ZX_max_lg(B) == 3);
-  init_modular_big(&S);
-  av = avma;
-  R = NULL;/*-Wall*/
-  H = NULL;
-  if (DEBUGLEVEL>5) err_printf("ZX_gcd: ");
-  while ((p = u_forprime_next(&S)))
-  {
-    if (g && !umodiu(g,p)) continue;
-    if (H && gc_needed(av,1))
-    {
-      if (DEBUGMEM>1) pari_warn(warnmem,"QX_gcd");
-      gerepileall(av, 2, &H, &q);
-    }
-    a = ZX_to_Flx(A, p);
-    b = ZX_to_Flx(B, p); Hp = Flx_gcd(a,b, p);
-    m = degpol(Hp);
-    if (m == 0) { /* coprime. DONE */
-      avma = ltop;
-      if (Anew) {
-        if (valA != valX) A = RgX_shift(A, valA - valX);
-        *Anew = A;
-      }
-      return pol_xn(valX, vA);
-    }
-    if (m > n) continue; /* p | Res(A/G, B/G). Discard */
-
-    if (!g) /* make sure lead(H) = g mod p */
-      Hp = Flx_normalize(Hp, p);
-    else
-    {
-      ulong t = Fl_mul(umodiu(g, p), Fl_inv(Hp[m+2],p), p);
-      Hp = Flx_Fl_mul(Hp, t, p);
-    }
-    if (m < n)
-    { /* First time or degree drop [all previous p were as above; restart]. */
-      H = ZX_init_CRT(Hp,p,vA);
-      q = utoipos(p); n = m;
-      if (!small) continue; /* if gcd is small, try our luck */
-    }
-    else
-    {
-      if (DEBUGLEVEL>5) err_printf("%ld ", expi(q));
-      if (!ZX_incremental_CRT(&H, Hp, &q, p)) continue;
-    }
-    /* H stable: check divisibility */
-    if (!ZX_divides(Bg, H)) continue;
-    R = ZX_divides(Ag, H);
-    if (R) break;
-
-  }
-  if (!p) pari_err_OVERFLOW("ZX_gcd_all [ran out of primes]");
-  if (DEBUGLEVEL>5) err_printf("done\n");
-  if (Anew) {
-    A = R;
-    if (valA != valX) A = RgX_shift(A, valA - valX);
-    *Anew = A;
-  }
-  return valX ? RgX_shift(H, valX): H;
+  pari_sp av = avma;
+  return gerepilecopy(av, ZX_gcd_all(A,B,NULL));
 }
-GEN
-ZX_gcd(GEN A, GEN B) { return ZX_gcd_all(A,B,NULL); }
+
 GEN
 ZX_radical(GEN A) { GEN B; (void)ZX_gcd_all(A,ZX_deriv(A),&B); return B; }
 
@@ -1138,7 +1210,7 @@ QX_gcd(GEN A0, GEN B0)
 
   D = ZX_gcd(Q_primitive_part(A0, &a), Q_primitive_part(B0, &b));
   av2 = avma; a = _gcd(a,b);
-  if (isint1(a)) avma = av2; else D = ZX_Q_mul(D, a);
+  if (isint1(a)) set_avma(av2); else D = ZX_Q_mul(D, a);
   return gerepileupto(av, D);
 }
 
@@ -1244,7 +1316,7 @@ BD(GEN f)
   return G;
 }
 
-/* Let f be a non-zero QX, return the (squarefree) product of cyclotomic
+/* Let f be a nonzero QX, return the (squarefree) product of cyclotomic
  * divisors of f */
 GEN
 polcyclofactors(GEN f)
@@ -1259,7 +1331,7 @@ polcyclofactors(GEN f)
     f = BD(ZX_radical(f));
     if (f) return gerepilecopy(av, f);
   }
-  avma = av; return cgetg(1,t_VEC);
+  set_avma(av); return cgetg(1,t_VEC);
 }
 
 /* return t*x mod T(x), T a monic ZX. Assume deg(t) < deg(T) */
@@ -1329,11 +1401,11 @@ BD_iscyclo(GEN f)
   if (degpol(f) == 1) return isint1(gel(f,2))? 2: 1;
   f1 = ZX_graeffe(f);
   /* f = product of Phi_n, n odd */
-  if (ZX_equal(f, f1)) { avma = av; return BD_odd_iscyclo(f); }
+  if (ZX_equal(f, f1)) return gc_long(av, BD_odd_iscyclo(f));
 
   fn = ZX_z_unscale(f, -1); /* f(-x) */
   /* f = product of Phi_n, n = 2 mod 4 */
-  if (ZX_equal(f1, fn)) return 2*BD_odd_iscyclo(fn);
+  if (ZX_equal(f1, fn)) return gc_long(av, 2*BD_odd_iscyclo(fn));
 
   if (issquareall(f1, &f2))
   {
@@ -1343,7 +1415,7 @@ BD_iscyclo(GEN f)
     c = BD_iscyclo(f2);
     return odd(c)? 0: 2*c;
   }
-  avma = av; return 0;
+  return gc_long(av, 0);
 }
 long
 poliscyclo(GEN f)
@@ -1373,5 +1445,5 @@ poliscycloprod(GEN f)
   }
   f = BD(f); if (!f) return 0;
   for (i = lg(f)-1; i; i--) d -= degpol(gel(f,i));
-  avma = av; return d == 0;
+  return gc_long(av, d == 0);
 }
