@@ -2300,9 +2300,9 @@ RgXn_powu_i(GEN x, ulong m, long n)
 {
   struct modXn S;
   long v;
-  if (n == 1) return x;
+  if (n == 0) return x;
   v = RgX_valrem(x, &x);
-  if (v) n -= m * v;
+  if (v) { n -= m * v; if (n <= 0) return pol_0(varn(x)); }
   S.v = varn(x); S.n = n;
   x = gen_powu_i(x, m, (void*)&S,_sqrXn,_mulXn);
   if (v) x = RgX_shift_shallow(x, m * v);
@@ -2312,7 +2312,7 @@ GEN
 RgXn_powu(GEN x, ulong m, long n)
 {
   pari_sp av;
-  if (n == 1) return gcopy(x);
+  if (n == 0) return gcopy(x);
   av = avma; return gerepilecopy(av, RgXn_powu_i(x, m, n));
 }
 
@@ -2713,7 +2713,7 @@ RgXQ_norm(GEN x, GEN T)
   pari_sp av;
   long dx = degpol(x);
   GEN L, y;
-
+  if (degpol(T)==0) return gpowgs(x,0);
   av = avma; y = resultant(T, x);
   L = leading_coeff(T);
   if (gequal1(L) || !signe(x)) return y;
