@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -100,7 +101,7 @@ ZM_supnorm(GEN x)
 /**                           MULTIPLICATION                       **/
 /**                                                                **/
 /********************************************************************/
-/* x non-empty ZM, y a compatible nc (dimension > 0). */
+/* x nonempty ZM, y a compatible nc (dimension > 0). */
 static GEN
 ZM_nc_mul_i(GEN x, GEN y, long c, long l)
 {
@@ -130,7 +131,7 @@ ZV_zc_mul(GEN x, GEN y)
   return gerepileuptoint(av,s);
 }
 
-/* x non-empty ZM, y a compatible zc (dimension > 0). */
+/* x nonempty ZM, y a compatible zc (dimension > 0). */
 static GEN
 ZM_zc_mul_i(GEN x, GEN y, long c, long l)
 {
@@ -154,7 +155,7 @@ ZM_zc_mul(GEN x, GEN y) {
   return ZM_zc_mul_i(x,y, l, lgcols(x));
 }
 
-/* y non-empty ZM, x a compatible zv (dimension > 0). */
+/* y nonempty ZM, x a compatible zv (dimension > 0). */
 GEN
 zv_ZM_mul(GEN x, GEN y) {
   long i,j, lx = lg(x), ly = lg(y);
@@ -358,8 +359,8 @@ ZM_mul_sw(GEN A, GEN B, long m, long n, long p)
   C22 = add_slices(m2, p2, V2, 0, m2, 0, p2, M3, 0, m2, 0, p2);
   if (gc_needed(av, 1))
     gerepileall(av, 4, &C11, &C12, &C21, &C22);  /* destroy V2, M3 */
-  C = mkmat2(mkcol2(C11, C21), mkcol2(C12, C22));
-  return gerepilecopy(av, shallowmatconcat(C));
+  C = shallowconcat(vconcat(C11, C21), vconcat(C12, C22));
+  return gerepilecopy(av, C);
 }
 
 /* x[i,]*y. Assume lg(x) > 1 and 0 < i < lgcols(x) */
@@ -400,16 +401,96 @@ ZM_mul_classical(GEN x, GEN y, long l, long lx, long ly)
   return z;
 }
 
+static GEN
+ZM_mul_slice(GEN A, GEN B, GEN P, GEN *mod)
+{
+  pari_sp av = avma;
+  long i, n = lg(P)-1;
+  GEN H, T;
+  if (n == 1)
+  {
+    ulong p = uel(P,1);
+    GEN a = ZM_to_Flm(A, p);
+    GEN b = ZM_to_Flm(B, p);
+    GEN Hp = gerepileupto(av, Flm_to_ZM(Flm_mul(a, b, p)));
+    *mod = utoi(p); return Hp;
+  }
+  T = ZV_producttree(P);
+  A = ZM_nv_mod_tree(A, P, T);
+  B = ZM_nv_mod_tree(B, P, T);
+  H = cgetg(n+1, t_VEC);
+  for(i=1; i <= n; i++)
+    gel(H,i) = Flm_mul(gel(A,i),gel(B,i),P[i]);
+  H = nmV_chinese_center_tree_seq(H, P, T, ZV_chinesetree(P, T));
+  *mod = gmael(T, lg(T)-1, 1);
+  gerepileall(av, 2, &H, mod);
+  return H;
+}
+
+GEN
+ZM_mul_worker(GEN P, GEN A, GEN B)
+{
+  GEN V = cgetg(3, t_VEC);
+  gel(V,1) = ZM_mul_slice(A, B, P, &gel(V,2));
+  return V;
+}
+
+static long
+ZC_expi(GEN x)
+{
+  long i, l = lg(x), m=-1;
+  for(i = 1; i < l; i++)
+  {
+    GEN xi = gel(x,i);
+    long e = signe(xi) ? expi(xi): -1;
+    if (e > m) m = e;
+  }
+  return m;
+}
+
+static long
+ZM_expi(GEN x)
+{
+  long i, l = lg(x), m=-1;
+  for(i = 1; i < l; i++)
+  {
+    long e = ZC_expi(gel(x,i));
+    if (e > m) m = e;
+  }
+  return m;
+}
+
+static GEN
+ZM_mul_fast(GEN A, GEN B, long lA, long lB)
+{
+  pari_sp av = avma;
+  forprime_t S;
+  GEN H, worker;
+  long h, mA = ZM_expi(A), mB = ZM_expi(B);
+  if (mA==-1 || mB==-1) return zeromat(nbrows(A),lB-1);
+  h = 3 + mA + mB + expu(lA-1);
+  init_modular_big(&S);
+  worker = snm_closure(is_entry("_ZM_mul_worker"), mkvec2(A,B));
+  H = gen_crt("ZM_mul", worker, &S, NULL, h, 0, NULL,
+              nmV_chinese_center, FpM_center);
+  return gerepileupto(av, H);
+}
+
 /* Strassen-Winograd used for dim >= ZM_sw_bound */
 static GEN
 ZM_mul_i(GEN x, GEN y, long l, long lx, long ly)
 {
-  long s = maxss(ZM_max_lg_i(x,lx,l), ZM_max_lg_i(y,ly,lx));
-  long ZM_sw_bound = s > 60 ? 2: s > 25 ? 4: s>15 ? 8 : s > 8 ? 16 : 32;
-  if (l <= ZM_sw_bound || lx <= ZM_sw_bound || ly <= ZM_sw_bound)
-    return ZM_mul_classical(x, y, l, lx, ly);
+  if (lx > 70 && ly > 70 && l > 70)
+    return ZM_mul_fast(x, y, lx, ly);
   else
-    return ZM_mul_sw(x, y, l - 1, lx - 1, ly - 1);
+  {
+    long s = maxss(ZM_max_lg_i(x,lx,l), ZM_max_lg_i(y,ly,lx));
+    long ZM_sw_bound = s > 60 ? 2: s > 25 ? 4: s>15 ? 8 : s > 8 ? 16 : 32;
+    if (l <= ZM_sw_bound || lx <= ZM_sw_bound || ly <= ZM_sw_bound)
+      return ZM_mul_classical(x, y, l, lx, ly);
+    else
+      return ZM_mul_sw(x, y, l - 1, lx - 1, ly - 1);
+  }
 }
 
 GEN
@@ -435,6 +516,16 @@ QM_mul(GEN x, GEN y)
   return z;
 }
 
+GEN
+QM_sqr(GEN x)
+{
+  GEN dx, nx = Q_primitive_part(x, &dx);
+  GEN z = ZM_sqr(nx);
+  if (dx)
+    z = ZM_Q_mul(z, gsqr(dx));
+  return z;
+}
+
 GEN
 QM_QC_mul(GEN x, GEN y)
 {
@@ -560,12 +651,17 @@ ZM_transmul(GEN x, GEN y)
 static GEN
 ZM_sqr_i(GEN x, long l)
 {
-  long s = ZM_max_lg_i(x,l,l);
-  long ZM_sw_bound = s > 60 ? 2: s > 25 ? 4: s>15 ? 8 : s > 8 ? 16 : 32;
-  if (l <= ZM_sw_bound)
-    return ZM_mul_classical(x, x, l, l, l);
+  if (l > 70)
+    return ZM_mul_fast(x, x, l, l);
   else
-    return ZM_mul_sw(x, x, l - 1, l - 1, l - 1);
+  {
+    long s = ZM_max_lg_i(x,l,l);
+    long ZM_sw_bound = s > 60 ? 2: s > 25 ? 4: s>15 ? 8 : s > 8 ? 16 : 32;
+    if (l <= ZM_sw_bound)
+      return ZM_mul_classical(x, x, l, l, l);
+    else
+      return ZM_mul_sw(x, x, l - 1, l - 1, l - 1);
+  }
 }
 
 GEN
@@ -698,14 +794,14 @@ ZM_pow(GEN x, GEN n)
 {
   pari_sp av = avma;
   if (!signe(n)) return matid(lg(x)-1);
-  return gerepileupto(av, gen_pow(x, n, NULL, &_ZM_sqr, &_ZM_mul));
+  return gerepilecopy(av, gen_pow_i(x, n, NULL, &_ZM_sqr, &_ZM_mul));
 }
 GEN
 ZM_powu(GEN x, ulong n)
 {
   pari_sp av = avma;
   if (!n) return matid(lg(x)-1);
-  return gerepileupto(av, gen_powu(x, n, NULL, &_ZM_sqr, &_ZM_mul));
+  return gerepilecopy(av, gen_powu_i(x, n, NULL, &_ZM_sqr, &_ZM_mul));
 }
 /********************************************************************/
 /**                                                                **/
@@ -793,6 +889,9 @@ ZM_sub(GEN x, GEN y)
 GEN
 ZC_Z_divexact(GEN x, GEN c)
 { pari_APPLY_type(t_COL, diviiexact(gel(x,i), c)) }
+GEN
+ZC_u_divexact(GEN x, ulong c)
+{ pari_APPLY_type(t_COL, diviuexact(gel(x,i), c)) }
 
 GEN
 ZM_Z_divexact(GEN x, GEN c)
@@ -1065,7 +1164,6 @@ ZM_hnfdivrem(GEN x, GEN y, GEN *Q)
   return R;
 }
 
-
 /********************************************************************/
 /**                                                                **/
 /**                               TESTS                            **/
@@ -1088,6 +1186,14 @@ ZV_equal0(GEN V)
     if (signe(gel(V,l))) return 0;
   return 1;
 }
+int
+ZMrow_equal0(GEN V, long i)
+{
+  long l = lg(V);
+  while (--l > 0)
+    if (signe(gcoeff(V,i,l))) return 0;
+  return 1;
+}
 
 static int
 ZV_equal_lg(GEN V, GEN W, long l)
@@ -1177,7 +1283,7 @@ ZM_isidentity(GEN x)
     for (i=1; i<j; )
       if (signe(gel(c,i++))) return 0;
     /* i = j */
-      if (!equali1(gel(c,i++))) return 0;
+    if (!equali1(gel(c,i++))) return 0;
     for (   ; i<lx; )
       if (signe(gel(c,i++))) return 0;
   }
@@ -1215,7 +1321,7 @@ ZM_isscalar(GEN x, GEN s)
     for (i=1; i<j; )
       if (signe(gel(c,i++))) return 0;
     /* i = j */
-      if (!equalii(gel(c,i++), s)) return 0;
+    if (!equalii(gel(c,i++), s)) return 0;
     for (   ; i<lx; )
       if (signe(gel(c,i++))) return 0;
   }
@@ -1282,7 +1388,7 @@ ZV_content(GEN x)
   for (i = 2; i < l; i++)
   {
     c = gcdii(c, gel(x,i));
-    if (is_pm1(c)) { avma = av; return gen_1; }
+    if (is_pm1(c)) { set_avma(av); return gen_1; }
   }
   return gerepileuptoint(av, c);
 }
@@ -1372,6 +1478,15 @@ zv_sum(GEN v)
   n = v[1]; for (i = 2; i < l; i++) n += v[i];
   return n;
 }
+/* assumes no overflow and 0 <= n <= #v */
+long
+zv_sumpart(GEN v, long n)
+{
+  long i, p;
+  if (!n) return 0;
+  p = v[1]; for (i = 2; i <= n; i++) p += v[i];
+  return p;
+}
 GEN
 ZV_sum(GEN v)
 {
