@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -485,7 +486,7 @@ minpoly_dvdslice(GEN M, long i, long j, long k)
   pari_sp av = avma;
   long r = signe(RgX_rem(minpoly_polslice(M, i, j-1, 0),
                         minpoly_polslice(M, j, k, 0)));
-  avma = av; return r==0;
+  return gc_bool(av, r == 0);
 }
 
 static void
@@ -698,6 +699,62 @@ RgXQ_minpoly_naive(GEN y, GEN P)
   GEN M = ker(RgXQ_matrix_pow(y,n,n,P));
   return content(RgM_to_RgXV(M,varn(P)));
 }
+
+static GEN
+RgXQ_minpoly_FpXQ(GEN x, GEN y, GEN p)
+{
+  pari_sp av = avma;
+  GEN r;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = uel(p, 2);
+    r = Flx_to_ZX_inplace(Flxq_minpoly(RgX_to_Flx(x, pp),
+                                   RgX_to_Flx(y, pp), pp));
+  }
+  else
+    r = FpXQ_minpoly(RgX_to_FpX(x, p), RgX_to_FpX(y, p), p);
+  return gerepileupto(av, FpX_to_mod(r, p));
+}
+
+static GEN
+RgXQ_minpoly_FpXQXQ(GEN x, GEN S, GEN pol, GEN p)
+{
+  pari_sp av = avma;
+  GEN r;
+  GEN T = RgX_to_FpX(pol, p);
+  if (signe(T)==0) pari_err_OP("minpoly",x,S);
+  if (lgefint(p) == 3)
+  {
+    ulong pp = uel(p, 2);
+    GEN Tp = ZX_to_Flx(T, pp);
+    r = FlxX_to_ZXX(FlxqXQ_minpoly(RgX_to_FlxqX(x, Tp, pp),
+                                   RgX_to_FlxqX(S, Tp, pp), Tp, pp));
+  }
+  else
+    r = FpXQXQ_minpoly(RgX_to_FpXQX(x, T, p), RgX_to_FpXQX(S, T, p), T, p);
+  return gerepileupto(av, FpXQX_to_mod(r, T, p));
+}
+
+#define code(t1,t2) ((t1 << 6) | t2)
+
+static GEN
+RgXQ_minpoly_fast(GEN x, GEN y)
+{
+  GEN p, pol;
+  long pa;
+  long t = RgX_type2(x,y, &p,&pol,&pa);
+  switch(t)
+  {
+    case t_INTMOD: return RgXQ_minpoly_FpXQ(x, y, p);
+    case t_FFELT:  return FFXQ_minpoly(x, y, pol);
+    case code(t_POLMOD, t_INTMOD):
+                   return RgXQ_minpoly_FpXQXQ(x, y, pol, p);
+    default:       return NULL;
+  }
+}
+
+#undef code
+
 static GEN
 easymin(GEN x, long v)
 {
@@ -716,6 +773,8 @@ easymin(GEN x, long v)
       if (varncmp(gvar(a), v) <= 0) pari_err_PRIORITY("minpoly", x, "<", v);
       return deg1pol(gen_1, gneg_i(a), v);
     }
+    R = RgXQ_minpoly_fast(a,b);
+    if (R) { setvarn(R, v); return R; }
     if (!issquarefree(b))
     {
       R = RgXQ_minpoly_naive(a, b);
@@ -735,15 +794,8 @@ minpoly(GEN x, long v)
   if (v < 0) v = 0;
   P = easymin(x,v);
   if (P) return gerepileupto(av,P);
-  avma = av;
-  if (typ(x) == t_POLMOD)
-  {
-    P = RgXQ_minpoly_naive(gel(x,2), gel(x,1));
-    setvarn(P,v); return gerepileupto(av,P);
-  }
-  if (typ(x) != t_MAT) pari_err_TYPE("minpoly",x);
-  if (lg(x) == 1) return pol_1(v);
-  return RgM_minpoly(x,v);
+  /* typ(x) = t_MAT */
+  set_avma(av); return RgM_minpoly(x,v);
 }
 
 /*******************************************************************/
@@ -754,7 +806,7 @@ minpoly(GEN x, long v)
 static int
 relative0(GEN a, GEN a0, long bit)
 { return gequal0(a) || (bit && gexpo(a)-gexpo(a0) < bit); }
-/* assume x a non-empty square t_MAT */
+/* assume x a nonempty square t_MAT */
 static GEN
 RgM_hess(GEN x0, long prec)
 {
@@ -927,14 +979,13 @@ carhess(GEN x, long v)
 /* Bound for sup norm of charpoly(M/dM), M integral: let B = |M|oo / |dM|,
  *   s = max_k binomial(n,k) (kB^2)^(k/2),
  * return ceil(log2(s)) */
-static double
-charpoly_bound(GEN M, GEN dM)
+static long
+charpoly_bound(GEN M, GEN dM, GEN N)
 {
   pari_sp av = avma;
-  GEN B = itor(ZM_supnorm(M), LOWDEFAULTPREC);
+  GEN B = itor(N, LOWDEFAULTPREC);
   GEN s = real_0(LOWDEFAULTPREC), bin, B2;
   long n = lg(M)-1, k;
-  double d;
   bin = gen_1;
   if (dM) B = divri(B, dM);
   B2 = sqrr(B);
@@ -944,65 +995,68 @@ charpoly_bound(GEN M, GEN dM)
     if (abscmprr(t, s) > 0) s = t;
     bin = diviuexact(muliu(bin, k), n-k+1);
   }
-  d = dbllog2(s); avma = av; return ceil(d);
+  return gc_long(av, ceil(dbllog2(s)));
 }
 
-/* Return char_{M/d}(X) = d^(-n) char_M(dX) modulo p. Assume dp = d mod p. */
 static GEN
-QM_charpoly_Flx(GEN M, ulong dp, ulong p)
+QM_charpoly_ZX_slice(GEN A, GEN dM, GEN P, GEN *mod)
 {
   pari_sp av = avma;
-  GEN H = Flm_charpoly_i(ZM_to_Flm(M,p), p);
-  if (dp) H = Flx_rescale(H, Fl_inv(dp,p), p);
-  return gerepileuptoleaf(av, H);
+  long i, n = lg(P)-1;
+  GEN H, T;
+  if (n == 1)
+  {
+    ulong p = uel(P,1), dp = dM ? umodiu(dM, p): 1;
+    GEN Hp, a = ZM_to_Flm(A, p);
+    Hp = Flm_charpoly_i(a, p);
+    if (dp != 1) Hp = Flx_rescale(Hp, Fl_inv(dp, p), p);
+    Hp = gerepileupto(av, Flx_to_ZX(Hp));
+    *mod = utoi(p); return Hp;
+  }
+  T = ZV_producttree(P);
+  A = ZM_nv_mod_tree(A, P, T);
+  H = cgetg(n+1, t_VEC);
+  for(i=1; i <= n; i++)
+  {
+    ulong p = uel(P,i), dp = dM ? umodiu(dM, p): 1;
+    gel(H,i) = Flm_charpoly(gel(A, i), p);
+    if (dp != 1) gel(H,i) = Flx_rescale(gel(H,i), Fl_inv(dp, p), p);
+  }
+  H = nxV_chinese_center_tree(H, P, T, ZV_chinesetree(P,T));
+  *mod = gmael(T, lg(T)-1, 1);
+  gerepileall(av, 2, &H, mod);
+  return H;
 }
 
-static int
-ZX_CRT(GEN *H, GEN Hp, GEN *q, ulong p, long bit)
+GEN
+QM_charpoly_ZX_worker(GEN P, GEN M, GEN dM)
 {
-  if (!*H)
-  {
-    *H = ZX_init_CRT(Hp, p, 0);
-    if (DEBUGLEVEL>5)
-      err_printf("charpoly mod %lu, bound = 2^%ld\n", p, expu(p));
-    if (expu(p) > bit) return 1;
-    *q = utoipos(p);
-  }
-  else
-  {
-    int stable = ZX_incremental_CRT(H, Hp, q,p);
-    if (DEBUGLEVEL>5)
-      err_printf("charpoly mod %lu (stable=%ld), bound = 2^%ld\n",
-                 p, stable, expi(*q));
-    if (stable && expi(*q) > bit) return 1;
-  }
-  return 0;
+  GEN V = cgetg(3, t_VEC);
+  gel(V,1) = QM_charpoly_ZX_slice(M, equali1(dM) ? NULL:dM, P, &gel(V,2));
+  return V;
 }
 
 /* Assume M a square ZM, dM integer. Return charpoly(M / dM) in Z[X] */
 static GEN
-QM_charpoly_ZX_i(GEN M, GEN dM, long bit)
+QM_charpoly_ZX_i(GEN M, GEN dM, long bound)
 {
   long n = lg(M)-1;
-  GEN q = NULL, H = NULL;
   forprime_t S;
-  ulong p;
+  GEN worker = snm_closure(is_entry("_QM_charpoly_ZX_worker"),
+                           mkvec2(M, dM? dM: gen_1));
   if (!n) return pol_1(0);
-
-  if (bit < 0) bit = (long)charpoly_bound(M, dM) + 1;
-  if (DEBUGLEVEL>5) err_printf("ZM_charpoly: bit-bound 2^%ld\n", bit);
-  init_modular_big(&S);
-  while ((p = u_forprime_next(&S)))
+  if (bound < 0)
   {
-    ulong dMp = 0;
-    GEN Hp;
-    if (dM && !(dMp = umodiu(dM, p))) continue;
-    Hp = QM_charpoly_Flx(M, dMp, p);
-    if (ZX_CRT(&H, Hp, &q,p, bit)) break;
+    GEN N = ZM_supnorm(M);
+    if (signe(N) == 0) return monomial(gen_1, n, 0);
+    bound = charpoly_bound(M, dM, N) + 1;
   }
-  if (!p) pari_err_OVERFLOW("charpoly [ran out of primes]");
-  return H;
+  if (DEBUGLEVEL>5) err_printf("ZM_charpoly: bound 2^%ld\n", bound);
+  init_modular_big(&S);
+  return gen_crt("QM_charpoly_ZX", worker, &S, dM, bound, 0, NULL,
+                 nxV_chinese_center, FpX_center);
 }
+
 GEN
 QM_charpoly_ZX_bound(GEN M, long bit)
 {
@@ -1374,7 +1428,7 @@ matcompanion(GEN x)
   { /* not monic. Hardly ever used */
     pari_sp av = avma;
     GEN d = gclone(gneg(gel(x,n+2)));
-    avma = av;
+    set_avma(av);
     for (j=1; j<=n; j++) gel(c,j) = gdiv(gel(x,j+1), d);
     gunclone(d);
   }
@@ -1428,7 +1482,6 @@ gtrace(GEN x)
       gel(y,2) = FF_trace(x);
       return y;
 
-
     case t_RFRAC:
       av = avma; return gerepileupto(av, gadd(x, conj_i(x)));
 
@@ -1472,7 +1525,7 @@ qfgaussred_positive(GEN a)
   for (k=1; k<n; k++)
   {
     GEN bk, p = gcoeff(b,k,k), invp;
-    if (gsigne(p)<=0) { avma = av; return NULL; } /* not positive definite */
+    if (gsigne(p)<=0) return gc_NULL(av); /* not positive definite */
     invp = ginv(p);
     bk = row(b, k);
     for (i=k+1; i<n; i++) gcoeff(b,k,i) = gmul(gel(bk,i), invp);
@@ -1493,7 +1546,7 @@ qfgaussred_positive(GEN a)
 
 /* Maximal pivot strategy: x is a suitable pivot if it is non zero and either
  * - an exact type, or
- * - it is maximal among remaining non-zero (t_REAL) pivots */
+ * - it is maximal among remaining nonzero (t_REAL) pivots */
 static int
 suitable(GEN x, long k, GEN *pp, long *pi)
 {
@@ -1512,7 +1565,7 @@ suitable(GEN x, long k, GEN *pp, long *pi)
 }
 
 /* Gauss reduction (arbitrary symetric matrix, only the part above the
- * diagonal is considered). If signature is non-zero, return only the
+ * diagonal is considered). If signature is nonzero, return only the
  * signature, in which case gsigne() should be defined for elements of a. */
 static GEN
 gaussred(GEN a, long signature)
@@ -1613,7 +1666,7 @@ gaussred(GEN a, long signature)
     }
   }
   if (!signature) return gerepilecopy(av, a);
-  avma = av; return mkvec2s(sp, sn);
+  set_avma(av); return mkvec2s(sp, sn);
 }
 
 GEN
@@ -1727,13 +1780,13 @@ jacobi(GEN a, long prec)
         e = expo(z); if (e > e2) { e2=e; p=j; q=i; }
       }
     }
-    avma = av2;
+    set_avma(av2);
   }
   /* sort eigenvalues from smallest to largest */
   c = indexsort(L);
   r2 = vecpermute(r, c); for (i=1; i<l; i++) gel(r,i) = gel(r2,i);
   L2 = vecpermute(L, c); for (i=1; i<l; i++) gel(L,i) = gel(L2,i);
-  avma = av1; return ja;
+  set_avma(av1); return ja;
 }
 
 /*************************************************************************/
@@ -1750,8 +1803,8 @@ matrixqz0(GEN x,GEN p)
   if (typ(p) != t_INT) pari_err_TYPE("matrixqz",p);
   if (signe(p)>=0) return QM_minors_coprime(x,p);
   if (!RgM_is_QM(x)) pari_err_TYPE("matrixqz", x);
-  if (absequaliu(p,1)) return QM_ImZ_hnf(x); /* p = -1 */
-  if (absequaliu(p,2)) return QM_ImQ_hnf(x); /* p = -2 */
+  if (absequaliu(p,1)) return QM_ImZ(x); /* p = -1 */
+  if (absequaliu(p,2)) return QM_ImQ(x); /* p = -2 */
   pari_err_FLAG("QM_minors_coprime");
   return NULL; /* LCOV_EXCL_LINE */
 }
@@ -1777,14 +1830,14 @@ QM_minors_coprime(GEN x, GEN D)
   {
     if (gequal0(ZM_det(x)))
       pari_err_DOMAIN("QM_minors_coprime", "rank(A)", "<",stoi(n),x);
-    avma = av; return matid(n);
+    set_avma(av); return matid(n);
   }
   /* m > n */
   if (!D || gequal0(D))
   {
     pari_sp av2 = avma;
     D = ZM_detmult(shallowtrans(x));
-    if (is_pm1(D)) { avma = av2; return ZM_copy(x); }
+    if (is_pm1(D)) { set_avma(av2); return ZM_copy(x); }
   }
   P = gel(Z_factor(D), 1); lP = lg(P);
   av1 = avma;
@@ -1815,7 +1868,7 @@ QM_minors_coprime(GEN x, GEN D)
 }
 
 static GEN
-QM_ImZ_hnfall_i(GEN A, GEN *U, long remove)
+QM_ImZ_all_i(GEN A, GEN *U, long remove, long hnf, long linindep)
 {
   GEN V = NULL, D;
   A = Q_remove_denom(A,&D);
@@ -1831,65 +1884,76 @@ QM_ImZ_hnfall_i(GEN A, GEN *U, long remove)
       A = ZM_Z_divexact(ZM_mul(A, V), D);
     }
   }
-  A = ZM_hnflll(A, U, remove);
-  if (U && V) *U = ZM_mul(V,*U);
+  if (!linindep && ZM_rank(A)==lg(A)-1) linindep = 1;
+  if (hnf || !linindep) A = ZM_hnflll(A, U, remove);
+  if (U && V)
+  {
+    if (hnf)    *U = ZM_mul(V,*U);
+    else        *U = V;
+  }
   return A;
 }
 GEN
-QM_ImZ_hnfall(GEN x, GEN *U, long remove)
+QM_ImZ_all(GEN x, GEN *U, long remove, long hnf)
 {
   pari_sp av = avma;
-  x = QM_ImZ_hnfall_i(x, U, remove);
-  if (U) gerepileall(av, 2, &x, &U); else x = gerepileupto(av, x);
+  x = QM_ImZ_all_i(x, U, remove, hnf, 0);
+  if (U) gerepileall(av, 2, &x, &U); else x = gerepilecopy(av, x);
   return x;
 }
 GEN
+QM_ImZ_hnfall(GEN x, GEN *U, long remove) { return QM_ImZ_all(x, U, remove, 1); }
+GEN
 QM_ImZ_hnf(GEN x) { return QM_ImZ_hnfall(x, NULL, 1); }
+GEN
+QM_ImZ(GEN x) { return QM_ImZ_all(x, NULL, 1, 0); }
 
 GEN
-QM_ImQ_hnfall(GEN x, GEN *U, long remove)
+QM_ImQ_all(GEN x, GEN *U, long remove, long hnf)
 {
-  pari_sp av = avma, av1;
-  long k, r, m, n = lg(x);
-  GEN c, V;
+  pari_sp av = avma;
+  long i, n = lg(x), m;
+  GEN ir, V, D, c, K = NULL;
 
   if (U) *U = matid(n-1);
   if (n==1) return gcopy(x);
-  m = lgcols(x); x = RgM_shallowcopy(x);
-  c = zero_zv(n-1); r = 1;
-  av1 = avma;
-  for (k = 1; k < m; k++)
-  {
-    long j = 1, a;
-    GEN p;
-    while (j<n && (c[j] || isintzero(gcoeff(x,k,j)))) j++;
-    if (j==n) continue;
+  m = lg(gel(x,1));
 
-    c[j]=k; p = gcoeff(x,k,j);
-    gel(x,j) = RgC_Rg_div(gel(x,j), p);
-    if (U) gel(*U,j) = RgC_Rg_div(gel(*U,j), p);
-    for (a = 1; a < n; a++)
-      if (a != j)
-      {
-        GEN t = gcoeff(x,k,a);
-        if (gequal0(t)) continue;
-        if (U) gel(*U,a) = RgC_sub(gel(*U,a), RgC_Rg_mul(gel(*U,j),t));
-        gel(x,a) = RgC_sub(gel(x,a), RgC_Rg_mul(gel(x,j),t));
-      }
-    if (++r == n) break;
-    if (gc_needed(av1,1))
-    {
-      if(DEBUGMEM>1) pari_warn(warnmem,"QM_ImQ_hnf");
-      gerepileall(av1, U? 2: 1, &x, U);
-    }
+  x = RgM_shallowcopy(x);
+  for (i=1; i<n; i++)
+  {
+    gel(x,i) = Q_primitive_part(gel(x,i), &c);
+    if (U && c && signe(c)) gcoeff(*U,i,i) = ginv(c);
   }
-  x = QM_ImZ_hnfall_i(x, U? &V: NULL, remove);
-  if (U) { *U = RgM_mul(*U,V); gerepileall(av,2,&x,U); }
-  else x = gerepileupto(av,x);
+
+  ir = ZM_indexrank(x);
+  if (U)
+  {
+    *U = vecpermute(*U, gel(ir,2));
+    if (remove < 2) K = ZM_ker(x);
+  }
+  x = vecpermute(x, gel(ir,2));
+
+  D = absi(ZM_det(rowpermute(x,gel(ir,1))));
+  x = RgM_Rg_div(x, D);
+  x = QM_ImZ_all_i(x, U? &V: NULL, remove, hnf, 1);
+
+  if (U)
+  {
+    *U = RgM_Rg_div(RgM_mul(*U,V),D);
+    if (remove < 2) *U = shallowconcat(K,*U);
+    if (!remove) x = shallowconcat(zeromatcopy(m-1,lg(K)-1), x);
+    gerepileall(av, 2, &x, U);
+  }
+  else x = gerepilecopy(av,x);
   return x;
 }
 GEN
+QM_ImQ_hnfall(GEN x, GEN *U, long remove) { return QM_ImQ_all(x, U, remove, 1); }
+GEN
 QM_ImQ_hnf(GEN x) { return QM_ImQ_hnfall(x, NULL, 1); }
+GEN
+QM_ImQ(GEN x) { return QM_ImQ_all(x, NULL, 1, 0); }
 
 GEN
 intersect(GEN x, GEN y)
