@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -88,7 +89,8 @@ shallowtrans(GEN x)
       dx = lgcols(x); y = cgetg(dx,t_MAT);
       for (i = 1; i < dx; i++) gel(y,i) = row_transpose(x,i);
       break;
-    default: pari_err_TYPE("shallowtrans",x); return NULL;
+    default: pari_err_TYPE("shallowtrans",x);
+      return NULL;/*LCOV_EXCL_LINE*/
   }
   return y;
 }
@@ -107,7 +109,8 @@ gtrans(GEN x)
       dx = lgcols(x); y = cgetg(dx,t_MAT);
       for (i = 1; i < dx; i++) gel(y,i) = row_transposecopy(x,i);
       break;
-    default: pari_err_TYPE("gtrans",x); return NULL;
+    default: pari_err_TYPE("gtrans",x);
+      return NULL;/*LCOV_EXCL_LINE*/
   }
   return y;
 }
@@ -350,12 +353,12 @@ extract0(GEN x, GEN l1, GEN l2)
   {
     if (typ(x) != t_MAT) pari_err_TYPE("extract",x);
     y = shallowextract(x,l2);
-    if (select_0(l1)) { avma = av; return zeromat(0, lg(y)-1); }
+    if (select_0(l1)) { set_avma(av); return zeromat(0, lg(y)-1); }
     if (lg(y) == 1 && lg(x) > 1)
     {
       if (!extract_selector_ok(lgcols(x), l1))
         pari_err_TYPE("vecextract [incorrect mask]", l1);
-      avma = av; return cgetg(1, t_MAT);
+      set_avma(av); return cgetg(1, t_MAT);
     }
     y = shallowextract(shallowtrans(y), l1);
     av2 = avma;
@@ -442,16 +445,27 @@ GEN
 vecslice0(GEN A, long y1, long y2)
 {
   long skip, lB, t = typ(A);
-  lB = vecslice_parse_arg(lg(A), &y1, &y2, &skip);
   switch(t)
   {
     case t_VEC: case t_COL:
+      lB = vecslice_parse_arg(lg(A), &y1, &y2, &skip);
       return vecslice_i(A, t,lB,y1,skip);
     case t_VECSMALL:
+      lB = vecslice_parse_arg(lg(A), &y1, &y2, &skip);
       return vecsmallslice_i(A, t,lB,y1,skip);
+    case t_LIST:
+      if (list_typ(A) == t_LIST_RAW)
+      {
+        GEN y, z = list_data(A);
+        long l = z? lg(z): 1;
+        lB = vecslice_parse_arg(l, &y1, &y2, &skip);
+        y = mklist(); if (!z) return y;
+        list_data(y) = vecslice_i(z, t_VEC,lB,y1,skip);
+        return y;
+      }
     default:
       pari_err_TYPE("_[_.._]",A);
-      return NULL;
+      return NULL;/*LCOV_EXCL_LINE*/
   }
 }
 
@@ -493,8 +507,7 @@ vecrange(GEN a, GEN b)
   l = itos(subii(b,a))+1;
   a = setloop(a);
   y = cgetg(l+1, t_VEC);
-  for (i=1; i<=l; a = incloop(a), i++)
-    gel(y,i) = icopy(a);
+  for (i=1; i<=l; a = incloop(a), i++) gel(y,i) = icopy(a);
   return y;
 }
 
@@ -506,8 +519,7 @@ vecrangess(long a, long b)
   if (a>b) return cgetg(1,t_VEC);
   l = b-a+1;
   y = cgetg(l+1, t_VEC);
-  for (i=1; i<=l; a++, i++)
-    gel(y,i) = stoi(a);
+  for (i=1; i<=l; a++, i++) gel(y,i) = stoi(a);
   return y;
 }
 
@@ -536,7 +548,7 @@ genindexselect(void *E, long (*f)(void* E, GEN x), GEN A)
   av = avma;
   for (i = lv = 1; i < l; i++) {
     if (f(E, gel(z,i))) v[lv++] = i;
-    avma = av;
+    set_avma(av);
   }
   clone_unlock_deep(A); fixlg(v, lv); return v;
 }
@@ -624,12 +636,12 @@ parselect(GEN C, GEN D, long flag)
   check_callgen1(C, "parselect");
   if (!is_vec_t(typ(D))) pari_err_TYPE("parselect",D);
   W = cgetg(l, t_VECSMALL); av = avma;
-  worker = strtoclosure("_parselect_worker", 1, C);
+  worker = snm_closure(is_entry("_parselect_worker"), mkvec(C));
   V = gen_parapply(worker, D);
   for (lv=1, i=1; i<l; i++)
     if (signe(gel(V,i))) W[lv++] = i;
   fixlg(W, lv);
-  avma = av;
+  set_avma(av);
   return flag? W: extract_copy(D, W);
 }
 
@@ -753,24 +765,68 @@ veccatselapply(void *Epred, long (*pred)(void* E, GEN x), void *Efun,
   return lg(v) == 1? v: gerepilecopy(av, shallowconcat1(v));
 }
 
+/* suitable for gerepileupto */
+GEN
+parapply_slice_worker(GEN D, GEN worker)
+{
+  long l, i;
+  GEN w = cgetg_copy(D, &l);
+  for (i = 1; i < l; i++) gel(w,i) = closure_callgen1(worker, gel(D,i));
+  return w;
+}
+
+/* B <- {A[i] : i = r (mod m)}, 1 <= r <= m */
+static void
+arithprogset(GEN B, GEN A, long r, long m)
+{
+  long i, k, l = lg(A);
+  for (k = 1, i = r; i < l; i += m, k++) gel(B, k) = gel(A, i);
+  setlg(B, k);
+}
+GEN
+gen_parapply_slice(GEN worker, GEN D, long mmin)
+{
+  long l, r, n = lg(D)-1, m = minss(mmin, n), pending = 0;
+  GEN L = cgetg(n / m + 2, t_VEC), va = mkvec(L), V = cgetg_copy(D, &l);
+  struct pari_mt pt;
+  mt_queue_start_lim(&pt, worker, m);
+  for (r = 1; r <= m || pending; r++)
+  {
+    long workid;
+    GEN done;
+    if (r <= m) arithprogset(L, D, r, m);
+    mt_queue_submit(&pt, r, r <= m? va: NULL);
+    done = mt_queue_get(&pt, &workid, &pending);
+    if (done)
+    {
+      long j, k, J = lg(done)-1;
+      for (j = 1, k = workid; j <= J; j++, k +=m) gel(V, k) = gel(done, j);
+    }
+  }
+  mt_queue_end(&pt); return V;
+}
+
 GEN
 gen_parapply(GEN worker, GEN D)
 {
-  long l, i, pending = 0, workid;
-  GEN V, W, done;
+  long l = lg(D), i, pending = 0;
+  GEN W, V;
   struct pari_mt pt;
+
+  if (l == 1) return cgetg(1, typ(D));
   W = cgetg(2, t_VEC);
-  V = cgetg_copy(D, &l);
+  V = cgetg(l, typ(D));
   mt_queue_start_lim(&pt, worker, l-1);
-  for (i=1; i<l || pending; i++)
+  for (i = 1; i < l || pending; i++)
   {
-    if (i<l) gel(W,1) = gel(D,i);
+    long workid;
+    GEN done;
+    if (i < l) gel(W,1) = gel(D,i);
     mt_queue_submit(&pt, i, i<l? W: NULL);
     done = mt_queue_get(&pt, &workid, &pending);
     if (done) gel(V,workid) = done;
   }
-  mt_queue_end(&pt);
-  return V;
+  mt_queue_end(&pt); return V;
 }
 
 GEN
@@ -917,6 +973,20 @@ diagonal_shallow(GEN x)
   return y;
 }
 
+GEN
+zv_diagonal(GEN x)
+{
+  long j, l = lg(x), n = l-1;
+  GEN y = cgetg(l,t_MAT);
+
+  for (j = 1; j < l; j++)
+  {
+    gel(y,j) = zero_Flv(n);
+    ucoeff(y,j,j) = uel(x,j);
+  }
+  return y;
+}
+
 /* compute m*diagonal(d) */
 GEN
 matmuldiagonal(GEN m, GEN d)
@@ -971,5 +1041,3 @@ RgM_diagonal(GEN m)
   for (i=1; i<lx; i++) gel(y,i) = gcopy(gcoeff(m,i,i));
   return y;
 }
-
-
