@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -183,7 +184,6 @@ polhermite(long n, long v)
   GEN q,a,r;
 
   if (v<0) v = 0;
-  if (n < 0) pari_err_DOMAIN("polhermite", "degree", "<", gen_0, stoi(n));
   if (n==0) return pol_1(v);
 
   q = cgetg(n+3, t_POL); r = q + n+2;
@@ -201,17 +201,37 @@ polhermite(long n, long v)
   q[1] = evalsigne(1) | evalvarn(v);
   return q;
 }
+static void
+err_hermite(long n)
+{ pari_err_DOMAIN("polhermite", "degree", "<", gen_0, stoi(n)); }
 GEN
-polhermite_eval(long n, GEN x)
+polhermite_eval0(long n, GEN x, long flag)
 {
   long i;
   pari_sp av, av2;
   GEN x2, u, v;
 
-  if (!x) return polhermite(n, 0);
-  if (gequalX(x)) return polhermite(n, varn(x));
-  if (n==0) return gen_1;
-  if (n==1) return gmul2n(x,1);
+  if (n < 0) err_hermite(n);
+  if (!x || gequalX(x))
+  {
+    long v = x? varn(x): 0;
+    if (flag)
+    {
+      if (!n) err_hermite(-1);
+      retmkvec2(polhermite(n-1,v),polhermite(n,v));
+    }
+    return polhermite(n, v);
+  }
+  if (n==0)
+  {
+    if (flag) err_hermite(-1);
+    return gen_1;
+  }
+  if (n==1)
+  {
+    if (flag) retmkvec2(gen_1, gmul2n(x,1));
+    return gmul2n(x,1);
+  }
   av = avma; x2 = gmul2n(x,1); v = gen_1; u = x2;
   av2= avma;
   for (i=1; i<n; i++)
@@ -221,8 +241,11 @@ polhermite_eval(long n, GEN x)
     t = gsub(gmul(x2, u), gmulsg(2*i,v));
     v = u; u = t;
   }
+  if (flag) return gerepilecopy(av, mkvec2(v, u));
   return gerepileupto(av, u);
 }
+GEN
+polhermite_eval(long n, GEN x) { return polhermite_eval0(n, x, 0); }
 
 /* Legendre polynomial
  * L0=1; L1=X; (n+1)*L(n+1)=(2*n+1)*X*L(n)-n*L(n-1)
@@ -257,20 +280,59 @@ pollegendre(long n, long v)
   q[1] = evalsigne(1) | evalvarn(v);
   return gerepileupto(av, gmul2n(q,-n));
 }
-
+/* q such that Ln * 2^n = q(x^2) [n even] or x q(x^2) [n odd] */
 GEN
-pollegendre_eval(long n, GEN x)
+pollegendre_reduced(long n, long v)
 {
-  long i;
+  long k, l, N;
   pari_sp av;
-  GEN u, v;
+  GEN a, r, q;
 
-  if (!x) return pollegendre(n, 0);
-  if (gequalX(x)) return pollegendre(n, varn(x));
+  if (v<0) v = 0;
   /* pollegendre(-n) = pollegendre(n-1) */
   if (n < 0) n = -n-1;
-  if (n==0) return gen_1;
-  if (n==1) return gcopy(x);
+  if (n<=1) return n? scalarpol_shallow(gen_2,v): pol_1(v);
+
+  N = n >> 1;
+  q = cgetg(N+3, t_POL); r = q + N+2;
+  gel(r--,0) = a = binomialuu(n<<1,n);
+  for (k=1,l=n; l>1; k++,l-=2)
+  { /* l = n-2*k+2 */
+    av = avma;
+    a = diviuuexact(muluui(l, l-1, a), 2*k, n+l-1);
+    togglesign(a);
+    gel(r--,0) = a = gerepileuptoint(av, a);
+  }
+  q[1] = evalsigne(1) | evalvarn(v);
+  return q;
+}
+
+GEN
+pollegendre_eval0(long n, GEN x, long flag)
+{
+  pari_sp av;
+  GEN u, v;
+  long i;
+
+  if (n < 0) n = -n-1; /* L(-n) = L(n-1) */
+  /* n >= 0 */
+  if (flag && flag != 1) pari_err_FLAG("pollegendre");
+  if (!x || gequalX(x))
+  {
+    long v = x? varn(x): 0;
+    if (flag) retmkvec2(pollegendre(n-1,v), pollegendre(n,v));
+    return pollegendre(n, v);
+  }
+  if (n==0)
+  {
+    if (flag) retmkvec2(gen_1, gcopy(x));
+    return gen_1;
+  }
+  if (n==1)
+  {
+    if (flag) retmkvec2(gcopy(x), gen_1);
+    return gcopy(x);
+  }
   av = avma; v = gen_1; u = x;
   for (i=1; i<n; i++)
   { /* u = P_i(x), v = P_{i-1}(x), compute t = P_{i+1}(x) */
@@ -279,8 +341,82 @@ pollegendre_eval(long n, GEN x)
     t = gdivgs(gsub(gmul(gmulsg(2*i+1,x), u), gmulsg(i,v)), i+1);
     v = u; u = t;
   }
+  if (flag) return gerepilecopy(av, mkvec2(v, u));
   return gerepileupto(av, u);
 }
+GEN
+pollegendre_eval(long n, GEN x) { return pollegendre_eval0(n, x, 0); }
+
+/* Laguerre polynomial
+ * L0^a = 1; L1^a = -X+a+1;
+ * (n+1)*L^a(n+1) = (-X+(2*n+a+1))*L^a(n) - (n+a)*L^a(n-1)
+ * L^a(n) = sum_{k=0}^n (-1)^k * binom(n+a,n-k) * x^k/k! */
+GEN
+pollaguerre(long n, GEN a, long v)
+{
+  pari_sp av = avma;
+  GEN L = cgetg(n+3, t_POL), c1 = gen_1, c2 = mpfact(n);
+  long i;
+
+  L[1] = evalsigne(1) | evalvarn(v);
+  if (odd(n)) togglesign_safe(&c2);
+  for (i = n; i >= 0; i--)
+  {
+    gel(L, i+2) = gdiv(c1, c2);
+    if (i)
+    {
+      c2 = divis(c2,-i);
+      c1 = gdivgs(gmul(c1, gaddsg(i,a)), n+1-i);
+    }
+  }
+  return gerepilecopy(av, L);
+}
+static void
+err_lag(long n)
+{ pari_err_DOMAIN("pollaguerre", "degree", "<", gen_0, stoi(n)); }
+GEN
+pollaguerre_eval0(long n, GEN a, GEN x, long flag)
+{
+  pari_sp av = avma;
+  long i;
+  GEN v, u;
+
+  if (n < 0) err_lag(n);
+  if (flag && flag != 1) pari_err_FLAG("pollaguerre");
+  if (!a) a = gen_0;
+  if (!x || gequalX(x))
+  {
+    long v = x? varn(x): 0;
+    if (flag)
+    {
+      if (!n) err_lag(-1);
+      retmkvec2(pollaguerre(n-1,a,v), pollaguerre(n,a,v));
+    }
+    return pollaguerre(n,a,v);
+  }
+  if (n==0)
+  {
+    if (flag) err_lag(-1);
+    return gen_1;
+  }
+  if (n==1)
+  {
+    if (flag) retmkvec2(gsub(gaddgs(a,1),x), gen_1);
+    return gsub(gaddgs(a,1),x);
+  }
+  av = avma; v = gen_1; u = gsub(gaddgs(a,1),x);
+  for (i=1; i<n; i++)
+  { /* u = P_i(x), v = P_{i-1}(x), compute t = P_{i+1}(x) */
+    GEN t;
+    if ((i & 0xff) == 0) gerepileall(av,2,&u, &v);
+    t = gdivgs(gsub(gmul(gsub(gaddsg(2*i+1,a),x), u), gmul(gaddsg(i,a),v)), i+1);
+    v = u; u = t;
+  }
+  if (flag) return gerepilecopy(av, mkvec2(v, u));
+  return gerepileupto(av, u);
+}
+GEN
+pollaguerre_eval(long n, GEN x, GEN a) { return pollaguerre_eval0(n, x, a, 0); }
 
 /* polcyclo(p) = X^(p-1) + ... + 1 */
 static GEN
@@ -350,7 +486,7 @@ polcyclo_eval(long n, GEN x)
   if (tx == t_INT) { /* shortcut */
     if (is_pm1(x))
     {
-      avma = av;
+      set_avma(av);
       if (signe(x) > 0 && l == 1) return utoipos(P[1]);
       return gen_1;
     }
@@ -482,6 +618,36 @@ matqpascal(long n, GEN q)
   return gerepilecopy(av, m);
 }
 
+GEN
+eulerianpol(long N, long v)
+{
+  pari_sp av = avma;
+  long n, n2, k = 0;
+  GEN A;
+  if (v < 0) v = 0;
+  if (N <= 0) pari_err_DOMAIN("eulerianpol", "index", "<=", gen_0, stoi(N));
+  if (N == 1) return pol_1(v);
+  if (N == 2) return deg1pol_shallow(gen_1, gen_1, v);
+  A = cgetg(N+1, t_VEC);
+  gel(A,1) = gen_1; gel(A,2) = gen_1; /* A_2 = x+1 */
+  for (n = 3; n <= N; n++)
+  { /* A(n,k) = (n-k)A(n-1,k-1) + (k+1)A(n-1,k) */
+    n2 = n >> 1;
+    if (odd(n)) gel(A,n2+1) = mului(n+1, gel(A,n2));
+    for (k = n2-1; k; k--)
+      gel(A,k+1) = addii(mului(n-k, gel(A,k)), mului(k+1, gel(A,k+1)));
+    if (gc_needed(av,1))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"eulerianpol, %ld/%ld",n,N);
+      for (k = odd(n)? n2+1: n2; k < N; k++) gel(A,k+1) = gen_0;
+      A = gerepilecopy(av, A);
+    }
+  }
+  k = N >> 1; if (odd(N)) k++;
+  for (; k < N; k++) gel(A,k+1) = gel(A, N-k);
+  return gerepilecopy(av, RgV_to_RgX(A, v));
+}
+
 /******************************************************************/
 /**                                                              **/
 /**                       PRECISION CHANGES                      **/
@@ -632,7 +798,8 @@ laplace(GEN x)
   {
     case t_POL: x = pollaplace(x); break;
     case t_SER: x = serlaplace(x); break;
-    default: pari_err_TYPE("laplace",x);
+    default: if (is_scalar_t(typ(x))) return gcopy(x);
+             pari_err_TYPE("laplace",x);
   }
   return gerepilecopy(av, x);
 }
@@ -812,7 +979,7 @@ binomial(GEN n, long k)
       GEN z = subiu(n,k);
       if (cmpis(z,k) < 0)
       {
-        k = itos(z); avma = av;
+        k = itos(z); set_avma(av);
         if (k <= 1)
         {
           if (k < 0) return gen_0;
@@ -883,7 +1050,7 @@ vecbinomial(long n)
 /**                  STIRLING NUMBERS                              **/
 /********************************************************************/
 /* Stirling number of the 2nd kind. The number of ways of partitioning
-   a set of n elements into m non-empty subsets. */
+   a set of n elements into m nonempty subsets. */
 GEN
 stirling2(ulong n, ulong m)
 {
@@ -983,6 +1150,18 @@ polrecip(GEN x)
 /**                  POLYNOMIAL INTERPOLATION                      **/
 /**                                                                **/
 /********************************************************************/
+static GEN
+RgV_polint_fast(GEN X, GEN Y, long v)
+{
+  GEN p, pol;
+  long t, pa;
+  if (X) t = RgV_type2(X,Y, &p, &pol, &pa);
+  else   t = Rg_type(Y, &p, &pol, &pa);
+  if (t != t_INTMOD) return NULL;
+  Y = RgC_to_FpC(Y, p);
+  X = X? RgC_to_FpC(X, p): identity_ZV(lg(Y)-1);
+  return FpX_to_mod(FpV_polint(X, Y, p, v), p);
+}
 /* allow X = NULL for [1,...,n] */
 GEN
 RgV_polint(GEN X, GEN Y, long v)
@@ -990,11 +1169,8 @@ RgV_polint(GEN X, GEN Y, long v)
   pari_sp av0 = avma, av;
   GEN Q, P = NULL;
   long i, l = lg(Y);
-  if (!X)
-  {
-    X = cgetg(l, t_VEC);
-    for (i=1; i<l; i++) gel(X,i) = utoipos(i);
-  }
+  if ((Q = RgV_polint_fast(X,Y,v))) return Q;
+  if (!X) X = identity_ZV(l-1);
   Q = roots_to_pol(X, v); av = avma;
   for (i=1; i<l; i++)
   {
@@ -1006,63 +1182,57 @@ RgV_polint(GEN X, GEN Y, long v)
     P = P? RgX_add(P, dP): dP;
     if (gc_needed(av,2))
     {
-      if (DEBUGMEM>1) pari_warn(warnmem,"FpV_polint");
+      if (DEBUGMEM>1) pari_warn(warnmem,"RgV_polint i = %ld/%ld", i, l-1);
       P = gerepileupto(av, P);
     }
   }
-  if (!P) { avma = av; return zeropol(v); }
+  if (!P) { set_avma(av); return zeropol(v); }
   return gerepileupto(av0, P);
 }
+static int
+inC(GEN x)
+{
+  switch(typ(x)) {
+    case t_INT: case t_REAL: case t_FRAC: case t_COMPLEX: case t_QUAD: return 1;
+    default: return 0;
+  }
+}
+static long
+check_dy(GEN X, GEN x, long n)
+{
+  GEN D = NULL;
+  long i, ns = 0;
+  if (!inC(x)) return -1;
+  for (i = 0; i < n; i++)
+  {
+    GEN t = gsub(x, gel(X,i));
+    if (!inC(t)) return -1;
+    t = gabs(t, DEFAULTPREC);
+    if (!D || gcmp(t,D) < 0) { ns = i; D = t; }
+  }
+  /* X[ns] is closest to x */
+  return ns;
+}
 /* X,Y are "spec" GEN vectors with n > 0 components ( at X[0], ... X[n-1] ) */
 GEN
-polint_i(GEN X, GEN Y, GEN x, long n, GEN *ptdy)
+polintspec(GEN X, GEN Y, GEN x, long n, long *pe)
 {
-  long i, m, ns = 0;
-  pari_sp av = avma;
+  long i, m, ns;
+  pari_sp av = avma, av2;
   GEN y, c, d, dy = NULL; /* gcc -Wall */
 
-  if (n == 1)
-  {
-    if (ptdy) *ptdy = gen_0;
-    return gmul(gel(Y,0), Rg_get_1(x));
-  }
-  if (!X)
-  {
-    X = cgetg(n+1, t_VEC);
-    for (i=1; i<=n; i++) gel(X,i) = utoipos(i);
-    X++;
-  }
-  switch(typ(x)) {
-    case t_INT: case t_REAL: case t_FRAC: case t_COMPLEX: case t_QUAD:
-    {
-      GEN D = NULL;
-      for (i=0; i<n; i++)
-      {
-        GEN t = gsub(x,gel(X,i));
-        switch(typ(t))
-        {
-          case t_INT: case t_REAL: case t_FRAC: case t_COMPLEX: case t_QUAD:
-            t = gabs(t, DEFAULTPREC);
-            if (!D || gcmp(t,D) < 0) { ns = i; D = t; }
-            break;
-          default:
-            goto NODY;
-        }
-      }
-      break;
-      /* X[ns] is closest to x */
-    }
-NODY:
-    default:
-      if (ptdy) { *ptdy = gen_0; ptdy = NULL; }
-  }
-  c = new_chunk(n);
-  d = new_chunk(n); for (i=0; i<n; i++) gel(c,i) = gel(d,i) = gel(Y,i);
-  y = gel(d,ns--);
+  if (pe) *pe = -HIGHEXPOBIT;
+  if (n == 1) return gmul(gel(Y,0), Rg_get_1(x));
+  if (!X) X = identity_ZV(n) + 1;
+  av2 = avma;
+  ns = check_dy(X, x, n); if (ns < 0) { pe = NULL; ns = 0; }
+  c = cgetg(n+1, t_VEC);
+  d = cgetg(n+1, t_VEC); for (i=0; i<n; i++) gel(c,i+1) = gel(d,i+1) = gel(Y,i);
+  y = gel(d,ns+1);
   /* divided differences */
-  for (m=1; m<n; m++)
+  for (m = 1; m < n; m++)
   {
-    for (i=0; i<n-m; i++)
+    for (i = 0; i < n-m; i++)
     {
       GEN ho = gsub(gel(X,i),x), hp = gsub(gel(X,i+m),x), den = gsub(ho,hp);
       if (gequal0(den))
@@ -1071,21 +1241,24 @@ NODY:
         char *x2 = stack_sprintf("X[%ld]", i+m+1);
         pari_err_DOMAIN("polinterpolate",x1,"=",strtoGENstr(x2), X);
       }
-      den = gdiv(gsub(gel(c,i+1),gel(d,i)), den);
-      gel(c,i) = gmul(ho,den);
-      gel(d,i) = gmul(hp,den);
+      den = gdiv(gsub(gel(c,i+2),gel(d,i+1)), den);
+      gel(c,i+1) = gmul(ho,den);
+      gel(d,i+1) = gmul(hp,den);
     }
-    dy = (2*(ns+1) < n-m)? gel(c,ns+1): gel(d,ns--);
+    dy = (2*ns < n-m)? gel(c,ns+1): gel(d,ns--);
     y = gadd(y,dy);
+    if (gc_needed(av2,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"polint, %ld/%ld",m,n-1);
+      gerepileall(av2, 4, &y, &c, &d, &dy);
+    }
   }
-  if (!ptdy) return gerepileupto(av, y);
-  *ptdy = dy;
-  gerepileall(av, 2, &y, ptdy);
-  return y;
+  if (pe && inC(dy)) *pe = gexpo(dy);
+  return gerepileupto(av, y);
 }
 
 GEN
-polint(GEN X, GEN Y, GEN t, GEN *ptdy)
+polint_i(GEN X, GEN Y, GEN t, long *pe)
 {
   long lx = lg(X), vt;
 
@@ -1100,7 +1273,7 @@ polint(GEN X, GEN Y, GEN t, GEN *ptdy)
     Y = X;
     X = NULL;
   }
-  if (ptdy) *ptdy = gen_0;
+  if (pe) *pe = -HIGHEXPOBIT;
   vt = t? gvar(t): 0;
   if (vt != NO_VARIABLE)
   { /* formal interpolation */
@@ -1120,7 +1293,15 @@ polint(GEN X, GEN Y, GEN t, GEN *ptdy)
   }
   /* numerical interpolation */
   if (lx == 1) return Rg_get_0(t);
-  return polint_i(X? X+1: NULL,Y+1,t,lx-1,ptdy);
+  return polintspec(X? X+1: NULL,Y+1,t,lx-1, pe);
+}
+GEN
+polint(GEN X, GEN Y, GEN t, GEN *pe)
+{
+  long e;
+  GEN p = polint_i(X, Y, t, &e);
+  if (pe) *pe = stoi(e);
+  return p;
 }
 
 /********************************************************************/
@@ -1277,7 +1458,7 @@ gen_sortspec_uniq(GEN v, long n, void *E, int (*cmp)(void*,GEN,GEN))
   }
   while (ix<=nx) w[m++] = x[ix++];
   while (iy<=ny) w[m++] = y[iy++]+NX;
-  avma = av;
+  set_avma(av);
   W = cgetg(m, t_VECSMALL);
   for (i = 1; i < m; i++) W[i] = w[i];
   return W;
@@ -1320,7 +1501,7 @@ gen_sortspec(GEN v, long n, void *E, int (*cmp)(void*,GEN,GEN))
       w[m++] = y[iy++]+nx;
   while (ix<=nx) w[m++] = x[ix++];
   while (iy<=ny) w[m++] = y[iy++]+nx;
-  avma = (pari_sp)w; return w;
+  set_avma((pari_sp)w); return w;
 }
 
 static void
@@ -1425,20 +1606,34 @@ gen_sort_inplace(GEN x, void *E, int (*cmp)(void*,GEN,GEN), GEN *perm)
     for (i=1; i<lx; i++) gel(z,i) = gel(x,y[i]);
     for (i=1; i<lx; i++) gel(x,i) = gel(z,i);
     *perm = y;
-    avma = (pari_sp)y;
+    set_avma((pari_sp)y);
   } else {
     for (i=1; i<lx; i++) gel(y,i) = gel(x,y[i]);
     for (i=1; i<lx; i++) gel(x,i) = gel(y,i);
-    avma = av;
+    set_avma(av);
   }
 }
+GEN
+gen_sort_shallow(GEN x, void *E, int (*cmp)(void*,GEN,GEN))
+{
+  long tx, lx, i;
+  pari_sp av;
+  GEN y, z;
+
+  init_sort(&x, &tx, &lx);
+  if (lx<=2) return x;
+  z = cgetg(lx, tx); av = avma;
+  y = gen_sortspec(x,lx-1, E, cmp);
+  for (i=1; i<lx; i++) gel(z,i) = gel(x,y[i]);
+  return gc_const(av, z);
+}
 
 static int
 closurecmp(void *data, GEN x, GEN y)
 {
   pari_sp av = avma;
   long s = gsigne(closure_callgen2((GEN)data, x,y));
-  avma = av; return s;
+  set_avma(av); return s;
 }
 static void
 check_positive_entries(GEN k)
@@ -1568,7 +1763,7 @@ vecsearch(GEN v, GEN x, GEN k)
     x = (GEN)itos(x);
   else if (!is_matvec_t(tv)) pari_err_TYPE("vecsearch", v);
   r = CMP? gen_search(v, x, 0, E, CMP): key_search(v, x, k);
-  avma = av; return r;
+  return gc_long(av, r);
 }
 
 GEN
@@ -1580,6 +1775,44 @@ ZV_sort_uniq(GEN L) { return gen_sort_uniq(L, (void*)&cmpii, &cmp_nodata); }
 void
 ZV_sort_inplace(GEN L) { gen_sort_inplace(L, (void*)&cmpii, &cmp_nodata,NULL); }
 
+GEN
+vec_equiv(GEN F)
+{
+  pari_sp av = avma;
+  long j, k, L = lg(F);
+  GEN w = cgetg(L, t_VEC);
+  GEN perm = gen_indexsort(F, (void*)&cmp_universal, cmp_nodata);
+  for (j = k = 1; j < L;)
+  {
+    GEN v = cgetg(L, t_VECSMALL);
+    long l = 1, o = perm[j];
+    v[l++] = o;
+    for (j++; j < L; v[l++] = perm[j++])
+      if (!gequal(gel(F,o), gel(F, perm[j]))) break;
+    setlg(v, l); gel(w, k++) = v;
+  }
+  setlg(w, k); return gerepilecopy(av,w);
+}
+
+GEN
+vec_reduce(GEN v, GEN *pE)
+{
+  GEN E, F, P = gen_indexsort(v, (void*)cmp_universal, cmp_nodata);
+  long i, m, l;
+  F = cgetg_copy(v, &l);
+  *pE = E = cgetg(l, t_VECSMALL);
+  for (i = m = 1; i < l;)
+  {
+    GEN u = gel(v, P[i]);
+    long k;
+    for(k = i + 1; k < l; k++)
+      if (cmp_universal(gel(v, P[k]), u)) break;
+    E[m] = k - i; gel(F, m) = u; i = k; m++;
+  }
+  setlg(F, m);
+  setlg(E, m); return F;
+}
+
 /********************************************************************/
 /**                      SEARCH IN SORTED VECTOR                   **/
 /********************************************************************/
@@ -1620,7 +1853,18 @@ long
 ZV_search(GEN x, GEN y) { return tablesearch(x, y, cmpii); }
 
 long
-zv_search(GEN x, long y) { return tablesearch(x, (GEN)y, cmp_small); }
+zv_search(GEN T, long x)
+{
+  long l = 1, u = lg(T)-1;
+  while (u>=l)
+  {
+    long i = (l+u)>>1;
+    if (x < T[i]) u = i-1;
+    else if (x > T[i]) l = i+1;
+    else return i;
+  }
+  return 0;
+}
 
 /********************************************************************/
 /**                   COMPARISON FUNCTIONS                         **/
@@ -1738,26 +1982,50 @@ merge_factor(GEN fx, GEN fy, void *data, int (*cmp)(void *,GEN,GEN))
 GEN
 merge_sort_uniq(GEN x, GEN y, void *data, int (*cmp)(void *,GEN,GEN))
 {
-  long ix, iy, m, lx = lg(x), ly = lg(y), l = lx+ly-1;
-  GEN M;
-
-  M = cgetg(l, t_COL);
-  m = ix = iy = 1;
-  while (ix<lx && iy<ly)
+  long i, j, k, lx = lg(x), ly = lg(y);
+  GEN z = cgetg(lx + ly - 1, typ(x));
+  i = j = k = 1;
+  while (i<lx && j<ly)
   {
-    int s = cmp(data, gel(x,ix), gel(y,iy));
+    int s = cmp(data, gel(x,i), gel(y,j));
     if (s < 0)
-    { gel(M,m) = gel(x,ix); ix++; }
-    else if (s == 0)
-    { gel(M,m) = gel(x,ix); iy++; ix++; }
+      gel(z,k++) = gel(x,i++);
+    else if (s > 0)
+      gel(z,k++) = gel(y,j++);
     else
-    { gel(M,m) = gel(y,iy); iy++; }
-    m++;
+    { gel(z,k++) = gel(x,i++); j++; }
   }
-  while (ix<lx) { gel(M,m) = gel(x,ix); ix++; m++; }
-  while (iy<ly) { gel(M,m) = gel(y,iy); iy++; m++; }
-  setlg(M, m); return M;
+  while (i<lx) gel(z,k++) = gel(x,i++);
+  while (j<ly) gel(z,k++) = gel(y,j++);
+  setlg(z, k); return z;
 }
+/* in case of equal keys in x,y, take the key from x */
+static GEN
+ZV_union_shallow_t(GEN x, GEN y, long t)
+{
+  long i, j, k, lx = lg(x), ly = lg(y);
+  GEN z = cgetg(lx + ly - 1, t);
+  i = j = k = 1;
+  while (i<lx && j<ly)
+  {
+    int s = cmpii(gel(x,i), gel(y,j));
+    if (s < 0)
+      gel(z,k++) = gel(x,i++);
+    else if (s > 0)
+      gel(z,k++) = gel(y,j++);
+    else
+    { gel(z,k++) = gel(x,i++); j++; }
+  }
+  while (i < lx) gel(z,k++) = gel(x,i++);
+  while (j < ly) gel(z,k++) = gel(y,j++);
+  setlg(z, k); return z;
+}
+GEN
+ZV_union_shallow(GEN x, GEN y)
+{ return ZV_union_shallow_t(x, y, t_VEC); }
+GEN
+ZC_union_shallow(GEN x, GEN y)
+{ return ZV_union_shallow_t(x, y, t_COL); }
 
 /* sort generic factorization, in place */
 GEN
@@ -1774,7 +2042,7 @@ sort_factor(GEN y, void *data, int (*cmp)(void *,GEN,GEN))
   w = gen_sortspec(a, n-1, data, cmp);
   for (i=1; i<n; i++) { long k=w[i]; gel(A,i) = gel(a,k); gel(B,i) = gel(b,k); }
   for (i=1; i<n; i++) { gel(a,i) = gel(A,i); gel(b,i) = gel(B,i); }
-  avma = av; return y;
+  set_avma(av); return y;
 }
 /* sort polynomial factorization, in place */
 GEN
@@ -1839,63 +2107,18 @@ setsearch(GEN T, GEN y, long flag)
   return gen_search(T,y,flag,(void*)cmp_universal,cmp_nodata);
 }
 
+GEN
+setunion_i(GEN x, GEN y)
+{ return merge_sort_uniq(x,y, (void*)cmp_universal, cmp_nodata); }
+
 GEN
 setunion(GEN x, GEN y)
 {
   pari_sp av = avma;
-  long i, j, k, lx = lg(x), ly = lg(y);
-  GEN z = cgetg(lx + ly - 1, t_VEC);
   if (typ(x) != t_VEC) pari_err_TYPE("setunion",x);
   if (typ(y) != t_VEC) pari_err_TYPE("setunion",y);
-  i = j = k = 1;
-  while (i<lx && j<ly)
-  {
-    int s = cmp_universal(gel(x,i), gel(y,j));
-    if (s < 0)
-      z[k++] = x[i++];
-    else if (s > 0)
-      z[k++] = y[j++];
-    else {
-      z[k++] = x[i++];
-      j++;
-    }
-  }
-  while (i<lx) z[k++] = x[i++];
-  while (j<ly) z[k++] = y[j++];
-  setlg(z, k);
-  return gerepilecopy(av, z);
+  return gerepilecopy(av, setunion_i(x, y));
 }
-/* in case of equal keys in x,y, take the key from x */
-static GEN
-ZV_union_shallow_t(GEN x, GEN y, long t)
-{
-  long i, j, k, lx = lg(x), ly = lg(y);
-  GEN z = cgetg(lx + ly - 1, t);
-  i = j = k = 1;
-  while (i<lx && j<ly)
-  {
-    int s = cmpii(gel(x,i), gel(y,j));
-    if (s < 0)
-      z[k++] = x[i++];
-    else if (s > 0)
-      z[k++] = y[j++];
-    else {
-      z[k++] = x[i++];
-      j++;
-    }
-  }
-  while (i<lx) z[k++] = x[i++];
-  while (j<ly) z[k++] = y[j++];
-  setlg(z, k); return z;
-}
-
-GEN
-ZV_union_shallow(GEN x, GEN y)
-{ return ZV_union_shallow_t(x, y, t_VEC); }
-
-GEN
-ZC_union_shallow(GEN x, GEN y)
-{ return ZV_union_shallow_t(x, y, t_COL); }
 
 GEN
 setintersect(GEN x, GEN y)
