@@ -1205,7 +1205,7 @@ ellnf_goodl_l(GEN E, GEN v)
     for (j=1; j<=g; j++)
     {
       GEN prj = gel(pr, j);
-      if (idealval(nf,disc,prj) > 0) {i--; continue;}
+      if (idealval(nf,disc,prj) != 0) {i--; continue;}
       else
       {
         long t = itos(ellap(E, prj));
@@ -1295,7 +1295,8 @@ ellnf_prime_degree(GEN E)
   GEN nf = ellnf_get_nf(E);
   GEN disc = ell_get_disc(E);
   GEN P, B = gen_0, rB;
-  GEN bad = mulii(nfnorm(nf, disc),nf_get_disc(nf));
+  GEN nor = nfnorm(nf, disc);
+  GEN bad = mulii(mulii(denom_i(nor),numer_i(nor)),nf_get_disc(nf));
   u_forprime_init(&T, 5UL,ULONG_MAX);
   for(i=1; i<=20; i++)
   {
