@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -33,15 +34,12 @@ ANY WARRANTY WHATSOEVER. */
 #include "pari.h"
 #include "paripriv.h"
 
-static GEN global_modular_eqn;
 static THREAD GEN modular_eqn;
 
 void
-pari_init_seadata(void)  { global_modular_eqn = NULL; }
-void
-pari_thread_init_seadata(void)  { modular_eqn = global_modular_eqn; }
-void
-pari_pthread_init_seadata(void)  { global_modular_eqn = modular_eqn; }
+pari_set_seadata(GEN mod)  { modular_eqn = mod; }
+GEN
+pari_get_seadata(void)  { return modular_eqn; }
 
 static char *
 seadata_filename(ulong ell)
@@ -61,7 +59,7 @@ get_seadata(ulong ell)
   { /* table of polynomials of small level */
     eqn = gp_readvec_stream(F->file);
     modular_eqn = eqn = gclone(eqn);
-    avma = av;
+    set_avma(av);
   }
   pari_fclose(F);
   return eqn;
@@ -149,6 +147,7 @@ ellmodulareqn(long ell, long vx, long vy)
 
 static GEN divpol(GEN t, GEN r2, long n, void *E, const struct bb_algebra *ff);
 
+/* f_n^2, return ff->(zero|one) or a clone */
 static GEN
 divpol_f2(GEN t, GEN r2, long n, void *E, const struct bb_algebra *ff)
 {
@@ -159,6 +158,7 @@ divpol_f2(GEN t, GEN r2, long n, void *E, const struct bb_algebra *ff)
   return gmael(t,2,n);
 }
 
+/* f_n f_{n-2}, return ff->zero or a clone */
 static GEN
 divpol_ff(GEN t, GEN r2, long n, void *E, const struct bb_algebra *ff)
 {
@@ -169,44 +169,43 @@ divpol_ff(GEN t, GEN r2, long n, void *E, const struct bb_algebra *ff)
   return gmael(t,3,n);
 }
 
+/* f_n, return ff->zero or a clone */
 static GEN
 divpol(GEN t, GEN r2, long n, void *E, const struct bb_algebra *ff)
 {
   long m = n/2;
   pari_sp av = avma;
-  GEN res;
+  GEN f;
   if (n==0) return ff->zero(E);
   if (gmael(t,1,n)) return gmael(t,1,n);
   switch(n)
   {
   case 1:
   case 2:
-    res = ff->one(E);
+    f = ff->one(E);
     break;
   default:
     if (odd(n))
       if (odd(m))
-        res = ff->sub(E, ff->mul(E, divpol_ff(t,r2,m+2,E,ff),
-                                    divpol_f2(t,r2,m,E,ff)),
-                         ff->mul(E, r2,
-                                    ff->mul(E,divpol_ff(t,r2,m+1,E,ff),
-                                              divpol_f2(t,r2,m+1,E,ff))));
+        f = ff->sub(E, ff->mul(E, divpol_ff(t,r2,m+2,E,ff),
+                                  divpol_f2(t,r2,m,E,ff)),
+                       ff->mul(E, r2,
+                                  ff->mul(E,divpol_ff(t,r2,m+1,E,ff),
+                                            divpol_f2(t,r2,m+1,E,ff))));
       else
-        res = ff->sub(E, ff->mul(E, r2,
-                                    ff->mul(E, divpol_ff(t,r2,m+2,E,ff),
-                                               divpol_f2(t,r2,m,E,ff))),
-                         ff->mul(E, divpol_ff(t,r2,m+1,E,ff),
-                                    divpol_f2(t,r2,m+1,E,ff)));
-    else
-      res = ff->sub(E, ff->mul(E, divpol_ff(t,r2,m+2,E,ff),
-                                  divpol_f2(t,r2,m-1,E,ff)),
-                       ff->mul(E, divpol_ff(t,r2,m,E,ff),
+        f = ff->sub(E, ff->mul(E, r2,
+                                  ff->mul(E, divpol_ff(t,r2,m+2,E,ff),
+                                             divpol_f2(t,r2,m,E,ff))),
+                       ff->mul(E, divpol_ff(t,r2,m+1,E,ff),
                                   divpol_f2(t,r2,m+1,E,ff)));
+    else
+      f = ff->sub(E, ff->mul(E, divpol_ff(t,r2,m+2,E,ff),
+                                divpol_f2(t,r2,m-1,E,ff)),
+                     ff->mul(E, divpol_ff(t,r2,m,E,ff),
+                                divpol_f2(t,r2,m+1,E,ff)));
   }
-  res = ff->red(E, res);
-  gmael(t,1,n) = gclone(res);
-  avma = av;
-  return gmael(t,1,n);
+  gmael(t,1,n) = f = gclone( ff->red(E, f) );
+  set_avma(av); return f;
 }
 
 static void
@@ -215,9 +214,9 @@ divpol_free(GEN t)
   long i, l = lg(gel(t,1));
   for (i=1; i<l; i++)
   {
-    if (gmael(t,1,i)) gunclone(gmael(t,1,i));
-    if (gmael(t,2,i)) gunclone(gmael(t,2,i));
-    if (gmael(t,3,i)) gunclone(gmael(t,3,i));
+    guncloneNULL(gmael(t,1,i));
+    guncloneNULL(gmael(t,2,i));
+    guncloneNULL(gmael(t,3,i));
   }
 }
 
@@ -288,16 +287,14 @@ static GEN
 rhs(GEN a4, GEN a6, long v)
 {
   GEN RHS = mkpoln(4, gen_1, gen_0, a4, a6);
-  setvarn(RHS, v);
-  return RHS;
+  setvarn(RHS, v); return RHS;
 }
 
 static GEN
 Flxq_rhs(GEN a4, GEN a6, long v, long vs)
 {
   GEN RHS = mkpoln(4, pol1_Flx(vs),  pol0_Flx(vs), a4, a6);
-  setvarn(RHS, v);
-  return RHS;
+  setvarn(RHS, v); return RHS;
 }
 
 struct divpolmod_red
@@ -365,15 +362,11 @@ Fq_elldivpolmod(GEN a4, GEN a6, long n, GEN h, GEN T, GEN p)
 
 GEN
 FpXQ_elldivpol(GEN a4, GEN a6, long n, GEN T, GEN p)
-{
-  return Fq_elldivpolmod(a4,a6,n,NULL,T,p);
-}
+{ return Fq_elldivpolmod(a4,a6,n,NULL,T,p); }
 
 GEN
 Fp_elldivpol(GEN a4, GEN a6, long n, GEN p)
-{
-  return Fq_elldivpolmod(a4,a6,n,NULL,NULL,p);
-}
+{ return Fq_elldivpolmod(a4,a6,n,NULL,NULL,p); }
 
 static GEN
 Fq_ellyn(struct divpolmod_red *d, long k)
@@ -400,20 +393,14 @@ Fq_ellyn(struct divpolmod_red *d, long k)
 
 static void
 Fq_elldivpolmod_close(struct divpolmod_red *d)
-{
-  divpol_free(d->t);
-}
+{ divpol_free(d->t); }
 static GEN
 Fq_elldivpol2(GEN a4, GEN a6, GEN T, GEN p)
-{
-  return mkpoln(4, utoi(4), gen_0, Fq_mulu(a4, 4, T, p), Fq_mulu(a6, 4, T, p));
-}
+{ return mkpoln(4, utoi(4), gen_0, Fq_mulu(a4, 4, T, p), Fq_mulu(a6, 4, T, p)); }
 
 static GEN
 Fq_elldivpol2d(GEN a4, GEN T, GEN p)
-{
-  return mkpoln(3, utoi(6), gen_0, Fq_mulu(a4, 2, T, p));
-}
+{ return mkpoln(3, utoi(6), gen_0, Fq_mulu(a4, 2, T, p)); }
 
 static GEN
 FqX_numer_isog_abscissa(GEN h, GEN a4, GEN a6, GEN T, GEN p, long vx)
@@ -457,69 +444,148 @@ Zq_sqrt(GEN b, GEN T, GEN q, GEN p, long e)
 
 static GEN
 Zq_divexact(GEN a, GEN b)
-{
-  return typ(a)==t_INT ? diviiexact(a, b): ZX_Z_divexact(a, b);
-}
+{ return typ(a)==t_INT ? diviiexact(a, b): ZX_Z_divexact(a, b); }
 
 static long
 Zq_pval(GEN a, GEN p)
-{
-  return typ(a)==t_INT ? Z_pval(a, p): ZX_pval(a, p);
-}
+{ return typ(a)==t_INT ? Z_pval(a, p): ZX_pval(a, p); }
 
 static GEN
-Zq_Z_div_safe(GEN a, GEN b, GEN T, GEN q, GEN p, long e)
+Zq_divu_safe(GEN a, ulong b, GEN T, GEN q, GEN p, long e)
 {
-  long v;
-  if (e==1) return Fq_div(a, b, T, q);
-  v = Z_pvalrem(b, p, &b);
-  if (v>0)
+  long v, w;
+  if (e==1) return Fq_div(a, utoi(b), T, q);
+  v = u_pvalrem(b, p, &b);
+  if (v > 0)
   {
-    long w = Z_pval(Q_content(a), p);
-    if (v>w) pari_err_INV("Zq_div",b);
+    if (signe(a)==0) return gen_0;
+    w = Zq_pval(a, p);
+    if (v > w) return NULL;
     a = Zq_divexact(a, powiu(p,v));
   }
-  return Fq_Fp_mul(a, Fp_inv(b, q), T, q);
+  return Fq_Fp_mul(a, Fp_inv(utoi(b), q), T, q);
 }
 
-/*Gives the first precS terms of the Weierstrass series related to */
-/*E: y^2 = x^3 + a4x + a6.  Assumes (precS-2)*(2precS+3) < ULONG_MAX, i.e.
- * precS < 46342 in 32-bit machines */
 static GEN
-find_coeff(GEN a4, GEN a6, GEN T, GEN p, long precS, GEN pp, long e)
+FqX_shift(GEN P,long n)
+{ return RgX_shift_shallow(P, n); }
+
+static GEN
+FqX_mulhigh_i(GEN f, GEN g, long n, GEN T, GEN p)
+{ return FqX_shift(FqX_mul(f,g,T, p),-n); }
+
+static GEN
+FqX_mulhigh(GEN f, GEN g, long n2, long n, GEN T, GEN p)
 {
-  GEN res, den;
-  long k, h;
-  if (e > 1) { p = sqri(p); e *= 2; }
-  res = cgetg(precS+1, t_VEC);
-  den = cgetg(precS+1, t_VECSMALL);
-  if (precS == 0) return res;
-  gel(res, 1) = Fq_div(a4, stoi(-5), T, p);
-  den[1] = 0;
-  if (precS == 1) return res;
-  gel(res, 2) = Fq_div(a6, stoi(-7), T, p);
-  den[2] = 0;
-  for (k = 3; k <= precS; ++k)
+  GEN F = RgX_blocks(f, n2, 2), fl = gel(F,1), fh = gel(F,2);
+  return FqX_add(FqX_mulhigh_i(fl, g, n2, T, p), FqXn_mul(fh, g, n - n2, T, p), T, p);
+}
+
+static GEN
+FqX_invlift1(GEN Q, GEN P, long t1, long t2, GEN T, GEN p)
+{
+  GEN H = FqXn_mul(FqX_mulhigh(Q, P, t1, t2, T, p), Q, t2-t1, T, p);
+  return FqX_sub(Q, FqX_shift(H, t1), T, p);
+}
+
+static GEN
+FqX_invsqrtlift1(GEN Q, GEN P, long t1, long t2, GEN T, GEN p)
+{
+  GEN D = FqX_mulhigh(P, FqX_sqr(Q, T, p), t1, t2, T, p);
+  GEN H = FqXn_mul(Q, FqX_halve(D, T, p), t2-t1, T, p);
+  return FqX_sub(Q, FqX_shift(H, t1), T, p);
+}
+
+/* Q(x^2) = intformal(subst(x^N*P,x,x^2)) */
+static GEN
+ZqX_integ2Xn(GEN P, long N, GEN T, GEN p, GEN pp, long e)
+{
+  long d = degpol(P), v = varn(P);
+  long k;
+  GEN Q;
+  if(d==-1) return pol_0(v);
+  Q = cgetg(d+3,t_POL);
+  Q[1] = evalsigne(1) | evalvarn(v);
+  for (k = 0; k <= d; k++)
   {
-    pari_sp btop = avma;
-    GEN a = gen_0, d;
-    long v=0;
-    if (e > 1)
-      for (h = 1; h <= k-2; h++)
-        v = maxss(v, den[h]+den[k-1-h]);
-    for (h = 1; h <= k-2; h++)
-    {
-      GEN b = Fq_mul(gel(res, h), gel(res, k-1-h), T, p);
-      if (v)
-        b = Fq_Fp_mul(b, powiu(pp, v-(den[h]+den[k-1-h])), T, p);
-      a = Fq_add(a, b, T, p);
-    }
-    v += Z_pvalrem(utoi((k-2) * (2*k + 3)), pp, &d);
-    a = Zq_div(gmulgs(a, 3), d, T, p, pp, e);
-    gel(res, k) = gerepileupto(btop, a);
-    den[k] = v;
+    GEN q = Zq_divu_safe(gel(P,2+k), 2*(k+N)+1, T, p, pp, e);
+    if (!q) return NULL;
+    gel(Q, 2+k) = q;
   }
-  return mkvec2(res, den);
+  return ZXX_renormalize(Q,d+3);
+}
+
+/* solution of G*(S'^2)=(S/x)*(HoS) mod x^m */
+static GEN
+Zq_Weierstrass(GEN a4, GEN a6, GEN b4, GEN b6, long m, GEN T, GEN p, GEN pp, long n)
+{
+  pari_sp av = avma;
+  long v = 0;
+  ulong mask = quadratic_prec_mask(m);
+  GEN iGdS2 = pol_1(v);
+  GEN G = mkpoln(4, a6, a4, gen_0, gen_1);
+  GEN GdS2 = G, S = pol_x(v), sG = pol_1(v), isG = sG, dS = sG;
+  long N = 1;
+  for (;mask>1;)
+  {
+    GEN S2, HS, K, dK, E;
+    long N2 = N, d;
+    N<<=1; if (mask & 1) N--;
+    mask >>= 1;
+    d = N-N2;
+    S2 = FqX_sqr(S, T, p);
+    HS = FqX_Fq_add(FqX_Fq_mul(S, b6, T, p), b4, T, p);
+    HS = FqX_Fq_add(FqXn_mul(S2, HS, N, T, p), gen_1, T, p);
+    HS = FqXn_mul(HS, FqX_shift(S,-1), N, T, p);
+    sG  = FqXn_mul(G, isG, N2, T, p);
+    /* (HS-Gds2)/(Gds2*sG) */
+    dK = FqXn_mul(FqX_shift(FqX_sub(HS, GdS2, T, p), -N2),
+                  FqXn_mul(iGdS2, isG, d, T, p), d, T, p);
+    K = ZqX_integ2Xn(dK, N2, T, p, pp, n);
+    if (!K) return gc_NULL(av);
+    E = FqXn_mul(FqXn_mul(K, sG, d, T, p), dS, d, T, p);
+    S = FqX_add(S, FqX_shift(E, N2+1), T, p);
+    if (mask <= 1) break;
+    isG = FqX_invsqrtlift1(isG, G, N2, N, T, p);
+    dS = FqX_deriv(S, T, p);
+    GdS2 = FqX_mul(G, FqX_sqr(dS, T, p), T, p);
+    iGdS2 = FqX_invlift1(iGdS2, GdS2, N2, N, T, p);
+  }
+  return gerepileupto(av, S);
+}
+
+static GEN
+ZqXn_WNewton(GEN S, long l, GEN a4, GEN a6, GEN pp1, GEN T, GEN p, GEN pp, long e)
+{
+  long d = degpol(S);
+  long k;
+  GEN Ge = cgetg(2+d,t_POL);
+  Ge[1] = evalsigne(1);
+  gel(Ge,2) = pp1;
+  if (d >= 2)
+  {
+    GEN g = Zq_divu_safe(Fq_sub(gel(S,4), Fq_mulu(a4,(l-1),T,p),T,p), 6,T,p,pp,e);
+    if (!g) return NULL;
+    gel(Ge, 3) = g;
+  }
+  if (d >= 3)
+  {
+    GEN g = Zq_divu_safe(Fq_sub(Fq_sub(gel(S,5),
+            Fq_mul(a4,Fq_mulu(pp1,6,T,p),T,p),T,p),
+            Fq_mulu(a6,(l-1)*2,T,p),T,p),10,T,p,pp,e);
+    if (!g) return NULL;
+    gel(Ge, 4) = g;
+  }
+  for (k = 4; k <= d; k++)
+  {
+    GEN g = Zq_divu_safe(Fq_sub(Fq_sub(gel(S,4+k-2),
+            Fq_mul(a4,Fq_mulu(gel(Ge,k-1),4*k-6,T,p),T,p),T,p),
+            Fq_mul(a6,Fq_mulu(gel(Ge,k-2),4*k-8,T,p),T,p),T,p),
+            4*k-2, T, p, pp, e);
+    if (!g) return NULL;
+    gel(Ge, k+1) = g;
+  }
+  return ZXX_renormalize(Ge, 2+d);
 }
 
 /****************************************************************************/
@@ -551,9 +617,7 @@ Zq_ellj(GEN a4, GEN a6, GEN T, GEN p, GEN pp, long e)
 
 static GEN
 Fq_to_Flx(GEN a4, GEN T, ulong p)
-{
-  return typ(a4)==t_INT ? Z_to_Flx(a4, p, get_Flx_var(T)): ZX_to_Flx(a4, p);
-}
+{ return typ(a4)==t_INT ? Z_to_Flx(a4, p, get_Flx_var(T)): ZX_to_Flx(a4, p); }
 
 static GEN
 Flxq_find_eigen_Frobenius(GEN a4, GEN a6, GEN h, GEN T, ulong p)
@@ -590,7 +654,7 @@ find_eigen_value_oneroot(GEN a4, GEN a6, ulong ell, GEN tr, GEN h, GEN T, GEN p)
   Dy = FqXQ_mul(Gy, gel(f,2), h, T, p);
   if (!gequal(gel(f,1), Dy)) t = ell-t;
   Fq_elldivpolmod_close(&d);
-  avma = ltop; return t;
+  return gc_ulong(ltop, t);
 }
 
 static ulong
@@ -617,7 +681,7 @@ Flxq_find_eigen_value_power(GEN a4, GEN a6, ulong ell, long k, ulong lambda,
   }
   if (DEBUGLEVEL>2) err_printf(" (%ld ms)",timer_delay(&ti));
   Fq_elldivpolmod_close(&d);
-  avma = ltop; return t;
+  return gc_ulong(ltop, t);
 }
 
 /*Finds the eigenvalue of the Frobenius modulo ell^k given E, ell, k, h factor
@@ -645,7 +709,7 @@ Fq_find_eigen_value_power(GEN a4, GEN a6, ulong ell, long k, ulong lambda, GEN h
   }
   if (DEBUGLEVEL>2) err_printf(" (%ld ms)",timer_delay(&ti));
   Fq_elldivpolmod_close(&d);
-  avma = ltop; return t;
+  return gc_ulong(ltop, t);
 }
 
 static ulong
@@ -663,69 +727,26 @@ find_eigen_value_power(GEN a4, GEN a6, ulong ell, long k, ulong lambda, GEN hq,
   return Fq_find_eigen_value_power(a4, a6, ell, k, lambda, hq, T, p);
 }
 
-/*Finds the kernel polynomial h, dividing the ell-division polynomial from the
-  isogenous curve Eb and trace term pp1. Uses CCR algorithm and returns h.
-  Return NULL if E and Eb are *not* isogenous. */
 static GEN
-find_kernel(GEN a4, GEN a6, ulong ell, GEN a4t, GEN a6t, GEN pp1, GEN T, GEN p, GEN pp, long e)
+find_kernel(GEN a4, GEN a6, long l, GEN b4, GEN b6, GEN pp1, GEN T, GEN p, GEN pp, long e)
 {
-  const long ext = 2;
-  pari_sp ltop = avma, btop;
-  GEN P, v, tlist, h;
-  long i, j, k;
-  long deg = (ell - 1)/2, dim = 2 + deg + ext;
-  GEN psi2 = Fq_elldivpol2(a4, a6, T, p);
-  GEN Dpsi2 = Fq_elldivpol2d(a4, T, p);
-  GEN C  = find_coeff(a4, a6, T, p, dim, pp, e);
-  GEN Ct = find_coeff(a4t, a6t, T, p, dim, pp, e);
-  GEN V = cgetg(dim+1, t_VEC);
-  for (k = 1; k <= dim; k++)
-  {
-    long v = mael(C,2,k);
-    GEN z = gmul(gsub(gmael(Ct,1,k), gmael(C,1,k)), shifti(mpfact(2*k), -1));
-    if (signe(z) && Zq_pval(z, pp) < v) return NULL;
-    gel(V, k) = Zq_divexact(z, powiu(pp, v));
-  }
-  btop = avma;
-  v = zerovec(dim);
-  gel(v, 1) = utoi(deg);
-  gel(v, 2) = pp1;
-  P = pol_x(0);
-  for (k = 3; k <= dim; k++)
-  {
-    GEN s, r = FqX_Fq_mul(Dpsi2, gel(P, 3), T, p);
-    for (j = 4; j < lg(P); j++)
-    {
-      long o = j - 2;
-      GEN D = FqX_add(RgX_shift_shallow(Dpsi2, 1), FqX_mulu(psi2, o-1, T, p), T, p);
-      GEN E = FqX_Fq_mul(D, Fq_mulu(gel(P, j), o, T, p), T, p);
-      r = FqX_add(r, RgX_shift_shallow(E, o-2), T, p);
-    }
-    P = r;
-    s = Fq_mul(gel(P, 2), gel(v, 1), T, p);
-    for (j = 3; j < lg(P)-1; j++)
-      s = Fq_add(s, Fq_mul(gel(P, j), gel(v, j-1), T, p), T, p);
-    gel(v, k) = Zq_Z_div_safe(Fq_sub(gel(V, k-2), s, T, p), gel(P, j), T, p, pp, e);
-    if (gc_needed(btop, 1))
-    {
-      if(DEBUGMEM>1) pari_warn(warnmem,"find_kernel");
-      gerepileall(btop, 2, &v, &P);
-    }
-  }
-  tlist = cgetg(dim, t_VEC);
-  gel(tlist, dim-1) = gen_1;
-  for (k = 1; k <= dim-2; k++)
-  {
-    pari_sp btop = avma;
-    GEN s = gel(v, k+1);
-    for (i = 1; i < k; i++)
-      s = Fq_add(s, Fq_mul(gel(tlist, dim-i-1), gel(v, k-i+1), T, p), T, p);
-    gel(tlist, dim-k-1) = gerepileupto(btop, Zq_Z_div_safe(s, stoi(-k), T, p, pp, e));
-  }
-  for (i = 1; i <= ext; i++)
-    if (signe(Fq_red(gel(tlist, i),T, pp))) { avma = ltop; return NULL; }
-  h = FqX_red(RgV_to_RgX(vecslice(tlist, ext+1, dim-1), 0),T,p);
-  return signe(Fq_elldivpolmod(a4, a6, ell, h, T, pp)) ? NULL: h;
+  GEN Ge, S, Sd;
+  const long ext = 1;
+  long d = (l+1)/2+ext;
+  if(l==3)
+    return deg1pol(gen_1, Fq_neg(pp1, T, p), 0);
+  S = Zq_Weierstrass(a4, a6, b4, b6, d + 1, T, p, pp, e);
+  if (S==NULL) return NULL;
+  S  = FqX_shift(S, -1);
+  Sd = FqXn_inv(S, d, T, p);
+  Ge = ZqXn_WNewton(Sd, l, a4, a6, pp1, T, p, pp, e);
+  if (!Ge) return NULL;
+  Ge = FqX_neg(Ge, T, p);
+  Ge = T && lgefint(pp)==3 ? ZlXQXn_expint(Ge, d, T, p, pp[2])
+                           : FqXn_expint(Ge, d, T, p);
+  Ge = RgX_recip(FqX_red(Ge, T, pp));
+  if (degpol(Ge)==(l-1)/2) return Ge;
+  return NULL;
 }
 
 static GEN
@@ -746,6 +767,21 @@ compute_u(GEN gprime, GEN Dxxg, GEN DxJg, GEN DJJg, GEN j, GEN pJ, GEN px, ulong
   return gerepileupto(ltop, Fq_add(Zq_div(g,px,T,p,pp,e), Fq_mulu(f,q,T,p), T, p));
 }
 
+static void
+a4a6t(GEN *a4t, GEN *a6t, ulong l, GEN E4t, GEN E6t, GEN T, GEN p)
+{
+  GEN l2 = modii(sqru(l), p), l4 = Fp_sqr(l2, p), l6 = Fp_mul(l4, l2, p);
+  *a4t = Fq_mul(E4t, Fp_muls(l4, -3, p), T, p);
+  *a6t = Fq_mul(E6t, Fp_muls(l6, -2, p), T, p);
+}
+static void
+a4a6t_from_J(GEN *a4t, GEN *a6t, ulong l, GEN C4t, GEN C6t, GEN T, GEN p)
+{
+  GEN l2 = modii(sqru(l), p), l4 = Fp_sqr(l2, p), l6 = Fp_mul(l4, l2, p);
+  GEN v = Fp_inv(stoi(-864), p), u = Fp_mulu(v, 18, p);
+  *a4t = Fq_mul(C4t, Fp_mul(u, l4, p), T, p);
+  *a6t = Fq_mul(C6t, Fp_mul(v, l6, p), T, p);
+}
 /* Finds the isogenous EC, and the sum of the x-coordinates of the points in
  * the kernel of the isogeny E -> Eb
  * E: elliptic curve, ell: a prime, meqn: Atkin modular equation
@@ -775,7 +811,7 @@ find_isogenous_from_Atkin(GEN a4, GEN a6, ulong ell, struct meqn *MEQN, GEN g, G
   if (!signe(Fq_red(dJ,T,pp)) || !signe(Fq_red(dx,T,pp)))
   {
     if (DEBUGLEVEL>0) err_printf("[A: d%c=0]",signe(dJ)? 'x': 'J');
-    avma = ltop; return NULL;
+    return gc_NULL(ltop);
   }
   a = Fq_mul(dJ, Fq_mul(g, E6, T, p), T, p);
   b = Fq_mul(E4, dx, T, p);
@@ -787,7 +823,7 @@ find_isogenous_from_Atkin(GEN a4, GEN a6, ulong ell, struct meqn *MEQN, GEN g, G
   Roots = FqX_roots(meqnx, T, pp);
 
   btop = avma;
-  for (k = lg(Roots)-1; k >= 1; k--, avma = btop)
+  for (k = lg(Roots)-1; k >= 1; k--, set_avma(btop))
   {
     GEN jt = gel(Roots, k);
     if (signe(FqX_eval(Dmeqnx, jt, T, pp))==0)
@@ -797,7 +833,7 @@ find_isogenous_from_Atkin(GEN a4, GEN a6, ulong ell, struct meqn *MEQN, GEN g, G
     if (signe(Fq_red(jt, T, pp)) == 0 || signe(Fq_sub(jt, utoi(1728), T, pp)) == 0)
     {
       if (DEBUGLEVEL>0) err_printf("[A: jt=%ld]",signe(Fq_red(jt,T,p))? 1728: 0);
-      avma = ltop; return NULL;
+      return gc_NULL(ltop);
     }
     else
     {
@@ -811,9 +847,9 @@ find_isogenous_from_Atkin(GEN a4, GEN a6, ulong ell, struct meqn *MEQN, GEN g, G
       GEN E6t = Zq_div(Fq_mul(u, E4t, T, p), v, T, p, pp, e);
       GEN u2 = compute_u(gprime, Dxxg, DxJg, DJJg, jt, pJstar, pxstar, ell, E4t, E6t, T, p, pp, e);
       GEN pp1 = Fq_mulu(Fq_sub(u1, u2, T, p), 3*ell, T, p);
-      GEN a4t = Fq_mul(mulsi(-3, powuu(ell,4)), E4t, T, p);
-      GEN a6t = Fq_mul(mulsi(-2, powuu(ell,6)), E6t, T, p);
-      GEN h = find_kernel(a4, a6, ell, a4t, a6t, pp1, T, p, pp, e);
+      GEN a4t, a6t, h;
+      a4a6t(&a4t, &a6t, ell, E4t, E6t, T, p);
+      h = find_kernel(a4, a6, ell, a4t, a6t, pp1, T, p, pp, e);
       if (h) return gerepilecopy(ltop, mkvec3(a4t, a6t, h));
     }
   }
@@ -853,11 +889,11 @@ find_isogenous_from_canonical(GEN a4, GEN a6, ulong ell, struct meqn *MEQN, GEN
   ulong tis = ugcd(12, ell-1), is = 12 / tis;
   GEN itis = Fq_inv(stoi(-tis), T, p);
   GEN deltal = Fq_div(Fq_mul(delta, Fq_powu(g, tis, T, p), T, p), powuu(ell, 12), T, p);
-  GEN E4l, E6l, a4tilde, a6tilde, p_1;
+  GEN E4l, E6l, a4t, a6t, p_1;
   if (signe(Fq_red(dx,T, pp))==0)
   {
     if (DEBUGLEVEL>0) err_printf("[C: dx=0]");
-    avma = ltop; return NULL;
+    return gc_NULL(ltop);
   }
   if (signe(Fq_red(dJ, T, pp))==0)
   {
@@ -865,13 +901,15 @@ find_isogenous_from_canonical(GEN a4, GEN a6, ulong ell, struct meqn *MEQN, GEN
     if (DEBUGLEVEL>0) err_printf("[C: dJ=0]");
     E4l = Fq_div(E4, sqru(ell), T, p);
     jl  = Zq_div(Fq_powu(E4l, 3, T, p), deltal, T, p, pp, e);
-    E6l = Zq_sqrt(Fq_mul(Fq_sub(jl, utoi(1728), T, p), deltal, T, p), T, p, pp, e);
+    E6l = Zq_sqrt(Fq_mul(Fq_sub(jl, utoi(1728), T, p),
+                         deltal, T, p), T, p, pp, e);
     p_1 = gen_0;
   }
   else
   {
     GEN jl, f, fd, Dgs, Djs, jld;
-    GEN E2s = Zq_div(Fq_mul(Fq_neg(Fq_mulu(E6, 12, T, p), T, p), dJ, T, p), Fq_mul(Fq_mulu(E4, is, T, p), dx, T, p), T, p, pp, e);
+    GEN E2s = Zq_div(Fq_mul(Fq_neg(Fq_mulu(E6, 12, T, p), T, p), dJ, T, p),
+                     Fq_mul(Fq_mulu(E4, is, T, p), dx, T, p), T, p, pp, e);
     GEN gd = Fq_mul(Fq_mul(E2s, itis, T, p), g, T, p);
     GEN jd = Zq_div(Fq_mul(Fq_neg(E42, T, p), E6, T, p), delta, T, p, pp, e);
     GEN E0b = Zq_div(E6, Fq_mul(E4, E2s, T, p), T, p, pp, e);
@@ -880,26 +918,26 @@ find_isogenous_from_canonical(GEN a4, GEN a6, ulong ell, struct meqn *MEQN, GEN
     GEN DJgJj = FqX_eval(FqX_deriv(DJg, T, p), j, T, p);
     GEN Djd = Fq_add(Fq_mul(jd, pJ, T, p), Fq_mul(j, Fq_add(Fq_mul(jd, DJgJj, T, p), Fq_mul(gd, ExJ, T, p), T, p), T, p), T, p);
     GEN E0bd = Zq_div(Fq_sub(Fq_mul(Dgd, itis, T, p), Fq_mul(E0b, Djd, T, p), T, p), dJ, T, p, pp, e);
-    E4l = Zq_div(Fq_sub(E4, Fq_mul(E2s, Fq_sub(Fq_sub(Fq_add(Zq_div(Fq_mulu(E0bd, 12, T, p), E0b, T, p, pp, e), Zq_div(Fq_mulu(E42, 6, T, p), E6, T, p, pp, e), T, p), Zq_div(Fq_mulu(E6, 4, T, p), E4, T, p, pp, e), T, p), E2s, T, p), T, p), T, p), sqru(ell), T, p, pp, e);
+    E4l = Fq_div(Fq_sub(E4, Fq_mul(E2s, Fq_sub(Fq_sub(Fq_add(Zq_div(Fq_mulu(E0bd, 12, T, p), E0b, T, p, pp, e), Zq_div(Fq_mulu(E42, 6, T, p), E6, T, p, pp, e), T, p), Zq_div(Fq_mulu(E6, 4, T, p), E4, T, p, pp, e), T, p), E2s, T, p), T, p), T, p), sqru(ell), T, p);
     jl = Zq_div(Fq_powu(E4l, 3, T, p), deltal, T, p, pp, e);
     if (signe(Fq_red(jl,T,pp))==0)
     {
       if (DEBUGLEVEL>0) err_printf("[C: jl=0]");
-      avma = ltop; return NULL;
+      return gc_NULL(ltop);
     }
     f =  Zq_div(powuu(ell, is), g, T, p, pp, e);
     fd = Fq_neg(Fq_mul(Fq_mul(E2s, f, T, p), itis, T, p), T, p);
     Dgs = FqXY_eval(Dx, f, jl, T, p);
     Djs = FqXY_eval(DJ, f, jl, T, p);
-    jld = Zq_div(Fq_mul(Fq_neg(fd, T, p), Dgs, T, p), Fq_mulu(Djs, ell, T, p), T, p, pp, e);
+    jld = Zq_div(Fq_mul(Fq_neg(fd, T, p), Dgs, T, p),
+                 Fq_mulu(Djs, ell, T, p), T, p, pp, e);
     E6l = Zq_div(Fq_mul(Fq_neg(E4l, T, p), jld, T, p), jl, T, p, pp, e);
     p_1 = Fq_neg(Fq_halve(Fq_mulu(E2s, ell, T, p), T, p),T,p);
   }
-  a4tilde = Fq_mul(Fq_mul(stoi(-3), powuu(ell,4), T, p), E4l, T, p);
-  a6tilde = Fq_mul(Fq_mul(stoi(-2), powuu(ell,6), T, p), E6l, T, p);
-  h = find_kernel(a4, a6, ell, a4tilde, a6tilde, p_1, T, p, pp, e);
+  a4a6t(&a4t, &a6t, ell, E4l, E6l, T, p);
+  h = find_kernel(a4, a6, ell, a4t, a6t, p_1, T, p, pp, e);
   if (!h) return NULL;
-  return gerepilecopy(ltop, mkvec3(a4tilde, a6tilde, h));
+  return gerepilecopy(ltop, mkvec3(a4t, a6t, h));
 }
 
 static GEN
@@ -907,7 +945,7 @@ corr(GEN c4, GEN c6, GEN T, GEN p, GEN pp, long e)
 {
   GEN c46 = Zq_div(Fq_sqr(c4, T, p), c6, T, p, pp, e);
   GEN c64 = Zq_div(c6, c4, T, p, pp, e);
-  GEN a = Fp_div(gen_2, utoi(3), p);
+  GEN a = Fp_divu(gen_2, 3, p);
   return Fq_add(Fq_halve(c46, T, p), Fq_mul(a, c64, T, p), T, p);
 }
 
@@ -1010,23 +1048,23 @@ find_isogenous_from_J(GEN a4, GEN a6, ulong ell, struct meqn *MEQN, GEN g, GEN T
   pari_sp ltop = avma;
   GEN meqn = MEQN->eval;
   GEN p = e==1 ? pp: powiu(pp, e);
-  GEN h;
+  GEN h, a4t, a6t;
   GEN C4, C6, C4t, C6t;
   GEN j, jp, jtp, jtp2, jtp3;
   GEN Py, Pxy, Pyy, Pxj, Pyj, Pxxj, Pxyj, Pyyj;
   GEN s0, s1, s2, s3;
-  GEN den, D, co, cot, c0, p_1, a4tilde, a6tilde;
+  GEN den, D, co, cot, c0, p_1;
   if (signe(g) == 0 || signe(Fq_sub(g, utoi(1728), T, p)) == 0)
   {
     if (DEBUGLEVEL>0) err_printf("[J: g=%ld]",signe(g)==0 ?0: 1728);
-    avma = ltop; return NULL;
+    return gc_NULL(ltop);
   }
   C4 = Fq_mul(a4, stoi(-48), T, p);
   C6 = Fq_mul(a6, stoi(-864), T, p);
   if (signe(C4)==0 || signe(C6)==0)
   {
     if (DEBUGLEVEL>0) err_printf("[J: C%ld=0]",signe(C4)==0 ?4: 6);
-    avma = ltop; return NULL;
+    return gc_NULL(ltop);
   }
   j = Zq_ellj(a4, a6, T, p, pp, e);
   jp = Fq_mul(j, Zq_div(C6, C4, T, p, pp, e), T, p);
@@ -1038,13 +1076,14 @@ find_isogenous_from_J(GEN a4, GEN a6, ulong ell, struct meqn *MEQN, GEN g, GEN T
   if (signe(Pxj)==0)
   {
     if (DEBUGLEVEL>0) err_printf("[J: Pxj=0]");
-    avma = ltop; return NULL;
+    return gc_NULL(ltop);
   }
   Pyj = FqX_eval(Py, g, T, p);
   Pxxj = FqX_eval(gel(meqn, 3), g, T, p);
   Pxyj = FqX_eval(Pxy, g, T, p);
   Pyyj = FqX_eval(Pyy, g, T, p);
-  jtp = Fq_div(Fq_mul(jp, Zq_div(Pxj, Pyj, T, p, pp, e), T, p), negi(utoi(ell)), T, p);
+  jtp = Fq_div(Fq_mul(jp, Zq_div(Pxj, Pyj, T, p, pp, e), T, p),
+               utoineg(ell), T, p);
   jtp2 = Fq_sqr(jtp,T,p);
   jtp3 = Fq_mul(jtp,jtp2,T,p);
   den = Fq_mul(Fq_sqr(g,T,p),Fq_sub(g,utoi(1728),T,p),T, p);
@@ -1058,18 +1097,29 @@ find_isogenous_from_J(GEN a4, GEN a6, ulong ell, struct meqn *MEQN, GEN g, GEN T
   cot = corr(C4t, C6t, T, p, pp, e);
   c0 = Fq_sub(co,Fq_mulu(cot,ell,T,p),T,p);
   p_1 = Fq_div(Fq_mulu(Fq_add(s3, c0, T, p),ell,T,p),stoi(-4),T,p);
-  a4tilde = Fq_mul(Fq_div(C4t, stoi(-48), T, p),powuu(ell,4), T, p);
-  a6tilde = Fq_mul(Fq_div(C6t, stoi(-864), T, p),powuu(ell,6), T, p);
-  h = find_kernel(a4, a6, ell, a4tilde, a6tilde, p_1, T, p, pp, e);
+  a4a6t_from_J(&a4t, &a6t, ell, C4t, C6t, T, p);
+  h = find_kernel(a4, a6, ell, a4t, a6t, p_1, T, p, pp, e);
   if (!h) return NULL;
-  return gerepilecopy(ltop, mkvec3(a4tilde, a6tilde, h));
+  return gerepilecopy(ltop, mkvec3(a4t, a6t, h));
+}
+
+static long
+newtonlogint(ulong n, ulong pp)
+{
+  long s = 0;
+  while (n > pp)
+  {
+    s += ulogint(n, pp);
+    n = (n+1)>>1;
+  }
+  return s;
 }
 
 static GEN
 find_isogenous(GEN a4,GEN a6, ulong ell, struct meqn *MEQN, GEN g, GEN T,GEN p)
 {
   ulong pp = itou_or_0(p);
-  long e = (pp && pp <= 2*ell+3) ? 2+factorial_lval(ell, pp): 1;
+  long e = pp ? newtonlogint(1+(ell>>1), pp) + ulogint(2*ell+4, pp) + 1: 1;
   if (signe(a4)==0 || signe(a6)==0)
   {
     if (DEBUGLEVEL>0) err_printf("[%c: j=%ld]",MEQN->type,signe(a4)==0 ?0: 1728);
@@ -1131,7 +1181,7 @@ find_kernel_power(GEN Eba4, GEN Eba6, GEN Eca4, GEN Eca6, ulong ell, struct meqn
   {
     GEN h;
     GEN tmp = find_isogenous(Eca4, Eca6, ell, MEQN, gel(mroots, i), T, p);
-    if (!tmp) { avma = ltop; return NULL; }
+    if (!tmp) return gc_NULL(ltop);
     a4t =  gel(tmp, 1);
     a6t =  gel(tmp, 2);
     gtmp = gel(tmp, 3);
@@ -1144,9 +1194,9 @@ find_kernel_power(GEN Eba4, GEN Eba6, GEN Eca4, GEN Eca6, ulong ell, struct meqn
       GEN kpoly_new = FqX_homogenous_eval(gtmp,   numer_i(Ic),denom_i(Ic), T,p);
       return gerepilecopy(ltop, mkvecn(5, a4t, a6t, kpoly_new, gtmp, Ic));
     }
-    avma = btop;
+    set_avma(btop);
   }
-  avma = ltop; return NULL;
+  return gc_NULL(ltop);
 }
 
 /****************************************************************************/
@@ -1243,7 +1293,7 @@ find_trace_Elkies_power(GEN a4, GEN a6, ulong ell, long *pt_k, struct meqn *MEQN
   Eba4 = a4;
   Eba6 = a6;
   tmp = find_isogenous(a4,a6, ell, MEQN, g, T, p);
-  if (!tmp) { avma = ltop; return NULL; }
+  if (!tmp) return gc_NULL(ltop);
   Eca4 =  gel(tmp, 1);
   Eca6 =  gel(tmp, 2);
   kpoly = gel(tmp, 3);
@@ -1255,8 +1305,8 @@ find_trace_Elkies_power(GEN a4, GEN a6, ulong ell, long *pt_k, struct meqn *MEQN
   {
     ulong pell = pellk%ell;
     ulong ap = Fl_add(lambda, Fl_div(pell, lambda, ell), ell);
-    if (Fl_sub(pell, ap, ell)==ell-1) { avma = ltop; return mkvecsmall(ap); }
-    if (smallfact < 0 && Fl_add(pell, ap, ell)==ell-1) { avma = ltop; return mkvecsmall(ap); }
+    if (Fl_sub(pell, ap, ell)==ell-1) { set_avma(ltop); return mkvecsmall(ap); }
+    if (smallfact < 0 && Fl_add(pell, ap, ell)==ell-1) { set_avma(ltop); return mkvecsmall(ap); }
   }
   btop = avma;
   for (cnt = 2; cnt <= k; cnt++)
@@ -1278,7 +1328,7 @@ find_trace_Elkies_power(GEN a4, GEN a6, ulong ell, long *pt_k, struct meqn *MEQN
     }
     if (DEBUGLEVEL>1) err_printf(" [%ld ms]", timer_delay(ti));
   }
-  avma = ltop;
+  set_avma(ltop);
   ellk = upowuu(ell, k);
   pellk = umodiu(q, ellk);
   *pt_k = k;
@@ -1299,7 +1349,7 @@ find_trace_Atkin(ulong ell, long r, GEN q)
   pari_sp btop = avma;
   if (r==2 && krouu(ell-pell, ell) < 0)
     val_pos[++nval] = 0;
-  for (teta = 1; teta < ell; teta++, avma = btop)
+  for (teta = 1; teta < ell; teta++, set_avma(btop))
   {
     ulong disc = Fl_sub(Fl_sqr(teta,ell), Fl_mul(4UL,pell,ell), ell);
     GEN a;
@@ -1311,7 +1361,7 @@ find_trace_Atkin(ulong ell, long r, GEN q)
     {
       pari_sp av = avma;
       long i, l=lg(P);
-      for (i = 2; i < l; i++, avma = av)
+      for (i = 2; i < l; i++, set_avma(av))
         if (Flx_equal1(Flxq_powu(U, r/P[i], S, ell))) break;
       if (i==l) val_pos[++nval] = teta;
     }
@@ -1372,7 +1422,7 @@ find_trace(GEN a4, GEN a6, GEN j, ulong ell, GEN q, GEN T, GEN p, long *ptr_kt,
     if (!tr)
     {
       if (DEBUGLEVEL) err_printf("[fail]\n");
-      avma = ltop; return NULL;
+      return gc_NULL(ltop);
     }
     break;
   case MTroots:
@@ -1385,7 +1435,7 @@ find_trace(GEN a4, GEN a6, GEN j, ulong ell, GEN q, GEN T, GEN p, long *ptr_kt,
     kt = 1;
     break;
   default: /* case MTpathological: */
-    avma = ltop; return NULL;
+    return gc_NULL(ltop);
   }
   if (DEBUGLEVEL) {
     long n = lg(tr)-1;
@@ -1448,7 +1498,7 @@ multiple_crt(GEN x, GEN y, GEN q, GEN P)
   a1 = mulii(P,u);
   a2 = mulii(q,v); A2X = ZC_Z_mul(x, a2);
   av = avma; affii(mulii(P,q), P);
-  for (i = 1, k = 1; i <= lx; i++, avma = av)
+  for (i = 1, k = 1; i <= lx; i++, set_avma(av))
   {
     GEN a2x = gel(A2X,i);
     for (j = 1; j <= ly; ++j)
@@ -1457,7 +1507,7 @@ multiple_crt(GEN x, GEN y, GEN q, GEN P)
       affii(t, gel(x, k++));
     }
   }
-  setlg(x, k); avma = ltop;
+  setlg(x, k); set_avma(ltop);
 }
 
 /****************************************************************************/
@@ -1534,7 +1584,7 @@ set_cost(GEN B, long b, GEN cost_vec, long *pi)
   long i = *pi;
   while (cmpii(costb, cost(B[i], cost_vec)) < 0) --i;
   B[++i] = b;
-  *pi = i; avma = av;
+  *pi = i; set_avma(av);
 }
 
 static GEN
@@ -1570,7 +1620,7 @@ champion(GEN atkin, long k, GEN bound_champ)
       long b1 = Bp[i1], b2 = Bp[i2]|(1L<<(j-1));
       if (cmpii(value(b1, atkin, k), value(b2, atkin, k)) < 0)
         { b = b1; i1++; } else { b = b2; i2++; }
-      avma = av;
+      set_avma(av);
       set_cost(B, b, cost_vec, &i);
     }
     for ( ; i2 <= n; i2++)
@@ -1682,7 +1732,7 @@ match_and_sort(GEN compile_atkin, GEN Mu, GEN u, GEN q, void *E, const struct bb
   if (DEBUGLEVEL>=2) timer_start(&ti);
   av1 = avma;
   best_i = separation( get_lgatkin(compile_atkin, k) );
-  avma = av1;
+  set_avma(av1);
 
   baby  = possible_traces(compile_atkin, utoi(best_i), &Mb, 1);
   giant = possible_traces(compile_atkin, subiu(int2n(k), best_i+1), &Mg, 0);
@@ -1690,7 +1740,7 @@ match_and_sort(GEN compile_atkin, GEN Mu, GEN u, GEN q, void *E, const struct bb
   lgiant = lg(giant);
   den = Fp_inv(Fp_mul(Mu, Mb, Mg), Mg);
   av2 = avma;
-  for (i = 1; i < lgiant; i++, avma = av2)
+  for (i = 1; i < lgiant; i++, set_avma(av2))
     affii(Fp_mul(gel(giant,i), den, Mg), gel(giant,i));
   ZV_sort_inplace(giant);
   Sg = Fp_mul(negi(u), den, Mg);
@@ -1699,7 +1749,7 @@ match_and_sort(GEN compile_atkin, GEN Mu, GEN u, GEN q, void *E, const struct bb
   togglesign(dec_inf); /* now, dec_inf = ceil(- (Mb/2 + Sg Mb/Mg) ) */
   div = mulii(truedivii(dec_inf, Mb), Mb);
   av2 = avma;
-  for (i = 1; i < lbaby; i++, avma = av2)
+  for (i = 1; i < lbaby; i++, set_avma(av2))
   {
     GEN b = addii(Fp_mul(Fp_sub(gel(baby,i), u, Mb), den, Mb), div);
     if (cmpii(b, dec_inf) < 0) b = addii(b, Mb);
@@ -1713,7 +1763,7 @@ match_and_sort(GEN compile_atkin, GEN Mu, GEN u, GEN q, void *E, const struct bb
 
   av2 = avma;
 MATCH_RESTART:
-  avma = av2;
+  set_avma(av2);
   nbcard = 0;
   P = grp->rand(E);
   point = grp->pow(E,P, Mu);
@@ -1740,7 +1790,7 @@ MATCH_RESTART:
       point = gerepileupto(av1, point);
     }
   }
-  avma = av1;
+  set_avma(av1);
   /* Precomputations for giants */
   pre = BSGS_pre(&diff, giant, Pg, E, grp);
 
@@ -1859,7 +1909,7 @@ Fq_ellcard_SEA(GEN a4, GEN a6, GEN q, GEN T, GEN p, long smallfact)
   if (odd(smallfact) && !mpodd(TR))
   {
     if (DEBUGLEVEL) err_printf("Aborting: #E(Fq) divisible by 2\n");
-    avma = ltop; return gen_0;
+    set_avma(ltop); return gen_0;
   }
   vy = fetch_var();
   vx = fetch_var_higher();
@@ -1898,7 +1948,7 @@ Fq_ellcard_SEA(GEN a4, GEN a6, GEN q, GEN T, GEN p, long smallfact)
                        tcard_mod_ell ? "" : "_twist", ell);
           delete_var();
           delete_var();
-          avma = ltop; return gen_0;
+          set_avma(ltop); return gen_0;
         }
       }
       (void)Z_incremental_CRT(&TR, t_mod_ellkt, &TR_mod, ellkt);
@@ -1957,6 +2007,4 @@ Fq_ellcard_SEA(GEN a4, GEN a6, GEN q, GEN T, GEN p, long smallfact)
 
 GEN
 Fp_ellcard_SEA(GEN a4, GEN a6, GEN p, long smallfact)
-{
-  return Fq_ellcard_SEA(a4, a6, p, NULL, p, smallfact);
-}
+{ return Fq_ellcard_SEA(a4, a6, p, NULL, p, smallfact); }
