@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -39,18 +40,13 @@ warn_coercion(GEN x, GEN y, GEN z)
 
 static long
 kro_quad(GEN x, GEN y)
-{
-  pari_sp av=avma;
-  long k = kronecker(quad_disc(x), y);
-  avma = av; return k;
-}
+{ pari_sp av=avma; return gc_long(av, kronecker(quad_disc(x), y)); }
 
 /* is -1 not a square in Zp, assume p prime */
 INLINE int
 Zp_nosquare_m1(GEN p) { return (mod4(p) & 2); /* 2 or 3 mod 4 */ }
 
 static GEN addsub_pp(GEN x, GEN y, GEN(*op)(GEN,GEN));
-static GEN addsub_frac(GEN x, GEN y, GEN (*op)(GEN,GEN));
 static GEN mulpp(GEN x, GEN y);
 static GEN divpp(GEN x, GEN y);
 /* Argument codes for inline routines
@@ -114,15 +110,15 @@ addqf(GEN x, GEN y, long prec) { pari_sp av = avma;
 static GEN
 mulrfrac(GEN x, GEN y)
 {
-  pari_sp av = avma;
+  pari_sp av;
   GEN z, a = gel(y,1), b = gel(y,2);
   if (is_pm1(a)) /* frequent special case */
   {
-    z = divri(x, b);
-    if (signe(a) < 0) togglesign(z);
+    z = divri(x, b); if (signe(a) < 0) togglesign(z);
     return z;
   }
-  return gerepileuptoleaf(av, divri(mulri(x,gel(y,1)), gel(y,2)));
+  av = avma;
+  return gerepileuptoleaf(av, divri(mulri(x,a), b));
 }
 static GEN
 mulqf(GEN x, GEN y, long prec) { pari_sp av = avma;
@@ -170,7 +166,7 @@ static GEN
 add_intmod_same(GEN z, GEN X, GEN x, GEN y) {
   if (lgefint(X) == 3) {
     ulong u = Fl_add(itou(x),itou(y), X[2]);
-    avma = (pari_sp)z; gel(z,2) = utoi(u);
+    set_avma((pari_sp)z); gel(z,2) = utoi(u);
   }
   else {
     GEN u = addii(x,y); if (cmpii(u, X) >= 0) u = subii(u, X);
@@ -182,7 +178,7 @@ static GEN
 sub_intmod_same(GEN z, GEN X, GEN x, GEN y) {
   if (lgefint(X) == 3) {
     ulong u = Fl_sub(itou(x),itou(y), X[2]);
-    avma = (pari_sp)z; gel(z,2) = utoi(u);
+    set_avma((pari_sp)z); gel(z,2) = utoi(u);
   }
   else {
     GEN u = subii(x,y); if (signe(u) < 0) u = addii(u, X);
@@ -195,7 +191,7 @@ static GEN
 mul_intmod_same(GEN z, GEN X, GEN x, GEN y) {
   if (lgefint(X) == 3) {
     ulong u = Fl_mul(itou(x),itou(y), X[2]);
-    avma = (pari_sp)z; gel(z,2) = utoi(u);
+    set_avma((pari_sp)z); gel(z,2) = utoi(u);
   }
   else
     gel(z,2) = gerepileuptoint((pari_sp)z, remii(mulii(x,y), X) );
@@ -207,7 +203,7 @@ div_intmod_same(GEN z, GEN X, GEN x, GEN y)
 {
   if (lgefint(X) == 3) {
     ulong m = uel(X,2), u = Fl_div(itou(x), itou(y), m);
-    avma = (pari_sp)z; gel(z,2) = utoi(u);
+    set_avma((pari_sp)z); gel(z,2) = utoi(u);
   }
   else
     gel(z,2) = gerepileuptoint((pari_sp)z, remii(mulii(x, Fp_inv(y,X)), X) );
@@ -370,22 +366,26 @@ Qdivii(GEN x, GEN y)
   pari_sp av = avma;
   GEN r, q;
 
+  if (lgefint(y) == 3)
+  {
+    q = Qdiviu(x, y[2]);
+    if (signe(y) > 0) return q;
+    if (typ(q) == t_INT) togglesign(q); else togglesign_safe(&gel(q,1));
+    return q;
+  }
   if (is_pm1(y)) return (signe(y) < 0)? negi(x): icopy(x);
-  if (equali1(x)) {
-    long s = signe(y);
-    if (!s) pari_err_INV("gdiv",y);
-    if (signe(x) < 0) s = -s;
-    q = cgetg(3, t_FRAC);
-    gel(q,1) = s<0? gen_m1: gen_1;
-    gel(q,2) = absi(y); return q;
+  if (equali1(x))
+  {
+    if (!signe(y)) pari_err_INV("gdiv",y);
+    retmkfrac(signe(y) < 0? gen_m1: gen_1, absi(y));
   }
   q = dvmdii(x,y,&r);
   if (r == gen_0) return q; /* gen_0 intended */
-  r = gcdii(y,r);
+  r = gcdii(y, r);
   if (lgefint(r) == 3)
   {
     ulong t = r[2];
-    avma = av;
+    set_avma(av);
     if (t == 1) q = mkfraccopy(x,y);
     else
     {
@@ -403,6 +403,56 @@ Qdivii(GEN x, GEN y)
   normalize_frac(q); return q;
 }
 
+/* x t_INT, return x/y in reduced form */
+GEN
+Qdiviu(GEN x, ulong y)
+{
+  pari_sp av = avma;
+  ulong r, t;
+  GEN q;
+
+  if (y == 1) return icopy(x);
+  if (!y) pari_err_INV("gdiv",gen_0);
+  if (equali1(x)) retmkfrac(gen_1, utoipos(y));
+  q = absdiviu_rem(x,y,&r);
+  if (!r)
+  {
+    if (signe(x) < 0) togglesign(q);
+    return q;
+  }
+  t = ugcd(y, r); set_avma(av);
+  if (t == 1) retmkfrac(icopy(x), utoipos(y));
+  retmkfrac(diviuexact(x,t), utoipos(y / t));
+}
+
+/* x t_INT, return x/y in reduced form */
+GEN
+Qdivis(GEN x, long y)
+{
+  pari_sp av = avma;
+  ulong r, t;
+  long s;
+  GEN q;
+
+  if (y > 0) return Qdiviu(x, y);
+  if (!y) pari_err_INV("gdiv",gen_0);
+  s = signe(x);
+  if (!s) return gen_0;
+  if (y < 0) { y = -y; s = -s; }
+  if (y == 1) { x = icopy(x); setsigne(x,s); return x; }
+  if (equali1(x)) retmkfrac(s > 0? gen_1: gen_m1, utoipos(y));
+  q = absdiviu_rem(x,y,&r);
+  if (!r)
+  {
+    if (s < 0) togglesign(q);
+    return q;
+  }
+  t = ugcd(y, r); set_avma(av); q = cgetg(3, t_FRAC);
+  if (t != 1) { x = diviuexact(x,t); y /= t; } else x = icopy(x);
+  gel(q,1) = x; setsigne(x, s);
+  gel(q,2) = utoipos(y); return q;
+}
+
 /*******************************************************************/
 /*                                                                 */
 /*                          CONJUGATION                            */
@@ -557,7 +607,6 @@ conjvec(GEN x,long prec)
   return z;
 }
 
-
 /********************************************************************/
 /**                                                                **/
 /**                           ADDITION                             **/
@@ -592,7 +641,7 @@ addsub_pp(GEN x, GEN y, GEN (*op)(GEN,GEN))
     u = swap? op(gel(y,4), gel(x,4)): op(gel(x,4), gel(y,4));
     if (!signe(u) || (c = Z_pvalrem(u,p,&u)) >= r)
     {
-      avma = av; return zeropadic(p, e+r);
+      set_avma(av); return zeropadic(p, e+r);
     }
     if (c)
     {
@@ -602,7 +651,7 @@ addsub_pp(GEN x, GEN y, GEN (*op)(GEN,GEN))
     }
   }
   u = modii(u, mod);
-  avma = av; z = cgetg(5,t_PADIC);
+  set_avma(av); z = cgetg(5,t_PADIC);
   z[1] = evalprecp(r) | evalvalp(e);
   gel(z,2) = icopy(p);
   gel(z,3) = icopy(mod);
@@ -612,7 +661,7 @@ addsub_pp(GEN x, GEN y, GEN (*op)(GEN,GEN))
 static GEN
 Q_to_Fp(GEN x, GEN p)
 { return mulii(gel(x,1), Fp_inv(gel(x,2),p)); }
-/* return x + y, where y t_PADIC and x is a non-zero t_INT or t_FRAC */
+/* return x + y, where y t_PADIC and x is a nonzero t_INT or t_FRAC */
 static GEN
 addQp(GEN x, GEN y)
 {
@@ -622,7 +671,7 @@ addQp(GEN x, GEN y)
 
   e = Q_pvalrem(x, p, &x);
   d = vy - e; r = d + py;
-  if (r <= 0) { avma = av; return gcopy(y); }
+  if (r <= 0) { set_avma(av); return gcopy(y); }
   mod = gel(y,3);
   u   = gel(y,4);
   (void)new_chunk(5 + ((lgefint(mod) + lgefint(p)*labs(d)) << 1));
@@ -648,7 +697,7 @@ addQp(GEN x, GEN y)
     u = addii(u, x);
     if (!signe(u) || (c = Z_pvalrem(u,p,&u)) >= r)
     {
-      avma = av; return zeropadic(p,e+r);
+      set_avma(av); return zeropadic(p,e+r);
     }
     if (c)
     {
@@ -657,7 +706,7 @@ addQp(GEN x, GEN y)
       e += c;
     }
   }
-  u = modii(u, mod); avma = av;
+  u = modii(u, mod); set_avma(av);
   z = cgetg(5,t_PADIC);
   z[1] = evalprecp(r) | evalvalp(e);
   gel(z,2) = icopy(p);
@@ -770,7 +819,7 @@ add_rfrac_scal(GEN y, GEN x)
   return gerepileupto(av, gred_rfrac_simple(n, gel(y,2)));
 }
 
-/* x "scalar", ty != t_MAT and non-scalar */
+/* x "scalar", ty != t_MAT and nonscalar */
 static GEN
 add_scal(GEN y, GEN x, long ty)
 {
@@ -788,97 +837,68 @@ add_scal(GEN y, GEN x, long ty)
   return NULL; /* LCOV_EXCL_LINE */
 }
 
+/* assumes z = cget(3, t_FRAC) comes first in stack, then a, then b */
+static GEN
+setfrac(GEN z, GEN a, GEN b)
+{
+  gel(z,1) = icopy_avma(a, (pari_sp)z);
+  gel(z,2) = icopy_avma(b, (pari_sp)gel(z,1));
+  set_avma((pari_sp)gel(z,2)); return z;
+}
+/* z <- a / (b*Q), (Q,a) = 1 */
+static GEN
+addsub_frac_i(GEN z, GEN Q, GEN a, GEN b)
+{
+  GEN q = Qdivii(a, b); /* != 0 */
+  if (typ(q) == t_INT)
+  {
+    gel(z,1) = gerepileuptoint((pari_sp)Q, q);
+    gel(z,2) = Q; return z;
+  }
+  return setfrac(z, gel(q,1), mulii(gel(q,2), Q));
+}
 static GEN
 addsub_frac(GEN x, GEN y, GEN (*op)(GEN,GEN))
 {
-  pari_sp av0 = avma;
-  GEN x1 = gel(x,1), x2 = gel(x,2), z = cgetg(3,t_FRAC);
-  GEN y1 = gel(y,1), y2 = gel(y,2), q, r, n, d, delta;
+  GEN x1 = gel(x,1), x2 = gel(x,2);
+  GEN y1 = gel(y,1), y2 = gel(y,2), z, Q, q, r, n, delta;
   int s = cmpii(x2, y2);
 
+  /* common denominator: (x1 op y1) / x2 */
   if (!s)
-  { /* common denominator: (x1 op y1) / x2 */
-    n = op(x1, y1);
-    if (!signe(n)) { avma = av0; return gen_0; }
-    d = x2;
-    q = dvmdii(n, d, &r);
-    if (r == gen_0) { avma = av0; return icopy(q); }
-    r = gcdii(d, r);
-    if (!equali1(r)) { n = diviiexact(n, r); d = diviiexact(d, r); }
-    avma = (pari_sp)z;
-    gel(z,1) = icopy(n);
-    gel(z,2) = icopy(d); return z;
+  {
+    pari_sp av = avma;
+    return gerepileupto(av, Qdivii(op(x1, y1), x2));
   }
+  z = cgetg(3, t_FRAC);
   if (s < 0)
   {
-    GEN Q = dvmdii(y2, x2, &r);
-    if (r == gen_0)
-    { /* y2 = Q x2: 1/x2 . (Q x1 op y1)/Q, where latter is in coprime form */
-      pari_sp av = avma;
-      n = op(mulii(Q,x1), y1);
-      q = dvmdii(n, x2, &r);
-      if (r == gen_0)
-      {
-        gel(z,1) = gerepileuptoint(av, q);
-        gel(z,2) = Q; return z;
-      }
-      r = gcdii(x2, r);
-      if (!equali1(r)) { n = diviiexact(n, r); x2 = diviiexact(x2, r); }
-      d = mulii(x2,Q); avma = (pari_sp)z;
-      gel(z,1) = icopy(n);
-      gel(z,2) = icopy(d); return z;
-    }
+    Q = dvmdii(y2, x2, &r);
+    /* y2 = Q x2: 1/x2 . (Q x1 op y1)/Q, where latter is in coprime form */
+    if (r == gen_0) return addsub_frac_i(z, Q, op(mulii(Q,x1), y1), x2);
     delta = gcdii(x2,r);
   }
   else
   {
-    GEN Q = dvmdii(x2, y2, &r);
-    if (r == gen_0)
-    { /* x2 = Q y2: 1/y2 . (x1 op Q y1)/Q, where latter is in coprime form */
-      pari_sp av = avma;
-      n = op(x1, mulii(Q,y1));
-      q = dvmdii(n, y2, &r);
-      if (r == gen_0)
-      {
-        gel(z,1) = gerepileuptoint(av, q);
-        gel(z,2) = Q; return z;
-      }
-      r = gcdii(y2, r);
-      if (!equali1(r)) { n = diviiexact(n, r); y2 = diviiexact(y2, r); }
-      d = mulii(y2,Q); avma=(pari_sp)z;
-      gel(z,1) = icopy(n);
-      gel(z,2) = icopy(d); return z;
-    }
+    Q = dvmdii(x2, y2, &r);
+    /* x2 = Q y2: 1/y2 . (x1 op Q y1)/Q, where latter is in coprime form */
+    if (r == gen_0) return addsub_frac_i(z, Q, op(x1, mulii(Q,y1)), y2);
     delta = gcdii(y2,r);
   }
   /* delta = gcd(x2,y2) */
   if (equali1(delta))
-  { /* numerator is non-zero */
+  { /* numerator is nonzero */
     gel(z,1) = gerepileuptoint((pari_sp)z, op(mulii(x1,y2), mulii(y1,x2)));
     gel(z,2) = mulii(x2,y2); return z;
   }
   x2 = diviiexact(x2,delta);
   y2 = diviiexact(y2,delta);
-  n = op(mulii(x1,y2), mulii(y1,x2));
-  if (!signe(n)) { avma = av0; return gen_0; }
-  d = mulii(x2, y2);
+  n = op(mulii(x1,y2), mulii(y1,x2)); /* != 0 */
   q = dvmdii(n, delta, &r);
-  if (r == gen_0)
-  {
-    if (equali1(d)) { avma = av0; return icopy(q); }
-    avma = (pari_sp)z;
-    gel(z,2) = icopy(d);
-    gel(z,1) = icopy(q); return z;
-  }
+  if (r == gen_0) return setfrac(z, q, mulii(x2, y2));
   r = gcdii(delta, r);
-  if (!equali1(r))
-  {
-    n     = diviiexact(n, r);
-    delta = diviiexact(delta, r);
-  }
-  d = mulii(d,delta); avma = (pari_sp)z;
-  gel(z,1) = icopy(n);
-  gel(z,2) = icopy(d); return z;
+  if (!equali1(r)) { n = diviiexact(n, r); delta = diviiexact(delta, r); }
+  return setfrac(z, n, mulii(mulii(x2, y2), delta));
 }
 
 /* assume x2, y2 are t_POLs in the same variable */
@@ -952,7 +972,7 @@ gadd(GEN x, GEN y)
       gel(z,2) = gadd(gel(x,2),gel(y,2));
       if (isintzero(gel(z,2)))
       {
-        avma = (pari_sp)(z+3);
+        set_avma((pari_sp)(z+3));
         return gadd(gel(x,1),gel(y,1));
       }
       gel(z,1) = gadd(gel(x,1),gel(y,1));
@@ -1159,7 +1179,7 @@ gadd(GEN x, GEN y)
         d = gel(y,2); vd = RgX_valrem(d, &d);
 
         l = lg(x) + valp(x) - (vn - vd);
-        if (l < 3) { avma = av; return gcopy(x); }
+        if (l < 3) { set_avma(av); return gcopy(x); }
 
         /* take advantage of y = t^n ! */
         if (degpol(d))
@@ -1257,7 +1277,7 @@ gsub(GEN x, GEN y)
       gel(z,2) = gsub(gel(x,2),gel(y,2));
       if (isintzero(gel(z,2)))
       {
-        avma = (pari_sp)(z+3);
+        set_avma((pari_sp)(z+3));
         return gsub(gel(x,1),gel(y,1));
       }
       gel(z,1) = gsub(gel(x,1),gel(y,1));
@@ -1312,16 +1332,14 @@ gsub(GEN x, GEN y)
 /********************************************************************/
 static GEN
 mul_ser_scal(GEN y, GEN x) {
-  long ly, i;
+  long l, i;
   GEN z;
-  if (isexactzero(x)) return gmul(Rg_get_0(y), x);
+  if (isrationalzero(x)) return gmul(Rg_get_0(y), x);
   if (ser_isexactzero(y))
-  {
-    if (lg(y) == 2) return gcopy(y);
-    return scalarser(gmul(x,gel(y,2)), varn(y), valp(y));
-  }
-  z = cgetg_copy(y, &ly); z[1] = y[1];
-  for (i = 2; i < ly; i++) gel(z,i) = gmul(x,gel(y,i));
+    return scalarser(lg(y) == 2? Rg_get_0(x)
+                               : gmul(gel(y,2), x), varn(y), valp(y));
+  z = cgetg_copy(y, &l); z[1] = y[1];
+  for (i = 2; i < l; i++) gel(z,i) = gmul(gel(y,i), x);
   return normalize(z);
 }
 /* (n/d) * x, x "scalar" or polynomial in the same variable as d
@@ -1694,7 +1712,7 @@ mulpp(GEN x, GEN y) {
   t = (precp(x) > precp(y))? y: x;
   z = cgetp(t); setvalp(z,l); av = avma;
   affii(remii(mulii(gel(x,4),gel(y,4)), gel(t,3)), gel(z,4));
-  avma = av; return z;
+  set_avma(av); return z;
 }
 /* x,y QUAD */
 static GEN
@@ -1841,10 +1859,9 @@ gmul(GEN x, GEN y)
       vy = varn(y);
       if (vx != vy) {
         if (varncmp(vx, vy) < 0) return mul_ser_scal(x, y);
-        else                     return mul_ser_scal(y, x);
+        return mul_ser_scal(y, x);
       }
-      lx = lg(x);
-      ly = lg(y); if (lx > ly) { lx = ly; swap(x, y); }
+      lx = minss(lg(x), lg(y));
       if (lx == 2) return zeroser(vx, valp(x)+valp(y));
       av = avma; z = cgetg(lx,t_SER);
       z[1] = evalvalp(valp(x)+valp(y)) | evalvarn(vx) | evalsigne(1);
@@ -1869,7 +1886,6 @@ gmul(GEN x, GEN y)
       }
       return z;
 
-
     default:
       pari_err_TYPE2("*",x,y);
   }
@@ -1890,7 +1906,7 @@ gmul(GEN x, GEN y)
           p1 = gcdii(x,gel(y,2));
           if (equali1(p1))
           {
-            avma = (pari_sp)z;
+            set_avma((pari_sp)z);
             gel(z,2) = icopy(gel(y,2));
             gel(z,1) = mulii(gel(y,1), x);
           }
@@ -2046,7 +2062,7 @@ gmul(GEN x, GEN y)
             p2 = gmul(p1,y);
             settyp(p1, t_VECSMALL); /* p1 left on stack */
           } else {
-            avma = av;
+            set_avma(av);
             p2 = mul_ser_scal(y, gel(x,2));
           }
           setvalp(p2, valp(p2) + vn);
@@ -2070,7 +2086,7 @@ gmul(GEN x, GEN y)
   return NULL; /* LCOV_EXCL_LINE */
 }
 
-/* return a non-normalized result */
+/* return a nonnormalized result */
 GEN
 sqr_ser_part(GEN x, long l1, long l2)
 {
@@ -2241,17 +2257,16 @@ static GEN
 div_rfrac(GEN x, GEN y)
 { return mul_rfrac(gel(x,1),gel(x,2), gel(y,2),gel(y,1)); }
 
+/* x != 0 */
 static GEN
-div_ser_scal(GEN x, GEN y) {
-  long i, lx;
+div_ser_scal(GEN y, GEN x) {
+  long i, l;
   GEN z;
-  if (ser_isexactzero(x))
-  {
-    if (lg(x) == 2) return gcopy(x);
-    return scalarser(gdiv(gel(x,2), y), varn(x), valp(x));
-  }
-  z = cgetg_copy(x, &lx); z[1] = x[1];
-  for (i=2; i<lx; i++) gel(z,i) = gdiv(gel(x,i),y);
+  if (ser_isexactzero(y))
+    return scalarser(lg(y) == 2? Rg_get_0(x)
+                               : gdiv(gel(y,2), x), varn(y), valp(y));
+  z = cgetg_copy(y, &l); z[1] = y[1];
+  for (i = 2; i < l; i++) gel(z,i) = gdiv(gel(y,i), x);
   return normalize(z);
 }
 GEN
@@ -2266,6 +2281,7 @@ ser_normalize(GEN x)
   return z;
 }
 
+/* y != 0 */
 static GEN
 div_T_scal(GEN x, GEN y, long tx) {
   switch(tx)
@@ -2288,16 +2304,8 @@ div_scal_pol(GEN x, GEN y) {
   return gerepileupto(av, gred_rfrac_simple(x,y));
 }
 static GEN
-div_scal_ser(GEN x, GEN y) { /* TODO: improve */
-  GEN z;
-  long ly, i;
-  if (gequal0(x)) { pari_sp av=avma; return gerepileupto(av, gmul(x, ginv(y))); }
-  ly = lg(y); z = (GEN)pari_malloc(ly*sizeof(long));
-  z[0] = evaltyp(t_SER) | evallg(ly);
-  z[1] = evalsigne(1) | _evalvalp(0) | evalvarn(varn(y));
-  gel(z,2) = x; for (i=3; i<ly; i++) gel(z,i) = gen_0;
-  y = gdiv(z,y); pari_free(z); return y;
-}
+div_scal_ser(GEN x, GEN y)
+{ pari_sp av = avma; return gerepileupto(av, gmul(x, ser_inv(y))); }
 static GEN
 div_scal_T(GEN x, GEN y, long ty) {
   switch(ty)
@@ -2310,25 +2318,35 @@ div_scal_T(GEN x, GEN y, long ty) {
   return NULL; /* LCOV_EXCL_LINE */
 }
 
+static GEN
+ser2pol_ii(GEN x, long lx, long y1)
+{
+  GEN y = cgetg(lx, t_POL);
+  long i;
+  for (i = lx-1; i > 1; i--) gel(y,i) = gel(x,i);
+  y[1] = y1; return y;
+}
+
 /* assume tx = ty = t_SER, same variable vx */
 static GEN
 div_ser(GEN x, GEN y, long vx)
 {
-  long i, j, l = valp(x) - valp(y), lx = lg(x), ly = lg(y);
-  GEN y_lead, p1, p2, z;
+  long v = valp(x) - valp(y), lx = lg(x), ly = lg(y);
+  GEN y_lead, z;
+  pari_sp av = avma;
 
   if (!signe(y)) pari_err_INV("div_ser", y);
   if (ser_isexactzero(x))
   {
-    if (lx == 2) return zeroser(vx, l);
-    return scalarser(gmul(gel(x,2),Rg_get_0(y)), varn(x), l);
+    if (lx == 2) return zeroser(vx, v);
+    return scalarser(gmul(gel(x,2),Rg_get_0(y)), varn(x), v);
   }
   y_lead = gel(y,2);
   if (gequal0(y_lead)) /* normalize denominator if leading term is 0 */
   {
     GEN y0 = y;
     pari_warn(warner,"normalizing a series with 0 leading term");
-    for (l--, ly--,y++; ly > 2; l--, ly--, y++)
+    for (v--, ly--,y++; ly > 2; v--, ly--, y++)
     {
       y_lead = gel(y,2);
       if (!gequal0(y_lead)) break;
@@ -2336,25 +2354,16 @@ div_ser(GEN x, GEN y, long vx)
     if (ly <= 2) pari_err_INV("div_ser", y0);
   }
   if (ly < lx) lx = ly;
-  p2 = cgetg(lx, t_VECSMALL); /* left on stack for efficiency */
-  for (i=3; i<lx; i++)
-  {
-    p1 = gel(y,i);
-    if (isrationalzero(p1)) p1 = NULL;
-    gel(p2,i) = p1;
+  z = cgetg(lx,t_SER); z[1] = evalvalp(v) | evalvarn(vx) | evalsigne(1);
+  x = ser2pol_i(x, lx);
+  y = ser2pol_ii(y, lx, z[1] & ~VALPBITS);
+  if (lx == 3) gel(z,2) = gdiv(gel(x,2), gel(y,2));
+  else
+  { /* FIXME: use Karp/Markstein */
+    y = RgXn_mul(x, RgXn_inv_i(y, lx-2), lx-2);
+    z = fill_ser(z,y);
   }
-  z = cgetg(lx,t_SER);
-  z[1] = evalvalp(l) | evalvarn(vx) | evalsigne(1);
-  gel(z,2) = gdiv(gel(x,2), y_lead);
-  for (i=3; i<lx; i++)
-  {
-    pari_sp av = avma;
-    p1 = gel(x,i);
-    for (j=2, l=i; j<i; j++, l--)
-      if (p2[l]) p1 = gsub(p1, gmul(gel(z,j), gel(p2,l)));
-    gel(z,i) = gerepileupto(av, gdiv(p1, y_lead));
-  }
-  return normalize(z);
+  return gerepilecopy(av, z);
 }
 /* x,y compatible PADIC */
 static GEN
@@ -2489,8 +2498,12 @@ gdiv(GEN x, GEN y)
       vx = varn(x);
       vy = varn(y);
       if (vx != vy) {
-        if (varncmp(vx, vy) < 0) return div_ser_scal(x, y);
-                            else return div_scal_ser(x, y);
+        if (varncmp(vx, vy) < 0)
+        {
+          if (!signe(y)) pari_err_INV("gdiv",y);
+          return div_ser_scal(x, y);
+        }
+        return div_scal_ser(x, y);
       }
       return div_ser(x, y, vx);
     case t_RFRAC:
@@ -2542,7 +2555,7 @@ gdiv(GEN x, GEN y)
         z = cgetg(3,t_FRAC); p1 = gcdii(x,gel(y,1));
         if (equali1(p1))
         {
-          avma = (pari_sp)z;
+          set_avma((pari_sp)z);
           gel(z,2) = icopy(gel(y,1));
           gel(z,1) = mulii(gel(y,2), x);
           normalize_frac(z);
@@ -2623,7 +2636,7 @@ gdiv(GEN x, GEN y)
         p1 = gcdii(y,gel(x,1));
         if (equali1(p1))
         {
-          avma = (pari_sp)z; tetpil = 0;
+          set_avma((pari_sp)z); tetpil = 0;
           gel(z,1) = icopy(gel(x,1));
         }
         else
@@ -2831,7 +2844,7 @@ gmulsg(long s, GEN y)
   switch(typ(y))
   {
     case t_INT:  return mulsi(s,y);
-    case t_REAL: return mulsr(s,y);
+    case t_REAL: return s? mulsr(s,y): gen_0; /* gmul semantic */
     case t_INTMOD: { GEN p = gel(y,1);
       z = cgetg(3,t_INTMOD);
       gel(z,2) = gerepileuptoint((pari_sp)z, modii(mulsi(s,gel(y,2)), p));
@@ -2855,7 +2868,9 @@ gmulsg(long s, GEN y)
       }
       return z;
 
-    case t_COMPLEX: z = cgetg(3, t_COMPLEX);
+    case t_COMPLEX:
+      if (!s) return gen_0;
+      z = cgetg(3, t_COMPLEX);
       gel(z,1) = gmulsg(s,gel(y,1));
       gel(z,2) = gmulsg(s,gel(y,2)); return z;
 
@@ -2911,7 +2926,7 @@ gdivgs(GEN x, long s)
 {
   long tx = typ(x), lx, i;
   pari_sp av;
-  GEN z, y;
+  GEN z;
 
   if (!s)
   {
@@ -2920,18 +2935,8 @@ gdivgs(GEN x, long s)
   }
   switch(tx)
   {
-    case t_INT:
-      av = avma; z = divis_rem(x,s,&i);
-      if (!i) return z;
-
-      i = cgcd(s, i);
-      avma=av; z = cgetg(3,t_FRAC);
-      if (i == 1) y = icopy(x); else { s /= i; y = diviuexact(x, i); }
-      gel(z,1) = y;
-      gel(z,2) = stoi(s); normalize_frac(z); return z;
-
-    case t_REAL:
-      return divrs(x,s);
+    case t_INT: return Qdivis(x, s);
+    case t_REAL: return divrs(x,s);
 
     case t_INTMOD:
       z = cgetg(3, t_INTMOD);
@@ -3148,8 +3153,7 @@ ginv(GEN x)
     case t_POLMOD: return inv_polmod(gel(x,1), gel(x,2));
     case t_FFELT: return FF_inv(x);
     case t_POL: return gred_rfrac_simple(gen_1,x);
-    case t_SER: return gdiv(gen_1,x);
-
+    case t_SER: return ser_inv(x);
     case t_RFRAC:
     {
       GEN n = gel(x,1), d = gel(x,2);
@@ -3159,7 +3163,7 @@ ginv(GEN x)
       n = simplify_shallow(n);
       if (typ(n) != t_POL || varn(n) != varn(d))
       {
-        if (gequal1(n)) { avma = av; return RgX_copy(d); }
+        if (gequal1(n)) { set_avma(av); return RgX_copy(d); }
         ltop = avma;
         z = RgX_Rg_div(d,n);
       } else {
