@@ -872,12 +872,14 @@ list_cmp(GEN x, GEN y, int cmp(GEN x, GEN y))
 {
   int t = list_typ(x);
   GEN vx, vy;
+  long lvx, lvy;
   if (list_typ(y)!=t) return 0;
   vx = list_data(x);
   vy = list_data(y);
-  if (!vx) return vy? 0: 1;
-  if (!vy) return 0;
-  if (lg(vx) != lg(vy)) return 0;
+  lvx = vx ? lg(vx): 1;
+  lvy = vy ? lg(vy): 1;
+  if (lvx==1 && lvy==1) return 1;
+  if (lvx != lvy) return 0;
   switch (t)
   {
   case t_LIST_MAP:
@@ -976,9 +978,9 @@ serequal(GEN x, GEN y)
   if (!signe(x) && !signe(y)) return 1;
   lx = lg(x); vx = valp(x); LX = lx + vx;
   ly = lg(y); vy = valp(y); LY = ly + vy;
-  if (LX > LY) { LX = LY; lx = LX - vx; } else ly = LX - vy;
-  for (lx--, ly--; lx >= 2 && ly >= 2; lx--, ly--)
-    if (!gequal(gel(x,lx), gel(y,ly))) return 0;
+  if (LX > LY) lx = LY - vx; else ly = LX - vy;
+  while (lx >= 3 && ly >= 3)
+    if (!gequal(gel(x,--lx), gel(y,--ly))) return 0;
   while(--ly >= 2) if (!gequal0(gel(y,ly))) return 0;
   while(--lx >= 2) if (!gequal0(gel(x,lx))) return 0;
   return 1;
