@@ -493,7 +493,7 @@ zind(GEN z, long ind)
 
 /* z not 0 or 1, c not a nonpositive integer */
 static long
-F21ind(GEN a, GEN b, GEN c, GEN z)
+F21ind(GEN a, GEN b, GEN c, GEN z, long bit)
 {
   GEN v = const_vec(6, mkoo());
   long ind = 0;
@@ -504,8 +504,8 @@ F21ind(GEN a, GEN b, GEN c, GEN z)
   if (!isnegint(cind(a,b,c, 4))) gel(v,4) = gabs(zind(z,4), LD);
   if (!isnegint(cind(a,b,c, 5))) gel(v,5) = gabs(zind(z,5), LD);
   if (!isnegint(cind(a,b,c, 6))) gel(v,6) = gabs(zind(z,6), LD);
-  ind = vecindexmin(v); /* |znew| close to 1 ? */
-  return (gcmp(gel(v,ind), mkfracss(49,50)) <= 0)? -ind: ind;
+  ind = vecindexmin(v); /* |znew| <= 1; close to 1 ? */
+  return (gexpo(gsubgs(gel(v,ind),1)) > -maxss(bit / 4, 32))? -ind: ind;
 }
 static GEN
 mul4(GEN a, GEN b, GEN c, GEN d) { return gmul(a,gmul(b, gmul(c, d))); }
@@ -545,7 +545,7 @@ F21lam(long N, GEN a, GEN c)
   gel(S,i+1) = gel(vb,N); return RgV_to_RgX(S,0);
 }
 
-/* F21finite stuff: case where a or b is a nonpositive integer */
+/* F(-m,b; c; z), m >= 0 */
 static GEN
 F21finitetaylor(long m, GEN b, GEN c, GEN z, long prec)
 {
@@ -585,12 +585,12 @@ F21finite_i(long m, GEN b, GEN c, GEN z, GEN B, GEN C, GEN coe, long prec)
               F21finitetaylor(m, b, c, z, prec));
 }
 
-/* c not a nonpositive integer */
+/* F(-m,b; c; z), m >= 0; c not a nonpositive integer */
 static GEN
 F21finite(long m, GEN b, GEN c, GEN z, long prec)
 {
   GEN a = stoi(-m), b1 = b, c1 = c, z1;
-  long ind = F21ind(a, b, c, z), inda = labs(ind);
+  long ind = F21ind(a, b, c, z, prec2nbits(prec)), inda = labs(ind);
   z1 = zind(z, inda);
   if (ind < 0)
   {
@@ -602,10 +602,10 @@ F21finite(long m, GEN b, GEN c, GEN z, long prec)
     case 1: return F21finite_i(m, b1, c1, z1, b, c, gsubsg(1,z), prec);
     case 2: return gmul(gpowgs(gsubsg(1,z), m),
                         F21finitetaylor(m, b1, c, z1, prec));
-    case 4: return F21finite_i(m, b, c1, z1, gsub(c,b), c, gen_1, prec);
+    case 4: return F21finite_i(m, b1, c1, z1, gsub(c,b), c, gen_1, prec);
     case 5: return F21finite_i(m, b1, c1, z1, gsub(c,b), c, z, prec);
     case 6: return F21finite_i(m, b1, c1, z1, b, c, gneg(z), prec);
-    default:return F21finitetaylor(m, b, c, z, prec);
+    default:return F21finitetaylor(m, b1, c1, z, prec);
   }
 }
 
@@ -651,22 +651,20 @@ Basic transforms:
   5: (1+a-c,b-c+a+1,1-1/z)
   6: (1+a-c,1+a-b,1/z)
 
-F21: calls F21_i and increase accuracy if too much cancellation
-F21_i:
-- z approx 0: return 1.
-- z=1: return Gauss multgam
-- a (or b) negative integer: call F21finite(-a,b)
-- c-a negative integer: call F21finite(a-c,c-b)
+F21: calls F21_i and increase prec if too much cancellation
+F21_i: c is not a non-positive integer
+- z ~ 0 or 1: return special value
+- if a, b, c-b or c-a a non-positive integer: use F21finite
 - compute index, value of z
-   If |z|<=0.98 call F21taylorind
-   else: if R(b)<=0, swap and/or recurse so may assume
-   R(b)>0 and R(a)>=R(b) and integrate.
+   if |z| < 1-epsilon return F21taylorind
+   if Re(b)<=0, swap and/or recurse
+   so may assume Re(b)>0 and Re(a)>=Re(b) and integrate.
 
 F21finite:
 - compute index, value of z
 - call F21finitetaylor
 
-F21ind: find best index (1 to 6, -1 to -6 if |z| < 0.98)
+F21ind: find best index (1 to 6, -1 to -6 if |z| < 1-epsilon)
 F21finitetaylor: a or b in Z_{<=0}; calls precFtaylor
 
 F21taylorind: in case 2, may lose accuracy, possible bug.
@@ -692,19 +690,19 @@ F21_i(GEN a, GEN b, GEN c, GEN z, long prec)
   if (is0(z, bitprec)) return real_1(prec);
   if (gequal1(z))
   {
-    GEN tmp = gsub(c, gadd(a, b)); check_hyp1(tmp);
-    return multgam(c, tmp, gsub(c,a), gsub(c,b), prec);
+    GEN x = gsub(c, gadd(a, b)); check_hyp1(x);
+    return multgam(c, x, gsub(c,a), gsub(c,b), prec);
   }
   if (isnegint2(b, &m)) return F21finite(m, a, c, z, prec);
   if (isnegint2(a, &m)) return F21finite(m, b, c, z, prec);
   if (isnegint(gsub(c, b))) swap(a, b);
   if (isnegint2(gsub(c, a), &m))
   {
-    GEN tmp = gpow(gsubsg(1, z), gneg(gaddsg(m, b)), prec);
-    return gmul(tmp, F21finite(m, gsub(c, b), c, z, prec));
+    GEN x = gpow(gsubsg(1, z), gneg(gaddsg(m, b)), prec);
+    return gmul(x, F21finite(m, gsub(c, b), c, z, prec));
   }
   /* Here a, b, c, c-a, c-b are not nonpositive integers */
-  ind = F21ind(a, b, c, z);
+  ind = F21ind(a, b, c, z, bitprec);
   if (ind < 0) return gmul(ggamma(c, prec), F21taylorind(a,b,c, z, ind, prec));
   if (gsigne(real_i(b)) <= 0)
   {
@@ -744,6 +742,7 @@ F21_i(GEN a, GEN b, GEN c, GEN z, long prec)
   return gmul(multgam(gen_1, c, b, gsub(c,b), prec), res);
 }
 
+/* c not a non-positive integer */
 static GEN
 F21(GEN a, GEN b, GEN c, GEN z, long prec)
 {
@@ -763,21 +762,23 @@ F21taylor(GEN a, GEN b, GEN c, GEN z, long prec)
 { return gdiv(Ftaylor(mkvec2(a,b), mkvec(c), z, prec), ggamma(c, prec)); }
 
 static GEN
-F21taylorlim(GEN N, long m, GEN z, long si, long prec)
+F21taylorlim(GEN N, long m, GEN z, GEN Z, long ind, long prec)
 {
   pari_sp av;
   GEN C, P, S, tmp;
-  long j, ct, pradd, mi, fl, bitmin, tol;
+  long j, ct, pradd, mi, fl, bitmin, tol, si = (ind == 5 || ind == 6)? -1: 1;
   pradd = precFtaylor(N, mkvec(stoi(m + 1)), z, &mi);
   if (pradd)
   {
     prec += pradd;
     N = gprec_wensure(N, prec);
     z = gprec_wensure(z, prec);
+    Z = gprec_wensure(Z, prec);
   }
   bitmin = -(prec2nbits(prec) + 10);
-  P = gadd(gneg(glog(gmulsg(si, z), prec)),
-           gsub(gpsi(stoi(m+1), prec), mpeuler(prec)));
+  P = glog(Z, prec);
+  if (ind == 4 || ind == 5) P = gneg(P);
+  P = gadd(P, gsub(gpsi(stoi(m+1), prec), mpeuler(prec)));
   tmp = gel(N, 2); if (si == -1) tmp = gsubsg(1, tmp);
   P = gsub(P, gadd(gpsi(gel(N, 1), prec), gpsi(tmp, prec)));
   C = real_1(prec); ct = 0; j = 0; tol = 0;
@@ -803,6 +804,7 @@ F21taylorlim(GEN N, long m, GEN z, long si, long prec)
   return gdiv(S, mpfact(m));
 }
 
+/* N = [a,b]; (m-1)! sum_{1 <= k < m} (-z)^k (a)_k (b)_k / k! (m-k)!*/
 static GEN
 F21finitelim(GEN N, long m, GEN z, long prec)
 {
@@ -812,7 +814,7 @@ F21finitelim(GEN N, long m, GEN z, long prec)
   a = gel(N,1);
   b = gel(N,2);
   S = C = real_1(prec + EXTRAPRECWORD);
-  for (j = 1; j < m; ++j)
+  for (j = 1; j < m; j++)
   {
     GEN v1 = gaddsg(j-1, a), v2 = gaddsg(j-1, b);
     C = gdivgs(gmul(C, gmul(gmul(v1, v2), z)), j*(j-m));
@@ -855,19 +857,20 @@ FBaux1(GEN v1, GEN g1, GEN c1, GEN v2, GEN g2, GEN c2, GEN z, GEN bma,
 
 static GEN
 FBaux2(GEN v1, GEN g1, GEN c1, long m, GEN z1, GEN c2, GEN g2, GEN v2, GEN z2,
-       long si, long prec)
+       GEN Z2, long ind, long prec)
 {
   GEN t1 = gdiv(c1, mulgammav2(g1, prec)), r1;
   GEN t2 = gdiv(c2, mulgammav2(g2, prec)), r2;
   r1 = gmul(t1, F21finitelim(v1, m, z1, prec));
-  r2 = gmul(t2, F21taylorlim(v2, m, z2, si, prec));
+  r2 = gmul(t2, F21taylorlim(v2, m, z2, Z2, ind, prec));
   return gadd(r1, r2);
 }
 
+/* 1 / (1-z) */
 static GEN
 F21taylor1(GEN a, GEN b, GEN c, GEN z, long prec)
 {
-  GEN bma = gsub(b, a), coe1, coe2, z1, g1, g2, v1, v2;
+  GEN bma = gsub(b, a), coe1, coe2, z1, g1, g2, v1, v2, Z;
   long m;
   if (!islong(bma,&m, prec))
   {
@@ -883,15 +886,17 @@ F21taylor1(GEN a, GEN b, GEN c, GEN z, long prec)
                   mkvec2(a,b1), coe2, z1, bma, prec, prec);
   }
   if (m < 0) { swap(a,b); m = -m; }
-  coe1 = gpow(gsubsg(1, z), gneg(a), prec);
+  Z = gsubsg(1, z);
+  coe1 = gpow(Z, gneg(a), prec);
   v2 = g1 = mkvec2(gaddgs(a, m), gsub(c, a));
   v1 = mkvec2(a, gsub(c, gaddgs(a, m)));
-  z1 = ginv(gsubsg(1, z));
+  z1 = ginv(Z);
   coe2 = gmul(coe1, gpowgs(z1, m)); if (m & 1L) coe2 = gneg(coe2);
-  g2 = mkvec2(a, gsub(c, gaddgs(a, m)));
-  return FBaux2(v1, g1, coe1, m, z1, coe2, g2, v2, z1, 1, prec);
+  g2 = mkvec2(a, gsub(c, gaddgs(a, m))); /* 15.8.9 */
+  return FBaux2(v1, g1, coe1, m, z1, coe2, g2, v2, z1, Z, 1, prec);
 }
 
+/* 1 - z */
 static GEN
 F21taylor4(GEN a, GEN b, GEN c, GEN z, long prec)
 {
@@ -916,10 +921,11 @@ F21taylor4(GEN a, GEN b, GEN c, GEN z, long prec)
   v2 = g1 = mkvec2(gaddgs(a,m), gaddgs(b,m));
   v1 = g2 = mkvec2(a, b);
   z1 = gsubgs(z, 1);
-  z2 = gneg(z1); coe2 = gpowgs(z1, m);
-  return FBaux2(v1, g1, gen_1, m, z1, coe2, g2, v2, z2, 1, prec);
+  z2 = gneg(z1); coe2 = gpowgs(z1, m); /* 15.8.10 */
+  return FBaux2(v1, g1, gen_1, m, z1, coe2, g2, v2, z2, z2, 4, prec);
 }
 
+/* 1 - 1/z */
 static GEN
 F21taylor5(GEN a, GEN b, GEN c, GEN z, long prec)
 {
@@ -953,14 +959,15 @@ F21taylor5(GEN a, GEN b, GEN c, GEN z, long prec)
   z2 = gneg(z1);
   g2 = mkvec2(a, b);
   coe1 = gpow(z, gneg(a), prec);
-  coe2 = gmul(coe1, gpowgs(z2, m));
-  return FBaux2(v1, g1, coe1, m, z1, coe2, g2, v2, z2, -1, prec);
+  coe2 = gmul(coe1, gpowgs(z2, m)); /* 15.8.11 */
+  return FBaux2(v1, g1, coe1, m, z1, coe2, g2, v2, z2, z1, 5, prec);
 }
 
+/* 1 / z */
 static GEN
 F21taylor6(GEN a, GEN b, GEN c, GEN z, long prec)
 {
-  GEN bma = gsub(b, a), cma, am, coe1, coe2, z1, g1, g2, v1, v2, z2;
+  GEN bma = gsub(b, a), cma, am, coe1, coe2, z1, g1, g2, v1, v2, z2, Z;
   long m;
   if (!islong(bma,&m, prec))
   {
@@ -980,15 +987,16 @@ F21taylor6(GEN a, GEN b, GEN c, GEN z, long prec)
   /* b - a ~ m */
   if (m < 0) { swap(a,b); m = -m; }
   cma = gsub(c, a); am = gaddgs(a,m);
-  coe1 = gpow(gneg(z), gneg(a), prec);
+  Z = gneg(z);
+  coe1 = gpow(Z, gneg(a), prec);
   coe2 = gdiv(coe1, gpowgs(z, m));
   g1 = mkvec2(am, cma);
   v1 = mkvec2(a, gsubsg(1, cma));
   g2 = mkvec2(a, gsubgs(cma, m));
   v2 = mkvec2(am, gsubsg(m+1, cma));
   z2 = ginv(z);
-  z1 = gneg(z2);
-  return FBaux2(v1, g1, coe1, m, z1, coe2, g2, v2, z2, -1, prec);
+  z1 = gneg(z2); /* 15.8.8 */
+  return FBaux2(v1, g1, coe1, m, z1, coe2, g2, v2, z2, Z, 6, prec);
 }
 
 /* (new b, new c, new z): given by bind, cind, zind
