@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -22,6 +23,10 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #define M_E 2.7182818284590452354
 #endif
 
+/* Handle complex Vga whose sum is real */
+static GEN
+sumVga(GEN Vga) { return real_i(vecsum(Vga)); }
+
 /* rough approximation to W0(a > -1/e), < 1% relative error */
 double
 dbllambertW0(double a)
@@ -115,110 +120,141 @@ dblcoro526(double a, double c, double B)
 
 static const double MELLININV_CUTOFF = 121.; /* C*C */
 
+/* x real */
 static GEN
-MOD2(GEN x) { GEN q = gdivent(x,gen_2); return gsub(x,gmul2n(q,1)); }
+RMOD2(GEN x) { return gsub(x, gmul2n(gdiventgs(x,2), 1)); }
+/* x real or complex, return canonical representative for x mod 2Z */
 static GEN
-RgV_MOD2(GEN v)
-{
-  long i, l;
-  GEN w = cgetg_copy(v,&l);
-  for (i=1; i<l; i++) gel(w,i) = MOD2(gel(v,i));
-  return w;
-}
+MOD2(GEN x)
+{ return typ(x) == t_COMPLEX? mkcomplex(RMOD2(gel(x,1)), gel(x,2)): RMOD2(x); }
+static GEN
+RgV_MOD2(GEN x)
+{ pari_APPLY_same(MOD2(gel(x,i))); }
 
-/* poles of the gamma factor */
+/* classes of poles of the gamma factor mod 2Z, sorted by increasing
+ * Re(s) mod 2 (in [0,2[).*/
 static GEN
-gammapoles(GEN Vga)
+gammapoles(GEN Vga, long *pdV, long bit)
 {
-  long i, m, d = lg(Vga)-1;
-  GEN P, V, B = RgV_MOD2(Vga);
-  P = gen_indexsort(B, (void*)gcmp, cmp_nodata);
-  V = cgetg(d+1, t_VEC);
-  for (i = 1, m = 1; i <= d;)
+  long i, m, emax, l = lg(Vga);
+  GEN P, B = RgV_MOD2(Vga), V = cgetg(l, t_VEC);
+  P = gen_indexsort(B, (void*)lexcmp, cmp_nodata);
+  for (i = m = 1; i < l;)
   {
     GEN u = gel(B, P[i]);
     long k;
-    for(k = i+1; k <= d; ++k)
-      if (!gequal(gel(B, P[k]), u)) break;
-    gel(V, m++) = vecpermute(Vga, vecslice(P,i,k-1));
+    for(k = i+1; k < l; k++)
+    {
+      GEN v = gsub(u, gel(B, P[k]));
+      if (!gequal0(v) && (!isinexactreal(v) || gexpo(v) > -bit)) break;
+    }
+    gel(V, m++) = vecslice(P,i,k-1);
     i = k;
   }
-  setlg(V, m); return V;
+  setlg(V, m); emax = 0;
+  for (i = 1; i < m; i++)
+  {
+    long j, e = 0, li = lg(gel(V,i))-1;
+    GEN b = gel(B, gel(V,i)[1]);
+    for (j = 1; j < m; j++)
+      if (j != i) e -= gexpo(gsub(gel(B, gel(V,j)[1]), b));
+    emax = maxss(emax, e * li);
+  }
+  for (i = 1; i < m; i++) gel(V,i) = vecpermute(Vga, gel(V,i));
+  *pdV = emax; return V;
 }
 
 static GEN
 sercoeff(GEN x, long n, long prec)
 {
   long N = n - valp(x);
-  return (N < 0)? gen_0: gtofp(gel(x, N+2), prec);
+  return (N < 0)? gen_0: gprec_wtrunc(gel(x, N+2), prec);
 }
 
+/* prod_i Gamma(s/2 + (m+LA[i])/2), set t *= prod_i (s/2 + (m+LA[i])/2) */
+static GEN
+get_gamma(GEN *pt, GEN LA, GEN m, int round, long precdl, long prec)
+{
+  long i, l = lg(LA);
+  GEN pr = NULL, t = *pt;
+  for (i = 1; i < l; i++)
+  {
+    GEN u, g, a = gmul2n(gadd(m, gel(LA,i)), -1);
+    if (round) a = ground(a);
+    u = deg1pol_shallow(ghalf, a, 0);
+    g = ggamma(RgX_to_ser(u, precdl), prec);
+    pr = pr? gmul(pr, g): g;
+    t = t? gmul(t, u): u;
+  }
+  *pt = t; return pr;
+}
 /* generalized power series expansion of inverse Mellin around x = 0;
  * m-th derivative */
 static GEN
-Kderivsmallinit(GEN Vga, long m, long bitprec)
+Kderivsmallinit(GEN ldata, GEN Vga, long m, long bit)
 {
-  pari_sp av = avma;
-  long d = lg(Vga)-1, N, j, l, limn, prec;
-  GEN LA, lj, mj, mat;
-  double C2 = MELLININV_CUTOFF;
+  const double C2 = MELLININV_CUTOFF;
+  long prec2, N, j, l, dLA, limn, d = lg(Vga)-1;
+  GEN piA, LA, L, M, mat;
 
-  LA = gammapoles(Vga);
-  N = lg(LA)-1;
-  lj = cgetg(N+1, t_VECSMALL);
-  mj = cgetg(N+1, t_VEC);
-  for (j = 1; j <= N; ++j)
-  {
-    GEN L = gel(LA,j);
-    lj[j] = lg(L)-1;
-    gel(mj,j) = gsubsg(2, vecmin(L));
-  }
-  prec = nbits2prec((long)(1+bitprec*(1+M_PI*d/C2)));
-  limn = ceil(2*M_LN2*bitprec/(d*dbllambertW0(C2/(M_PI*M_E))));
+  LA = gammapoles(Vga, &dLA, bit); N = lg(LA)-1;
+  prec2 = nbits2prec(dLA + bit * (1 + M_PI*d/C2));
+#if BITS_IN_LONG == 32
+  prec2 += prec2 & 1L;
+#endif
+  if (ldata) Vga = ldata_get_gammavec(ldata_newprec(ldata, prec2));
+  L = cgetg(N+1, t_VECSMALL);
+  M = cgetg(N+1, t_VEC);
   mat = cgetg(N+1, t_VEC);
+  limn = ceil(2*M_LN2*bit / (d * dbllambertW0(C2/(M_PI*M_E))));
   l = limn + 2;
-  for (j=1; j <= N; j++)
+  for (j = 1; j <= N; j++)
   {
-    GEN C, c, mjj = gel(mj,j), pr = gen_1, t = gen_1;
-    long i, k, n, ljj = lj[j], lprecdl = ljj+3;
-    for (i=1; i <= d; i++)
-    {
-      GEN a = gmul2n(gadd(mjj, gel(Vga,i)), -1);
-      GEN u = deg1pol_shallow(ghalf, a, 0);
-      pr = gmul(pr, ggamma(RgX_to_ser(u, lprecdl), prec));
-      t = gmul(t, u);
-    }
-    c = cgetg(limn+2,t_COL);
-    gel(c,1) = pr;
-    for (n=1; n <= limn; n++)
-      gel(c,n+1) = gdiv(gel(c,n), RgX_translate(t, stoi(-2*n)));
+    GEN S = gel(LA,j);
+    GEN C, c, mj, G = NULL, t = NULL, tj = NULL;
+    long i, k, n, jj, lj = L[j] = lg(S)-1, precdl = lj+3;
 
-    gel(mat, j) = C = cgetg(ljj+1, t_COL);
-    for (k = 1; k <= ljj; k++)
+    gel(M,j) = mj = gsubsg(2, gel(S, vecindexmin(real_i(S))));
+    for (jj = 1; jj <= N; jj++)
+    {
+      GEN g;
+      if (jj == j) /* poles come from this class only */
+        g = get_gamma(&tj, gel(LA,jj), mj, 1, precdl, prec2);
+      else
+        g = get_gamma(&t, gel(LA,jj), mj, 0, precdl, prec2);
+      G = G? gmul(G, g): g;
+    }
+    c = cgetg(limn+2,t_COL); gel(c,1) = G;
+    for (n=1; n <= limn; n++)
+    {
+      GEN A = utoineg(2*n), T = RgX_translate(tj, A);
+      /* T = exact polynomial, may vanish at 0 (=> pole in c[n+1]) */
+      if (t) T = RgX_mul(T, RgX_translate(t, A)); /* no pole here */
+      gel(c,n+1) = gdiv(gel(c,n), T);
+    }
+    gel(mat, j) = C = cgetg(lj+1, t_COL);
+    for (k = 1; k <= lj; k++)
     {
       GEN L = cgetg(l, t_POL);
-      L[1] = evalsigne(1)|evalvarn(0);
-      for (n = 2; n < l; n++) gel(L,n) = sercoeff(gel(c,n), -k, prec);
-      gel(C,k) = L;
+      for (n = 2; n < l; n++) gel(L,n) = sercoeff(gel(c,n), -k, prec2);
+      L[1] = evalsigne(1)|evalvarn(0); gel(C,k) = L;
     }
-    /* C[k] = \sum_n c_{j,k} t^n =: C_k(t) in Dokchitser's Algo 3.3 */
-    /* Take m-th derivative of t^(-mj+2) sum_k (-ln t)^k/k! C_k(t^2) */
+    /* C[k] = \sum_n c_{j,k} t^n =: C_k(t) in Dokchitser's Algo 3.3
+     * m-th derivative of t^(-M+2) sum_k (-ln t)^k/k! C_k(t^2) */
     if (m)
     {
-      mjj = gsubgs(mjj, 2);
-      for (i = 1; i <= m; i++, mjj = gaddgs(mjj, 1))
-      {
-        for (k = 1; k <= ljj; k++)
+      mj = gsubgs(mj, 2);
+      for (i = 1; i <= m; i++, mj = gaddgs(mj,1))
+        for (k = 1; k <= lj; k++)
         {
           GEN c = gel(C,k), d = RgX_shift_shallow(gmul2n(RgX_deriv(c),1), 1);
-          c = RgX_Rg_mul(c, mjj);
-          if (k < ljj) c = RgX_add(c, gel(C,k+1));
+          c = RgX_Rg_mul(c, mj);
+          if (k < lj) c = RgX_add(c, gel(C,k+1));
           gel(C,k) = RgX_sub(d, c);
         }
-      }
-      gel(mj,j) = gaddgs(mjj,2);
+      gel(M,j) = gaddgs(mj,2);
     }
-    for (k = 1; k <= ljj; k++)
+    for (k = 1; k <= lj; k++)
     {
       GEN c = gel(C,k);
       if (k > 2) c = RgX_Rg_div(c, mpfact(k-1));
@@ -226,7 +262,9 @@ Kderivsmallinit(GEN Vga, long m, long bitprec)
     }
   }
   /* Algo 3.3: * \phi^(m)(t) = sum_j t^m_j sum_k (-ln t)^k mat[j,k](t^2) */
-  return gerepilecopy(av, mkvec3(lj,RgV_neg(mj),mat));
+  piA = gsubsg(m*d, sumVga(Vga));
+  if (!gequal0(piA)) piA = powPis(gmul2n(piA,-1), prec2);
+  return mkvec5(L, RgV_neg(M), mat, mkvecsmall(prec2), piA);
 }
 
 /* Evaluate a vector considered as a polynomial using Horner. Unstable!
@@ -255,8 +293,10 @@ get_tmax(long bitprec)
 static GEN
 GMi_get_Vga(GEN K) { return gel(K,2); }
 static long
+GMi_get_degree(GEN K) { return lg(gel(K,2))-1; }
+static long
 GMi_get_m(GEN K) { return itos( gel(K,3) ); }
-static GEN /* [lj,mj,mat], Kderivsmall only */
+static GEN /* [lj,mj,mat,prec2], Kderivsmall only */
 GMi_get_VS(GEN K) { return gel(K,4); }
 static GEN /* [Ms,cd,A2], Kderivlarge only */
 GMi_get_VL(GEN K) { return gel(K,5); }
@@ -270,22 +310,18 @@ GMi_get_tmax(GEN K, long bitprec)
 static GEN
 Kderivsmall(GEN K, GEN x, GEN x2d, long bitprec)
 {
-  pari_sp ltop = avma;
-  GEN Vga = GMi_get_Vga(K), VS = GMi_get_VS(K);
-  GEN lj = gel(VS,1), mj = gel(VS,2), mat = gel(VS,3);
-  GEN d2, Lx, x2, x2i, A, S, pi;
-  long prec, d, N, j, k, limn, m = GMi_get_m(K);
-  double Ed, xd, Wd;
+  GEN VS = GMi_get_VS(K), L = gel(VS,1), M = gel(VS,2), mat = gel(VS,3);
+  GEN d2, Lx, x2, x2i, S, pi, piA = gel(VS,5);
+  long prec = gel(VS,4)[1], d = GMi_get_degree(K);
+  long j, k, limn, N = lg(L)-1;
+  double xd, Wd, Ed = M_LN2*bitprec / d;
 
-  N = lg(lj)-1; d = lg(Vga)-1; A = vecsum(Vga);
-  Ed = M_LN2*bitprec / d;
   xd = maxdd(M_PI*dblmodulus(x2d), 1E-13); /* pi |x|^2/d unless x tiny */
   if (xd > Ed) pari_err_BUG("Kderivsmall (x2d too large)");
   /* Lemma 5.2.6 (2), a = 1 + log(Pi x^(2/d)) = log(e / xd),
    * B = log(2)*bitprec / d = Ed */
   Wd = dbllambertW0( Ed / (M_E*xd) ); /* solution of w exp(w) = B exp(-a)*/
   limn = (long) ceil(2*Ed/Wd);
-  prec = nbits2prec((long) ceil(bitprec+d*xd/M_LN2));
   pi = mppi(prec);
   d2 = gdivsg(d,gen_2);
   if (x)
@@ -294,20 +330,19 @@ Kderivsmall(GEN K, GEN x, GEN x2d, long bitprec)
     x = gpow(gmul(gtofp(x2d,prec),pi), d2, prec);
   /* at this stage, x has been replaced by pi^(d/2) x */
   x2 = gsqr(x);
-  Lx = gpowers(gneg(glog(x,prec)), vecsmall_max(lj));
+  Lx = gpowers(gneg(glog(x,prec)), vecsmall_max(L));
   x2i = (gcmp(gnorml2(x2), gen_1) <= 0)? NULL: ginv(x2);
   S = gen_0;
   for (j = 1; j <= N; ++j)
   {
-    long ljj = lj[j];
+    long lj = L[j];
     GEN s = gen_0;
-    for (k = 1; k <= ljj; k++)
+    for (k = 1; k <= lj; k++)
       s = gadd(s, gmul(gel(Lx,k), evalvec(gmael(mat,j,k), limn, x2, x2i)));
-    S = gadd(S, gmul(gpow(x, gel(mj,j), prec), s));
+    S = gadd(S, gmul(gpow(x, gel(M,j), prec), s));
   }
-  A = gsubsg(m*d, A);
-  if (!gequal0(A)) S = gmul(S, gsqrt(gpow(pi, A, prec), prec));
-  return gerepileupto(ltop, gtofp(S, nbits2prec(bitprec)));
+  if (!gequal0(piA)) S = gmul(S, piA);
+  return S;
 }
 
 /* In Klarge, we conpute K(t) as (asymptotic) * F(z), where F ~ 1 is given by
@@ -328,8 +363,8 @@ get_D(long d) { return d <= 2 ? 157. : 180.; }
 static void
 Kderivlarge_optim(GEN K, long abs, GEN t2d,GEN gcd, long *pbitprec, long *pnlim)
 {
-  GEN Vga = GMi_get_Vga(K), VL = GMi_get_VL(K), A2 = gel(VL,3);
-  long bitprec = *pbitprec, d = lg(Vga)-1;
+  GEN VL = GMi_get_VL(K), A2 = gel(VL,3);
+  long bitprec = *pbitprec, d = GMi_get_degree(K);
   const double D = get_D(d), td = dblmodulus(t2d), cd = gtodouble(gcd);
   double a, rtd, E = M_LN2*bitprec;
 
@@ -356,9 +391,8 @@ Kderivlarge_optim(GEN K, long abs, GEN t2d,GEN gcd, long *pbitprec, long *pnlim)
 static GEN
 Kderivlarge(GEN K, GEN t, GEN t2d, long bitprec0)
 {
-  pari_sp ltop = avma;
-  GEN tdA, P, S, pi, z, Vga = GMi_get_Vga(K);
-  const long d = lg(Vga)-1;
+  GEN tdA, P, S, pi, z;
+  const long d = GMi_get_degree(K);
   GEN M, VL = GMi_get_VL(K), Ms = gel(VL,1), cd = gel(VL,2), A2 = gel(VL,3);
   long status, prec, nlim, m = GMi_get_m(K), bitprec = bitprec0;
 
@@ -397,31 +431,80 @@ Kderivlarge(GEN K, GEN t, GEN t2d, long bitprec0)
     }
     if (status == 1) S = gmul(S, gsubsg(1, ginv(gmul(z, pi))));
   }
-  return gerepileupto(ltop, gmul(P, S));
+  return gmul(P, S);
 }
 
 /* Dokchitser's coefficients used for asymptotic expansion of inverse Mellin
- * 2 <= p <= min(n+1, d) */
+ * 2 <= p <= min(n+1, d), c = 2n-p+1; sh = (sh(x)/x)^(d-p) */
 static GEN
-fun_vp(long p, long n, long d, GEN SM, GEN vsinh)
+vp(long p, long c, GEN SMd, GEN sh)
 {
-  pari_sp ltop = avma;
+  GEN s, ve = cgetg(p+2, t_VEC);
   long m, j, k;
-  GEN s = gen_0;
-  for (m = 0; m <= p; ++m)
+
+  gel(ve,1) = gen_1; gel(ve,2) = utoipos(c);
+  for (j = 2; j <= p; j++) gel(ve,j+1) = gdivgs(gmulgs(gel(ve,j), c), j);
+  s = gel(SMd, 1);
+  for (m = 1; m <= p; m++)
   {
-    GEN pr = gen_1, s2 = gen_0, sh = gel(vsinh, d-p+1);/* (sh(x)/x)^(d-p) */
-    long pm = p-m;
-    for (j = m; j < p; ++j) pr = muliu(pr, d-j);
-    for (k = 0; k <= pm; k+=2)
-    {
-      GEN e = gdiv(powuu(2*n-p+1, pm-k), mpfact(pm-k));
-      s2 = gadd(s2, gmul(e, RgX_coeff(sh, k)));
-    }
-    s = gadd(s, gmul(gmul(gel(SM, m+1), pr), s2));
-    if (gc_needed(ltop, 1)) s = gerepilecopy(ltop, s);
+    GEN t, c = gel(SMd, m+1);
+    if (gequal0(c)) continue;
+    t = gel(ve, m+1);
+    for (k = 1; k <= m/2; k++)
+      t = gadd(t, gmul(gel(ve, m-2*k+1), RgX_coeff(sh, k)));
+    s = gadd(s, gmul(c, t));
   }
-  return gerepileupto(ltop, gmul(gdivsg(-d, powuu(2*d, p)), s));
+  return s;
+}
+
+static GEN
+get_SM(GEN Vga)
+{
+  long k, m, d = lg(Vga)-1;
+  GEN pol, nS1, SM, C, t = vecsum(Vga);
+
+  pol = roots_to_pol(gmulgs(Vga, -d), 0); /* deg(pol) = d */
+  SM = cgetg(d+2, t_VEC); gel(SM,1) = gen_1;
+  if (gequal0(t))
+  { /* shortcut */
+    for (m = 1; m <= d; m++) gel(SM,m+1) = gel(pol,d+2-m);
+    return SM;
+  }
+  nS1 = gpowers(gneg(t), d); C = matpascal(d);
+  for (m = 1; m <= d; m++)
+  {
+    pari_sp av = avma;
+    GEN s = gmul(gel(nS1, m+1), gcoeff(C, d+1, m+1));
+    for (k = 1; k <= m; k++)
+    {
+      GEN e = gmul(gel(nS1, m-k+1), gcoeff(C, d-k+1, m-k+1));
+      s = gadd(s, gmul(e, RgX_coeff(pol, d-k)));
+    }
+    gel(SM, m+1) = gerepileupto(av, s);
+  }
+  return SM;
+}
+
+static GEN
+get_SMd(GEN Vga)
+{
+  GEN M, SM = get_SM(Vga);
+  long p, m, d = lg(Vga)-1;
+
+  M = cgetg(d, t_VEC);
+  for (p = 2; p <= d; p++)
+  {
+    GEN a = gen_1, c;
+    long D = d - p;
+    gel(M, p-1) = c = cgetg(p+2, t_COL);
+    gel(c, 1) = gel(SM, p+1);
+    for (m = 1; m <= p; m++)
+    {
+      a = muliu(a, D + m);
+      gel(c, m+1) = gmul(gel(SM, p-m+1), a);
+    }
+  }
+  return M;
 }
 
 /* Asymptotic expansion of inverse Mellin, to length nlimmax. Set status = 0
@@ -431,59 +514,43 @@ fun_vp(long p, long n, long d, GEN SM, GEN vsinh)
  * If status = 2, the asymptotic expansion is finite so return only
  * the necessary number of terms nlim <= nlimmax + d. */
 static GEN
-Klargeinit0(GEN Vga, long nlimmax, long *status)
+Klargeinit(GEN Vga, long nlimmax, long *status)
 {
-  const long prec = LOWDEFAULTPREC;
-  const long d = lg(Vga)-1;
-  long k, n, m, cnt;
-  GEN pol, SM, nS1, se, vsinh, M, dk;
+  long d = lg(Vga) - 1, p, n, cnt;
+  GEN M, SMd, se, vsinh, vd;
 
-  if (d==1 || (d==2 && gequal1(gabs(gsub(gel(Vga,1), gel(Vga,2)), prec))))
-  { /* shortcut */
-    *status = 2; return mkvec(gen_1);
-  }
+  if (Vgaeasytheta(Vga)) { *status = 2; return mkvec(gen_1); }
   /* d >= 2 */
   *status = 0;
-  pol = roots_to_pol(gneg(Vga), 0); /* deg(pol) = d */
-  nS1 = gpowers(gneg(RgX_coeff(pol, d-1)), d);
-  dk = gpowers(utoi(d), d-1);
-  SM = cgetg(d+3, t_VEC);
-  for (m = 0; m <= d; ++m)
-  {
-    pari_sp btop = avma;
-    GEN s = gmul(gdivgs(gel(nS1, m+1), d), binomialuu(d, m));
-    for (k = 1; k <= m; ++k)
-    {
-      GEN e = gmul(gel(nS1, m-k+1), gel(dk, k));
-      s = gadd(s, gmul(gmul(e, binomialuu(d-k, m-k)), RgX_coeff(pol, d-k)));
-    }
-    gel(SM, m+1) = gerepileupto(btop, s);
-  }
-  se = gdiv(gsinh(RgX_to_ser(pol_x(0), d+2), prec), pol_x(0));
+  SMd = get_SMd(Vga);
+  se = gsinh(RgX_to_ser(pol_x(0), d+2), 0); setvalp(se,0);
+  se = gdeflate(se, 0, 2); /* se(x^2) = sinh(x)/x */
   vsinh = gpowers(se, d);
-  M = vectrunc_init(nlimmax + d);
-  vectrunc_append(M, gen_1);
-  for (n=2, cnt=0; (n <= nlimmax) || cnt; ++n)
+  vd = gpowers(utoipos(2*d), d);
+  M = cgetg(nlimmax + d + 1, t_VEC); gel(M,1) = gen_1;
+  for (n = 2, cnt = 0; n <= nlimmax || cnt; n++)
   {
-    pari_sp btop = avma;
-    long p, ld = minss(d, n);
+    pari_sp av = avma;
+    long ld = minss(d, n);
     GEN s = gen_0;
-    for (p = 2; p <= ld; ++p)
-      s = gadd(s, gmul(fun_vp(p, n-1, d, SM, vsinh), gel(M, n+1-p)));
-    s = gerepileupto(btop, gdivgs(s, n-1));
-    vectrunc_append(M, s);
-    if (!isintzero(s))
+    for (p = 2; p <= ld; p++)
     {
-      if (n >= nlimmax) break;
-      cnt = 0;
+      GEN z = vp(p, 2*n-1-p, gel(SMd, p-1), gel(vsinh, d-p+1));
+      s = gadd(s, gmul(gdiv(z, gel(vd, p+1)), gel(M, n+1-p)));
+    }
+    gel(M,n) = s = gerepileupto(av, gdivgs(s, 1-n));
+    if (gequal0(s))
+    {
+      cnt++; *status = 1;
+      if (cnt >= d-1) { *status = 2; n -= d-2; break; }
     }
     else
     {
-      cnt++; *status = 1;
-      if (cnt >= d-1) { *status = 2; setlg(M, lg(M) - (d-1)); break; }
+      if (n >= nlimmax) { n++; break; }
+      cnt = 0;
     }
   }
-  return M;
+  setlg(M, n); return M;
 }
 
 /* remove trailing zeros from vector. */
@@ -502,30 +569,40 @@ stripzeros(GEN M)
 static GEN
 gammamellininvasymp_i(GEN Vga, long nlimmax, long m, long *status)
 {
-  pari_sp ltop = avma;
   GEN M, A, Aadd;
   long d, i, nlim, n;
 
-  M = Klargeinit0(Vga, nlimmax, status);
-  if (!m) return gerepilecopy(ltop, M);
+  M = Klargeinit(Vga, nlimmax, status);
+  if (!m) return M;
   d = lg(Vga)-1;
   /* half the exponent of t in asymptotic expansion. */
-  A = gdivgs(gaddsg(1-d, vecsum(Vga)), 2*d);
+  A = gdivgs(gaddsg(1-d, sumVga(Vga)), 2*d);
   if (*status == 2) M = shallowconcat(M, zerovec(m));
   nlim = lg(M)-1;
-  Aadd = gdivgs(stoi(2-d), 2*d); /* (1/d) - (1/2) */
+  Aadd = sstoQ(2-d, 2*d); /* (1/d) - (1/2) */
   for (i = 1; i <= m; i++, A = gadd(A,Aadd))
     for (n = nlim-1; n >= 1; --n)
       gel(M, n+1) = gsub(gel(M, n+1),
-                         gmul(gel(M, n), gsub(A, gdivgs(stoi(n-1), d))));
-  stripzeros(M);
-  return gerepilecopy(ltop, M);
+                         gmul(gel(M, n), gsub(A, sstoQ(n-1, d))));
+  stripzeros(M); return M;
+}
+static GEN
+get_Vga(GEN x, GEN *ldata)
+{
+  *ldata = lfunmisc_to_ldata_shallow_i(x);
+  if (*ldata) x = ldata_get_gammavec(*ldata);
+  return x;
 }
 GEN
-gammamellininvasymp(GEN Vga, long nlimmax, long m)
-{ long status;
-  if (!is_vec_t(typ(Vga))) pari_err_TYPE("gammamellininvinit",Vga);
-  return gammamellininvasymp_i(Vga, nlimmax, m, &status);
+gammamellininvasymp(GEN Vga, long nlim, long m)
+{
+  pari_sp av = avma;
+  long status;
+  GEN ldata;
+  Vga = get_Vga(Vga, &ldata);
+  if (!is_vec_t(typ(Vga)) || lg(Vga) == 1)
+    pari_err_TYPE("gammamellininvasymp",Vga);
+  return gerepilecopy(av, gammamellininvasymp_i(Vga, nlim, m, &status));
 }
 
 /* Does the continued fraction of the asymptotic expansion M at oo of inverse
@@ -553,18 +630,19 @@ ishankelspec(GEN Vga, GEN M)
 GEN
 gammamellininvinit(GEN Vga, long m, long bitprec)
 {
+  const double C2 = MELLININV_CUTOFF;
   pari_sp ltop = avma;
-  GEN A2, M, VS, VL, cd;
-  long d = lg(Vga)-1, status;
-  const double C2 = MELLININV_CUTOFF, D = get_D(d);
+  GEN A2, M, VS, VL, cd, ldata;
+  long nlimmax, status, d, prec = nbits2prec((4*bitprec)/3);
   double E = M_LN2*bitprec, tmax = get_tmax(bitprec); /* = E/C2 */
-  const long nlimmax = ceil(E*log2(1+M_PI*tmax)*C2/D);
 
-  if (!is_vec_t(typ(Vga))) pari_err_TYPE("gammamellininvinit",Vga);
-  A2 = gaddsg(m*(2-d) + 1-d, vecsum(Vga));
+  Vga = get_Vga(Vga, &ldata); d = lg(Vga)-1;
+  if (!is_vec_t(typ(Vga)) || !d) pari_err_TYPE("gammamellininvinit",Vga);
+  nlimmax = ceil(E * log2(1+M_PI*tmax) * C2 / get_D(d));
+  A2 = gaddsg(m*(2-d) + 1-d, sumVga(Vga));
   cd = (d <= 2)? gen_2: gsqrt(gdivgs(int2n(d+1), d), nbits2prec(bitprec));
   /* if in Klarge, we have |t| > tmax = E/C2, thus nlim < E*C2/D. */
-  M = gammamellininvasymp_i(Vga, nlimmax, m, &status);
+  M = gammamellininvasymp_i(RgV_gtofp(Vga, prec), nlimmax, m, &status);
   if (status == 2)
   {
     tmax = -1.; /* only use Klarge */
@@ -572,8 +650,7 @@ gammamellininvinit(GEN Vga, long m, long bitprec)
   }
   else
   {
-    long prec = nbits2prec((4*bitprec)/3);
-    VS = Kderivsmallinit(Vga, m, bitprec);
+    VS = Kderivsmallinit(ldata, Vga, m, bitprec);
     if (status == 0 && ishankelspec(Vga, M)) status = 1;
     if (status == 1)
     { /* a Hankel determinant vanishes => contfracinit is undefined.
@@ -612,10 +689,10 @@ gammamellininv(GEN K, GEN s, long m, long bitprec)
   pari_sp av = avma;
   GEN z, s2d;
   long d;
-  if (!is_vec_t(typ(K))) pari_err_TYPE("gammamellininv",K);
-  if (lg(K) != 6 || !is_vec_t(typ(GMi_get_Vga(K))))
+
+  if (!is_vec_t(typ(K)) || lg(K) != 6 || !is_vec_t(typ(GMi_get_Vga(K))))
     K = gammamellininvinit(K, m, bitprec);
-  d = lg(GMi_get_Vga(K))-1;
+  d = GMi_get_degree(K);
   s2d = gpow(s, gdivgs(gen_2, d), nbits2prec(bitprec));
   if (dblmodulus(s2d) < GMi_get_tmax(K, bitprec))
     z = Kderivsmall(K, s, s2d, bitprec);
