@@ -2157,7 +2157,7 @@ ZXQX_resultant_slice(GEN A, GEN B, GEN U, GEN dB, GEN P, GEN *mod)
     GEN u = ZX_to_Flx(U, p);
     GEN Hp = ZXQX_resultant_prime(a, b, dB, degA, degB, u, p);
     if (!Hp) { set_avma(av); *mod = gen_1; return pol_0(v); }
-    set_avma(av); *mod = utoipos(p); return Flx_to_ZX(Hp);
+    Hp = gerepileupto(av, Flx_to_ZX(Hp)); *mod = utoipos(p); return Hp;
   }
   T = ZV_producttree(P);
   A = ZXX_nv_mod_tree(A, P, T, v);
@@ -2695,7 +2695,7 @@ ZX_ZXY_resultant(GEN A, GEN B)
   if (DEBUGLEVEL>4) err_printf("bound for resultant coeffs: 2^%ld\n",bound);
   worker = snm_closure(is_entry("_ZX_ZXY_resultant_worker"),
                        mkvec4(A, B, dB? dB: gen_0,
-                              mkvecsmall5(degA, degB,dres, vY, sX)));
+                              mkvecsmall5(degA, degB, dres, sX, vY)));
   init_modular_big(&S);
   H = gen_crt("ZX_ZXY_resultant_all", worker, &S, dB, bound, 0, NULL,
                nxV_chinese_center, FpX_center_i);
@@ -2749,7 +2749,7 @@ ZX_ZXY_rnfequation(GEN A, GEN B, long *lambda)
   if (lambda)
   {
     *lambda = ZX_ZXY_rnfequation_lambda(A, B, *lambda);
-    B = RgX_translate(B, monomial(stoi(*lambda), 1, varn(A)));
+    if (*lambda) B = RgX_translate(B, monomial(stoi(*lambda), 1, varn(A)));
   }
   return ZX_ZXY_resultant(A,B);
 }
