@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -90,7 +91,7 @@ test_curve_order(norm_eqn_t ne, ulong a4, ulong a6,
         swapspec(n0, n1, N0, N1);
         swapped = 1; continue;
       }
-      avma = ltop; return 0;
+      return gc_long(ltop,0);
     }
 
     m0 *= n_s;
@@ -100,11 +101,11 @@ test_curve_order(norm_eqn_t ne, ulong a4, ulong a6,
     for ( ; x <= hasse_high; x += m0)
       if ((x % m1) == a1 && x != N0 && x != N1) break;
     /* every x in N was either N0 or N1, so we return true */
-    if (x > hasse_high) { avma = ltop; return 1; }
+    if (x > hasse_high) return gc_long(ltop,1);
 
     lswap(a4, a4t);
     lswap(a6, a6t);
-    lswap(m0, m1); avma = av;
+    lswap(m0, m1); set_avma(av);
   }
 }
 
@@ -765,9 +766,9 @@ find_j_inv_with_given_trace(
         found = 1; break;
       }
     }
-    avma = av;
+    set_avma(av);
   }
-  avma = ltop; return curves_tested;
+  return gc_long(ltop, curves_tested);
 }
 
 /*
@@ -789,7 +790,7 @@ next_generator(GEN DD, long D, ulong u, long filter, GEN *genred, long *P)
       /* If gen is in the principal class, skip it */
       *genred = redimag(gen);
       if (!equali1(gel(*genred,1))) { *P = (long)p; return gen; }
-      avma = av;
+      set_avma(av);
     }
   }
 }
@@ -880,7 +881,7 @@ evec_inverse(long e2[], const long e1[], const long n[], const long r[], long k)
     evec_reduce(e3, n, r, k);
   }
   evec_copy(e2, e4, k);
-  avma = av;
+  set_avma(av);
 }
 
 /* e1 and e2 may overlap */
@@ -913,7 +914,7 @@ evec_order(const long e[], const long n[], const long r[], long k)
     evec_reduce(f, n, r, k);
     o *= m;
   }
-  avma = av; return o;
+  return gc_long(av,o);
 }
 
 /* Computes orders o[] for each generator using relative orders n[]
@@ -930,7 +931,7 @@ evec_orders(long o[], const long n[], const long r[], long k)
     if (i) e[i - 1] = 0;
     o[i] = evec_order(e, n, r, k);
   }
-  avma = av;
+  set_avma(av);
 }
 
 INLINE int
@@ -961,7 +962,6 @@ evec_n_to_m(long m[], const long n[], long k)
   for (i = 1; i < k; ++i) m[i] = m[i - 1] * n[i];
 }
 
-
 /* Based on logfac() in Sutherland's classpoly package.
  * Ramanujan approximation to log(n!), accurate to O(1/n^3) */
 INLINE double
@@ -981,7 +981,7 @@ upper_bound_on_classpoly_coeffs(long D, long h, GEN qfinorms)
   pari_sp ltop = avma;
   GEN C = dbltor(2114.567);
   double Mk, m, logbinom;
-  GEN tmp = mulrr(mppi(LOWDEFAULTPREC), sqrtr(stor(-D, LOWDEFAULTPREC)));
+  GEN tmp = mulrr(mppi(LOWDEFAULTPREC), sqrtr(utor(-D, LOWDEFAULTPREC)));
   /* We treat this case separately since the table is not initialised when
    * h = 1. This is the same as in the for loop below but with ak = 1. */
   double log2Mk = dbllog2r(mpadd(mpexp(tmp), C));
@@ -1006,7 +1006,7 @@ upper_bound_on_classpoly_coeffs(long D, long h, GEN qfinorms)
     log2Mk = dbllog2r(mpadd(mpexp(divru(tmp, ak)), C));
     res += log2Mk;
     if (ak > maxak) { maxak = ak; log2Mh = log2Mk; }
-    avma = btop;
+    set_avma(btop);
   }
 
   Mk = pow(2.0, log2Mh);
@@ -1017,7 +1017,7 @@ upper_bound_on_classpoly_coeffs(long D, long h, GEN qfinorms)
   logbinom = (m > 0 && m < h)
     ? LOG2E * (logfac(h) - logfac(m) - logfac(h - m))
     : 0;
-  avma = ltop;
+  set_avma(ltop);
   return res + logbinom - m * log2Mh + 2.0;
 }
 
@@ -1039,20 +1039,20 @@ distinct_inverses(const long f[], const long ef[], const long ei[],
   for (j = i + 1; j < k; ++j) e2[j] = 0;
   evec_reduce(e2, n, r, k);
 
-  if (evec_equal(ef, e2, k)) { avma = av; return 0; }
+  if (evec_equal(ef, e2, k)) return gc_long(av,0);
 
   e3 = new_chunk(k);
   evec_inverse_o(e3, ef, n, o, r, k);
-  if (evec_equal(e2, e3, k)) { avma = av; return 0; }
+  if (evec_equal(e2, e3, k)) return gc_long(av,0);
 
   if (f) {
     evec_compose(e3, f, ei, n, r, k);
-    if (evec_equal(e2, e3, k)) { avma = av; return 0; }
+    if (evec_equal(e2, e3, k)) return gc_long(av,0);
 
     evec_inverse_o(e3, e3, n, o, r, k);
-    if (evec_equal(e2, e3, k)) { avma = av; return 0; }
+    if (evec_equal(e2, e3, k)) return gc_long(av,0);
   }
-  avma = av; return 1;
+  return gc_long(av,1);
 }
 
 INLINE long
@@ -1075,7 +1075,7 @@ next_prime_evec(long *qq, long f[], const long m[], long k,
   if (!he) pari_err_BUG("next_prime_evec");
   idx = itos((GEN) he->val);
   index_to_evec(f, idx, m, k);
-  avma = av; return 1;
+  return gc_long(av,1);
 }
 
 /* Return 1 on success, 0 on failure. */
@@ -1141,7 +1141,7 @@ orient_pcp(classgp_pcp_t G, long *ni, long D, long u, hashtable *tbl)
     if (!he) pari_err_BUG("orient_pcp");
     *ni = itos((GEN) he->val);
   }
-  avma = av; return 1;
+  return gc_bool(av,1);
 }
 
 /* We must avoid situations where L_i^{+/-2} = L_j^2 (or = L_0*L_j^2
@@ -1186,7 +1186,7 @@ classgp_pcp_check_generators(const long *n, long *r, long k, long L0)
       if (j == i) return i;
     }
   }
-  avma = av; return -1;
+  return gc_long(av,-1);
 }
 
 static void
@@ -1249,7 +1249,7 @@ classgp_make_pcp(
     v = const_vecsmall(1, 1);
     *height = upper_bound_on_classpoly_coeffs(D, h, v);
     /* NB: No need to set *ni when h = 1 */
-    avma = av; return;
+    set_avma(av); return;
   }
 
   DD = stoi(D);
@@ -1325,7 +1325,7 @@ classgp_make_pcp(
       pari_err_IMPL("classgp_pcp");
     else
       G->Lfilter *= L[i];
-    avma = bv;
+    set_avma(bv);
   }
 
   v = cgetg(h + 1, t_VECSMALL);
@@ -1345,25 +1345,21 @@ classgp_make_pcp(
   if (G->L0 && (G->L[0] != G->L0 || G->o[0] != 2))
     pari_err_BUG("classgp_pcp");
 
-  avma = av; return;
+  set_avma(av); return;
 }
 
 INLINE ulong
 classno_wrapper(long D)
 {
   pari_sp av = avma;
-  GEN clsgp = quadclassunit0(stoi(D), 0, NULL, DEFAULTPREC);
-  ulong h = itou(gel(clsgp, 1));
-  avma = av; return h;
+  GEN G = quadclassunit0(stoi(D), 0, NULL, DEFAULTPREC);
+  return gc_ulong(av, itou(abgrp_get_no(G)));
 }
 
 /*
  * SECTION: Functions for calculating class polynomials.
  */
 
-/* NB: Sutherland defines V_MAX to be 1200 with saying why. */
-#define V_MAX 1200
-
 #define NSMALL_PRIMES 11
 static const long SMALL_PRIMES[11] = {
   2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31
@@ -1398,107 +1394,83 @@ hclassno_wrapper(long D, long h)
     hurwitz = (double) h;
   else
     hurwitz = rtodbl(gtofp(hclassno(utoi(abs_D)), DEFAULTPREC));
-  avma = av; return hurwitz;
+  set_avma(av); return hurwitz;
 }
 
-
-/* This is Sutherland 2009, Algorithm 2.1 (p8).
- * NB: This function is not gerepileupto-safe. */
+/* This is Sutherland 2009, Algorithm 2.1 (p8); delta > 0 */
 static GEN
-select_classpoly_prime_pool(
-  double min_prime_bits, double delta, classgp_pcp_t G)
-{
-  pari_sp av;
-  double prime_bits = 0.0, hurwitz, z;
-  ulong i;
-  /* t_min[v] will hold the lower bound of the t we need to look at
-   * for a given v. */
-  ulong t_min[V_MAX], t_size_lim;
-  GEN res;
-  long D = G->D, inv = G->inv;
+select_classpoly_prime_pool(double min_bits, double delta, classgp_pcp_t G)
+{ /* Sutherland defines V_MAX to be 1200 without saying why */
+  const long V_MAX = 1200;
+  pari_sp av = avma;
+  double bits = 0.0, hurwitz, z;
+  ulong t_size_lim;
+  long ires, D = G->D, inv = G->inv;
+  GEN res, t_min; /* t_min[v] = lower bound for the t we look at for that v */
 
-  if (delta <= 0) pari_err_BUG("select_suitable_primes");
   hurwitz = hclassno_wrapper(D, G->h);
 
-  res = cgetg(1, t_VEC);
+  res = cgetg(128+1, t_VEC);
+  ires = 1;
   /* Initialise t_min to be all 2's.  This avoids trace 0 and trace 1 curves */
-  for (i = 0; i < V_MAX; ++i) t_min[i] = 2;
+  t_min = const_vecsmall(V_MAX-1, 2);
 
   /* maximum possible trace = sqrt(2^BIL - D) */
   t_size_lim = 2.0 * sqrt((double)((1UL << (BITS_IN_LONG - 2)) - (((ulong)-D) >> 2)));
 
-  av = avma;
-  for (z = -D / (2.0 * hurwitz); ; z *= delta + 1.0) {
-    /* v_bound_aux = -4 z H(-D). */
+  for (z = -D / (2.0 * hurwitz); ; z *= delta + 1.0)
+  { /* v_bound_aux = -4 z H(-D) */
     double v_bound_aux = -4.0 * z * hurwitz;
     ulong v;
     dbg_printf(1)("z = %.2f\n", z);
-    for (v = 1; ; ++v) {
-      ulong pcount = 0, t, t_max, vfactors;
-      ulong m_vsqr_D = v * v * (ulong)(-D);
+    for (v = 1; v < V_MAX; v++)
+    {
+      ulong p, t, t_max, vfactors, m_vsqr_D = v * v * (ulong)(-D);
       /* hurwitz_ratio_bound = 11 * log(log(v + 4))^2 */
       double hurwitz_ratio_bound = log(log(v + 4.0)), max_p, H;
+      long ires0;
       hurwitz_ratio_bound *= 11.0 * hurwitz_ratio_bound;
 
-      if (v >= v_bound_aux * hurwitz_ratio_bound / D || v >= V_MAX) break;
-
-      if ( ! is_smooth_enough(&vfactors, v)) continue;
+      if (v >= v_bound_aux * hurwitz_ratio_bound / D) break;
+      if (!is_smooth_enough(&vfactors, v)) continue;
       H = hclassno_wrapper(m_vsqr_D, 0);
 
       /* t <= 2 sqrt(p) and p <= z H(-v^2 D) and
-       *
        *   H(-v^2 D) < vH(-D) (11 log(log(v + 4))^2)
-       *
        * This last term is v * hurwitz * hurwitz_ratio_bound. */
       max_p = z * v * hurwitz * hurwitz_ratio_bound;
-      t_max = 2.0 * mindd(sqrt((double)((1UL << (BITS_IN_LONG - 2)) - (m_vsqr_D >> 2))),
-                          sqrt(max_p));
-      for (t = t_min[v]; t <= t_max; ++t) {
-        ulong possible_4p = t * t + m_vsqr_D;
-        if (possible_4p % 4 == 0) {
-          ulong possible_p = possible_4p / 4;
-          if (uisprime(possible_p) && modinv_good_prime(inv, possible_p)) {
-            long p = possible_p;
-            double rho_inv = p / H;
-            GEN hit;
-
-            hit = mkvecsmall5(p, t, v, (long)rho_inv, vfactors);
-            /* FIXME: Avoid doing GC for every prime as here. */
-            res = gerepileupto(av, gconcat(res, hit));
-            prime_bits += log2(p);
-            ++pcount;
-          }
+      t_max = 2.0 * sqrt(mindd((1UL<<(BITS_IN_LONG-2)) - (m_vsqr_D>>2), max_p));
+      t = t_min[v]; if ((t & 1) != (m_vsqr_D & 1)) t++;
+      p = (t * t + m_vsqr_D) >> 2;
+      ires0 = ires;
+      for (; t <= t_max; p += t+1, t += 2) /* 4p = t^2 - v^2*D */
+        if (modinv_good_prime(inv,p) && uisprime(p))
+        {
+          if (ires == lg(res)) res = vec_lengthen(res, lg(res) << 1);
+          gel(res, ires++) = mkvecsmall5(p, t, v, (long)(p / H), vfactors);
+          bits += log2(p);
         }
-      }
-      t_min[v] = t_max + 1;
+      t_min[v] = t;
 
-      if (pcount) {
-        dbg_printf(2)("  Found %lu primes for v = %lu.\n", pcount, v);
-        if (gc_needed(av, 2))
-          res = gerepilecopy(av, res);
+      if (ires - ires0) {
+        dbg_printf(2)("  Found %lu primes for v = %lu.\n", ires - ires0, v);
       }
-      if (prime_bits > min_prime_bits) {
-        dbg_printf(1)("Found %ld primes; total size %.2f bits.\n",
-                    glength(res), prime_bits);
-        return gerepilecopy(av, res);
+      if (bits > min_bits) {
+        dbg_printf(1)("Found %ld primes; total size %.2f bits.\n", ires-1,bits);
+        setlg(res, ires); return gerepilecopy(av, res);
       }
     }
-    /* Have we exhausted all possible solutions that fit in machine words? */
-    if (t_min[1] >= t_size_lim) {
+    if (uel(t_min,1) >= t_size_lim) {
+      /* exhausted all solutions that fit in ulong */
       char *err = stack_sprintf("class polynomial of discriminant %ld", D);
       pari_err(e_ARCH, err);
     }
   }
 }
 
-INLINE int
-cmp_small(long a, long b)
-{ return a>b? 1: (a<b? -1: 0); }
-
 static int
 primecmp(void *data, GEN v1, GEN v2)
-{ (void)data; return cmp_small(v1[4], v2[4]); }
-
+{ (void)data; return cmpss(v1[4], v2[4]); }
 
 static long
 height_margin(long inv, long D)
@@ -1602,9 +1574,9 @@ oneroot_of_classpoly(
     else if (lvl_diff > 0)
       /* otherwise j's level is greater than v(u) so we descend */
       j = descend_volcano(phi, j, p, pi, jlvl, L, depth, lvl_diff);
-    avma = bv;
+    set_avma(bv);
   }
-  avma = av;
+  set_avma(av);
   /* At this point the probability that j has the wrong endomorphism
    * ring is about \sum_{p|u_compl} 1/p (and u_compl must be bigger
    * than L_bound, so pretty big), so just return it and rely on
@@ -1758,7 +1730,7 @@ polclass_roots_modp(
     if ( ! res && endo_cert) pari_err_BUG("polclass_roots_modp");
     if (res && ! endo_cert && vecsmall_isin_skip(res, res[1], 2))
     {
-      avma = av;
+      set_avma(av);
       res = NULL;
     }
   } while (!res);
@@ -1787,8 +1759,7 @@ verify_edge(ulong j0, ulong j1, ulong p, ulong pi, long L, GEN fdb)
   pari_sp av = avma;
   GEN phi = polmodular_db_getp(fdb, L, p);
   GEN f = Flm_Fl_polmodular_evalx(phi, L, j1, p, pi);
-  ulong r = Flx_eval_pre(f, j0, p, pi);
-  avma = av; return !r;
+  return gc_long(av, Flx_eval_pre(f, j0, p, pi) == 0);
 }
 
 INLINE long
@@ -1803,7 +1774,7 @@ verify_2path(
   GEN d = Flx_gcd(f, g, p);
   long n = degpol(d);
   if (n >= 2) n = Flx_nbroots(d, p);
-  avma = av; return n;
+  return gc_long(av, n);
 }
 
 static long
@@ -1874,7 +1845,7 @@ oriented_n_action(
   /* Orient representation of [N] relative to the torsor <signs, rels> */
   for (i = 0; i < k; ++i) e[i] = (signs[i] < 0 ? o[i] - ni[i] : ni[i]);
   evec_reduce(e, n, rels, k);
-  avma = av; return evec_to_index(e, m, k);
+  return gc_long(av, evec_to_index(e,m,k));
 }
 
 /* F = double_eta_raw(inv) */
@@ -1883,7 +1854,7 @@ adjust_orientation(GEN F, long inv, GEN v, long e, ulong p, ulong pi)
 {
   ulong j0 = uel(v, 1), je = uel(v, e);
 
-  if (!modinv_j_from_2double_eta(F, inv, NULL, j0, je, p, pi)) {
+  if (!modinv_j_from_2double_eta(F, inv, j0, je, p, pi)) {
     if (modinv_inverted_involution(inv)) Flv_inv_pre_inplace(v, p, pi);
     if (modinv_negated_involution(inv)) Flv_neg_inplace(v, p);
   }
