@@ -1771,7 +1771,7 @@ all_roots(GEN p, long bit)
   pd = RgX_deflate_max(p, &h); elc = gexpo(leading_coeff(pd));
   fb = fujiwara_bound(pd);
   e = (fb < 0)? 0: (long)(2 * fb);
-  bit2 = bit + gexpo(pd) + (long)log2(n/h)+1+e;
+  bit2 = bit + maxss(0, gexpo(pd)) + (long)log2(n/h)+1+e;
   e = 0;
   for (av=avma,i=1;; i++,set_avma(av))
   {
@@ -2320,15 +2320,20 @@ usp(GEN Q0, long flag, long bitprec)
     { /* exactly one root */
       GEN s = gen_0;
       if (flag == 0)
+      {
         s = mkvec2(gmul2n(c,-k), gmul2n(addiu(c,1),-k));
+        s = gerepilecopy(av2, s);
+      }
       else if (flag == 1) /* Caveat: Qremapped is the reciprocal polynomial */
       {
         s = polsolve(*pQremapped, bitprec+1);
         s = addir(c, divrr(s, addsr(1, s)));
         shiftr_inplace(s, -k);
         if (realprec(s) != prec) s = rtor(s, prec);
+        s = gerepileupto(av2, s);
       }
-      gel(sol, ++nbr) = gerepileupto(av2, s);
+      else set_avma(av2);
+      gel(sol, ++nbr) = s;
     }
     else if (nb)
     { /* unknown, add two nodes to refine */
