@@ -797,12 +797,17 @@ subfields_of_given_degree(blockdata *B)
   return gc_const(av,L);
 }
 
+static void
+setvarn2(GEN t, long v) { setvarn(gel(t,1),v); setvarn(gel(t,2),v); }
 static GEN
-fix_var(GEN x, long v)
+fix_var(GEN x, long v, long fl)
 {
   long i, l = lg(x);
   if (!v) return x;
-  for (i=1; i<l; i++) { GEN t = gel(x,i); setvarn(gel(t,1),v); setvarn(gel(t,2),v); }
+  if (fl)
+    for (i = 1; i < l; i++) setvarn(gel(x,i), v);
+  else
+    for (i = 1; i < l; i++) setvarn2(gel(x,i), v);
   return x;
 }
 
@@ -877,7 +882,7 @@ subfieldsall(GEN nf0, long fl)
   (void)delete_var(); /* from init_primedata() */
   LSB = shallowconcat(LSB, fl==1? mkvec(T):_subfield(T, pol_x(0)));
   if (DEBUGLEVEL) err_printf("\n***** Leaving subfields\n\n");
-  return fix_var(gerepilecopy(av, LSB), v);
+  return fix_var(gerepilecopy(av, LSB), v, fl);
 }
 
 GEN
@@ -933,7 +938,7 @@ nfsubfields0(GEN nf0, long d, long fl)
   set_avma(av);
   if (!LSB) return cgetg(1, t_VEC);
   G = gcopy(LSB); gunclone(LSB);
-  return fix_var(G, v0);
+  return fix_var(G, v0, fl);
 }
 
 GEN
@@ -1017,6 +1022,7 @@ RgXY_to_RgC(GEN P, long dx, long dy)
 {
   GEN res, c;
   long i, j, k, d = degpol(P);
+  if (d > dy) pari_err_BUG("RgXY_to_RgC [incorrect degree]");
   res = cgetg((dx+1)*(dy+1)+1, t_COL);
   k = 1;
   for (i=0; i<=d; i++)
@@ -1025,6 +1031,7 @@ RgXY_to_RgC(GEN P, long dx, long dy)
     if (typ(c)==t_POL)
     {
       long dc = degpol(c);
+      if (dc > dx) pari_err_BUG("RgXY_to_RgC [incorrect degree]");
       for (j=0; j<=dc; j++)
         gel(res,k++) = gel(c,j+2);
     } else
@@ -1048,6 +1055,8 @@ twoembequation(GEN pol, GEN fa, GEN lambda)
   long i,j, lfa = lg(fa), dx = degpol(pol);
   long vx = varn(pol), vy = varn(gel(fa,1)); /* vx < vy ! */
 
+  if (varncmp(vx,vy) <= 0) pari_err_BUG("twoembequation [incorrect variable priorities]");
+
   lambda = shallowcopy(lambda);
   fa = shallowcopy(fa);
   j = 1;
@@ -1113,6 +1122,12 @@ subfields_cleanup(GEN* nf, GEN* pol, long* n, GEN* fa)
     *pol = nf_get_pol(*nf);
     *n = degpol(*pol);
   }
+  if(*fa)
+  {
+    long v = varn(*pol);
+    GEN o = gcoeff(*fa,1,1);
+    if (varncmp(varn(o),v) >= 0) pari_err_PRIORITY("nfsubfields_fa", o, "<=", v);
+  }
 }
 
 static GEN
@@ -1163,29 +1178,18 @@ subfields_get_fa(GEN pol, GEN nf, GEN fa)
 static long
 subfields_get_ero(GEN pol, GEN nf)
 {
-  GEN ro;
-  long ero, e, i;
-  if (nf)   ro = nf_get_roots(nf);
-  else      ro = roots(pol, LOWDEFAULTPREC);
-  ero = 0;
-  for (i=1; i<lg(ro); i++)
-  {
-    e = 1+gexpo(gel(ro,i));
-    if (e > ero) ero = e;
-  }
-  return ero+1;
-};
+  return 1 + gexpo(nf? nf_get_roots(nf):
+                       QX_complex_roots(pol, LOWDEFAULTPREC));
+}
 
 static GEN
 try_imag(GEN x, GEN c, GEN pol, long v, ulong p, GEN emb, GEN galpol, long fl)
 {
-  GEN a;
-  a = Q_primpart(RgX_sub(RgX_RgXQ_eval(x,c,pol),x));
+  GEN a = Q_primpart(RgX_sub(RgX_RgXQ_eval(x,c,pol),x));
   if (Flx_is_squarefree(Flxq_charpoly(ZX_to_Flx(a,p),ZX_to_Flx(pol,p),p),p))
   {
-    emb = RgX_RgXQ_eval(a, emb, galpol);
     pol = ZXQ_charpoly(a, pol, v);
-    return fl ? pol : mkvec2(pol, emb);
+    return fl ? pol : mkvec2(pol, RgX_RgXQ_eval(a, emb, galpol));
   }
   return NULL;
 }
@@ -1194,7 +1198,7 @@ static GEN
 galoissubfieldcm(GEN G, long fl)
 {
   pari_sp av = avma;
-  GEN c, H, elts, g, Hset, c2, gene, sub, pol, M, emb, a, galpol, B, b;
+  GEN c, H, elts, g, Hset, c2, gene, sub, pol, emb, a, galpol, B, b;
   long n, i, j, nH, ind, v, d;
   ulong p = 1009;
 
@@ -1261,16 +1265,24 @@ galoissubfieldcm(GEN G, long fl)
   c = galoispermtopol(G,c);
   if (d<n)
   {
-    c = RgX_to_RgC(c,n);
-    M = cgetg(d+1,t_MAT);
-    gel(M,1) = col_ei(n,1);
-    a = pol_1(v);
+    GEN M = cgetg(d+1,t_MAT), contc, contM;
+    gel(M,1) = col_ei(n,1); a = pol_1(v);
     for (i=2; i<=d; i++)
     {
-      a = ZX_rem(ZX_mul(a,emb), galpol);
+      a = RgX_rem(QX_mul(a,emb), galpol);
       gel(M,i) = RgX_to_RgC(a,n);
     }
-    c = RgM_RgC_invimage(M, c);
+    c = RgX_RgXQ_eval(emb,c,galpol);
+    c = Q_primitive_part(c,&contc);
+    c = RgX_to_RgC(c,n);
+    M = Q_primitive_part(M,&contM);
+    c = RgM_RgC_invimage(M,c);
+    if (contc)
+    {
+      if (contM) contc = gdiv(contc,contM);
+      c = RgV_Rg_mul(c, contc);
+    }
+    else if (contM) c = RgV_Rg_mul(c, ginv(contM));
     c = RgV_to_RgX(c, v);
   }
 
