@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -51,7 +52,7 @@ exiterr(const char *str)
 {
   term_color(c_ERR);
   err_printf("\n  *** X fatal error: %s\n",str);
-  term_color(c_NONE); exit(1);
+  term_color(c_NONE); _exit(1);
 }
 
 static long
@@ -249,7 +250,7 @@ draw(PARI_plot *T, GEN w, GEN x, GEN y)
 EXIT:
         XUnloadFont(display,font_info->fid);
         XFreeGC(display,gc);
-        XCloseDisplay(display); exit(0);
+        XCloseDisplay(display); _exit(0);
 
       case KeyRelease:
         /* Mod4 == Super on "std" Linux */
