@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -61,9 +62,44 @@ extern const long CATCH_ALL;
 #define pari_CATCH_reset() pari_CATCH2_reset(__iferr_old)
 #define pari_ENDCATCH pari_ENDCATCH2(__iferr_old)
 
-extern const double LOG2, LOG10_2, LOG2_10;
+#define pari_APPLY_same(EXPR)\
+  { \
+    long i, _l; \
+    GEN _y = cgetg_copy(x, &_l);\
+    for (i=1; i<_l; i++) gel(_y,i) = EXPR;\
+    return _y;\
+  }
+
+#define pari_APPLY_type(TYPE, EXPR)\
+  { \
+    long i, _l = lg(x); \
+    GEN _y = cgetg(_l, TYPE);\
+    for (i=1; i<_l; i++) gel(_y,i) = EXPR;\
+    return _y;\
+  }
+
+#define pari_APPLY_long(EXPR)\
+  { \
+    long i, _l = lg(x); \
+    GEN _y = cgetg(_l, t_VECSMALL);\
+    for (i=1; i<_l; i++) _y[i] = EXPR;\
+    return _y;\
+  }
+
+#define pari_APPLY_ulong(EXPR)\
+  { \
+    long i, _l = lg(x); \
+    GEN _y = cgetg(_l, t_VECSMALL);\
+    for (i=1; i<_l; i++) ((ulong*)_y)[i] = EXPR;\
+    return _y;\
+  }
+
+extern const double LOG10_2, LOG2_10;
 #ifndef  M_PI
-#  define M_PI 3.14159265358979323846
+#  define M_PI 3.14159265358979323846264338327950288
+#endif
+#ifndef  M_LN2
+#  define M_LN2 0.693147180559945309417232121458176568
 #endif
 
 /* Common global variables: */
@@ -71,8 +107,7 @@ extern int new_galois_format, factor_add_primes, factor_proven;
 extern ulong DEBUGFILES, DEBUGLEVEL, DEBUGMEM, precdl;
 extern long DEBUGVAR;
 extern ulong pari_mt_nbthreads;
-extern THREAD GEN  bernzone;
-extern GEN primetab;
+extern THREAD GEN  zetazone, bernzone, primetab;
 extern GEN gen_m1,gen_1,gen_2,gen_m2,ghalf,gen_0,gnil,err_e_STACK;
 extern THREAD VOLATILE int PARI_SIGINT_block, PARI_SIGINT_pending;
 
@@ -104,11 +139,11 @@ enum {
 
 #ifndef HAS_EXP2
 #  undef exp2
-#  define exp2(x) (exp((double)(x)*LOG2))
+#  define exp2(x) (exp((double)(x)*M_LN2))
 #endif
 #ifndef HAS_LOG2
 #  undef log2
-#  define log2(x) (log((double)(x))/LOG2)
+#  define log2(x) (log((double)(x))/M_LN2)
 #endif
 
 #define ONLY_REM ((GEN*)0x1L)
