@@ -5,7 +5,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -62,6 +63,16 @@ pari_kernel_init(void)
   mp_set_memory_functions((void *(*)(size_t)) pari_malloc, pari_gmp_realloc, pari_gmp_free);
 }
 
+const char *
+pari_kernel_version(void)
+{
+#ifdef gmp_version
+  return gmp_version;
+#else
+  return "";
+#endif
+}
+
 void
 pari_kernel_close(void)
 {
@@ -371,7 +382,7 @@ shiftispec(GEN x, long nx, long n)
       mpn_rshift((mp_limb_t *) yd, (mp_limb_t *) (x + d), nx - d, m);
       if (yd[ny - 1] == 0)
       {
-        if (ny == 1) { avma = (pari_sp)(yd + 1); return gen_0; }
+        if (ny == 1) return gc_const((pari_sp)(yd + 1), gen_0);
         ny--;
       }
     }
@@ -419,8 +430,8 @@ mantissa2nr(GEN x, long n)
       shift_right(y,x, 2,ly, 0,m);
       if (y[2] == 0)
       {
-        if (ly==3) { avma = (pari_sp)(y+3); return gen_0; }
-        ly--; avma = (pari_sp)(++y);
+        if (ly==3) return gc_const((pari_sp)(y+3), gen_0);
+        ly--; set_avma((pari_sp)(++y));
       }
     } else {
       for (i=2; i<ly; i++) y[i]=x[i];
@@ -448,7 +459,7 @@ truncr(GEN x)
     GEN z=cgeti(d);
     for (i=2; i<d; i++) z[d-i+1]=x[i];
     mpn_rshift(LIMBS(y),LIMBS(z),d-2,BITS_IN_LONG-m);
-    avma=(pari_sp)y;
+    set_avma((pari_sp)y);
   }
   return y;
 }
@@ -540,7 +551,7 @@ addumului(ulong a, ulong b, GEN y)
   hi=mpn_addmul_1(LIMBS(z), LIMBS(y), NLIMBS(y), b);
   if (hi) z[lz-1]=hi; else lz--;
   z[1] = evalsigne(1) | evallgefint(lz);
-  avma=(pari_sp)z; return z;
+  return gc_const((pari_sp)z, z);
 }
 
 /***********************************************************************/
@@ -656,8 +667,7 @@ divrr_with_gmp(GEN x, GEN y)
   else { w[2] = HIGHBIT; e++; }
   if (sy < 0) sx = -sx;
   w[1] = evalsigne(sx) | evalexpo(e);
-  avma=(pari_sp) w;
-  return w;
+  return gc_const((pari_sp)w, w);
 }
 
 /* We keep llx bits of x and lly bits of y*/
@@ -694,8 +704,7 @@ divri_with_gmp(GEN x, GEN y)
   else { w[2] = HIGHBIT; e++; }
   if (sy < 0) sx = -sx;
   w[1] = evalsigne(sx) | evalexpo(e);
-  avma=(pari_sp) w;
-  return w;
+  return gc_const((pari_sp)w, w);
 }
 
 GEN
@@ -871,7 +880,7 @@ dvmdii(GEN x, GEN y, GEN *z)
     if (z) *z = icopy(x);
     return gen_0;
   }
-DIVIDE: /* quotient is non-zero */
+DIVIDE: /* quotient is nonzero */
   av=avma; if (sx<0) sy = -sy;
   if (ly==3)
   {
@@ -882,8 +891,8 @@ DIVIDE: /* quotient is non-zero */
     if (q[lq - 1] == 0) lq--;
     if (z == ONLY_REM)
     {
-      avma=av; if (!si) return gen_0;
-      r=cgeti(3);
+      if (!si) return gc_const(av, gen_0);
+      set_avma(av); r = cgeti(3);
       r[1] = evalsigne(sx) | evallgefint(3);
       r[2] = si; return r;
     }
@@ -904,10 +913,10 @@ DIVIDE: /* quotient is non-zero */
     if (!r[lr - 1])
     {
       while(lr>2 && !r[lr - 1]) lr--;
-      if (lr == 2) {avma=av; return gen_0;} /* exact division */
+      if (lr == 2) return gc_const(av, gen_0); /* exact division */
     }
     r[1] = evalsigne(sx) | evallgefint(lr);
-    avma = (pari_sp) r; return r;
+    return gc_const((pari_sp)r, r);
   }
   else
   {
@@ -918,14 +927,14 @@ DIVIDE: /* quotient is non-zero */
     mpn_tdiv_qr(LIMBS(q), LIMBS(r),0, LIMBS(x), NLIMBS(x), LIMBS(y), NLIMBS(y));
     if (q[lq - 1] == 0) lq--;
     q[1] = evalsigne(sy) | evallgefint(lq);
-    if (!z) { avma = (pari_sp)q; return q; }
+    if (!z) return gc_const((pari_sp)q, q);
     if (!r[lr - 1])
     {
       while(lr>2 && !r[lr - 1]) lr--;
-      if (lr == 2) {avma=(pari_sp) q; *z=gen_0; return q;} /* exact division */
+      if (lr == 2) { *z = gen_0; return gc_const((pari_sp)q, q); } /* exact */
     }
     r[1] = evalsigne(sx) | evallgefint(lr);
-    avma = (pari_sp) r; *z = r; return q;
+    *z = gc_const((pari_sp)r, r); return q;
   }
 }
 
@@ -1010,7 +1019,7 @@ red_montgomery(GEN T, GEN N, ulong inv)
   /* copy result */
   Td = (GEN)av - 1; /* *Td = high word of final result */
   while (*Td == 0 && Te < Td) Td--; /* strip leading 0s */
-  k = Td - Te; if (!k) { avma = av; return gen_0; }
+  k = Td - Te; if (!k) return gc_const(av, gen_0);
   Td = (GEN)av - k; /* will write mantissa there */
   (void)memmove(Td, Te+1, k*sizeof(long));
   Td -= 2;
@@ -1026,7 +1035,7 @@ red_montgomery(GEN T, GEN N, ulong inv)
     || cmpii(Td, addii(shifti(T, -s), N)) >= 0) pari_err_BUG("red_montgomery");
 }
 #endif
-  avma = (pari_sp)Td; return Td;
+  return gc_const((pari_sp)Td, Td);
 }
 
 /* EXACT INTEGER DIVISION */
@@ -1138,7 +1147,6 @@ diviuuexact(GEN x, ulong y, ulong z)
   return diviiexact(x, tmp);
 }
 
-
 /********************************************************************/
 /**                                                                **/
 /**               INTEGER MULTIPLICATION                           **/
@@ -1304,7 +1312,7 @@ sqrtremi(GEN a, GEN *r)
     GEN R = cgeti(2 + na);
     nr = mpn_sqrtrem(LIMBS(S), LIMBS(R), LIMBS(a), na);
     if (nr) R[1] = evalsigne(1) | evallgefint(nr+2);
-    else    { avma = (pari_sp)S; R = gen_0; }
+    else    { set_avma((pari_sp)S); R = gen_0; }
     *r = R;
   }
   else
@@ -1346,10 +1354,10 @@ sqrtr_abs(GEN a)
       mpn_add_1(c,c,l,1);
   }
   xmpn_mirrorcopy(RLIMBS(res),c,l);
-  avma = (pari_sp)res; return res;
+  return gc_const((pari_sp)res, res);
 }
 
-/* Normalize a non-negative integer */
+/* Normalize a nonnegative integer */
 GEN
 int_normalize(GEN x, long known_zero_words)
 {
