@@ -2,40 +2,38 @@
 #define __AGM_ATAN_LIMIT                 60
 #define __DIVRR_GMP_LIMIT                4
 #define __EXPNEWTON_LIMIT                66
-#define __F2x_MUL_KARATSUBA_LIMIT        15
+#define __EXTGCD_HALFGCD_LIMIT           -1
+#define __F2x_MUL_KARATSUBA_LIMIT        11
+#define __F2x_MUL_MULII_LIMIT            41
 #define __F2xqX_BARRETT_LIMIT            48
-#define __F2xqX_DIVREM_BARRETT_LIMIT     292
-#define __F2xqX_INVBARRETT_LIMIT         46
-#define __F2xqX_REM_BARRETT_LIMIT        292
-#define __Flx_BARRETT_HALFMULII_LIMIT    21
-#define __Flx_BARRETT_KARATSUBA_LIMIT    1172
-#define __Flx_BARRETT_MULII2_LIMIT       16
-#define __Flx_BARRETT_MULII_LIMIT        448
-#define __Flx_BARRETT_QUARTMULII_LIMIT   23
-#define __Flx_DIVREM_BARRETT_LIMIT       768
-#define __Flx_EXTGCD_LIMIT               241
-#define __Flx_GCD_LIMIT                  1017
-#define __Flx_HALFGCD_HALFMULII_LIMIT    48
-#define __Flx_HALFGCD_KARATSUBA_LIMIT    77
-#define __Flx_HALFGCD_MULII2_LIMIT       25
-#define __Flx_HALFGCD_MULII_LIMIT        71
-#define __Flx_HALFGCD_QUARTMULII_LIMIT   52
-#define __Flx_INVBARRETT_HALFMULII_LIMIT 231
-#define __Flx_INVBARRETT_KARATSUBA_LIMIT 5067
-#define __Flx_INVBARRETT_MULII2_LIMIT    26
-#define __Flx_INVBARRETT_MULII_LIMIT     1154
-#define __Flx_INVBARRETT_QUARTMULII_LIMIT 139
-#define __Flx_MUL_HALFMULII_LIMIT        5
-#define __Flx_MUL_KARATSUBA_LIMIT        142
-#define __Flx_MUL_MULII2_LIMIT           5
-#define __Flx_MUL_MULII_LIMIT            7
-#define __Flx_MUL_QUARTMULII_LIMIT       5
-#define __Flx_REM_BARRETT_LIMIT          1266
-#define __Flx_SQR_HALFSQRI_LIMIT         3
-#define __Flx_SQR_KARATSUBA_LIMIT        316
-#define __Flx_SQR_QUARTSQRI_LIMIT        3
-#define __Flx_SQR_SQRI2_LIMIT            7
-#define __Flx_SQR_SQRI_LIMIT             5
+#define __F2xqX_DIVREM_BARRETT_LIMIT     97
+#define __F2xqX_EXTGCD_LIMIT             97
+#define __F2xqX_GCD_LIMIT                605
+#define __F2xqX_HALFGCD_LIMIT            127
+#define __F2xqX_INVBARRETT_LIMIT         50
+#define __F2xqX_REM_BARRETT_LIMIT        101
+#define __Flx_BARRETT2_LIMIT             31
+#define __Flx_BARRETT_LIMIT              92
+#define __Flx_DIVREM2_BARRETT_LIMIT      14
+#define __Flx_DIVREM_BARRETT_LIMIT       161
+#define __Flx_EXTGCD2_LIMIT              61
+#define __Flx_EXTGCD_LIMIT               158
+#define __Flx_GCD2_LIMIT                 1409
+#define __Flx_GCD_LIMIT                  852
+#define __Flx_HALFGCD2_LIMIT             36
+#define __Flx_HALFGCD_LIMIT              120
+#define __Flx_INVBARRETT2_LIMIT          22
+#define __Flx_INVBARRETT_LIMIT           200
+#define __Flx_MUL2_KARATSUBA_LIMIT       11
+#define __Flx_MUL2_MULII_LIMIT           8
+#define __Flx_MUL_KARATSUBA_LIMIT        33
+#define __Flx_MUL_MULII_LIMIT            30
+#define __Flx_REM2_BARRETT_LIMIT         89
+#define __Flx_REM_BARRETT_LIMIT          159
+#define __Flx_SQR2_KARATSUBA_LIMIT       15
+#define __Flx_SQR2_SQRI_LIMIT            14
+#define __Flx_SQR_KARATSUBA_LIMIT        93
+#define __Flx_SQR_SQRI_LIMIT             37
 #define __FlxqX_BARRETT_LIMIT            17
 #define __FlxqX_DIVREM_BARRETT_LIMIT     46
 #define __FlxqX_EXTGCD_LIMIT             44
@@ -59,6 +57,8 @@
 #define __FpX_REM_BARRETT_LIMIT          111
 #define __Fp_POW_BARRETT_LIMIT           127
 #define __Fp_POW_REDC_LIMIT              17
+#define __GCD_HALFGCD_LIMIT              -1
+#define __HALFGCD_LIMIT                  3
 #define __INVMOD_GMP_LIMIT               3
 #define __INVNEWTON_LIMIT                75
 #define __LOGAGMCX_LIMIT                 22
@@ -75,40 +75,38 @@
 #define __AGM_ATAN_LIMIT                 89
 #define __DIVRR_GMP_LIMIT                4
 #define __EXPNEWTON_LIMIT                197
-#define __F2x_MUL_KARATSUBA_LIMIT        23
-#define __F2xqX_BARRETT_LIMIT            17
-#define __F2xqX_DIVREM_BARRETT_LIMIT     46
-#define __F2xqX_INVBARRETT_LIMIT         22
-#define __F2xqX_REM_BARRETT_LIMIT        48
-#define __Flx_BARRETT_HALFMULII_LIMIT    23
-#define __Flx_BARRETT_KARATSUBA_LIMIT    905
-#define __Flx_BARRETT_MULII2_LIMIT       647
-#define __Flx_BARRETT_MULII_LIMIT        433
-#define __Flx_BARRETT_QUARTMULII_LIMIT   20
-#define __Flx_DIVREM_BARRETT_LIMIT       1289
-#define __Flx_EXTGCD_LIMIT               632
-#define __Flx_GCD_LIMIT                  2514
-#define __Flx_HALFGCD_HALFMULII_LIMIT    78
-#define __Flx_HALFGCD_KARATSUBA_LIMIT    139
-#define __Flx_HALFGCD_MULII2_LIMIT       537
-#define __Flx_HALFGCD_MULII_LIMIT        91
-#define __Flx_HALFGCD_QUARTMULII_LIMIT   37
-#define __Flx_INVBARRETT_HALFMULII_LIMIT 240
-#define __Flx_INVBARRETT_KARATSUBA_LIMIT 3600
-#define __Flx_INVBARRETT_MULII2_LIMIT    1815
-#define __Flx_INVBARRETT_MULII_LIMIT     1293
-#define __Flx_INVBARRETT_QUARTMULII_LIMIT 73
-#define __Flx_MUL_HALFMULII_LIMIT        7
-#define __Flx_MUL_KARATSUBA_LIMIT        90
-#define __Flx_MUL_MULII2_LIMIT           152
-#define __Flx_MUL_MULII_LIMIT            8
-#define __Flx_MUL_QUARTMULII_LIMIT       7
-#define __Flx_REM_BARRETT_LIMIT          689
-#define __Flx_SQR_HALFSQRI_LIMIT         4
-#define __Flx_SQR_KARATSUBA_LIMIT        159
-#define __Flx_SQR_QUARTSQRI_LIMIT        4
-#define __Flx_SQR_SQRI2_LIMIT            470
-#define __Flx_SQR_SQRI_LIMIT             5
+#define __EXTGCD_HALFGCD_LIMIT           -1
+#define __F2x_MUL_KARATSUBA_LIMIT        13
+#define __F2x_MUL_MULII_LIMIT            774
+#define __F2xqX_BARRETT_LIMIT            48
+#define __F2xqX_DIVREM_BARRETT_LIMIT     127
+#define __F2xqX_EXTGCD_LIMIT             127
+#define __F2xqX_GCD_LIMIT                884
+#define __F2xqX_HALFGCD_LIMIT            89
+#define __F2xqX_INVBARRETT_LIMIT         40
+#define __F2xqX_REM_BARRETT_LIMIT        127
+#define __Flx_BARRETT2_LIMIT             52
+#define __Flx_BARRETT_LIMIT              164
+#define __Flx_DIVREM2_BARRETT_LIMIT      111
+#define __Flx_DIVREM_BARRETT_LIMIT       470
+#define __Flx_EXTGCD2_LIMIT              184
+#define __Flx_EXTGCD_LIMIT               469
+#define __Flx_GCD2_LIMIT                 1281
+#define __Flx_GCD_LIMIT                  2817
+#define __Flx_HALFGCD2_LIMIT             181
+#define __Flx_HALFGCD_LIMIT              586
+#define __Flx_INVBARRETT2_LIMIT          397
+#define __Flx_INVBARRETT_LIMIT           501
+#define __Flx_MUL2_KARATSUBA_LIMIT       9
+#define __Flx_MUL2_MULII_LIMIT           8
+#define __Flx_MUL_KARATSUBA_LIMIT        57
+#define __Flx_MUL_MULII_LIMIT            146
+#define __Flx_REM2_BARRETT_LIMIT         89
+#define __Flx_REM_BARRETT_LIMIT          388
+#define __Flx_SQR2_KARATSUBA_LIMIT       18
+#define __Flx_SQR2_SQRI_LIMIT            20
+#define __Flx_SQR_KARATSUBA_LIMIT        112
+#define __Flx_SQR_SQRI_LIMIT             183
 #define __FlxqX_BARRETT_LIMIT            17
 #define __FlxqX_DIVREM_BARRETT_LIMIT     46
 #define __FlxqX_EXTGCD_LIMIT             44
@@ -132,6 +130,8 @@
 #define __FpX_REM_BARRETT_LIMIT          127
 #define __Fp_POW_BARRETT_LIMIT           11
 #define __Fp_POW_REDC_LIMIT              3
+#define __GCD_HALFGCD_LIMIT              -1
+#define __HALFGCD_LIMIT                  22
 #define __INVMOD_GMP_LIMIT               3
 #define __INVNEWTON_LIMIT                93
 #define __LOGAGMCX_LIMIT                 32
