@@ -5,7 +5,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -98,6 +99,24 @@ affrr_fixlg(GEN y, GEN z) { fixlg(z, lg(y)); affrr(y, z); }
 /*                       ALLOCATE ON STACK                         */
 /*                                                                 */
 /*******************************************************************/
+INLINE void
+set_avma(ulong av) { avma = av; }
+
+INLINE double
+gc_double(pari_sp av, double d) { set_avma(av); return d; }
+INLINE long
+gc_long(pari_sp av, long s) { set_avma(av); return s; }
+INLINE ulong
+gc_ulong(pari_sp av, ulong s) { set_avma(av); return s; }
+INLINE int
+gc_bool(pari_sp av, int s) { set_avma(av); return s; }
+INLINE int
+gc_int(pari_sp av, int s) { set_avma(av); return s; }
+INLINE GEN
+gc_NULL(pari_sp av) { set_avma(av); return NULL; }
+INLINE GEN
+gc_const(pari_sp av, GEN x) { set_avma(av); return x; }
+
 INLINE GEN
 new_chunk(size_t x) /* x is a number of longs */
 {
@@ -105,7 +124,7 @@ new_chunk(size_t x) /* x is a number of longs */
   CHECK_CTRLC
   if (x > (avma-pari_mainstack->bot) / sizeof(long))
     new_chunk_resize(x);
-  avma = (pari_sp)z;
+  set_avma((pari_sp)z);
 #ifdef MEMSTEP
   if (DEBUGMEM>1 && pari_mainstack->memused != DISABLE_MEMUSED) {
     long d = (long)pari_mainstack->memused - (long)z;
@@ -130,8 +149,9 @@ stack_malloc(size_t N)
 INLINE char *
 stack_malloc_align(size_t N, long k)
 {
-  ulong d = ((ulong)avma) % k;
+  ulong d = ((ulong)avma) % k, e = ((ulong)N) % k;
   if (d) (void)new_chunk(d/sizeof(long));
+  if (e) N += k-e;
   return (char*) new_chunk(nchar2nlong(N));
 }
 
@@ -142,6 +162,15 @@ stack_calloc(size_t N)
   memset(p, 0, N); return p;
 }
 
+INLINE char *
+stack_calloc_align(size_t N, long k)
+{
+  ulong d = ((ulong)avma) % k, e = ((ulong)N) % k;
+  if (d) (void)new_chunk(d/sizeof(long));
+  if (e) N += k-e;
+  return stack_calloc(N);
+}
+
 /* cgetg(lg(x), typ(x)), set *lx. Implicit unsetisclone() */
 INLINE GEN
 cgetg_copy(GEN x, long *plx) {
@@ -247,8 +276,6 @@ INLINE GEN
 mpneg(GEN x) { GEN y = leafcopy(x); togglesign(y); return y; }
 INLINE GEN negi(GEN x) { return mpneg(x); }
 INLINE GEN negr(GEN x) { return mpneg(x); }
-INLINE GEN gmax_shallow(GEN x, GEN y) { return gcmp(x,y)<0? y: x; }
-INLINE GEN gmin_shallow(GEN x, GEN y) { return gcmp(x,y)<0? x: y; }
 
 /* negate in place */
 INLINE void
@@ -425,6 +452,14 @@ INLINE ulong int_bit(GEN x, long n)
 /*                           COMPARISON                            */
 /*                                                                 */
 /*******************************************************************/
+INLINE int
+cmpss(long a, long b)
+{ return a>b? 1: (a<b? -1: 0); }
+
+INLINE int
+cmpuu(ulong a, ulong b)
+{ return a>b? 1: (a<b? -1: 0); }
+
 INLINE int
 cmpir(GEN x, GEN y)
 {
@@ -437,7 +472,7 @@ cmpir(GEN x, GEN y)
     if (expo(y) >= expi(x)) return 0;
     return signe(x);
   }
-  av=avma; z = itor(x, realprec(y)); avma=av;
+  av=avma; z = itor(x, realprec(y)); set_avma(av);
   return cmprr(z,y); /* cmprr does no memory adjustment */
 }
 INLINE int
@@ -449,7 +484,7 @@ cmpsr(long x, GEN y)
   GEN z;
 
   if (!x) return -signe(y);
-  av=avma; z = stor(x, LOWDEFAULTPREC); avma=av;
+  av=avma; z = stor(x, LOWDEFAULTPREC); set_avma(av);
   return cmprr(z,y);
 }
 INLINE int
@@ -681,7 +716,7 @@ remii(GEN a, GEN b) { return dvmdii(a,b,ONLY_REM); }
 INLINE GEN
 divss(long x, long y) { return stoi(x / y); }
 INLINE GEN
-modss(long x, long y) { return stoi(smodss(x, y)); }
+modss(long x, long y) { return utoi(smodss(x, y)); }
 INLINE GEN
 remss(long x, long y) { return stoi(x % y); }
 INLINE long
@@ -820,15 +855,14 @@ INLINE long
 smodis(GEN x, long y)
 {
   pari_sp av = avma;
-  long r;
-  (void)divis_rem(x,y, &r); avma = av; return (r >= 0) ? r: labs(y) + r;
+  long r; (void)divis_rem(x,y, &r);
+  return gc_long(av, (r >= 0)? r: labs(y) + r);
 }
 INLINE GEN
 modis(GEN x, long y) { return stoi(smodis(x,y)); }
 INLINE GEN
 modsi(long x, GEN y) {
-  long r;
-  (void)sdivsi_rem(x, y, &r);
+  long r; (void)sdivsi_rem(x, y, &r);
   return (r >= 0)? stoi(r): addsi_sign(r, y, 1);
 }
 
@@ -853,7 +887,7 @@ remis(GEN x, long y)
 {
   pari_sp av = avma;
   long r;
-  (void)divis_rem(x,y, &r); avma = av; return stoi(r);
+  (void)divis_rem(x,y, &r); set_avma(av); return stoi(r);
 }
 
 INLINE GEN
@@ -862,7 +896,7 @@ rdivis(GEN x, long y, long prec)
   GEN z = cgetr(prec);
   pari_sp av = avma;
   affrr(divrs(itor(x,prec), y),z);
-  avma = av; return z;
+  set_avma(av); return z;
 }
 INLINE GEN
 rdivsi(long x, GEN y, long prec)
@@ -870,7 +904,7 @@ rdivsi(long x, GEN y, long prec)
   GEN z = cgetr(prec);
   pari_sp av = avma;
   affrr(divsr(x, itor(y,prec)), z);
-  avma = av; return z;
+  set_avma(av); return z;
 }
 INLINE GEN
 rdivss(long x, long y, long prec)
@@ -878,57 +912,61 @@ rdivss(long x, long y, long prec)
   GEN z = cgetr(prec);
   pari_sp av = avma;
   affrr(divrs(stor(x, prec), y), z);
-  avma = av; return z;
+  set_avma(av); return z;
 }
 
 INLINE void
 rdiviiz(GEN x, GEN y, GEN z)
 {
-  pari_sp av = avma;
-  long prec = realprec(z);
-  affir(x, z);
-  if (!is_bigint(y)) {
-    affrr(divrs(z, y[2]), z);
-    if (signe(y) < 0) togglesign(z);
+  long prec = realprec(z), lx = lgefint(x), ly = lgefint(y);
+  if (lx == 2) { affur(0, z); return; }
+  if (ly == 3)
+  {
+    affir(x, z); if (signe(y) < 0) togglesign(z);
+    affrr(divru(z, y[2]), z);
+  }
+  else if (lx > prec + 1 || ly > prec + 1)
+  {
+    affir(x,z); affrr(divri(z, y), z);
   }
   else
-    affrr(divrr(z, itor(y,prec)), z);
-  avma = av;
+  {
+    long b = bit_accuracy(prec) + expi(y) - expi(x) + 1;
+    GEN q = divii(b > 0? shifti(x, b): x, y);
+    affir(q, z); if (b > 0) shiftr_inplace(z, -b);
+  }
+  set_avma((ulong)z);
 }
 INLINE GEN
 rdivii(GEN x, GEN y, long prec)
-{
-  GEN z = itor(x, prec);
-  pari_sp av = avma;
-  if (lg(y) == 3) {
-    affrr(divru(z, y[2]), z);
-    if (signe(y) < 0) togglesign(z);
-  }
-  else
-    affrr(divrr(z, itor(y,prec)), z);
-  avma = av; return z;
-}
+{ GEN z = cgetr(prec); rdiviiz(x, y, z); return z; }
 INLINE GEN
-fractor(GEN x, long prec) { return rdivii(gel(x,1), gel(x,2), prec); }
+fractor(GEN x, long prec)
+{ return rdivii(gel(x,1), gel(x,2), prec); }
 
 INLINE int
 dvdii(GEN x, GEN y)
 {
-  pari_sp av=avma;
-  GEN r = remii(x,y);
-  avma = av; return r == gen_0;
+  pari_sp av = avma;
+  GEN r;
+  if (!signe(x)) return 1;
+  if (!signe(y)) return 0;
+  r = remii(x,y);
+  return gc_bool(av, r == gen_0);
 }
 INLINE int
 dvdsi(long x, GEN y)
 {
-  if (!signe(y)) return x == 0;
+  if (x == 0) return 1;
+  if (!signe(y)) return 0;
   if (lgefint(y) != 3) return 0;
   return x % y[2] == 0;
 }
 INLINE int
 dvdui(ulong x, GEN y)
 {
-  if (!signe(y)) return x == 0;
+  if (x == 0) return 1;
+  if (!signe(y)) return 0;
   if (lgefint(y) != 3) return 0;
   return x % y[2] == 0;
 }
@@ -945,7 +983,7 @@ dvdisz(GEN x, long y, GEN z)
   const pari_sp av = avma;
   long r;
   GEN p1 = divis_rem(x,y, &r);
-  avma = av; if (r) return 0;
+  set_avma(av); if (r) return 0;
   affii(p1,z); return 1;
 }
 INLINE int
@@ -954,18 +992,16 @@ dvdiuz(GEN x, ulong y, GEN z)
   const pari_sp av = avma;
   ulong r;
   GEN p1 = absdiviu_rem(x,y, &r);
-  avma = av; if (r) return 0;
+  set_avma(av); if (r) return 0;
   affii(p1,z); return 1;
 }
 INLINE int
 dvdiiz(GEN x, GEN y, GEN z)
 {
   const pari_sp av=avma;
-  GEN p2;
-  const GEN p1=dvmdii(x,y,&p2);
-
-  if (signe(p2)) { avma=av; return 0; }
-  affii(p1,z); avma=av; return 1;
+  GEN p2, p1 = dvmdii(x,y,&p2);
+  if (signe(p2)) return gc_bool(av,0);
+  affii(p1,z); return gc_bool(av,1);
 }
 
 INLINE ulong
@@ -1142,8 +1178,10 @@ Fl_sqr(ulong a, ulong p)
   if (!hiremainder) return x % p;
   (void)divll(x,p); return hiremainder;
 }
+/* don't assume that p is prime: can't special case a = 0 */
 INLINE ulong
-Fl_div(ulong a, ulong b, ulong p) { return Fl_mul(a, Fl_inv(b, p), p); }
+Fl_div(ulong a, ulong b, ulong p)
+{ return Fl_mul(a, Fl_inv(b, p), p); }
 
 /*******************************************************************/
 /*                                                                 */
@@ -1194,7 +1232,6 @@ vali(GEN x)
   return vals(*xp) + i * BITS_IN_LONG;
 }
 
-
 /* assume x > 0 */
 INLINE long
 expu(ulong x) { return (BITS_IN_LONG-1) - (long)bfffo(x); }
@@ -1344,116 +1381,116 @@ mpaff(GEN x, GEN y) { if (typ(x)==t_INT) affiz(x, y); else affrr(x,y); }
 /*******************************************************************/
 
 INLINE void addiiz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affii(addii(x,y),z); avma = av; }
+{ pari_sp av = avma; affii(addii(x,y),z); set_avma(av); }
 INLINE void addirz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(addir(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(addir(x,y),z); set_avma(av); }
 INLINE void addriz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(addri(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(addri(x,y),z); set_avma(av); }
 INLINE void addrrz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(addrr(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(addrr(x,y),z); set_avma(av); }
 INLINE void addsiz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affii(addsi(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(addsi(s,y),z); set_avma(av); }
 INLINE void addsrz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affrr(addsr(s,y),z); avma = av; }
+{ pari_sp av = avma; affrr(addsr(s,y),z); set_avma(av); }
 INLINE void addssz(long s, long y, GEN z)
-{ pari_sp av = avma; affii(addss(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(addss(s,y),z); set_avma(av); }
 
 INLINE void diviiz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affii(divii(x,y),z); avma = av; }
+{ pari_sp av = avma; affii(divii(x,y),z); set_avma(av); }
 INLINE void divirz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(divir(x,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(divir(x,y),z); set_avma(av); }
 INLINE void divisz(GEN x, long y, GEN z)
-{ pari_sp av = avma; affii(divis(x,y),z); avma = av; }
+{ pari_sp av = avma; affii(divis(x,y),z); set_avma(av); }
 INLINE void divriz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(divri(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(divri(x,y),z); set_avma(av); }
 INLINE void divrrz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(divrr(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(divrr(x,y),z); set_avma(av); }
 INLINE void divrsz(GEN y, long s, GEN z)
-{ pari_sp av = avma; affrr(divrs(y,s),z); avma = av; }
+{ pari_sp av = avma; affrr(divrs(y,s),z); set_avma(av); }
 INLINE void divsiz(long x, GEN y, GEN z)
 { long junk; affsi(sdivsi_rem(x,y,&junk), z); }
 INLINE void divsrz(long s, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(divsr(s,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(divsr(s,y),z); set_avma(av); }
 INLINE void divssz(long x, long y, GEN z)
 { affsi(x/y, z); }
 
 INLINE void modisz(GEN y, long s, GEN z)
 { affsi(smodis(y,s),z); }
 INLINE void modsiz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affii(modsi(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(modsi(s,y),z); set_avma(av); }
 INLINE void modssz(long s, long y, GEN z)
 { affsi(smodss(s,y),z); }
 
 INLINE void mpaddz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(mpadd(x,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(mpadd(x,y),z); set_avma(av); }
 INLINE void mpsubz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(mpsub(x,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(mpsub(x,y),z); set_avma(av); }
 INLINE void mpmulz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(mpmul(x,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(mpmul(x,y),z); set_avma(av); }
 
 INLINE void muliiz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affii(mulii(x,y),z); avma = av; }
+{ pari_sp av = avma; affii(mulii(x,y),z); set_avma(av); }
 INLINE void mulirz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(mulir(x,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(mulir(x,y),z); set_avma(av); }
 INLINE void mulriz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(mulri(x,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(mulri(x,y),z); set_avma(av); }
 INLINE void mulrrz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(mulrr(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(mulrr(x,y),z); set_avma(av); }
 INLINE void mulsiz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affii(mulsi(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(mulsi(s,y),z); set_avma(av); }
 INLINE void mulsrz(long s, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(mulsr(s,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(mulsr(s,y),z); set_avma(av); }
 INLINE void mulssz(long s, long y, GEN z)
-{ pari_sp av = avma; affii(mulss(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(mulss(s,y),z); set_avma(av); }
 
 INLINE void remiiz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affii(remii(x,y),z); avma = av; }
+{ pari_sp av = avma; affii(remii(x,y),z); set_avma(av); }
 INLINE void remisz(GEN y, long s, GEN z)
-{ pari_sp av = avma; affii(remis(y,s),z); avma = av; }
+{ pari_sp av = avma; affii(remis(y,s),z); set_avma(av); }
 INLINE void remsiz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affii(remsi(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(remsi(s,y),z); set_avma(av); }
 INLINE void remssz(long s, long y, GEN z)
-{ pari_sp av = avma; affii(remss(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(remss(s,y),z); set_avma(av); }
 
 INLINE void subiiz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affii(subii(x,y),z); avma = av; }
+{ pari_sp av = avma; affii(subii(x,y),z); set_avma(av); }
 INLINE void subirz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(subir(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(subir(x,y),z); set_avma(av); }
 INLINE void subisz(GEN y, long s, GEN z)
-{ pari_sp av = avma; affii(addsi(-s,y),z); avma = av; }
+{ pari_sp av = avma; affii(addsi(-s,y),z); set_avma(av); }
 INLINE void subriz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(subri(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(subri(x,y),z); set_avma(av); }
 INLINE void subrrz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(subrr(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(subrr(x,y),z); set_avma(av); }
 INLINE void subrsz(GEN y, long s, GEN z)
-{ pari_sp av = avma; affrr(addsr(-s,y),z); avma = av; }
+{ pari_sp av = avma; affrr(addsr(-s,y),z); set_avma(av); }
 INLINE void subsiz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affii(subsi(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(subsi(s,y),z); set_avma(av); }
 INLINE void subsrz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affrr(subsr(s,y),z); avma = av; }
+{ pari_sp av = avma; affrr(subsr(s,y),z); set_avma(av); }
 INLINE void subssz(long x, long y, GEN z) { addssz(x,-y,z); }
 
 INLINE void
 dvmdssz(long x, long y, GEN z, GEN t) {
   pari_sp av = avma;
   long r;
-  affii(divss_rem(x,y, &r), z); avma = av; affsi(r,t);
+  affii(divss_rem(x,y, &r), z); set_avma(av); affsi(r,t);
 }
 INLINE void
 dvmdsiz(long x, GEN y, GEN z, GEN t) {
   pari_sp av = avma;
   long r;
-  affii(divsi_rem(x,y, &r), z); avma = av; affsi(r,t);
+  affii(divsi_rem(x,y, &r), z); set_avma(av); affsi(r,t);
 }
 INLINE void
 dvmdisz(GEN x, long y, GEN z, GEN t) {
   pari_sp av = avma;
   long r;
-  affii(divis_rem(x,y, &r),z); avma = av; affsi(r,t);
+  affii(divis_rem(x,y, &r),z); set_avma(av); affsi(r,t);
 }
 INLINE void
 dvmdiiz(GEN x, GEN y, GEN z, GEN t) {
   pari_sp av = avma;
   GEN r;
-  affii(dvmdii(x,y,&r),z); affii(r,t); avma=av;
+  affii(dvmdii(x,y,&r),z); affii(r,t); set_avma(av);
 }
