@@ -2,40 +2,38 @@
 long AGM_ATAN_LIMIT                 = __AGM_ATAN_LIMIT;
 long DIVRR_GMP_LIMIT                = __DIVRR_GMP_LIMIT;
 long EXPNEWTON_LIMIT                = __EXPNEWTON_LIMIT;
+long EXTGCD_HALFGCD_LIMIT           = __EXTGCD_HALFGCD_LIMIT;
 long F2x_MUL_KARATSUBA_LIMIT        = __F2x_MUL_KARATSUBA_LIMIT;
-long Flx_BARRETT_QUARTMULII_LIMIT   = __Flx_BARRETT_QUARTMULII_LIMIT;
-long Flx_BARRETT_HALFMULII_LIMIT    = __Flx_BARRETT_HALFMULII_LIMIT;
-long Flx_BARRETT_KARATSUBA_LIMIT    = __Flx_BARRETT_KARATSUBA_LIMIT;
-long Flx_BARRETT_MULII2_LIMIT       = __Flx_BARRETT_MULII2_LIMIT;
-long Flx_BARRETT_MULII_LIMIT        = __Flx_BARRETT_MULII_LIMIT;
-long Flx_DIVREM_BARRETT_LIMIT       = __Flx_DIVREM_BARRETT_LIMIT;
-long Flx_EXTGCD_LIMIT               = __Flx_EXTGCD_LIMIT;
-long Flx_GCD_LIMIT                  = __Flx_GCD_LIMIT;
-long Flx_HALFGCD_QUARTMULII_LIMIT   = __Flx_HALFGCD_QUARTMULII_LIMIT;
-long Flx_HALFGCD_HALFMULII_LIMIT    = __Flx_HALFGCD_HALFMULII_LIMIT;
-long Flx_HALFGCD_KARATSUBA_LIMIT    = __Flx_HALFGCD_KARATSUBA_LIMIT;
-long Flx_HALFGCD_MULII2_LIMIT       = __Flx_HALFGCD_MULII2_LIMIT;
-long Flx_HALFGCD_MULII_LIMIT        = __Flx_HALFGCD_MULII_LIMIT;
-long Flx_INVBARRETT_QUARTMULII_LIMIT = __Flx_INVBARRETT_QUARTMULII_LIMIT;
-long Flx_INVBARRETT_HALFMULII_LIMIT = __Flx_INVBARRETT_HALFMULII_LIMIT;
-long Flx_INVBARRETT_KARATSUBA_LIMIT = __Flx_INVBARRETT_KARATSUBA_LIMIT;
-long Flx_INVBARRETT_MULII2_LIMIT    = __Flx_INVBARRETT_MULII2_LIMIT;
-long Flx_INVBARRETT_MULII_LIMIT     = __Flx_INVBARRETT_MULII_LIMIT;
-long Flx_MUL_QUARTMULII_LIMIT       = __Flx_MUL_QUARTMULII_LIMIT;
-long Flx_MUL_HALFMULII_LIMIT        = __Flx_MUL_HALFMULII_LIMIT;
-long Flx_MUL_KARATSUBA_LIMIT        = __Flx_MUL_KARATSUBA_LIMIT;
-long Flx_MUL_MULII2_LIMIT           = __Flx_MUL_MULII2_LIMIT;
-long Flx_MUL_MULII_LIMIT            = __Flx_MUL_MULII_LIMIT;
-long Flx_REM_BARRETT_LIMIT          = __Flx_REM_BARRETT_LIMIT;
-long Flx_SQR_QUARTSQRI_LIMIT        = __Flx_SQR_QUARTSQRI_LIMIT;
-long Flx_SQR_HALFSQRI_LIMIT         = __Flx_SQR_HALFSQRI_LIMIT;
-long Flx_SQR_KARATSUBA_LIMIT        = __Flx_SQR_KARATSUBA_LIMIT;
-long Flx_SQR_SQRI2_LIMIT            = __Flx_SQR_SQRI2_LIMIT;
-long Flx_SQR_SQRI_LIMIT             = __Flx_SQR_SQRI_LIMIT;
+long F2x_MUL_MULII_LIMIT            = __F2x_MUL_MULII_LIMIT;
 long F2xqX_BARRETT_LIMIT            = __F2xqX_BARRETT_LIMIT;
 long F2xqX_DIVREM_BARRETT_LIMIT     = __F2xqX_DIVREM_BARRETT_LIMIT;
+long F2xqX_EXTGCD_LIMIT             = __F2xqX_EXTGCD_LIMIT;
+long F2xqX_GCD_LIMIT                = __F2xqX_GCD_LIMIT;
+long F2xqX_HALFGCD_LIMIT            = __F2xqX_HALFGCD_LIMIT;
 long F2xqX_INVBARRETT_LIMIT         = __F2xqX_INVBARRETT_LIMIT;
 long F2xqX_REM_BARRETT_LIMIT        = __F2xqX_REM_BARRETT_LIMIT;
+long Flx_BARRETT2_LIMIT             = __Flx_BARRETT2_LIMIT;
+long Flx_BARRETT_LIMIT              = __Flx_BARRETT_LIMIT;
+long Flx_DIVREM2_BARRETT_LIMIT      = __Flx_DIVREM2_BARRETT_LIMIT;
+long Flx_DIVREM_BARRETT_LIMIT       = __Flx_DIVREM_BARRETT_LIMIT;
+long Flx_EXTGCD2_LIMIT              = __Flx_EXTGCD2_LIMIT;
+long Flx_EXTGCD_LIMIT               = __Flx_EXTGCD_LIMIT;
+long Flx_GCD2_LIMIT                 = __Flx_GCD2_LIMIT;
+long Flx_GCD_LIMIT                  = __Flx_GCD_LIMIT;
+long Flx_HALFGCD2_LIMIT             = __Flx_HALFGCD2_LIMIT;
+long Flx_HALFGCD_LIMIT              = __Flx_HALFGCD_LIMIT;
+long Flx_INVBARRETT2_LIMIT          = __Flx_INVBARRETT2_LIMIT;
+long Flx_INVBARRETT_LIMIT           = __Flx_INVBARRETT_LIMIT;
+long Flx_MUL2_KARATSUBA_LIMIT       = __Flx_MUL2_KARATSUBA_LIMIT;
+long Flx_MUL2_MULII_LIMIT           = __Flx_MUL2_MULII_LIMIT;
+long Flx_MUL_KARATSUBA_LIMIT        = __Flx_MUL_KARATSUBA_LIMIT;
+long Flx_MUL_MULII_LIMIT            = __Flx_MUL_MULII_LIMIT;
+long Flx_REM2_BARRETT_LIMIT         = __Flx_REM2_BARRETT_LIMIT;
+long Flx_REM_BARRETT_LIMIT          = __Flx_REM_BARRETT_LIMIT;
+long Flx_SQR2_KARATSUBA_LIMIT       = __Flx_SQR2_KARATSUBA_LIMIT;
+long Flx_SQR2_SQRI_LIMIT            = __Flx_SQR2_SQRI_LIMIT;
+long Flx_SQR_KARATSUBA_LIMIT        = __Flx_SQR_KARATSUBA_LIMIT;
+long Flx_SQR_SQRI_LIMIT             = __Flx_SQR_SQRI_LIMIT;
 long FlxqX_BARRETT_LIMIT            = __FlxqX_BARRETT_LIMIT;
 long FlxqX_DIVREM_BARRETT_LIMIT     = __FlxqX_DIVREM_BARRETT_LIMIT;
 long FlxqX_EXTGCD_LIMIT             = __FlxqX_EXTGCD_LIMIT;
@@ -43,13 +41,6 @@ long FlxqX_GCD_LIMIT                = __FlxqX_GCD_LIMIT;
 long FlxqX_HALFGCD_LIMIT            = __FlxqX_HALFGCD_LIMIT;
 long FlxqX_INVBARRETT_LIMIT         = __FlxqX_INVBARRETT_LIMIT;
 long FlxqX_REM_BARRETT_LIMIT        = __FlxqX_REM_BARRETT_LIMIT;
-long FpX_BARRETT_LIMIT              = __FpX_BARRETT_LIMIT;
-long FpX_DIVREM_BARRETT_LIMIT       = __FpX_DIVREM_BARRETT_LIMIT;
-long FpX_EXTGCD_LIMIT               = __FpX_EXTGCD_LIMIT;
-long FpX_GCD_LIMIT                  = __FpX_GCD_LIMIT;
-long FpX_HALFGCD_LIMIT              = __FpX_HALFGCD_LIMIT;
-long FpX_INVBARRETT_LIMIT           = __FpX_INVBARRETT_LIMIT;
-long FpX_REM_BARRETT_LIMIT          = __FpX_REM_BARRETT_LIMIT;
 long FpXQX_BARRETT_LIMIT            = __FpXQX_BARRETT_LIMIT;
 long FpXQX_DIVREM_BARRETT_LIMIT     = __FpXQX_DIVREM_BARRETT_LIMIT;
 long FpXQX_EXTGCD_LIMIT             = __FpXQX_EXTGCD_LIMIT;
@@ -57,8 +48,17 @@ long FpXQX_GCD_LIMIT                = __FpXQX_GCD_LIMIT;
 long FpXQX_HALFGCD_LIMIT            = __FpXQX_HALFGCD_LIMIT;
 long FpXQX_INVBARRETT_LIMIT         = __FpXQX_INVBARRETT_LIMIT;
 long FpXQX_REM_BARRETT_LIMIT        = __FpXQX_REM_BARRETT_LIMIT;
+long FpX_BARRETT_LIMIT              = __FpX_BARRETT_LIMIT;
+long FpX_DIVREM_BARRETT_LIMIT       = __FpX_DIVREM_BARRETT_LIMIT;
+long FpX_EXTGCD_LIMIT               = __FpX_EXTGCD_LIMIT;
+long FpX_GCD_LIMIT                  = __FpX_GCD_LIMIT;
+long FpX_HALFGCD_LIMIT              = __FpX_HALFGCD_LIMIT;
+long FpX_INVBARRETT_LIMIT           = __FpX_INVBARRETT_LIMIT;
+long FpX_REM_BARRETT_LIMIT          = __FpX_REM_BARRETT_LIMIT;
 long Fp_POW_BARRETT_LIMIT           = __Fp_POW_BARRETT_LIMIT;
 long Fp_POW_REDC_LIMIT              = __Fp_POW_REDC_LIMIT;
+long GCD_HALFGCD_LIMIT              = __GCD_HALFGCD_LIMIT;
+long HALFGCD_LIMIT                  = __HALFGCD_LIMIT;
 long INVMOD_GMP_LIMIT               = __INVMOD_GMP_LIMIT;
 long INVNEWTON_LIMIT                = __INVNEWTON_LIMIT;
 long LOGAGMCX_LIMIT                 = __LOGAGMCX_LIMIT;
