@@ -572,6 +572,13 @@ get_ret_type(const char **p, long arity, Gtype *t, long *flag)
   *t=Ggen; return arity==2?OCcallgen2:OCcallgen;
 }
 
+static void
+U_compile_err(const char *s)
+{ compile_err("this should be a small non-negative integer",s); }
+static void
+L_compile_err(const char *s)
+{ compile_err("this should be a small integer",s); }
+
 /*supported types:
  * type: Gusmall, Gsmall, Ggen, Gvoid, Gvec, Gclosure
  * mode: Gusmall, Gsmall, Ggen, Gvar, Gvoid
@@ -585,14 +592,12 @@ compilecast_loc(int type, int mode, const char *loc)
   case Gusmall:
     if (type==Ggen)        op_push_loc(OCitou,-1,loc);
     else if (type==Gvoid)  op_push_loc(OCpushlong,0,loc);
-    else if (type!=Gsmall)
-      compile_err("this should be a small integer >=0",loc);
+    else if (type!=Gsmall) U_compile_err(loc);
     break;
   case Gsmall:
     if (type==Ggen)        op_push_loc(OCitos,-1,loc);
     else if (type==Gvoid)  op_push_loc(OCpushlong,0,loc);
-    else if (type!=Gusmall)
-      compile_err("this should be a small integer",loc);
+    else if (type!=Gusmall) L_compile_err(loc);
     break;
   case Ggen:
     if (type==Gsmall)      op_push_loc(OCstoi,0,loc);
@@ -922,8 +927,7 @@ compilesmall(long n, long x, long mode)
     op_push(OCpushstoi, x, n);
   else
   {
-    if (mode==Gusmall && x < 0)
-      compile_err("this should be a small integer >=0",tree[n].str);
+    if (mode==Gusmall && x < 0) U_compile_err(tree[n].str);
     op_push(OCpushlong, x, n);
     compilecast(n,Gsmall,mode);
   }
@@ -1062,7 +1066,12 @@ compilecall(long n, int mode, entree *ep)
       compile_err("unexpected ';'", tree[x].str+tree[x].len);
     else if (f==Findarg)
     {
-      compilenode(tree[arg[j]].x, Ggen,FLnocopy);
+      long a = tree[arg[j]].x;
+      entree *ep = getlvalue(a);
+      long vn = getmvar(ep);
+      if (vn)
+        op_push(OCcowvarlex, vn, a);
+      compilenode(a, Ggen,FLnocopy);
       op_push(OClock,0,n);
     } else if (tx==CSTmember)
     {
@@ -1507,10 +1516,7 @@ compilefunc(entree *ep, long n, int mode, long flag)
             entree *ep = getlvalue(a);
             long vn = getmvar(ep);
             if (vn)
-            {
-              access_push(vn);
               op_push(OCcowvarlex, vn, a);
-            }
             else op_push(OCcowvardyn, (long)ep, a);
             compilenode(a, Ggen,FLnocopy);
             j++;
@@ -2152,8 +2158,8 @@ compilenode(long n, int mode, long flag)
         if (mode==Gvoid) return;
         if (mode==Gvar) compile_varerr(tree[n].str);
       }
-      if (mode==Gsmall)
-        compile_err("this should be a small integer", tree[n].str);
+      if (mode==Gsmall) L_compile_err(tree[n].str);
+      if (mode==Gusmall && tree[n].x != CSTint) U_compile_err(tree[n].str);
       switch(tree[n].x)
       {
       case CSTreal:
