@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -49,7 +50,7 @@ initprimes1(ulong size, long *lenp, ulong *lastp, byteptr p1)
   *r++ = 0;
   *lenp = r - p1;
   *lastp = ((s - p) << 1) + 1;
-  avma = av;
+  set_avma(av);
 }
 
 /*  Timing in ms (Athlon/850; reports 512K of secondary cache; looks
@@ -419,11 +420,13 @@ initprimes0(ulong maxnum, long *lenp, ulong *lastp, byteptr p1)
   *curdiff++ = 0; /* sentinel */
   *lenp = curdiff - p1;
   *lastp = last;
-  if (alloced) pari_free(p); else avma = av;
+  if (alloced) pari_free(p); else set_avma(av);
 }
 
 ulong
 maxprime(void) { return diffptr ? _maxprime : 0; }
+ulong
+maxprimeN(void) { return diffptr ? diffptrlen-1: 0; }
 
 void
 maxprime_check(ulong c) { if (_maxprime < c) pari_err_MAXPRIME(c); }
@@ -522,9 +525,9 @@ sieve_init(forprime_t *T, ulong a, ulong b)
 {
   T->sieveb = b;
   T->chunk = optimize_chunk(a, b);
+  /* >> 1 [only odds] + 3 [convert from bits to bytes] */
   T->isieve = (unsigned char*)stack_malloc(((T->chunk+2) >> 4) + 1);
   T->cache[0] = 0;
-  /* >> 1 [only odds] + 3 [convert from bits to bytes] */
   T->a = a;
   T->end = minuu(a + T->chunk, b);
   T->pos = T->maxpos = 0;
@@ -556,7 +559,7 @@ arith_set(forprime_t *T)
   if (T->c > r) d = subiu(d, T->q);
   /* d = c mod q,  d = c > r? p-r+c-q: p-r+c, so that
    *  d <= p  and  d+q = c>r? p-r+c  : p-r+c+q > p */
-  T->p = itou_or_0(d); avma = av; /* d = 0 is impossible */
+  T->p = itou_or_0(d); set_avma(av); /* d = 0 is impossible */
 }
 
 /* run through primes in arithmetic progression = c (mod q) */
@@ -786,8 +789,7 @@ void
 init_modular_big(forprime_t *S)
 {
 #ifdef LONG_IS_64BIT
-  ulong a = (1UL<<(BITS_IN_LONG-1))+1;
-  u_forprime_init(S, a, ULONG_MAX);
+  u_forprime_init(S, HIGHBIT + 1, ULONG_MAX);
 #else
   u_forprime_sieve_init(S, &pari_sieve_modular, ULONG_MAX);
 #endif
@@ -910,8 +912,33 @@ NEXT_CHUNK:
     if (T->q == 1)
     {
 #ifdef LONG_IS_64BIT
-      if (T->p == (1UL<<63)) return T->p = 9223372036854775837UL;
-      if (T->p == 9223372036854775837UL) return T->p = 9223372036854775907UL;
+      switch(T->p)
+      {
+#define retp(x) return T->p = (HIGHBIT+x <= T->b)? HIGHBIT+x: 0
+        case HIGHBIT: retp(29);
+        case HIGHBIT + 29: retp(99);
+        case HIGHBIT + 99: retp(123);
+        case HIGHBIT +123: retp(131);
+        case HIGHBIT +131: retp(155);
+        case HIGHBIT +155: retp(255);
+        case HIGHBIT +255: retp(269);
+        case HIGHBIT +269: retp(359);
+        case HIGHBIT +359: retp(435);
+        case HIGHBIT +435: retp(449);
+        case HIGHBIT +449: retp(453);
+        case HIGHBIT +453: retp(485);
+        case HIGHBIT +485: retp(491);
+        case HIGHBIT +491: retp(543);
+        case HIGHBIT +543: retp(585);
+        case HIGHBIT +585: retp(599);
+        case HIGHBIT +599: retp(753);
+        case HIGHBIT +753: retp(849);
+        case HIGHBIT +849: retp(879);
+        case HIGHBIT +879: retp(885);
+        case HIGHBIT +885: retp(903);
+        case HIGHBIT +903: retp(995);
+#undef retp
+      }
 #endif
       T->p = unextprime(T->p + 1);
     }
@@ -946,12 +973,12 @@ forprime_next(forprime_t *T)
   if (T->q == 1)
   {
     p = nextprime(addiu(p, 1));
-    if (T->bb && abscmpii(p, T->bb) > 0) { avma = av; return NULL; }
+    if (T->bb && abscmpii(p, T->bb) > 0) return gc_NULL(av);
   } else do {
     p = addiu(p, T->q);
-    if (T->bb && abscmpii(p, T->bb) > 0) { avma = av; return NULL; }
+    if (T->bb && abscmpii(p, T->bb) > 0) return gc_NULL(av);
   } while (!BPSW_psp(p));
-  affii(p, T->pp); avma = av; return T->pp;
+  affii(p, T->pp); return gc_const(av, T->pp);
 }
 
 void
@@ -960,7 +987,7 @@ forprimestep(GEN a, GEN b, GEN q, GEN code)
   pari_sp av = avma;
   forprime_t T;
 
-  if (!forprimestep_init(&T, a,b,q)) { avma = av; return; }
+  if (!forprimestep_init(&T, a,b,q)) { set_avma(av); return; }
 
   push_lex(T.pp,code);
   while(forprime_next(&T))
@@ -970,7 +997,7 @@ forprimestep(GEN a, GEN b, GEN q, GEN code)
     if (get_lex(-1) != T.pp)
       pari_err(e_MISC,"prime index read-only: was changed to %Ps", get_lex(-1));
   }
-  pop_lex(1); avma = av;
+  pop_lex(1); set_avma(av);
 }
 void
 forprime(GEN a, GEN b, GEN code) { return forprimestep(a,b,NULL,code); }
@@ -995,8 +1022,7 @@ forcomposite_init(forcomposite_t *C, GEN a, GEN b)
   if (!forprime_init(&C->T, a,b) && cmpii(a,b) > 0)
   {
     C->n = gen_1; /* in case caller forgets to check the return value */
-    C->b = gen_0;
-    avma = av; return 0;
+    C->b = gen_0; return gc_bool(av,0);
   }
   C->n = setloop(a);
   C->b = b;
@@ -1040,5 +1066,5 @@ forcomposite(GEN a, GEN b, GEN code)
     if (get_lex(-1) != n)
       pari_err(e_MISC,"index read-only: was changed to %Ps", get_lex(-1));
   }
-  pop_lex(1); avma = av;
+  pop_lex(1); set_avma(av);
 }
