@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -131,9 +132,7 @@ static int
 compare_str(char **s1, char **s2) { return strcmp(*s1, *s2); }
 
 /* Print all elements of list in columns, pausing every nbli lines
- * if nbli is non-zero.
- * list is a NULL terminated list of function names
- */
+ * if nbli is nonzero. list is a NULL terminated list of function names */
 void
 print_fun_list(char **list, long nbli)
 {
@@ -210,7 +209,7 @@ pari_center(const char *s)
   for (i=0; i<pad; i++) *u++ = ' ';
   while (*s) *u++ = *s++;
   *u++ = '\n'; *u = 0;
-  pari_puts(buf); avma = av;
+  pari_puts(buf); set_avma(av);
 }
 
 static void
@@ -352,7 +351,7 @@ cyc  : cyclic decomposition (SNF)    bid,     clgp,ell,   bnf,bnr\n\
 diff, codiff: different and codifferent                nf,bnf,bnr\n\
 disc : discriminant                                ell,nf,bnf,bnr,rnf\n\
 e, f : inertia/residue  degree           prid\n\
-fu   : fundamental units                                  bnf,bnr\n\
+fu   : fundamental units                                  bnf\n\
 gen  : generators                    bid,prid,clgp,ell,   bnf,bnr,    gal\n\
 group: group                                       ell,               gal\n\
 index: index                                           nf,bnf,bnr\n\
@@ -364,15 +363,15 @@ nf   : number field                                    nf,bnf,bnr,rnf\n\
 no   : number of elements            bid,     clgp,ell,   bnf,bnr\n\
 omega, eta: [w1,w2] and [eta1, eta2]               ell\n\
 orders: relative orders of generators                                 gal\n\
-p    : rational prime                    prid,     ell,           rnf,gal\n\
+p    : rational prime                    prid,     ell,nf,bnr,bnr,rnf,gal\n\
 pol  : defining polynomial                             nf,bnf,bnr,    gal\n\
 polabs: defining polynomial over Q                                rnf\n\
-reg  : regulator                                          bnf,bnr\n\
+reg  : regulator                                          bnf\n\
 roots: roots                                       ell,nf,bnf,bnr,    gal\n\
 sign,r1,r2 : signature                                 nf,bnf,bnr\n\
 t2   : t2 matrix                                       nf,bnf,bnr\n\
 tate : Tate's [u^2, u, q, [a,b], L, Ei]            ell\n\
-tu   : torsion unit and its order                         bnf,bnr\n\
+tu   : torsion unit and its order                         bnf\n\
 zk   : integral basis                                  nf,bnf,bnr,rnf\n\
 zkst : structure of (Z_K/m)*         bid,                     bnr\n");
 }
@@ -427,7 +426,7 @@ external_help(const char *s, int num)
 {
   long nbli = term_height()-3, li = 0;
   char buf[256], *str;
-  const char *opt = "", *ar = "", *cdir = "";
+  const char *opt = "", *ar = "";
   char *t, *help = GP_DATA->help;
   pariFILE *z;
   FILE *f;
@@ -442,15 +441,14 @@ external_help(const char *s, int num)
   else if (t[strlen(t)-1] != '@')
     ar = stack_sprintf("@%d",num);
 #ifdef _WIN32
-  if (*help=='@')
+  if (*help == '@')
   {
     const char *basedir = win32_basedir();
-    help++;
-    cdir = stack_sprintf("%c:& cd %s & ", *basedir, basedir);
+    help = stack_sprintf("%c:& cd %s & %s", *basedir, basedir, help+1);
   }
 #endif
-  str=stack_sprintf("%s%s -fromgp %s %c%s%s%c",cdir,help,opt,
-                                               SHELL_Q,t,ar,SHELL_Q);
+  str = stack_sprintf("%s -fromgp %s %c%s%s%c",
+                      help, opt, SHELL_Q, t, ar, SHELL_Q);
   z = try_pipe(str,0); f = z->file;
   pari_free(t);
   while (fgets(buf, numberof(buf), f))
@@ -565,7 +563,8 @@ help(const char *s0, int flag)
   if (isdigit((int)*s)) { digit_help(s,flag); return; }
   if (flag & h_APROPOS) { external_help(s,-1); return; }
   /* Get meaningful answer on '\ps 5' (e.g. from <F1>) */
-  if (*s == '\\') { char *t = s+1; pari_skip_alpha(&t); *t = '\0'; }
+  if (*s == '\\' && isalpha((int)*(s+1)))
+  { char *t = s+1; pari_skip_alpha(&t); *t = '\0'; }
   if (isalpha((int)*s))
   {
     char *t = s;
@@ -652,7 +651,7 @@ gp_help(const char *s, long flag)
   }
   term_color(c_HELP); help(s,flag); term_color(c_NONE);
   if ((flag & h_RL) == 0) pari_putc('\n');
-  avma = av;
+  set_avma(av);
 }
 
 /********************************************************************/
@@ -679,12 +678,11 @@ what_cc(void)
   char *s;
 #ifdef GCC_VERSION
 #  ifdef __cplusplus
-#    define Format "(C++) %s"
-#  else
-#    define Format "%s"
-#  endif
   s = stack_malloc(6 + strlen(GCC_VERSION) + 1);
-  (void)sprintf(s, Format, GCC_VERSION);
+  (void)sprintf(s, "(C++) %s", GCC_VERSION);
+#  else
+  s = stack_strdup(GCC_VERSION);
+#  endif
 #else
 #  ifdef _MSC_VER
   s = stack_malloc(32);
@@ -696,7 +694,7 @@ what_cc(void)
   return s;
 }
 
-static void
+static char *
 convert_time(char *s, long delay)
 {
   if (delay >= 3600000)
@@ -720,23 +718,27 @@ convert_time(char *s, long delay)
     }
   }
   sprintf(s, "%ld ms", delay); s+=strlen(s);
+  return s;
 }
 
 /* Format a time of 'delay' ms */
 const char *
 gp_format_time(long delay)
 {
-  static char buf[64];
-  char *s = buf;
-
+  char *buf = stack_malloc(64), *s = buf;
   term_get_color(s, c_TIME);
-  convert_time(s + strlen(s), delay);
-  s+=strlen(s);
-  term_get_color(s, c_NONE);
-  s+=strlen(s);
-  s[0] = '.';
-  s[1] = '\n';
-  s[2] = 0; return buf;
+  s = convert_time(s + strlen(s), delay);
+  term_get_color(s, c_NONE); return buf;
+}
+
+GEN
+strtime(long delay)
+{
+  long n = nchar2nlong(64);
+  GEN x = cgetg(n+1, t_STR);
+  char *buf = GSTR(x), *t = buf + 64, *s = convert_time(buf, delay);
+  s++; while (s < t) *s++ = 0; /* pacify valgrind */
+  return x;
 }
 
 /********************************************************************/
@@ -773,7 +775,7 @@ static FILE *
 gprc_chk(const char *s)
 {
   FILE *f = fopen(s, "r");
-  if (f && !(GP_DATA->flags & gpd_QUIET)) err_printf("Reading GPRC: %s ...", s);
+  if (f && !(GP_DATA->flags & gpd_QUIET)) err_printf("Reading GPRC: %s\n", s);
   return f;
 }
 
@@ -993,7 +995,7 @@ gp_initrc(pari_stack *p_A)
   }
   pari_stack_delete(&s_env);
   pop_buffer();
-  if (!(GP_DATA->flags & gpd_QUIET)) err_printf("Done.\n\n");
+  if (!(GP_DATA->flags & gpd_QUIET)) err_printf("GPRC Done.\n\n");
   fclose(file);
 }
 
@@ -1115,7 +1117,7 @@ update_logfile(const char *prompt, const char *s)
   av = avma;
   p = strip_prompt(prompt); /* raw prompt */
 
-  switch (logstyle) {
+  switch (pari_logstyle) {
     case logstyle_TeX:
       fprintf(pari_logfile,
               "\\PARIpromptSTART|%s\\PARIpromptEND|%s\\PARIinputEND|%%\n",
@@ -1130,7 +1132,7 @@ update_logfile(const char *prompt, const char *s)
                                           term_get_color(NULL,c_NONE));
       break;
   }
-  avma = av;
+  set_avma(av);
 }
 
 void
@@ -1158,9 +1160,9 @@ get_line_from_file(const char *prompt, filtre_t *F, FILE *file)
 
   IM.file = (void*)file;
   if (file==stdin && cb_pari_fgets_interactive)
-    IM.fgets = (fgets_t)cb_pari_fgets_interactive;
+    IM.myfgets = (fgets_t)cb_pari_fgets_interactive;
   else
-    IM.fgets = (fgets_t)&fgets;
+    IM.myfgets = (fgets_t)&fgets;
   IM.getline = &file_input;
   IM.free = 0;
   if (! input_loop(F,&IM))
@@ -1209,7 +1211,7 @@ gp_read_line(filtre_t *F, const char *PROMPT)
     BLOCK_EH_END
   }
   else
-  { /* in case UI fakes non-interactivity, e.g. TeXmacs */
+  { /* in case UI fakes noninteractivity, e.g. TeXmacs */
     if (cb_pari_start_output && cb_pari_get_line_interactive)
       res = cb_pari_get_line_interactive(p, GP_DATA->prompt_cont, F);
     else
@@ -1243,12 +1245,18 @@ gp_alarm_fun(void) {
 
 void
 gp_sigint_fun(void) {
-  char buf[64];
+  char buf[150];
 #if defined(_WIN32)
   if (win32alrm) { win32alrm = 0; gp_alarm_fun(); return;}
 #endif
   if (cb_pari_start_output) cb_pari_start_output();
   convert_time(buf, timer_get(GP_DATA->T));
+  if (pari_mt_nbthreads > 1)
+  {
+    sprintf(buf + strlen(buf), " cpu time, ");
+    convert_time(buf + strlen(buf), walltimer_get(GP_DATA->Tw));
+    sprintf(buf + strlen(buf), " real time");
+  }
   pari_sigint(buf);
 }
 
@@ -1402,7 +1410,7 @@ tex2mail_output(GEN z, long n)
         fprintf(out, "\\%%%ld = ", n);
     }
     if (log) {
-      switch (logstyle) {
+      switch (pari_logstyle) {
       case logstyle_plain:
         fprintf(log, "%%%ld = ", n);
         break;
@@ -1414,14 +1422,14 @@ tex2mail_output(GEN z, long n)
         break;
       }
     }
-    avma = av;
+    set_avma(av);
   }
   /* output */
   fputGEN_pariout(z, &T, out);
   /* flush and restore, output to logfile */
   prettyp_wait(out);
   if (log) {
-    if (logstyle == logstyle_TeX) {
+    if (pari_logstyle == logstyle_TeX) {
       T.TeXstyle |= TEXSTYLE_BREAK;
       fputGEN_pariout(z, &T, log);
       fputc('%', log);
@@ -1457,6 +1465,7 @@ sd_graphcolormap(const char *v, long flag)
 
   if (v)
   {
+    pari_sp av = avma;
     char *t = gp_filter(v);
     if (*t != '[' || t[strlen(t)-1] != ']')
       pari_err(e_SYNTAX, "incorrect value for graphcolormap", t, t);
@@ -1521,7 +1530,7 @@ sd_graphcolormap(const char *v, long flag)
       default:
         pari_err(e_SYNTAX, "incorrect value for graphcolormap", p, t);
       }
-    pari_free(t);
+    set_avma(av);
   }
   if (flag == d_RETURN || flag == d_ACKNOWLEDGE)
   {
@@ -1658,11 +1667,14 @@ pari_print_version(void)
 {
   pari_sp av = avma;
   char *buf, *ver = what_cc();
+  const char *kver = pari_kernel_version();
   const char *date = paricfg_compiledate;
 
   pari_center(paricfg_version);
-  pari_center(paricfg_buildinfo);
-  buf = stack_malloc(strlen(date) +  32 + (ver? strlen(ver): 0));
+  buf = stack_malloc(strlen(paricfg_buildinfo) + 2 + strlen(kver));
+  (void)sprintf(buf, paricfg_buildinfo, kver);
+  pari_center(buf);
+  buf = stack_malloc(strlen(date) + 32 + (ver? strlen(ver): 0));
   if (ver) (void)sprintf(buf, "compiled: %s, %s", date, ver);
   else     (void)sprintf(buf, "compiled: %s", date);
   pari_center(buf);
@@ -1672,7 +1684,7 @@ pari_print_version(void)
   buf = stack_malloc(strlen(ver) + 64);
   (void)sprintf(buf, "(readline %s, extended help%s enabled)", ver,
                 has_ext_help()? "": " not");
-  pari_center(buf); avma = av;
+  pari_center(buf); set_avma(av);
 }
 
 static int
@@ -1718,7 +1730,7 @@ print_all_user_fun(int member)
       pari_printf("%s =\n  %Ps\n\n", ep->name, ep->value);
     }
   }
-  avma = av;
+  set_avma(av);
 }
 
 static char *
@@ -1785,8 +1797,8 @@ escape(const char *tch, int ismain)
       s = get_name(s);
       if (*s)
       {
-        (void)sd_logfile(s,d_ACKNOWLEDGE);
-        if (pari_logfile) break;
+        if (pari_logfile) { (void)sd_logfile(s,d_ACKNOWLEDGE);break; }
+        (void)sd_logfile(s,d_SILENT);
       }
       (void)sd_log(pari_logfile?"0":"1",d_ACKNOWLEDGE);
       break;
@@ -1818,9 +1830,9 @@ escape(const char *tch, int ismain)
         {
           long i, l = lg(x);
           pari_warn(warner,"setting %ld history entries", l-1);
-          for (i=1; i<l; i++) pari_add_hist(gel(x,i), 0);
+          for (i=1; i<l; i++) pari_add_hist(gel(x,i), 0, 0);
         }
-        avma = av;
+        set_avma(av);
       }
       break;
     case 's': dbg_pari_heap(); break;
@@ -1846,7 +1858,13 @@ chron(const char *s)
     if (*s == '#') s++;
     if (*s) return 0;
     t = gp_format_time(pari_get_histtime(0));
-    pari_printf("  ***   last result computed in %s", t);
+    if (pari_mt_nbthreads==1)
+      pari_printf("  ***   last result computed in %s.\n", t);
+    else
+    {
+      const char *r = gp_format_time(pari_get_histrtime(0));
+      pari_printf("  ***   last result: cpu time %s, real time %s.\n", t,r);
+    }
   }
   else { GP_DATA->chrono ^= 1; (void)sd_timer(NULL,d_ACKNOWLEDGE); }
   return 1;
