@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -14,13 +15,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
-static const long EXTRAPREC =
-#ifdef LONG_IS_64BIT
-  1;
-#else
-  2;
-#endif
-
 static GEN
 intlin(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec);
 
@@ -50,16 +44,15 @@ interp(GEN h, GEN s, long L, long bit, long D)
 {
   pari_sp av = avma;
   long e1,e2;
-  GEN dss, ss = polint_i(h + L-D,s + L-D, gen_0, D+1, &dss);
+  GEN ss = polintspec(h + L-D, s + L-D, gen_0, D+1, &e2);
 
   e1 = gexpo(ss);
-  e2 = gexpo(dss);
   if (DEBUGLEVEL>2)
   {
     err_printf("romb: iteration %ld, guess: %Ps\n", L,ss);
     err_printf("romb: relative error < 2^-%ld [target %ld bits]\n",e1-e2,bit);
   }
-  if (e1-e2 <= bit && (L <= 10 || e1 >= -bit)) { avma = av; return NULL; }
+  if (e1-e2 <= bit && (L <= 10 || e1 >= -bit)) return gc_NULL(av);
   return cxtoreal(ss);
 }
 
@@ -212,38 +205,97 @@ intnumromb0_bitprec(GEN a, GEN b, GEN code, long flag, long bit)
 /********************************************************************/
 /**             NUMERICAL INTEGRATION (Gauss-Legendre)             **/
 /********************************************************************/
-GEN
-intnumgaussinit(long n, long prec)
+/* N > 1; S[n] = n^2. If !last, Newton step z - P_N(z) / P'_N(z),
+* else [z, (N-1)!P_{N-1}(z)]. */
+static GEN
+Legendrenext(long N, GEN z, GEN S, int last)
 {
-  pari_sp ltop = avma;
-  GEN L, dp1, p1, p2, R, W;
-  long prec0 = prec + EXTRAPRECWORD;
-  long bitprec = prec2nbits(prec), i, d1;
-  if (n <= 0) n = (long)(bitprec*0.2258);
-  if (odd(n)) n++;
-  if (n == 2) n = 4;
-  /* n even >= 4, p1 is even */
-  prec = nbits2prec(3*bitprec/2 + 32);
-  L = pollegendre(n, 0); /* L_n = p1(x^2) */
-  p1 = Q_remove_denom(RgX_deflate(L, 2), &dp1);
-  d1 = vali(dp1);
-  p2 = ZX_deriv(p1); /* L_n' = 2x p2(x^2) / 2^d1 */
-  R = ZX_Uspensky(p1, gen_0, 1, 3*bitprec/2 + 32); /* positive roots of p1 */
-  n >>= 1;
-  W = cgetg(n+1, t_VEC);
-  for (i = 1; i <= n; ++i)
+  GEN q, Z, a, b, z2 = sqrr(z);
+  long n, n0, u;
+
+  q = subrs(mulur(3, z2), 1);
+  if (odd(N))
   {
-    GEN t, r2 = gel(R,i);
-    if (typ(r2) != t_REAL) r2 = gtofp(r2, prec);
-    gel(R,i) = sqrtr_abs(r2); /* positive root of L_n */
-    /* 2 / (L'(r)^2(1-r^2)) =  2^(2d1 - 1) / (1-r2)r2 (p2(r2))^2 */
-    t = mulrr(subrr(r2, sqrr(r2)), sqrr(poleval(p2, r2)));
-    shiftr_inplace(t,1-2*d1);
-    gel(W,i) = invr(t);
+    n0 = 3;
+    a = mulrr(z2, subrs(mulur(5, q), 4));
+    b = q;
   }
-  R = gprec_wtrunc(R,prec0);
-  W = gprec_wtrunc(W,prec0);
-  return gerepilecopy(ltop, mkvec2(R,W));
+  else
+  {
+    n0 = 2;
+    a = mulrr(q, z);
+    b = z;
+  }
+  for (n = n0, u = 2*n0 + 1; n < N; n += 2, u += 4)
+  {
+    b = subrr(mulur(u, a), mulir(gel(S,n), b));
+    a = subrr(mulur(u + 2, mulrr(z2, b)), mulir(gel(S,n+1), a));
+  }
+  q = divrr(a, subrr(a, mulur(N, b)));
+  Z = subrr(z, divrr(mulrr(subrs(z2, 1), q), mulur(N, z)));
+  return last? mkvec2(Z, mulrr(b, addrs(q, 1))): Z;
+}
+/* root ~ dz of Legendre polynomial P_N */
+static GEN
+Legendreroot(long N, double dz, GEN S, long bit)
+{
+  GEN z = dbltor(dz);
+  long e = - dblexpo(1 - dz*dz), n = expu(bit + 32 - e) - 2;
+  long j, pr = 1 + e + ((bit - e) >> n);
+  for (j = 1; j <= n; j++)
+  {
+    pr = 2 * pr - e;
+    z = Legendrenext(N, rtor(z, nbits2prec(pr)), S, j == n);
+  }
+  return z;
+}
+GEN
+intnumgaussinit(long N, long prec)
+{
+  pari_sp av;
+  long N2, j, k, l, bit;
+  GEN res, V, W, F, S;
+  double d;
+
+  prec += EXTRAPREC64;
+  bit = prec2nbits(prec);
+  if (N <= 0) { N = bit >> 2; if (odd(N)) N++; }
+  if (N == 1) retmkvec2(mkvec(gen_0), mkvec(gen_2));
+  res = cgetg(3, t_VEC);
+  if (N == 2)
+  {
+    GEN z = cgetr(prec);
+    gel(res,1) = mkvec(z);
+    gel(res,2) = mkvec(gen_1);
+    av = avma; affrr(divru(sqrtr(utor(3,prec)), 3), z);
+    return gc_const(av, res);
+  }
+  N2 = N >> 1; l = (N+3)>> 1;
+  gel(res,1) = V = cgetg(l, t_VEC);
+  gel(res,2) = W = cgetg(l, t_VEC);
+  gel(V,1) = odd(N)? gen_0: cgetr(prec);
+  gel(W,1) = cgetr(prec);
+  for (k = 2; k < l; k++) { gel(V,k) = cgetr(prec); gel(W,k) = cgetr(prec); }
+  av = avma; S = vecpowuu(N, 2); F = sqrr(mpfactr(N-1, prec));
+  if (!odd(N)) k = 1;
+  else
+  {
+    GEN c = divrr(sqrr(sqrr(mpfactr(N2, prec))), mulri(F, gel(S,N)));
+    shiftr_inplace(c, 2*N-1);
+    affrr(c, gel(W,1)); k = 2;
+  }
+  F = divri(shiftr(F, 1), gel(S,N));
+  d = 1 - (N-1) / pow((double)(2*N),3);
+  for (j = 4*N2-1; j >= 3; k++, j -= 4)
+  {
+    pari_sp av2 = avma;
+    GEN zw = Legendreroot(N, d * cos(M_PI * j / (4*N+2)), S, bit);
+    GEN z = gel(zw,1), w = gel(zw,2);
+    affrr(z, gel(V,k));
+    w = mulrr(F, divrr(subsr(1, sqrr(z)), sqrr(w)));
+    affrr(w, gel(W,k)); set_avma(av2);
+  }
+  return gc_const(av, res);
 }
 
 GEN
@@ -251,12 +303,15 @@ intnumgauss(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec)
 {
   pari_sp ltop = avma;
   GEN R, W, bma, bpa, S;
-  long n, i, prec2 = prec + EXTRAPREC;
+  long n, i, prec2 = prec + EXTRAPREC64;
   if (!tab)
     tab = intnumgaussinit(0,prec);
   else if (typ(tab) != t_INT)
   {
-    if (typ(tab) != t_VEC || lg(tab) != 3)
+    if (typ(tab) != t_VEC || lg(tab) != 3
+        || typ(gel(tab,1)) != t_VEC
+        || typ(gel(tab,2)) != t_VEC
+        || lg(gel(tab,1)) != lg(gel(tab,2)))
       pari_err_TYPE("intnumgauss",tab);
   }
   else
@@ -264,16 +319,25 @@ intnumgauss(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec)
 
   R = gel(tab,1); n = lg(R)-1;
   W = gel(tab,2);
-  a = gprec_w(a, prec2);
-  b = gprec_w(b, prec2);
+  a = gprec_wensure(a, prec2);
+  b = gprec_wensure(b, prec2);
   bma = gmul2n(gsub(b,a), -1); /* (b-a)/2 */
   bpa = gadd(bma, a); /* (b+a)/2 */
-  S = gen_0;
-  for (i = 1; i <= n; ++i)
+  if (!signe(gel(R,1)))
+  { /* R[1] = 0, use middle node only once */
+    S = gmul(gel(W,1), eval(E, bpa));
+    i = 2;
+  }
+  else
   {
-    GEN r = gel(R,i);
-    GEN P = eval(E, gadd(bpa, gmul(bma, r)));
-    GEN M = eval(E, gsub(bpa, gmul(bma, r)));
+    S = gen_0;
+    i = 1;
+  }
+  for (; i <= n; ++i)
+  {
+    GEN h = gmul(bma, gel(R,i)); /* != 0 */
+    GEN P = eval(E, gadd(bpa, h));
+    GEN M = eval(E, gsub(bpa, h));
     S = gadd(S, gmul(gel(W,i), gadd(P,M)));
     S = gprec_wensure(S, prec2);
   }
@@ -289,7 +353,7 @@ intnumgauss0(GEN a, GEN b, GEN code, GEN tab, long prec)
 /********************************************************************/
 
 typedef struct _intdata {
-  long eps;  /* bit accuracy of current precision */
+  long bit;  /* bit accuracy of current precision */
   long l; /* table lengths */
   GEN tabx0; /* abscissa phi(0) for t = 0 */
   GEN tabw0; /* weight phi'(0) for t = 0 */
@@ -367,7 +431,7 @@ intinit_start(intdata *D, long m, double TUNE, long prec)
   h = divru(nh, n);
   if (m > 0) { h = gmul2n(h,-m); n <<= m; }
   D->h = h;
-  D->eps = bitprec;
+  D->bit = bitprec;
   D->l = l = n+1;
   D->tabxp = cgetg(l, t_VEC);
   D->tabwp = cgetg(l, t_VEC);
@@ -398,30 +462,29 @@ divr2_ip(GEN x) { shiftr_inplace(x, -1); return x; }
 static GEN
 inittanhsinh(long m, long prec)
 {
-  GEN et, ex, pi = mppi(prec);
+  GEN e, ei, ek, eik, pi = mppi(prec);
   long k, nt = -1;
   intdata D;
 
   intinit_start(&D, m, 1.86, prec);
   D.tabx0 = real_0(prec);
   D.tabw0 = Pi2n(-1,prec);
-  et = ex = mpexp(D.h);
+  e = mpexp(D.h); ek = mulrr(pi, e);
+  ei = invr(e); eik = mulrr(pi, ei);
   for (k = 1; k < D.l; k++)
   {
     GEN xp, wp, ct, st, z;
     pari_sp av;
     gel(D.tabxp,k) = cgetr(prec);
     gel(D.tabwp,k) = cgetr(prec); av = avma;
-    ct = divr2_ip(addrr(et, invr(et))); /* ch(kh) */
-    st = subrr(et, ct); /* sh(kh) */
-    z = invr( addrs(mpexp(mulrr(pi, st)), 1) );
-    shiftr_inplace(z, 1);
+    ct = divr2_ip(addrr(ek, eik)); /* Pi ch(kh) */
+    st = subrr(ek, ct); /* Pi sh(kh) */
+    z = invr( addrs(mpexp(st), 1) );
+    shiftr_inplace(z, 1); if (expo(z) < -D.bit) { nt = k-1; break; }
     xp = subsr(1, z);
-    wp = divr2_ip(mulrr(mulrr(pi,ct), mulrr(z, subsr(2, z))));
-    if (expo(wp) < -D.eps) { nt = k-1; break; }
-    affrr(xp, gel(D.tabxp,k));
-    if (absrnz_equal1(gel(D.tabxp,k))) { nt = k-1; break; }
-    affrr(wp, gel(D.tabwp,k)); et = gerepileuptoleaf(av, mulrr(et, ex));
+    wp = divr2_ip(mulrr(ct, subsr(1, sqrr(xp))));
+    affrr(xp, gel(D.tabxp,k)); mulrrz(ek, e, ek);
+    affrr(wp, gel(D.tabwp,k)); mulrrz(eik, ei, eik); set_avma(av);
   }
   return intinit_end(&D, nt, 0);
 }
@@ -451,7 +514,7 @@ initsinhsinh(long m, long prec)
     exu = invr(ext);
     xp = divr2_ip(subrr(ext, exu));
     wp = divr2_ip(mulrr(ct, addrr(ext, exu)));
-    if (expo(wp) - 2*expo(xp) < -D.eps) { nt = k-1; break; }
+    if (expo(wp) - 2*expo(xp) < -D.bit) { nt = k-1; break; }
     affrr(xp, gel(D.tabxp,k));
     affrr(wp, gel(D.tabwp,k)); et = gerepileuptoleaf(av, mulrr(et, ex));
   }
@@ -478,7 +541,7 @@ initsinh(long m, long prec)
     eti = invr(et);
     xp = subrr(et, eti);
     wp = addrr(et, eti);
-    if (cmprs(xp, (long)(M_LN2*(expo(wp)+D.eps) + 1)) > 0) { nt = k-1; break; }
+    if (cmprs(xp, (long)(M_LN2*(expo(wp)+D.bit) + 1)) > 0) { nt = k-1; break; }
     affrr(xp, gel(D.tabxp,k));
     affrr(wp, gel(D.tabwp,k)); et = gerepileuptoleaf(av, mulrr(et, ex));
   }
@@ -508,7 +571,7 @@ initexpsinh(long m, long prec)
     gel(D.tabwp,k) = mulrr(xp, t);
     gel(D.tabxm,k) = invr(xp);
     gel(D.tabwm,k) = mulrr(gel(D.tabxm,k), t);
-    if (expo(gel(D.tabxm,k)) < -D.eps) { nt = k-1; break; }
+    if (expo(gel(D.tabxm,k)) < -D.bit) { nt = k-1; break; }
   }
   return intinit_end(&D, nt, nt);
 }
@@ -537,8 +600,8 @@ initexpexp(long m, long prec)
     xp = mpexp(subrr(kh, et));
     xm = mpexp(negr(addrr(kh, eti)));
     wp = mulrr(xp, addsr(1, et));
+    if (expo(xm) < -D.bit && cmprs(xp, (long)(M_LN2*(expo(wp)+D.bit) + 1)) > 0) { nt = k-1; break; }
     wm = mulrr(xm, addsr(1, eti));
-    if (expo(xm) < -D.eps && cmprs(xp, (long)(M_LN2*(expo(wp)+D.eps) + 1)) > 0) { nt = k-1; break; }
     affrr(xp, gel(D.tabxp,k));
     affrr(wp, gel(D.tabwp,k));
     affrr(xm, gel(D.tabxm,k));
@@ -584,7 +647,7 @@ initnumsine(long m, long prec)
     wp = mulrr(subrr(extm1, mulrr(kct, extm)), mulrr(pi, sqrr(extm2)));
     xm = mulrr(negr(kpi), extp2); /* phi(-kh) */
     wm = mulrr(addrr(extp1, mulrr(kct, extp)), mulrr(pi, sqrr(extp2)));
-    if (expo(wm) < -D.eps && expo(extm) + exh + expu(10 * k) < -D.eps) { nt = k-1; break; }
+    if (expo(wm) < -D.bit && expo(extm) + exh + expu(10 * k) < -D.bit) { nt = k-1; break; }
     affrr(xp, gel(D.tabxp,k));
     affrr(wp, gel(D.tabwp,k));
     affrr(xm, gel(D.tabxm,k));
@@ -796,16 +859,14 @@ intninfinf(void *E, GEN (*eval)(void*, GEN), GEN tab)
 static GEN
 f_getycplx(GEN a, long prec)
 {
+  GEN a2R, a2I;
   long s;
-  GEN tmp, a2R, a2I;
 
   if (lg(a) == 2 || gequal0(gel(a,2))) return gen_1;
   a2R = real_i(gel(a,2));
   a2I = imag_i(gel(a,2));
   s = gsigne(a2I); if (s < 0) a2I = gneg(a2I);
-  tmp = s ? ginv(a2I) : ginv(a2R);
-  if (gprecision(tmp) < prec) tmp = gprec_w(tmp, prec);
-  return tmp;
+  return ginv(gprec_wensure(s ? a2I : a2R, prec));
 }
 
 static void
@@ -960,7 +1021,7 @@ intnuminit_i(GEN a, GEN b, long m, long prec)
 
   if (m > 30) pari_err_OVERFLOW("intnuminit [m]");
   if (m < 0) pari_err_DOMAIN("intnuminit", "m", "<", gen_0, stoi(m));
-  l = prec+EXTRAPREC;
+  l = prec+EXTRAPREC64;
   codea = transcode(a, "a"); if (codea == f_SER) codea = f_REG;
   codeb = transcode(b, "b"); if (codeb == f_SER) codeb = f_REG;
   if (codea == f_SINGSER || codeb == f_SINGSER)
@@ -1046,7 +1107,7 @@ intnuminit0(GEN a, GEN b, GEN tab, long prec)
 }
 
 /* Assigns the values of the function weighted by w[k] at quadrature points x[k]
- * [replacing the weights]. Return the index of the last non-zero coeff */
+ * [replacing the weights]. Return the index of the last nonzero coeff */
 static long
 weight(void *E, GEN (*eval)(void *, GEN), GEN x, GEN w)
 {
@@ -1097,7 +1158,7 @@ intfuncinit(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, long m, long prec)
 static GEN
 intnum_i(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec)
 {
-  GEN S = gen_0, res1, res2, kma, kmb;
+  GEN S = gen_0, kma, kmb;
   long sb, sgns = 1, codea = transcode(a, "a"), codeb = transcode(b, "b");
 
   if (codea == f_REG && typ(a) == t_VEC) a = gel(a,1);
@@ -1113,9 +1174,8 @@ intnum_i(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec)
     else
     {
       GEN c = gmul2n(gadd(gel(a,1), gel(b,1)), -1);
-      res1 = intnsing(E, eval, a, c, gel(tab,1));
-      res2 = intnsing(E, eval, b, c, gel(tab,2));
-      S = gsub(res1, res2);
+      S = gsub(intnsing(E, eval, a, c, gel(tab,1)),
+               intnsing(E, eval, b, c, gel(tab,2)));
     }
     return (sgns < 0) ? gneg(S) : S;
   }
@@ -1131,7 +1191,7 @@ intnum_i(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec)
   if (is_fin_f(codea))
   { /* either codea == f_SING  or codea == f_REG and codeb = f_YOSCC
      * or (codeb == f_YOSCS and !gequal0(a)) */
-    GEN c = real_i(codea == f_SING? gel(a,1): a);
+    GEN S2, c = real_i(codea == f_SING? gel(a,1): a);
     switch(codeb)
     {
       case f_YOSCC: case f_YOSCS:
@@ -1149,12 +1209,12 @@ intnum_i(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec)
         c = sb > 0? addiu(gceil(c), 1): subiu(gfloor(c), 1);
         break;
     }
-    res1 = codea==f_SING? intnsing(E, eval, a, c, gel(tab,1))
-                        : intn    (E, eval, a, c, gel(tab,1));
-    res2 = intninfpm(E, eval, c, sb*codeb, gel(tab,2));
-    if (sb < 0) res2 = gneg(res2);
-    res1 = gadd(res1, res2);
-    return sgns < 0 ? gneg(res1) : res1;
+    S = codea==f_SING? intnsing(E, eval, a, c, gel(tab,1))
+                     : intn    (E, eval, a, c, gel(tab,1));
+    S2 = intninfpm(E, eval, c, sb*codeb, gel(tab,2));
+    if (sb < 0) S2 = gneg(S2);
+    S = gadd(S, S2);
+    return sgns < 0 ? gneg(S) : S;
   }
   /* now a and b are infinite */
   if (codea * sb > 0)
@@ -1198,14 +1258,14 @@ GEN
 intnum(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec)
 {
   pari_sp ltop = avma;
-  long l = prec+EXTRAPREC;
+  long l = prec+EXTRAPREC64;
   GEN na = NULL, nb = NULL, S;
 
   if (transcode(a,"a") == f_SINGSER) {
     long v = gvar(gel(a,1));
     if (v != NO_VARIABLE) {
       na = cgetg(3,t_VEC);
-      gel(na,1) = polcoeff0(gel(a,1),0,v);
+      gel(na,1) = polcoef(gel(a,1),0,v);
       gel(na,2) = gel(a,2);
     }
     a = gel(a,1);
@@ -1214,7 +1274,7 @@ intnum(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec)
     long v = gvar(gel(b,1));
     if (v != NO_VARIABLE) {
       nb = cgetg(3,t_VEC);
-      gel(nb,1) = polcoeff0(gel(b,1),0,v);
+      gel(nb,1) = polcoef(gel(b,1),0,v);
       gel(nb,2) = gel(b,2);
     }
     b = gel(b,1);
@@ -1228,7 +1288,7 @@ intnum(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec)
     return gerepilecopy(ltop, S);
   }
   tab = intnuminit0(a, b, tab, prec);
-  S = intnum_i(E, eval, gprec_w(a, l), gprec_w(b, l), tab, prec);
+  S = intnum_i(E, eval, gprec_wensure(a, l), gprec_wensure(b, l), tab, prec);
   return gerepilecopy(ltop, gprec_wtrunc(S, prec));
 }
 
@@ -1372,42 +1432,39 @@ intnumdoub0(GEN a, GEN b, int nc, int nd, int nf, GEN tabext, GEN tabint, long p
 }
 #endif
 
-
-/* The quotient-difference algorithm. Given a vector M, convert the series
- * S = \sum_{n >= 0} M[n+1]z^n into a continued fraction.
- * Compute the c[n] such that
- * S = c[1] / (1 + c[2]z / (1+c[3]z/(1+...c[lim]z))),
- * Compute A[n] and B[n] such that
- * S = M[1]/ (1+A[1]*z+B[1]*z^2 / (1+A[2]*z+B[2]*z^2/ (1+...1/(1+A[lim\2]*z)))),
- * Assume lim <= #M.
- * Does not work for certain M. */
-
-/* Given a continued fraction CF output by the quodif program,
-convert it into an Euler continued fraction A(n), B(n), where
-$1/(1+c[2]z/(1+c[3]z/(1+..c[lim]z)))
-=1/(1+A[1]*z+B[1]*z^2/(1+A[2]*z+B[2]*z^2/(1+...1/(1+A[lim\2]*z)))). */
+/* Given a continued fraction C output by QD convert it into an Euler
+ * continued fraction A(n), B(n), where
+ * 1 / (1 + C[2]z / (1+C[3]z / (1+..C[lim]z)))
+ * = 1 / (1+A[1]*z+B[1]*z^2/(1+A[2]*z+B[2]*z^2/(1+...1/(1+A[lim\2]*z)))). */
 static GEN
-contfrac_Euler(GEN CF)
+contfrac_Euler(GEN C)
 {
-  long lima, limb, i, lim = lg(CF)-1;
-  GEN A, B;
-  lima = lim/2;
-  limb = (lim - 1)/2;
-  A = cgetg(lima+1, t_VEC);
-  B = cgetg(limb+1, t_VEC);
-  gel (A, 1) = gel(CF, 2);
-  for (i=2; i <= lima; ++i) gel(A,i) = gadd(gel(CF, 2*i), gel(CF, 2*i-1));
-  for (i=1; i <= limb; ++i) gel(B,i) = gneg(gmul(gel(CF, 2*i+1), gel(CF, 2*i)));
+  long i, n = lg(C) - 1, a = n/2, b = (n - 1)/2;
+  GEN A = cgetg(a+1, t_VEC), B = cgetg(b+1, t_VEC);
+  gel(A,1) = gel(C,2);
+  if (!b) return mkvec2(A, B);
+  gel(B,1) = gneg(gmul(gel(C,3), gel(C,2)));
+  for (i = 2; i <= b; i++)
+  {
+    GEN u = gel(C,2*i);
+    gel(A,i) = gadd(u, gel(C, 2*i-1));
+    gel(B,i) = gneg(gmul(gel(C, 2*i+1), u));
+  }
+  if (a != b) gel(A,a) = gadd(gel(C, 2*a), gel(C, 2*a-1));
   return mkvec2(A, B);
 }
 
+/* The quotient-difference algorithm. Given a vector M, convert the series
+ * S = sum_{n >= 0} M[n+1] z^n into a continued fraction.
+ * Compute the c[n] such that
+ * S = c[1] / (1 + c[2]z / (1+c[3]z/(1+...c[lim]z))),
+ * Assume lim <= #M. Does not work for certain M. */
 static GEN
-contfracinit_i(GEN M, long lim)
+QD(GEN M, long lim)
 {
   pari_sp av;
   GEN e, q, c;
-  long lim2;
-  long j, k;
+  long lim2, j, k;
   e = zerovec(lim);
   c = zerovec(lim+1); gel(c, 1) = gel(M, 1);
   q = cgetg(lim+1, t_VEC);
@@ -1432,37 +1489,49 @@ contfracinit_i(GEN M, long lim)
   return c;
 }
 
-GEN
-contfracinit(GEN M, long lim)
+static GEN
+quodif_i(GEN M, long n)
 {
-  pari_sp ltop = avma;
-  GEN c;
   switch(typ(M))
   {
     case t_RFRAC:
-      if (lim < 0) pari_err_TYPE("contfracinit",M);
-      M = gadd(M, zeroser(gvar(M), lim + 2)); /*fall through*/
+      if (n < 0) pari_err_TYPE("contfracinit",M);
+      M = gtoser(M, varn(gel(M,2)), n+3); /*fall through*/
     case t_SER: M = gtovec(M); break;
-    case t_POL: M = gtovecrev(M); break;
+    case t_POL: M = RgX_to_RgC(M, degpol(M)+1); break;
     case t_VEC: case t_COL: break;
     default: pari_err_TYPE("contfracinit", M);
   }
-  if (lim < 0)
-    lim = lg(M)-2;
-  else if (lg(M)-1 <= lim)
-    pari_err_COMPONENT("contfracinit", "<", stoi(lg(M)-1), stoi(lim));
-  if (lim < 0) retmkvec2(cgetg(1,t_VEC),cgetg(1,t_VEC));
-  c = contfracinit_i(M, lim);
-  return gerepilecopy(ltop, contfrac_Euler(c));
+  if (n < 0)
+  {
+    n = lg(M)-2;
+    if (n < 0) return cgetg(1,t_VEC);
+  }
+  else if (lg(M)-1 <= n)
+    pari_err_COMPONENT("contfracinit", "<", stoi(lg(M)-1), stoi(n));
+  return QD(M, n);
+}
+GEN
+quodif(GEN M, long n)
+{
+  pari_sp av = avma;
+  return gerepilecopy(av, quodif_i(M, n));
+}
+GEN
+contfracinit(GEN M, long n)
+{
+  pari_sp av = avma;
+  GEN C = quodif_i(M, n);
+  if (lg(C) < 3) { set_avma(av); retmkvec2(cgetg(1,t_VEC), cgetg(1,t_VEC)); }
+  return gerepilecopy(av, contfrac_Euler(C));
 }
 
 /* Evaluate at 1/tinv the nlim first terms of the continued fraction output by
- * contfracinit. */
-/* Not stack clean */
+ * contfracinit. Shallow. */
 GEN
 contfraceval_inv(GEN CF, GEN tinv, long nlim)
 {
-  pari_sp btop;
+  pari_sp av;
   long j;
   GEN S = gen_0, S1, S2, A, B;
   if (typ(CF) != t_VEC || lg(CF) != 3) pari_err_TYPE("contfraceval", CF);
@@ -1474,7 +1543,7 @@ contfraceval_inv(GEN CF, GEN tinv, long nlim)
     pari_err_COMPONENT("contfraceval", ">", stoi(lg(A)-1), stoi(nlim));
   if (lg(B)+1 <= nlim)
     pari_err_COMPONENT("contfraceval", ">", stoi(lg(B)), stoi(nlim));
-  btop = avma;
+  av = avma;
   if (nlim <= 1) return lg(A)==1? gen_0: gdiv(tinv, gadd(gel(A, 1), tinv));
   switch(nlim % 3)
   {
@@ -1496,7 +1565,7 @@ contfraceval_inv(GEN CF, GEN tinv, long nlim)
     S2 = gadd(gmul(gadd(gel(A, j-1), tinv), S1), gel(B, j-1));
     S3 = gadd(gmul(gadd(gel(A, j-2), tinv), S2), gmul(gel(B, j-2), S1));
     S = gdiv(gmul(gel(B, j-3), S2), S3);
-    if (gc_needed(btop, 3)) S = gerepilecopy(btop, S);
+    if (gc_needed(av, 3)) S = gerepilecopy(av, S);
   }
   return gdiv(tinv, gadd(gadd(gel(A, 1), tinv), S));
 }
@@ -1504,8 +1573,8 @@ contfraceval_inv(GEN CF, GEN tinv, long nlim)
 GEN
 contfraceval(GEN CF, GEN t, long nlim)
 {
-  pari_sp ltop = avma;
-  return gerepileupto(ltop, contfraceval_inv(CF, ginv(t), nlim));
+  pari_sp av = avma;
+  return gerepileupto(av, contfraceval_inv(CF, ginv(t), nlim));
 }
 
 /* MONIEN SUMMATION */
@@ -1524,7 +1593,7 @@ monrefine(GEN Q, GEN QP, GEN z, long prec)
     if (gcmp(gabs(prnew, prec), gabs(pr, prec)) >= 0) break;
     pr = prnew;
   }
-  z = gprec_w(z, 2*prec-2);
+  z = gprec_wensure(z, 2*prec-2);
   z = gsub(z, gdiv(poleval(Q, z), poleval(QP, z)));
   return gerepileupto(av, z);
 }
@@ -1551,7 +1620,7 @@ Pade(GEN M, GEN *pP, GEN *pQ)
 {
   pari_sp av = avma;
   long n = lg(M)-2, i;
-  GEN v = contfracinit_i(M, n), P = pol_0(0), Q = pol_1(0);
+  GEN v = QD(M, n), P = pol_0(0), Q = pol_1(0);
   /* evaluate continued fraction => Pade approximants */
   for (i = n-1; i >= 1; i--)
   { /* S = P/Q: S -> v[i]*x / (1+S) */
@@ -1575,8 +1644,8 @@ veczetaprime(GEN a, GEN b, long N, long prec)
   long l = nbits2prec(pr), e = fpr / 2;
   GEN eps, A, B;
   newprec = nbits2prec(pr + BITS_IN_LONG);
-  a = gprec_w(a, newprec);
-  b = gprec_w(b, newprec);
+  a = gprec_wensure(a, newprec);
+  b = gprec_wensure(b, newprec);
   eps = real2n(-e, l);
   A = veczeta(a, gsub(b, eps), N, newprec);
   B = veczeta(a, gadd(b, eps), N, newprec);
@@ -1644,7 +1713,7 @@ sumnummonieninit_i(GEN a, GEN b, GEN w, GEN n0, long prec)
   double bit = 2*prec2nbits(prec) / gtodouble(ga), D = bit*M_LN2;
   double da = maxdd(1., gtodouble(a));
   long n = (long)ceil(D / (da*(log(D)-1)));
-  long j, prec2, prec0 = prec + EXTRAPRECWORD;
+  long j, prec2, prec0 = prec + EXTRAPREC64;
   double bit0 = ceil((2*n+1)*LOG2_10);
   int neg = 1;
   struct mon_w S;
@@ -1654,8 +1723,8 @@ sumnummonieninit_i(GEN a, GEN b, GEN w, GEN n0, long prec)
   prec = nbits2prec(maxdd(2.05*da*bit, bit0));
   prec2 = nbits2prec(maxdd(1.3*da*bit, bit0));
   S.w = w;
-  S.a = a = gprec_w(a, 2*prec-2);
-  S.b = b = gprec_w(b, 2*prec-2);
+  S.a = a = gprec_wensure(a, 2*prec-2);
+  S.b = b = gprec_wensure(b, 2*prec-2);
   S.n = n;
   S.j = 1;
   S.prec = prec;
@@ -1738,6 +1807,7 @@ sumnummonieninit(GEN asymp, GEN w, GEN n0, long prec)
       b = asymp;
     }
     if (gsigne(a) <= 0) pari_err_DOMAIN(fun, "a", "<=", gen_0, a);
+    if (!isinR(b)) pari_err_TYPE(fun, b);
     if (gcmpgs(gadd(a,b), 1) <= 0)
       pari_err_DOMAIN(fun, "a+b", "<=", gen_1, mkvec2(a,b));
   }
@@ -1779,7 +1849,7 @@ sumnummonien(void *E, GEN (*eval)(void*,GEN), GEN n0, GEN tab, long prec)
     S = gadd(S, gmul(gel(vwt,i), eval(E, gel(vabs,i))));
     S = gprec_wensure(S, prec);
   }
-  return gerepilecopy(av, gprec_w(S, prec));
+  return gerepilecopy(av, gprec_wtrunc(S, prec));
 }
 
 static GEN
@@ -1799,7 +1869,7 @@ sumnuminit(GEN fast, long prec)
   N = (long)ceil(M_LN2*bitprec/(w*(1+w))+5);
   k = (long)ceil(N*w); if (k&1) k--;
 
-  prec += EXTRAPRECWORD;
+  prec += EXTRAPREC64;
   k2 = k/2;
   s = RgX_to_ser(monomial(real_1(prec),1,0), k+3);
   s = gmul2n(gasinh(s, prec), 2); /* asinh(x)/d, d = 1/4 */
@@ -1826,7 +1896,7 @@ sumnuminit(GEN fast, long prec)
   gel(res, 2) = utoi(N);
   gel(res, 3) = utoi(k);
   if (!fast) fast = get_oo(gen_0);
-  gel(res, 5) = intnuminit(gel(res,2), fast, 0, prec - EXTRAPRECWORD);
+  gel(res, 5) = intnuminit(gel(res,2), fast, 0, prec - EXTRAPREC64);
   return res;
 }
 
@@ -1861,7 +1931,7 @@ sumnum(void *E, GEN (*eval)(void*, GEN), GEN a, GEN tab, long prec)
   k = itos(gel(tab,3));
   v = gel(tab,4);
   tabint = gel(tab,5);
-  prec2 = prec+EXTRAPRECWORD;
+  prec2 = prec+EXTRAPREC64;
   av2 = avma;
   S = gmul(eval(E, stoi(N)), real2n(-1,prec2));
   for (m = as; m < N; m++)
@@ -1887,7 +1957,7 @@ sumnum(void *E, GEN (*eval)(void*, GEN), GEN a, GEN tab, long prec)
     S = gprec_wensure(S, prec2);
   }
   S = gadd(S, intnum(E, eval,stoi(N), fast, tabint, prec2));
-  return gerepilecopy(av, gprec_w(S, prec));
+  return gerepilecopy(av, gprec_wtrunc(S, prec));
 }
 
 GEN
@@ -1904,12 +1974,13 @@ intnumgauexpinit(long prec)
 {
   pari_sp ltop = avma;
   GEN V, N, E, P, Q, R, vabs, vwt;
-  long l, n, k, j, prec2, prec0 = prec + EXTRAPRECWORD, bit = prec2nbits(prec);
+  long l, n, k, j, prec2, prec0 = prec + EXTRAPREC64, bit = prec2nbits(prec);
 
   n = (long)ceil(bit*0.226);
   n |= 1; /* make n odd */
   prec = nbits2prec(1.5*bit + 32);
   prec2 = maxss(prec0, nbits2prec(1.15*bit + 32));
+  constbern(n+3);
   V = cgetg(n + 4, t_VEC);
   for (k = 1; k <= n + 3; ++k)
     gel(V, k) = gtofp(gdivgs(bernfrac(2*k), odd(k)? 2*k: -2*k), prec);
@@ -1949,7 +2020,7 @@ intnumgauexp(void *E, GEN (*eval)(void*,GEN), GEN gN, GEN tab, long prec)
     S = gadd(S, gmul(gdiv(w,x), cxtoreal(t)));
     S = gprec_wensure(S, prec);
   }
-  return gerepilecopy(av, gprec_w(S, prec));
+  return gerepilecopy(av, gprec_wtrunc(S, prec));
 }
 
 GEN
@@ -2013,33 +2084,30 @@ GEN
 sumnumap0(GEN a, GEN code, GEN tab, long prec)
 { EXPR_WRAP(code, sumnumap(EXPR_ARG, a, tab, prec)); }
 
-
 /* max (1, |zeros|), P a t_POL or scalar */
-static GEN
+static double
 polmax(GEN P)
 {
-  GEN r;
-  if (typ(P) != t_POL || degpol(P) <= 0) return gen_1;
-  r = polrootsbound(P, NULL);
-  if (gcmpgs(r, 1) <= 0) return gen_1;
-  return r;
+  pari_sp av = avma;
+  double r;
+  if (typ(P) != t_POL || degpol(P) <= 0) return 1.0;
+  r = gtodouble(polrootsbound(P, NULL));
+  return gc_double(av, maxdd(r, 1.0));
 }
 
 /* max (1, |poles|), F a t_POL or t_RFRAC or scalar */
-static GEN
+static double
 ratpolemax(GEN F)
 {
-  if (typ(F) == t_POL) return gen_1;
-  return polmax(gel(F, 2));
+  if (typ(F) == t_POL) return 1.0;
+  return polmax(gel(F,2));
 }
-/* max (1, |poles|, |zeros|), sets *p = max(1, |poles|)) */
-static GEN
-ratpolemax2(GEN F, GEN *p)
+/* max (1, |poles|, |zeros|)) */
+static double
+ratpolemax2(GEN F)
 {
-  GEN t;
-  if (typ(F) == t_POL) { if (p) *p = gen_1; return polmax(F); }
-  t = polmax(gel(F,2)); if (p) *p = t;
-  return gmax_shallow(t, polmax(gel(F,1)));
+  if (typ(F) == t_POL) return polmax(F);
+  return maxdd(polmax(gel(F,1)), polmax(gel(F,2)));
 }
 
 static GEN
@@ -2058,16 +2126,16 @@ intnumainfrat(GEN F, long N, double r, long prec)
   GEN S, ser;
   pari_sp av = avma;
 
-  lim = (long)ceil(B/log2(N/r)) + 1;
-  ser = gmul(F, real_1(prec + EXTRAPREC));
-  ser = rfracrecip_to_ser_absolute(ser, lim);
+  lim = (long)ceil(B/log2(N/r));
+  ser = gmul(F, real_1(prec + EXTRAPREC64));
+  ser = rfracrecip_to_ser_absolute(ser, lim+2);
   v = valp(ser);
   S = gdivgs(sercoeff(ser,lim+1), lim*N);
   /* goes down to 2, but coeffs are 0 in degree < v */
   for (k = lim; k >= v; k--) /* S <- (S + coeff(ser,k)/(k-1)) / N */
     S = gdivgs(gadd(S, gdivgs(sercoeff(ser,k), k-1)), N);
   if (v-2) S = gdiv(S, powuu(N, v-2));
-  return gerepilecopy(av, gprec_w(S, prec));
+  return gerepilecopy(av, gprec_wtrunc(S, prec));
 }
 
 static GEN
@@ -2120,7 +2188,7 @@ get_kN(long r, long B, long *pk, long *pN)
   long k = maxss(50, (long)ceil(0.241*B));
   GEN z;
   if (k&1L) k++;
-  *pk = k;
+  *pk = k; constbern(k >> 1);
   z = sqrtnr_abs(gmul2n(gtofp(bernfrac(k), LOWDEFAULTPREC), B), k);
   *pN = maxss(2*r, r + 1 + itos(gceil(z)));
 }
@@ -2134,7 +2202,7 @@ sumnumrat_i(GEN F, GEN F0, GEN vF, long prec)
   double r;
   if (poldegree(F, -1) > -2) pari_err(e_MISC, "sum diverges in sumnumrat");
   vx = varn(gel(F,2));
-  r = gtodouble(ratpolemax(F));
+  r = ratpolemax(F);
   get_kN((long)ceil(r), B, &k,&N);
   intf = intnumainfrat(F, N, r, prec);
   /* N > ratpolemax(F) is not a pole */
@@ -2178,7 +2246,7 @@ sumnumrat(GEN F, GEN a, long prec)
         GEN F2 = gsubst(F, vx, RgX_neg(pol_x(vx)));
         vF = mkvec2(F,F2);
         F = gadd(F, F2);
-        if (gequal0(F)) { avma = av; return real_0(prec); }
+        if (gequal0(F)) { set_avma(av); return real_0(prec); }
         F0 = rfrac_eval0(gel(vF,1));
         break;
       }
@@ -2188,11 +2256,15 @@ sumnumrat(GEN F, GEN a, long prec)
   }
   return gerepileupto(av, sumnumrat_i(F, F0, vF, prec));
 }
-static GEN
-_mpmul(GEN x, GEN y)
+/* deg ((a / b) - 1), assuming b a t_POL of positive degree in main variable  */
+static long
+rfracm1_degree(GEN a, GEN b)
 {
-  if (!x) return y;
-  return y? mpmul(x, y): x;
+  long da, db;
+  if (typ(a) != t_POL || varn(a) != varn(b)) return 0;
+  da = degpol(a);
+  db = degpol(b); if (da != db) return maxss(da - db, 0);
+  return degpol(RgX_sub(a,b)) - db;
 }
 
 /* prod_{n >= a} F(n) */
@@ -2201,20 +2273,21 @@ prodnumrat(GEN F, long a, long prec)
 {
   pari_sp ltop = avma;
   long B = prec2nbits(prec), j, k, m, N, vx;
-  GEN S, S1, S2, intf, G, F1 = gsubgs(F,1);
+  GEN S, S1, S2, intf, G;
   double r;
 
-  switch(typ(F1))
+  switch(typ(F))
   {
     case t_RFRAC: break;
     case t_INT: case t_REAL: case t_COMPLEX: case t_POL:
-      if (gequal0(F1)) return real_1(prec);
+      if (gequal1(F)) return real_1(prec);
     default: pari_err_TYPE("prodnumrat",F);
   }
-  if (poldegree(F1,-1) > -2) pari_err(e_MISC, "product diverges in prodnumrat");
+  if (rfracm1_degree(gel(F,1), gel(F,2)) > -2)
+    pari_err(e_MISC, "product diverges in prodnumrat");
   vx = varn(gel(F,2));
   if (a) F = gsubst(F, vx, gaddgs(pol_x(vx), a));
-  r = gtodouble(ratpolemax2(F, NULL));
+  r = ratpolemax2(F);
   get_kN((long)ceil(r), B, &k,&N);
   G = gdiv(deriv(F, vx), F);
   intf = intnumainfrat(gmul(pol_x(vx),G), N, r, prec);
@@ -2229,54 +2302,92 @@ prodnumrat(GEN F, long a, long prec)
   return gerepileupto(ltop, gmul(S1, gexp(gsub(intf, S2), prec)));
 }
 
+/* fan = factoru(n); sum_{d | n} mu(d)/d * s[n/d] */
 static GEN
-sdmob(GEN ser, long n)
+sdmob(GEN s, long n, GEN fan)
 {
-  GEN D = divisorsu(n), S = gen_0;
+  GEN D = divisorsu_moebius(gel(fan,1)), S = sercoeff(s, n); /* d = 1 */
   long i, l = lg(D);
-  for (i = 1; i < l; ++i)
-  {
-    long d = D[i], mob = moebiusu(d);
-    if (mob) S = gadd(S, gdivgs(sercoeff(ser, n/d), mob*d));
-  }
+  for (i = 2; i < l; i++)
+    S = gadd(S, gdivgs(sercoeff(s, n/labs(D[i])), D[i]));
   return S;
 }
+/* log (zeta(s) * prod_i (1 - P[i]^-s) */
 static GEN
 logzetan(GEN s, GEN P, long prec)
 {
-  GEN negs = gneg(s), Z = gzeta(gprec_w(s,prec), prec);
+  GEN Z = gzeta(s, prec);
   long i, l = lg(P);
-  for (i = 1; i < l; i++) Z = gmul(Z, gsubsg(1, gpow(gel(P,i), negs, prec)));
+  for (i = 1; i < l; i++) Z = gsub(Z, gdiv(Z, gpow(gel(P,i), s, prec)));
   return glog(Z, prec);
 }
 static GEN
-sumlogzeta(GEN ser, GEN s, long N, long vF, long lim, long prec)
+sumlogzeta(GEN ser, GEN s, GEN P, double rs, double lN, long vF, long lim,
+           long prec)
 {
-  GEN z = gen_0, P = primes_interval(gen_2, utoipos(N));
-  long n;
-  for (n = lim; n >= vF; n--)
+  GEN z = gen_0, v = vecfactoru_i(vF,lim);
+  pari_sp av = avma;
+  long i, n;
+  if (typ(s) == t_INT) constbern((itos(s) * lim + 1) >> 1);
+  for (n = lim, i = lg(v)-1; n >= vF; n--, i--)
   {
-    GEN t = sdmob(ser, n);
+    GEN t = sdmob(ser, n, gel(v,i));
     if (!gequal0(t))
-    {
-      long e = gexpo(t), prec2 = e <= 0? prec: prec + nbits2prec(e);
-      z = gadd(z, gmul(logzetan(gmulsg(n,s), P, prec2), t));
-      z = gprec_w(z, prec);
+    { /* (n Re(s) - 1) log2(N) bits cancel in logzetan */
+      long prec2 = prec + nbits2extraprec((n*rs-1) * lN);
+      GEN L = logzetan(gmulsg(n,gprec_wensure(s,prec2)), P, prec2);
+      z = gerepileupto(av, gadd(z, gmul(L, t)));
+      z = gprec_wensure(z, prec);
     }
   }
-  return z;
+  return gprec_wtrunc(z, prec);
 }
 
+static GEN
+rfrac_evalfp(GEN F, GEN x, long prec)
+{
+  GEN N = gel(F,1), D = gel(F,2), a, b = poleval(D, x);
+  a = (typ(N) == t_POL && varn(N) == varn(D))? poleval(N, x): N;
+  if (typ(a) != t_INT || typ(b) != t_INT ||
+      (lgefint(a) <= prec && lgefint(b) <= prec)) return gdiv(a, b);
+  return rdivii(a, b, prec + EXTRAPRECWORD);
+}
+
+/* { F(p^s): p in P, p >= a }, F t_RFRAC */
+static GEN
+vFps(GEN P, long a, GEN F, GEN s, long prec)
+{
+  long i, j, l = lg(P);
+  GEN v = cgetg(l, t_VEC);
+  for (i = j = 1; i < l; i++)
+  {
+    GEN p = gel(P,i); if (cmpiu(p, a) < 0) continue;
+    gel(v, j++) = rfrac_evalfp(F, gpow(p, s, prec), prec);
+  }
+  setlg(v, j); return v;
+}
+
+static void
+euler_set_Fs(GEN *F, GEN *s)
+{
+  if (!*s) *s = gen_1;
+  if (typ(*F) == t_RFRAC)
+  {
+    long m;
+    *F = rfrac_deflate_max(*F, &m);
+    if (m != 1) *s = gmulgs(*s, m);
+  }
+}
 /* sum_{p prime, p >= a} F(p^s), F rational function */
 GEN
 sumeulerrat(GEN F, GEN s, long a, long prec)
 {
   pari_sp av = avma;
-  forprime_t T;
-  GEN r, ser, res, rsg;
-  double rs, RS;
-  long B = prec2nbits(prec), vx, vF, p, N, lim;
+  GEN ser, z, P;
+  double r, rs, RS, lN;
+  long B = prec2nbits(prec), prec2 = prec + EXTRAPREC64, vF, N, lim;
 
+  euler_set_Fs(&F, &s);
   switch(typ(F))
   {
     case t_RFRAC: break;
@@ -2284,25 +2395,32 @@ sumeulerrat(GEN F, GEN s, long a, long prec)
       if (gequal0(F)) return real_0(prec);
     default: pari_err_TYPE("sumeulerrat",F);
   }
-  if (!s) s = gen_1;
+  /* F t_RFRAC */
   if (a < 2) a = 2;
-  vx = varn(gel(F,2));
   vF = -poldegree(F, -1);
-  rsg = real_i(s);
-  rs = gtodouble(rsg);
-  r = ratpolemax(F);
-  RS = maxdd(1./vF, dbllog2(r) / log2((double)a));
+  rs = gtodouble(real_i(s));
+  r = polmax(gel(F,2));
+  N = maxss(30, a); lN = log2((double)N);
+  RS = maxdd(1./vF, log2(r) / lN);
   if (rs <= RS)
     pari_err_DOMAIN("sumeulerrat", "real(s)", "<=",  dbltor(RS), dbltor(rs));
-  N = maxss(maxss(30, a), (long)ceil(2*gtodouble(r)));
-  lim = (long)ceil(B / dbllog2(gdiv(gpow(stoi(N), rsg, LOWDEFAULTPREC), r)))+1;
-  ser = gmul(real_1(prec + EXTRAPREC), F);
-  ser = rfracrecip_to_ser_absolute(ser, lim);
-  res = sumlogzeta(ser, s, N, vF, lim, prec);
-  u_forprime_init(&T, a, N);
-  while ( (p = u_forprime_next(&T)) )
-    res = gadd(res, gsubst(F, vx, gpow(utoipos(p), s, prec)));
-  return gerepilecopy(av, gprec_w(res, prec));
+  lim = (long)ceil(B / (rs*lN - log2(r)));
+  ser = gmul(real_1(prec2), F);
+  ser = rfracrecip_to_ser_absolute(ser, lim+1);
+  P = primes_interval(gen_2, utoipos(N));
+  z = sumlogzeta(ser, s, P, rs, lN, vF, lim, prec);
+  z = gadd(z, vecsum(vFps(P, a, F, s, prec)));
+  return gerepilecopy(av, gprec_wtrunc(z, prec));
+}
+
+/* F = N/D; return F'/F. Assume D a t_POL */
+static GEN
+rfrac_logderiv(GEN N, GEN D)
+{
+  if (typ(N) != t_POL || varn(N) != varn(D)) return gdiv(gneg(RgX_deriv(D)), D);
+  if (!degpol(D)) return gdiv(RgX_deriv(N), N);
+  return gdiv(RgX_sub(RgX_mul(RgX_deriv(N), D), RgX_mul(RgX_deriv(D), N)),
+              RgX_mul(N, D));
 }
 
 /* prod_{p prime, p >= a} F(p^s), F rational function */
@@ -2310,38 +2428,38 @@ GEN
 prodeulerrat(GEN F, GEN s, long a, long prec)
 {
   pari_sp ltop = avma;
-  forprime_t T;
-  GEN F1, r, r1, ser, res, rsg;
-  double rs, RS;
-  long B = prec2nbits(prec), vx = gvar(F), vF, p, N, lim;
+  GEN DF, NF, ser, P, z;
+  double r, rs, RS, lN;
+  long B = prec2nbits(prec), prec2 = prec + EXTRAPREC64, vF, N, lim;
 
-  F1 = gsubgs(F, 1);
+  euler_set_Fs(&F, &s);
   switch(typ(F))
   {
     case t_RFRAC: break;
     case t_INT: case t_REAL: case t_COMPLEX: case t_POL:
-      if (gequal0(F1)) return real_1(prec);
+      if (gequal1(F)) return real_1(prec);
     default: pari_err_TYPE("prodeulerrat",F);
-  }
-  if (!s) s = gen_1;
-  vF = -poldegree(F1, -1);
-  rsg = real_i(s);
-  rs = gtodouble(rsg);
-  r = ratpolemax2(F, &r1);
-  RS = maxdd(1./vF, dbllog2(r1) / log2((double)a));
+  } /* F t_RFRAC */
+  NF = gel(F, 1);
+  DF = gel(F, 2);
+  rs = gtodouble(real_i(s));
+  vF = - rfracm1_degree(NF, DF);
+  if (rs * vF <= 1) pari_err(e_MISC, "product diverges in prodeulerrat");
+  r = ratpolemax2(F);
+  N = maxss(maxss(30, a), (long)ceil(2*r)); lN = log2((double)N);
+  RS = maxdd(1./vF, log2(r) / lN);
   if (rs <= RS)
     pari_err_DOMAIN("prodeulerrat", "real(s)", "<=",  dbltor(RS), dbltor(rs));
-  N = maxss(maxss(30, a), (long)ceil(2*gtodouble(r)));
-  lim = (long)ceil(B / dbllog2(gdiv(gpow(stoi(N), rsg, LOWDEFAULTPREC), r)))+1;
-  ser = gmul(real_1(prec + EXTRAPREC), F1);
-  ser = gaddsg(1, rfracrecip_to_ser_absolute(ser, lim));
-  ser = glog(ser, 0);
-  res = sumlogzeta(ser, s, N, vF, lim, prec);
-  res = gexp(res, prec);
-  u_forprime_init(&T, a, N);
-  while ( (p = u_forprime_next(&T)) )
-    res = gmul(res, gsubst(F, vx, gpow(utoipos(p), s, prec)));
-  return gerepilecopy(ltop, gprec_w(res, prec));
+  lim = (long)ceil(B / (rs*lN - log2(r)));
+  (void)rfracrecip(&NF, &DF); /* returned value is 0 */
+  if (!RgX_is_ZX(DF) || !is_pm1(gel(DF,2))
+      || lim * log2(r) > 4 * B) NF = gmul(NF, real_1(prec2));
+  ser = integser(rfrac_to_ser(rfrac_logderiv(NF,DF), lim+3));
+  /* ser = log f, f = F(1/x) + O(x^(lim+1)) */
+  P = primes_interval(gen_2, utoipos(N));
+  z = gexp(sumlogzeta(ser, s, P, rs, lN, vF, lim, prec), prec);
+  z = gmul(z, vecprod(vFps(P, a, F, s, prec)));
+  return gerepilecopy(ltop, gprec_wtrunc(z, prec));
 }
 
 /* Compute $\sum_{n\ge a}c(n)$ using Lagrange extrapolation.
@@ -2399,6 +2517,12 @@ sumnumlagrange2init(GEN c1, long flag, long prec)
   return gerepilecopy(av, mkvec4(gen_2, stoi(prec2), gen_1, V));
 }
 
+static GEN
+_mpmul(GEN x, GEN y)
+{
+  if (!x) return y;
+  return y? mpmul(x, y): x;
+}
 /* Used only for al = 2, 1, 1/2, 1/3, 1/4. */
 static GEN
 sumnumlagrangeinit_i(GEN al, GEN c1, long flag, long prec)
@@ -2545,7 +2669,7 @@ sumnumlagrange(void *E, GEN (*eval)(void*,GEN,long), GEN a, GEN tab, long prec)
     s = gprec_wensure(s, prec);
   }
   if (!gequal1(S)) s = gdiv(s,S);
-  return gerepilecopy(av, gprec_w(s, prec));
+  return gerepilecopy(av, gprec_wtrunc(s, prec));
 }
 
 GEN
