@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -34,11 +35,10 @@ typedef struct {
 
 static long isin_G_H(buildroot *BR, long n1, long n2);
 
-static long *par_vec;
 /* k-1 entries filled so far
  * m = maximal allowed value, n = sum to reach with remaining elements */
 static void
-do_par(GEN T, long k, long n, long m)
+do_par(GEN T, long k, long n, long m, long *par_vec)
 {
   long i;
   if (n <= 0)
@@ -48,7 +48,7 @@ do_par(GEN T, long k, long n, long m)
     gel(T, ++T[0]) = t; return;
   }
   if (n < m) m = n;
-  for (i=1; i<=m; i++) { par_vec[k] = i; do_par(T, k+1, n-i, i); }
+  for (i=1; i<=m; i++) { par_vec[k] = i; do_par(T, k+1, n-i, i, par_vec); }
 }
 
 /* compute the partitions of n, as decreasing t_VECSMALLs */
@@ -57,7 +57,7 @@ partitions_galois(long n)
 {
   pari_sp av;
   long i, p;
-  GEN T;
+  GEN T, par_vec;
 
   switch(n) /* optimized for galoismoduloX ... */
   {
@@ -66,11 +66,11 @@ partitions_galois(long n)
     case 10:p = 42; break;
     default:
       if (n < 0) pari_err_TYPE("partitions_galois", stoi(n));
-      av = avma; p = itos( numbpart(stoi(n)) ); avma = av; break;
+      av = avma; p = itos( numbpart(stoi(n)) ); set_avma(av); break;
   }
   T = new_chunk(p + 1); T[0] = 0;
   par_vec = cgetg(n+1, t_VECSMALL); /* not Garbage Collected later */
-  do_par(T,1,n,n);
+  do_par(T,1,n,n,par_vec);
   if (DEBUGLEVEL > 7)
   {
     err_printf("Partitions of %ld (%ld)\n",n, p);
@@ -118,10 +118,9 @@ _typ(long l,...)
 static PERM
 _cr(long N, long a,...)
 {
-  static long x[NMAX+1];
   va_list args;
   long i;
-
+  GEN x = new_chunk(NMAX+1);
   va_start(args, a); x[0] = N; x[1] = a;
   for (i=2; i<=N; i++) x[i] = va_arg(args,int);
   va_end(args); return x;
@@ -296,7 +295,7 @@ galopen(const char *pre, long n, long n1, long n2)
   (void)sprintf(s, "%s/galdata/%s%ld_%ld_%ld", pari_datadir, pre, n, n1, n2);
   f = pari_fopengz(s);
   if (!f) pari_err_FILE("galois file",s);
-  avma = av; return f;
+  set_avma(av); return f;
 }
 
 static char
@@ -415,8 +414,8 @@ gpolynomial(GEN r, resolv *R)
     }
   }
   setlg(IM, k);
-  RE = vecpermute(RE, gen_indexsort(RE, (void*)&abscmprr, cmp_nodata));
-  IM = vecpermute(IM, gen_indexsort(IM, (void*)&abscmprr, cmp_nodata));
+  gen_sort_inplace(RE, (void*)&abscmprr, cmp_nodata, NULL);
+  gen_sort_inplace(IM, (void*)&abscmprr, cmp_nodata, NULL);
   re = gel(RE,1);
   for (i = 2; i <= R->nm; i++) re = addrr(re, gel(RE,i));
   if (k == 1) return re;
@@ -668,7 +667,7 @@ gpoly(GEN rr, long n1, long n2)
     }
   }
   pari_err_TYPE("gpoly [undefined invariant polynomial]", mkvecsmall2(n1,n2));
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* a is a t_VECSMALL representing a polynomial */
@@ -758,14 +757,13 @@ moreprec(buildroot *BR)
 
     if (d < BIGDEFAULTPREC-2) d = BIGDEFAULTPREC-2;
     BR->prmax = maxss(BR->prmax+d, (long)(BR->prmax * 1.2));
-    if (DEBUGLEVEL)
-      { err_printf("$$$$$ New prec = %ld\n",BR->prmax); err_flush(); }
+    if (DEBUGLEVEL) err_printf("$$$$$ New prec = %ld\n",BR->prmax);
     ro = sortroots(QX_complex_roots(BR->p,BR->prmax), gel(BR->r,1));
     delete_roots(BR);
     vectrunc_append(BR->r, gclone(ro));
     for (d = 2; d < l; d++)
       vectrunc_append(BR->r, new_pol(BR->N, ro, gel(BR->coef,d)));
-    avma = av;
+    set_avma(av);
   }
   fixprec(BR);
 }
@@ -833,7 +831,6 @@ dbg_rac(long nri,long nbracint,long numi[],GEN racint[],long multi[])
   for (k = nri+1; k <= nbracint; k++) err_printf(" %ld^%ld", numi[k], multi[k]);
   err_printf("\n");
   for (k = nri+1; k <= nbracint; k++) err_printf("\t%2ld: %Ps\n", numi[k], racint[k]);
-  err_flush();
 }
 
 #define M 2521
@@ -844,7 +841,7 @@ check_isin(buildroot *BR, resolv *R, GROUP tau, GROUP ss)
   long nogr, nocos, init, i, j, k, l, d;
   pari_sp av1 = avma, av2;
   long nbgr,nbcos,nbracint,nbrac,lastnbri,lastnbrm;
-  static long numi[M],numj[M],lastnum[M],multi[M],norac[M],lastnor[M];
+  long numi[M],numj[M],lastnum[M],multi[M],norac[M],lastnor[M];
   GEN  racint[M], roint;
 
   if (getpreci(BR) != BR->pr) fixprec(BR);
@@ -861,7 +858,7 @@ check_isin(buildroot *BR, resolv *R, GROUP tau, GROUP ss)
       if (!init)
       {
         av2 = avma; nbrac = nbracint = 0;
-        for (nocos=1; nocos<=nbcos; nocos++, avma = av2)
+        for (nocos=1; nocos<=nbcos; nocos++, set_avma(av2))
         {
           roint = get_ro_perm(T, ss[nocos], d, R, BR);
           if (!roint) continue;
@@ -870,7 +867,7 @@ check_isin(buildroot *BR, resolv *R, GROUP tau, GROUP ss)
           if (nbrac >= M)
           {
             pari_warn(warner, "more than %ld rational integer roots\n", M);
-            avma = av1; goto NEXT;
+            set_avma(av1); goto NEXT;
           }
           for (j=1; j<=nbracint; j++)
             if (equalii(roint,racint[j])) { multi[j]++; break; }
@@ -889,7 +886,7 @@ check_isin(buildroot *BR, resolv *R, GROUP tau, GROUP ss)
       else
       {
         nbrac = nbracint = 0;
-        for (l=1; l<=lastnbri; l++, avma = av1)
+        for (l=1; l<=lastnbri; l++, set_avma(av1))
         {
           long nri = nbracint;
           av2 = avma;
@@ -898,7 +895,7 @@ check_isin(buildroot *BR, resolv *R, GROUP tau, GROUP ss)
             {
               nocos = lastnum[k];
               roint = get_ro_perm(T, ss[nocos], d, R, BR);
-              if (!roint) { avma = av2; continue; }
+              if (!roint) { set_avma(av2); continue; }
 
               nbrac++;
               for (j=nri+1; j<=nbracint; j++)
@@ -915,7 +912,7 @@ check_isin(buildroot *BR, resolv *R, GROUP tau, GROUP ss)
             if (multi[i]==1) return permmul(T, ss[numi[i]]);
         }
       }
-      avma = av1; if (!nbracint) break;
+      set_avma(av1); if (!nbracint) break;
 
       lastnbri = nbracint; lastnbrm = nbrac;
       for (j=1; j<=nbrac; j++) { lastnum[j] = numj[j]; lastnor[j] = norac[j]; }
@@ -1490,10 +1487,8 @@ galoismodulo8(long EVEN, GEN pol, GEN dpol)
   GR[46]= _gr( 10, 1,3,5,6,8,9,12,13,15,22);
   GR[47]= _gr( 16, 1,2,3,4,5,6,7,8,9,11,12,13,14,15,20,22);
   GR[48]= _gr( -8, 1,3,5,9,12,15,20,21);
-
   gr[0]=51; res = galmodp(EVEN,pol,dpol,TYP,gr,GR);
-  avma=av; if (!res) return 0;
-  return EVEN? 49: 50;
+  return gc_long(av, res? (EVEN? 49: 50): 0);
 }
 
 /* DEGREE 9 */
@@ -1731,7 +1726,7 @@ galoismodulo9(long EVEN, GEN pol, GEN dpol)
   GR[32]= _gr( -7, 1,5,10,12,25,27,30);
 
   gr[0]=35; res = galmodp(EVEN,pol,dpol,TYP,gr,GR);
-  avma=av; if (!res) return 0;
+  set_avma(av); if (!res) return 0;
   return EVEN? 33: 34;
 }
 
@@ -2164,10 +2159,8 @@ galoismodulo10(long EVEN, GEN pol, GEN dpol)
   GR[42]= _gr(-17, 1,3,5,7,9,11,13,16,18,20,22,24,26,27,30,35,40);
   GR[43]= _gr( 32, 1,2,3,4,5,6,7,8,9,10,11,12,13,15,16,17,18,19,20,22,23,24,25,26,27,28,29,30,33,35,40,42);
   GR[44]= _gr(-22, 1,3,5,7,9,11,13,14,16,18,20,22,24,26,27,30,32,35,36,38,40,41);
-
   gr[0]=46; res = galmodp(EVEN,pol,dpol,TYP,gr,GR);
-  avma=av; if (!res) return 0;
-  return EVEN? 44: 45;
+  return gc_long(av, res? (EVEN? 44:45): 0);
 }
 
 /* DEGREE 11 */
@@ -2241,17 +2234,14 @@ galoismodulo11(long EVEN, GEN pol, GEN dpol)
     gel(TYP,5) = _typ(11, 1,1,1,1,1,1,1,1,1,1,1);
   }
   res = galmodp(EVEN,pol,dpol,TYP,gr,NULL);
-  avma=av; if (!res) return 0;
-  return EVEN? 7: 8;
+  return gc_long(av, res? (EVEN? 7: 8): 0);
 }
 
 static void
 init_isin(long N, long n1, long n2, GROUP *tau, PERM *s0, resolv *R)
 {
   int fl = 1;
-  if (DEBUGLEVEL) {
-    err_printf("\n*** Entering isin_%ld_G_H_(%ld,%ld)\n",N,n1,n2); err_flush();
-  }
+  if (DEBUGLEVEL) err_printf("\n*** Entering isin_%ld_G_H_(%ld,%ld)\n",N,n1,n2);
   switch(N)
   {
     case 8:
@@ -2317,7 +2307,6 @@ isin_G_H(buildroot *BR, long n1, long n2)
     {
       err_printf("\n    Output of isin_%ld_G_H(%ld,%ld): %ld",N,n1,n2,n2);
       err_printf("\n    Reordering of the roots: "); printperm(s0);
-      err_flush();
     }
     for (i = 1; i < l; i++)
     {
@@ -2325,17 +2314,14 @@ isin_G_H(buildroot *BR, long n1, long n2)
       for (j=1; j<=N; j++) gel(z,j) = gel(p1,s0[j]);
       for (j=1; j<=N; j++) gel(p1,j) = gel(z,j);
     }
-    avma = av; return n2;
+    return gc_long(av, n2);
   }
   if (DEBUGLEVEL)
-  {
     err_printf("    Output of isin_%ld_G_H(%ld,%ld): not included.\n",N,n1,n2);
-    err_flush();
-  }
-  avma = av; return 0;
+  return gc_long(av, 0);
 }
 
-GEN
+static GEN
 polgaloisnamesbig(long n, long k)
 {
   pari_sp av = avma;
@@ -2345,12 +2331,7 @@ polgaloisnamesbig(long n, long k)
 
   (void)sprintf(s, "%s/galdata/NAM%ld", pari_datadir, n);
   f = pari_fopengz(s);
-  if (!f)
-  {
-    pari_warn(warner,"Galois names files not available, please upgrade galdata\n[missing %s]",s);
-    avma = av; return strtoGENstr("");
-  }
-  V = gp_read_stream(f->file);
+  V = f? gp_read_stream(f->file): NULL;
   if (!V || typ(V)!=t_VEC || k>=lg(V)) pari_err_FILE("galois file %s",s);
   pari_fclose(f);
   return gerepilecopy(av, gel(V,k));
@@ -2380,7 +2361,7 @@ galoisbig(GEN pol, long prec)
   if (DEBUGLEVEL)
   {
     err_printf("Galoisbig: polynomial #1 = %Ps\n", pol);
-    err_printf("%s group\n", EVEN? "EVEN": "ODD"); err_flush();
+    err_printf("%s group\n", EVEN? "EVEN": "ODD");
   }
   switch(N)
   {
@@ -2389,7 +2370,7 @@ galoisbig(GEN pol, long prec)
     case 10:t = galoismodulo10(EVEN,pol,dpol); tab=tab10; break;
     case 11:t = galoismodulo11(EVEN,pol,dpol); tab=tab11; break;
     default: pari_err_IMPL("galois in degree > 11");
-      return NULL; /* not reached */
+      return NULL; /* LCOV_EXCL_LINE */
   }
   if (!t)
   {
@@ -2418,7 +2399,7 @@ galoisbig(GEN pol, long prec)
     }
     for (i = 1; i < lg(BR.r); i++) gunclone(gel(BR.r,i));
   }
-  avma = av;
+  set_avma(av);
   res    = cgetg(5,t_VEC);
   gel(res,1) = stoi(tab[t]);
   gel(res,2) = stoi(EVEN? 1: -1);
@@ -2510,11 +2491,11 @@ polgalois(GEN x, long prec)
 
   if (n<4)
   {
-    if (n == 1) { avma = av; return galois_res(n,1, 1,1); }
-    if (n == 2) { avma = av; return galois_res(n,2,-1,1); }
+    if (n == 1) { set_avma(av); return galois_res(n,1, 1,1); }
+    if (n == 2) { set_avma(av); return galois_res(n,2,-1,1); }
     /* n = 3 */
     f = Z_issquare(ZX_disc(x));
-    avma = av;
+    set_avma(av);
     return f? galois_res(n,3,1,1):
               galois_res(n,6,-1,2);
   }
@@ -2543,12 +2524,12 @@ polgalois(GEN x, long prec)
         p2 = gel(ZX_factor(p5),1);
         switch(lg(p2)-1)
         {
-          case 1: f = Z_issquare(ZX_disc(x)); avma = av;
+          case 1: f = Z_issquare(ZX_disc(x)); set_avma(av);
             return f? galois_res(n,12,1,4): galois_res(n,24,-1,5);
 
-          case 2: avma = av; return galois_res(n,8,-1,3);
+          case 2: set_avma(av); return galois_res(n,8,-1,3);
 
-          case 3: avma = av;
+          case 3: set_avma(av);
             return (degpol(gel(p2,1))==2)? galois_res(n,4,1,2)
                                          : galois_res(n,4,-1,1);
 
@@ -2586,10 +2567,10 @@ polgalois(GEN x, long prec)
           f=Z_issquare(ZX_disc(x));
           if (lg(p3)-1==1)
           {
-            avma = av;
+            set_avma(av);
             return f? galois_res(n,60,1,4): galois_res(n,120,-1,5);
           }
-          if (!f) { avma = av; return galois_res(n,20,-1,3); }
+          if (!f) { set_avma(av); return galois_res(n,20,-1,3); }
 
           pr = - (prec2nbits(prec) >> 1);
           for (l=1; l<=6; l++)
@@ -2608,7 +2589,7 @@ polgalois(GEN x, long prec)
           if (e <= -10)
           {
             if (gequal0(p4)) goto tchi;
-            f = Z_issquare(p4); avma = av;
+            f = Z_issquare(p4); set_avma(av);
             return f? galois_res(n,5,1,1): galois_res(n,10,1,2);
           }
           prec = precdbl(prec);
@@ -2660,7 +2641,7 @@ polgalois(GEN x, long prec)
                 if (!ZX_is_squarefree(p5)) goto tchi;
                 p2 = gel(ZX_factor(p5),1);
                 f = Z_issquare(ZX_disc(x));
-                avma = av;
+                set_avma(av);
                 if (lg(p2)-1==1)
                   return f? galois_res(n,360,1,15): galois_res(n,720,-1,16);
                 else
@@ -2672,17 +2653,17 @@ polgalois(GEN x, long prec)
               switch(l2)
               {
                 case 1: f = Z_issquare(ZX_disc(x));
-                  avma = av;
+                  set_avma(av);
                   return f? galois_res(n,60,1,12): galois_res(n,120,-1,14);
                 case 2: f = Z_issquare(ZX_disc(x));
-                  if (f) { avma = av; return galois_res(n,24,1,7); }
+                  if (f) { set_avma(av); return galois_res(n,24,1,7); }
                   p3 = (degpol(gel(p2,1))==2)? gel(p2,2): gel(p2,1);
                   f = Z_issquare(ZX_disc(p3));
-                  avma = av;
+                  set_avma(av);
                   return f? galois_res(n,24,-1,6): galois_res(n,48,-1,11);
                 case 3: f = Z_issquare(ZX_disc(gel(p2,1)))
                          || Z_issquare(ZX_disc(gel(p2,2)));
-                  avma = av;
+                  set_avma(av);
                   return f? galois_res(n,18,-1,5): galois_res(n,36,-1,9);
               }
             case 3:
@@ -2690,15 +2671,15 @@ polgalois(GEN x, long prec)
                 if (degpol(gel(p2,l2)) >= 3) p3 = gel(p2,l2);
               if (degpol(p3) == 3)
               {
-                f = Z_issquare(ZX_disc(p3)); avma = av;
+                f = Z_issquare(ZX_disc(p3)); set_avma(av);
                 return f? galois_res(n,6,-1,1): galois_res(n,12,-1,3);
               }
               else
               {
-                f = Z_issquare(ZX_disc(x)); avma = av;
+                f = Z_issquare(ZX_disc(x)); set_avma(av);
                 return f? galois_res(n,12,1,4): galois_res(n,24,-1,8);
               }
-            case 4: avma = av; return galois_res(n,6,-1,2);
+            case 4: set_avma(av); return galois_res(n,6,-1,2);
             default: pari_err_BUG("galois (bug3)");
           }
         }
@@ -2721,16 +2702,16 @@ polgalois(GEN x, long prec)
         p2=gel(ZX_factor(p5),1);
         switch(lg(p2)-1)
         {
-          case 1: f = Z_issquare(ZX_disc(x)); avma = av;
+          case 1: f = Z_issquare(ZX_disc(x)); set_avma(av);
             return f? galois_res(n,2520,1,6): galois_res(n,5040,-1,7);
-          case 2: f = degpol(gel(p2,1)); avma = av;
+          case 2: f = degpol(gel(p2,1)); set_avma(av);
             return (f==7 || f==28)? galois_res(n,168,1,5): galois_res(n,42,-1,4);
-          case 3: avma = av; return galois_res(n,21,1,3);
-          case 4: avma = av; return galois_res(n,14,-1,2);
-          case 5: avma = av; return galois_res(n,7,1,1);
+          case 3: set_avma(av); return galois_res(n,21,1,3);
+          case 4: set_avma(av); return galois_res(n,14,-1,2);
+          case 5: set_avma(av); return galois_res(n,7,1,1);
           default: pari_err_BUG("galois (bug2)");
         }
     }
-    tchi: avma = av1; x = tschirnhaus(x1);
+    tchi: set_avma(av1); x = tschirnhaus(x1);
   }
 }
