@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -45,9 +46,10 @@ galoisgetpol(long a, long b, long sig)
       pari_err_DOMAIN("galoisgetpol", "s", ">", gen_1, stoi(sig));
     default:
       pari_err_FLAG("galoisgetpol");
-      return NULL;
+      return NULL;/*LCOV_EXCL_LINE*/
   }
-  s = pari_sprintf("%s/galpol/%ld/%ld/%s", pari_datadir, a,b,si);
+  /* left on stack */
+  s = stack_sprintf("%s/galpol/%ld/%ld/%s", pari_datadir, a,b,si);
   F = pari_fopengz(s);
   if (!F)
   {
@@ -56,7 +58,6 @@ galoisgetpol(long a, long b, long sig)
       pari_err_DOMAIN("galoisgetpol", "group index", ">", stoi(n), stoi(b));
     else pari_err_FILE("galpol file", s);
   }
-  pari_free(s);
   V = gp_read_stream(F->file);
   if (!V || typ(V)!=t_VEC) pari_err_FILE("galpol file", F->name);
   pari_fclose(F); return V;
@@ -71,7 +72,8 @@ galoisgetgroup(long a, long b)
   if (a<=0) pari_err_DOMAIN("galoisgetgroup", "degree", "<=", gen_0, stoi(a));
   if (b<0) pari_err_DOMAIN("galoisgetgroup", "index", "<", gen_0, stoi(b));
   if (!b) return galoisnbpol(a);
-  s = pari_sprintf("%s/galpol/%ld/%ld/group", pari_datadir, a,b);
+  /* left on stack */
+  s = stack_sprintf("%s/galpol/%ld/%ld/group", pari_datadir, a,b);
   F = pari_fopengz(s);
   if (!F)
   {
@@ -80,7 +82,6 @@ galoisgetgroup(long a, long b)
       pari_err_DOMAIN("galoisgetgroup", "group index", ">", stoi(n), stoi(b));
     else pari_err_FILE("galpol file", s);
   }
-  pari_free(s);
   V = gp_read_stream(F->file);
   if (!V || typ(V)!=t_VEC) pari_err_FILE("galpol file", F->name);
   pari_fclose(F); return V;
@@ -94,7 +95,8 @@ galoisgetname(long a, long b)
   char *s;
   if (a<=0) pari_err_DOMAIN("galoisgetname", "degree", "<=", gen_0, stoi(a));
   if (b<0) pari_err_DOMAIN("galoisgetname", "index", "<", gen_0, stoi(b));
-  s = pari_sprintf("%s/galpol/%ld/%ld/name", pari_datadir, a,b);
+  /* left on stack */
+  s = stack_sprintf("%s/galpol/%ld/%ld/name", pari_datadir, a,b);
   F = pari_fopengz(s);
   if (!F)
   {
@@ -103,7 +105,6 @@ galoisgetname(long a, long b)
       pari_err_DOMAIN("galoisgetname", "group index", ">", stoi(n), stoi(b));
     else pari_err_FILE("galpol file", s);
   }
-  pari_free(s);
   V = gp_read_stream(F->file);
   if (!V || typ(V)!=t_STR) pari_err_FILE("galpol file", F->name);
   pari_fclose(F); return V;
