@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -284,7 +285,7 @@ DivideByPi(FAD_t *fdata, GEN pp, GEN ppp, GEN pol)
   {
     GEN r;
     gel(P,i) = dvmdii(gel(P,i), fdata->p, &r);
-    if (r != gen_0) { avma = av; return NULL; }
+    if (r != gen_0) return gc_NULL(av);
   }
   return FpX_red(P, pp);
 }
@@ -361,13 +362,12 @@ FreeRootTable(GEN T)
   if (T)
   {
     long j, l = lg(T);
-    for (j = 1; j < l; j++)
-      if (gel(T,j)) gunclone(gel(T,j));
+    for (j = 1; j < l; j++) guncloneNULL(gel(T,j));
   }
 }
 
 /* Return the number of roots of pol congruent to alpha modulo pi working
-   modulo pp (all roots if alpha is NULL); if flag is non-zero, return 1
+   modulo pp (all roots if alpha is NULL); if flag is nonzero, return 1
    as soon as a root is found. (Note: ppp = pp*p for DivideByPi) */
 static long
 RootCongruents(KRASNER_t *data, FAD_t *fdata, GEN pol, GEN alpha, GEN pp, GEN ppp, long sl, long flag)
@@ -424,7 +424,7 @@ RootCountingAlgorithm(KRASNER_t *data, FAD_t *fdata, GEN pol, long flag)
   return RootCongruents(data, fdata, P, NULL, diviiexact(data->pr, data->p), data->pr, 0, flag);
 }
 
-/* Return non-zero if the field given by fdata defines a field isomorphic to
+/* Return nonzero if the field given by fdata defines a field isomorphic to
  * the one defined by pol */
 static long
 IsIsomorphic(KRASNER_t *data, FAD_t *fdata, GEN pol)
@@ -438,11 +438,11 @@ IsIsomorphic(KRASNER_t *data, FAD_t *fdata, GEN pol)
   {
     GEN p1 = FqX_FpXQ_eval(pol, fdata->z, fdata->top, data->pr);
     nb = RootCountingAlgorithm(data, fdata, p1, 1);
-    if (nb) { avma = av; return nb; }
+    if (nb) return gc_long(av, nb);
     if (j < data->f)
       pol = FqX_FpXQ_eval(pol, data->frob, data->T, data->pr);
   }
-  avma = av; return 0;
+  return gc_long(av,0);
 }
 
 /* Compute the number of conjugates fields of the field given by fdata */
@@ -455,7 +455,7 @@ NbConjugateFields(KRASNER_t *data, FAD_t *fdata)
 
   if (RgX_is_ZX(pol)) { /* split for efficiency; contains the case f = 1 */
     fdata->cj = data->e / RootCountingAlgorithm(data, fdata, pol, 0);
-    avma = av; return;
+    set_avma(av); return;
   }
 
   nb = 0;
@@ -467,7 +467,7 @@ NbConjugateFields(KRASNER_t *data, FAD_t *fdata)
     if (j < data->f)
       pol = FqX_FpXQ_eval(pol, data->frob, data->T, data->pr);
   }
-  avma = av;
+  set_avma(av);
   fdata->cj = data->e * data->f / nb;
   return;
 }
@@ -665,7 +665,7 @@ WildlyRamifiedCase(KRASNER_t *data)
         err_printf("%ld more extension%s\t(%ld/%ld, %ldms)\n",
                    nb, (nb == 1)? "": "s", fd, nbext, timer_delay(&T));
     }
-    avma = av2;
+    set_avma(av2);
   }
 
   rep = cgetg(ct+1, t_VEC);
@@ -845,7 +845,7 @@ checkpols(GEN p, GEN EFJ, GEN pols)
   {
     pol = gel(pols, i);
     if (typ(pol) == t_VEC) pol = gel(pol, 1);
-    if (!isirreducible(pol)) pari_err_BUG("Polynomial is reducible");
+    if (!polisirreducible(pol)) pari_err_BUG("Polynomial is reducible");
     p1 = poldisc0(pol, -1);
     if (gvaluation(p1, p) != f*(e+j-1)) pari_err_BUG("Discriminant is incorrect");
     /* only compute a p-maximal order */
@@ -860,7 +860,7 @@ checkpols(GEN p, GEN EFJ, GEN pols)
   if (l == 2) return;
   if (e*f > 20) return;
 
-  /* TODO: check that (random) distinct polynomials give non-isomorphic extensions */
+  /* TODO: check that (random) distinct polynomials give nonisomorphic extensions */
   for (i = 1; i < 3*l; i++)
   {
     c1 = random_Fl(l-1)+1;
@@ -890,7 +890,7 @@ pols_from_efj(pari_sp av, GEN EFJ, GEN p, long flag)
 {
   long i, l;
   GEN L = cgetg_copy(EFJ, &l);
-  if (l == 1) { avma = av; return flag == 2? gen_0: cgetg(1, t_VEC); }
+  if (l == 1) { set_avma(av); return flag == 2? gen_0: cgetg(1, t_VEC); }
   for (i = 1; i < l; i++)
   {
     gel(L,i) = GetRamifiedPol(p, gel(EFJ,i), flag);
