@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -43,7 +44,7 @@ struct mt_pstate
   pthread_mutex_t pmut;
 };
 
-static THREAD long mt_thread_no = -1;
+static THREAD long mt_thread_no = -1, mt_issingle = 0;
 static struct mt_pstate *pari_mt;
 
 #define LOCK(x) pthread_mutex_lock(x); do
@@ -80,6 +81,7 @@ mt_err_recover(long er)
     } UNLOCK(mq->pmut);
     pthread_exit((void*)1);
   }
+  else if (mt_issingle) mtsingle_err_recover(er);
 }
 
 void
@@ -97,7 +99,29 @@ mt_is_parallel(void)
 int
 mt_is_thread(void)
 {
-  return mt_thread_no>=0;
+  return mt_thread_no>=0 ? 1: mt_issingle ? mtsingle_is_thread(): 0;
+}
+
+long
+mt_nbthreads(void)
+{
+  return pari_mt ? 1: pari_mt_nbthreads;
+}
+
+void
+mt_export_add(const char *str, GEN val)
+{
+  if (pari_mt)
+    pari_err(e_MISC,"export() not allowed during parallel sections");
+  export_add(str, val);
+}
+
+void
+mt_export_del(const char *str)
+{
+  if (pari_mt)
+    pari_err(e_MISC,"unexport() not allowed during parallel sections");
+  export_del(str);
 }
 
 void mt_broadcast(GEN code) {(void) code;}
@@ -105,6 +129,7 @@ void mt_broadcast(GEN code) {(void) code;}
 void pari_mt_init(void)
 {
   pari_mt = NULL;
+  mt_issingle = 0;
 #ifdef _SC_NPROCESSORS_CONF
   if (!pari_mt_nbthreads) pari_mt_nbthreads = sysconf(_SC_NPROCESSORS_CONF);
 #elif defined(_WIN32)
@@ -152,7 +177,7 @@ mt_queue_run(void *arg)
       pthread_cleanup_pop(0);
     } UNLOCK(&mq->mut);
     pari_mainstack = mq->mainstack;
-    avma = mq->avma;
+    set_avma(mq->avma);
     work = mq->input;
     pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);
     done = closure_callgenvec(mq->worker,work);
@@ -260,12 +285,12 @@ mtpthread_queue_submit(struct mt_state *junk, long workid, GEN work)
       pari_sp av = avma;
       struct pari_mainstack *st = pari_mainstack;
       pari_mainstack = mq->mainstack;
-      avma = mq->avma;
+      set_avma(mq->avma);
       mq->input = gcopy(work);
       mq->avma = avma;
       mq->mainstack = pari_mainstack;
       pari_mainstack = st;
-      avma = av;
+      set_avma(av);
     }
     BLOCK_SIGINT_END
     pthread_cond_signal(&mq->cond);
@@ -320,7 +345,10 @@ mt_queue_start_lim(struct pari_mt *pt, GEN worker, long lim)
   if (lim==0) lim = pari_mt_nbthreads;
   else        lim = minss(pari_mt_nbthreads, lim);
   if (pari_mt || lim <= 1)
+  {
+    mt_issingle = 1;
     mtsingle_queue_start(pt, worker);
+  }
   else
   {
     struct mt_pstate *mt =
@@ -339,12 +367,11 @@ mt_queue_start_lim(struct pari_mt *pt, GEN worker, long lim)
     mt->last = 0;
     pthread_cond_init(&mt->pcond,NULL);
     pthread_mutex_init(&mt->pmut,NULL);
-    pari_thread_sync();
     for (i=0;i<lim;i++)
     {
       struct mt_queue *mq = mt->mq+i;
       mq->no     = i;
-      mq->avma   = 0;
+      mq->avma = 0;
       mq->mainstack = NULL;
       mq->worker = worker;
       mq->input  = NULL;
@@ -363,6 +390,7 @@ mt_queue_start_lim(struct pari_mt *pt, GEN worker, long lim)
     for (i=0;i<lim;i++)
       pthread_create(&mt->th[i],NULL, &mt_queue_run, (void*)&mt->pth[i]);
     pari_mt = mt;
+    mt_issingle = 0;
     BLOCK_SIGINT_END
     pt->get=&mtpthread_queue_get;
     pt->submit=&mtpthread_queue_submit;
