@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -16,24 +17,14 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
  *
  * (GMU MP Library is Copyright Free Software Foundation, Inc.) */
 #define PARI_TUNE
-#include <pari.h>
-#include <paripriv.h>
+#include "pari.h"
+#include "paripriv.h"
 
 int option_trace = 0;
 double Step_Factor = .01; /* small steps by default */
-ulong DFLT_mod1, DFLT_hmod, DFLT_qmod, DFLT_mod2;
+ulong DFLT_mod, DFLT_mod2, DFLT_deg;
 GEN LARGE_mod;
 
-#ifdef LONG_IS_64BIT
-#  define DFLT_mod DFLT_mod1
-#  define Fmod_MUL_MULII_LIMIT Flx_MUL_MULII_LIMIT
-#  define Fmod_SQR_SQRI_LIMIT  Flx_SQR_SQRI_LIMIT
-#  else
-#  define DFLT_mod DFLT_mod2
-#  define Fmod_MUL_MULII_LIMIT Flx_MUL_MULII2_LIMIT
-#  define Fmod_SQR_SQRI_LIMIT  Flx_SQR_SQRI2_LIMIT
-#endif
-
 typedef struct {
   ulong reps, type;
   long *var, *var_disable, *var_enable, var_enable_min,  size, enabled;
@@ -102,7 +93,7 @@ static GEN
 rand_F2x(long n)
 {
   GEN x;
-  do x = random_F2x(BITS_IN_LONG*n, 0); while (lgpol(x) < n);
+  do x = random_F2x(BITS_IN_LONG*(n+1), 0); while (degpol(x) < n);
   return x;
 }
 /* Flx, degree n */
@@ -180,16 +171,10 @@ rand_NFpXQX(long n, GEN T)
 }
 
 #define t_F2x     100
-#define t_Fqx     200
-#define t_Fhx     201
-#define t_Flx     202
-#define t_Fl1x    203
-#define t_Fl2x    204
-#define t_NFqx    210
-#define t_NFhx    211
-#define t_NFlx    212
-#define t_NFl1x   213
-#define t_NFl2x   214
+#define t_Flx     200
+#define t_Fl2x    201
+#define t_NFlx    210
+#define t_NFl2x   211
 #define t_FpX     300
 #define t_NFpX    310
 #define t_F2xqX   400
@@ -207,15 +192,9 @@ rand_g(speed_param *s)
     case t_INT:  return rand_INT(n);
     case t_REAL: return rand_REAL(n);
     case t_F2x:  return rand_F2x(n);
-    case t_Fqx:  return rand_Flx(n,DFLT_qmod);
-    case t_Fhx:  return rand_Flx(n,DFLT_hmod);
     case t_Flx:  return rand_Flx(n,DFLT_mod);
-    case t_Fl1x: return rand_Flx(n,DFLT_mod1);
     case t_Fl2x: return rand_Flx(n,DFLT_mod2);
-    case t_NFqx: return rand_NFlx(n,DFLT_qmod);
-    case t_NFhx: return rand_NFlx(n,DFLT_hmod);
     case t_NFlx: return rand_NFlx(n,DFLT_mod);
-    case t_NFl1x: return rand_NFlx(n,DFLT_mod1);
     case t_NFl2x: return rand_NFlx(n,DFLT_mod2);
     case t_FpX:    return rand_FpX(n);
     case t_NFpX:   return rand_NFpX(n);
@@ -230,11 +209,26 @@ rand_g(speed_param *s)
 }
 
 static void
-dft_Flxq(speed_param *s)
+dft_F2xq(speed_param *s)
 {
+  pari_sp av = avma;
   do
   {
-    s->T = rand_NFlx(10, s->l);
+    avma = av;
+    s->T = random_F2x(BITS_IN_LONG*2, 0);
+  } while (!F2x_is_irred(s->T));
+  s->T[1] = evalvarn(1);
+  s->T = F2x_get_red(s->T);
+}
+
+static void
+dft_Flxq(speed_param *s)
+{
+  pari_sp av = avma;
+  do
+  {
+    avma = av;
+    s->T = rand_NFlx(DFLT_deg, s->l);
   } while (!Flx_is_irred(s->T, s->l));
   s->T[1] = evalvarn(1);
   s->T = Flx_get_red(s->T, s->l);
@@ -243,7 +237,12 @@ dft_Flxq(speed_param *s)
 static void
 dft_FpXQ(speed_param *s)
 {
-  s->T = rand_NFpX(10);
+  pari_sp av = avma;
+  do
+  {
+    avma = av;
+    s->T = rand_NFpX(DFLT_deg);
+  } while (!FpX_is_irred(s->T, LARGE_mod));
   setvarn(s->T, 1);
   s->T = FpX_get_red(s->T, s->p);
 }
@@ -252,20 +251,14 @@ static void
 dftmod(speed_param *s)
 {
   switch (s->type) {
-    case t_Fqx:  s->l=DFLT_qmod; return;
-    case t_Fhx:  s->l=DFLT_hmod; return;
     case t_Flx:  s->l=DFLT_mod;  return;
-    case t_Fl1x: s->l=DFLT_mod1; return;
     case t_Fl2x: s->l=DFLT_mod2; return;
-    case t_NFqx: s->l=DFLT_qmod;  return;
-    case t_NFhx: s->l=DFLT_hmod;  return;
     case t_NFlx: s->l=DFLT_mod;  return;
-    case t_NFl1x: s->l=DFLT_mod1;  return;
     case t_NFl2x: s->l=DFLT_mod2;  return;
     case t_FpX:  s->p=LARGE_mod; return;
     case t_NFpX: s->p=LARGE_mod; return;
-    case t_F2xqX:  s->l=2;  dft_Flxq(s); return;
-    case t_NF2xqX: s->l=2;  dft_Flxq(s); return;
+    case t_F2xqX:  s->l=2;  dft_F2xq(s); return;
+    case t_NF2xqX: s->l=2;  dft_F2xq(s); return;
     case t_FlxqX:  s->l=DFLT_mod;  dft_Flxq(s); return;
     case t_NFlxqX: s->l=DFLT_mod;  dft_Flxq(s); return;
     case t_FpXQX:  s->p=LARGE_mod; dft_FpXQ(s); return;
@@ -280,7 +273,7 @@ dftmod(speed_param *s)
     int i;                               \
     speed_starttime();                   \
     i = (s)->reps;                       \
-    do { call; avma = av; } while (--i); \
+    do { call; set_avma(av); } while (--i); \
   }                                      \
   return speed_endtime();                \
 }
@@ -315,6 +308,15 @@ static double speed_mulii(speed_param *s)
 static double speed_sqri (speed_param *s)
 { TIME_FUN(sqri(s->x)); }
 
+static double speed_extgcdii(speed_param *s)
+{ GEN u,v; TIME_FUN(bezout(s->x, s->y, &u, &v)); }
+
+static double speed_gcdii(speed_param *s)
+{ TIME_FUN(gcdii(s->x, s->y)); }
+
+static double speed_halfgcdii(speed_param *s)
+{ TIME_FUN(halfgcdii(s->x, s->y)); }
+
 static double speed_exp(speed_param *s)
 { TIME_FUN(mpexp(s->x)); }
 
@@ -430,6 +432,15 @@ static double speed_F2xqXQ_red(speed_param *s) {
   TIME_FUN(F2xqX_rem(x, q, s->T));
 }
 
+static double speed_F2xqX_halfgcd(speed_param *s)
+{ TIME_FUN(F2xqX_halfgcd(s->x, s->y, s->T)); }
+
+static double speed_F2xqX_extgcd(speed_param *s)
+{ GEN u,v; TIME_FUN(F2xqX_extgcd(s->x, s->y, s->T, &u, &v)); }
+
+static double speed_F2xqX_gcd(speed_param *s)
+{ TIME_FUN(F2xqX_gcd(s->x, s->y, s->T)); }
+
 static double speed_FlxqX_inv(speed_param *s)
 { TIME_FUN(FlxqX_invBarrett(s->x, s->T, s->l)); }
 
@@ -512,6 +523,9 @@ static tune_param param[] = {
 {PARI,var(SQRI_KARATSUBA_LIMIT),   t_INT, 4,0, speed_sqri,0,0,&SQRI_FFT_LIMIT},
 {PARI,var(MULII_FFT_LIMIT),        t_INT, 500,0, speed_mulii,0.02},
 {PARI,var(SQRI_FFT_LIMIT),         t_INT, 500,0, speed_sqri,0.02},
+{0,   var(HALFGCD_LIMIT),          t_INT, 3,0, speed_halfgcdii,0.02},
+{PARI,var(EXTGCD_HALFGCD_LIMIT),   t_INT, 5,0, speed_extgcdii,0.02},
+{PARI,var(GCD_HALFGCD_LIMIT),      t_INT, 5,0, speed_gcdii,0.02},
 {0,   var(MULRR_MULII_LIMIT),      t_REAL,4,0, speed_mulrr},
 {0,   var(SQRR_SQRI_LIMIT),        t_REAL,4,0, speed_sqrr},
 {0,   var(Fp_POW_REDC_LIMIT),      t_INT, 3,100, speed_Fp_pow,0,0,&Fp_POW_BARRETT_LIMIT},
@@ -523,55 +537,37 @@ static tune_param param[] = {
 {0,   var(LOGAGMCX_LIMIT),         t_REAL,3,0, speed_logcx,0.05},
 {0,   var(AGM_ATAN_LIMIT),         t_REAL,20,0, speed_atan,0.05},
 {GMP, var(INVMOD_GMP_LIMIT),       t_INT, 3,0, speed_invmod},
-{0,   var(F2x_MUL_KARATSUBA_LIMIT),t_F2x,3,0, speed_F2x_mul},
-{0,   var(Flx_MUL_KARATSUBA_LIMIT),t_Flx,5,0, speed_Flx_mul,0,0,&Fmod_MUL_MULII_LIMIT},
-{0,   var(Flx_SQR_KARATSUBA_LIMIT),t_Flx,5,0, speed_Flx_sqr,0,0,&Fmod_SQR_SQRI_LIMIT},
-{0,   var(Flx_MUL_QUARTMULII_LIMIT),t_Fqx,3,0, speed_Flx_mul},
-{0,   var(Flx_SQR_QUARTSQRI_LIMIT), t_Fqx,3,0, speed_Flx_sqr},
-{0,   var(Flx_MUL_HALFMULII_LIMIT),t_Fhx,3,0, speed_Flx_mul},
-{0,   var(Flx_SQR_HALFSQRI_LIMIT), t_Fhx,3,0, speed_Flx_sqr},
-{0,   var(Flx_MUL_MULII_LIMIT),    t_Fl1x,5,0, speed_Flx_mul},
-{0,   var(Flx_SQR_SQRI_LIMIT),     t_Fl1x,5,0, speed_Flx_sqr},
-{0,   var(Flx_MUL_MULII2_LIMIT),   t_Fl2x,5,20000, speed_Flx_mul,0.05},
-{0,   var(Flx_SQR_SQRI2_LIMIT),    t_Fl2x,5,20000, speed_Flx_sqr,0.05},
-{0,  var(Flx_INVBARRETT_KARATSUBA_LIMIT), t_NFlx,5,20000,
-            speed_Flx_inv,0,0,&Fmod_MUL_MULII_LIMIT,&Flx_MUL_KARATSUBA_LIMIT},
-{0,  var(Flx_INVBARRETT_QUARTMULII_LIMIT), t_NFqx,5,0,
-            speed_Flx_inv,0,0,NULL,&Flx_MUL_QUARTMULII_LIMIT},
-{0,  var(Flx_INVBARRETT_HALFMULII_LIMIT), t_NFhx,5,0,
-            speed_Flx_inv,0,0,NULL,&Flx_MUL_HALFMULII_LIMIT},
-{0,  var(Flx_INVBARRETT_MULII_LIMIT), t_NFl1x,5,0,
-            speed_Flx_inv,0,0,NULL,&Flx_MUL_MULII_LIMIT},
-{0,  var(Flx_INVBARRETT_MULII2_LIMIT),t_NFl2x,5,0,
-            speed_Flx_inv,0,0,NULL,&Flx_MUL_MULII2_LIMIT},
-{0,  var(Flx_DIVREM_BARRETT_LIMIT),t_NFlx,10,0, speed_Flx_divrem,0.05},
-{0,  var(Flx_REM_BARRETT_LIMIT),  t_NFlx,10,0, speed_Flx_rem,0.05},
-{0,  var(Flx_BARRETT_KARATSUBA_LIMIT), t_NFlx,5,0,
-            speed_Flxq_red,0,0,&Fmod_MUL_MULII_LIMIT,&Flx_MUL_KARATSUBA_LIMIT},
-{0,  var(Flx_BARRETT_QUARTMULII_LIMIT), t_NFqx,5,0,
-            speed_Flxq_red,0,0,NULL,&Flx_MUL_QUARTMULII_LIMIT},
-{0,  var(Flx_BARRETT_HALFMULII_LIMIT), t_NFhx,5,0,
-            speed_Flxq_red,0,0,NULL,&Flx_MUL_HALFMULII_LIMIT},
-{0,  var(Flx_BARRETT_MULII_LIMIT), t_NFl1x,5,0,
-            speed_Flxq_red,0,0,NULL,&Flx_MUL_MULII_LIMIT},
-{0,  var(Flx_BARRETT_MULII2_LIMIT),t_NFl2x,5,0,
-            speed_Flxq_red,0,0,NULL,&Flx_MUL_MULII2_LIMIT},
-{0,  var(Flx_HALFGCD_KARATSUBA_LIMIT), t_Flx,10,0,
-            speed_Flx_halfgcd,0,0,&Fmod_MUL_MULII_LIMIT,&Flx_MUL_KARATSUBA_LIMIT},
-{0,  var(Flx_HALFGCD_QUARTMULII_LIMIT), t_Fqx,10,0,
-            speed_Flx_halfgcd,0,0,NULL,&Flx_MUL_QUARTMULII_LIMIT},
-{0,  var(Flx_HALFGCD_HALFMULII_LIMIT), t_Fhx,10,0,
-            speed_Flx_halfgcd,0,0,NULL,&Flx_MUL_HALFMULII_LIMIT},
-{0,  var(Flx_HALFGCD_MULII_LIMIT), t_Fl1x,10,0,
-            speed_Flx_halfgcd,0,0,NULL,&Flx_MUL_MULII_LIMIT},
-{0,  var(Flx_HALFGCD_MULII2_LIMIT),t_Fl2x,10,0,
-            speed_Flx_halfgcd,0,0,NULL,&Flx_MUL_MULII2_LIMIT},
-{0,  var(Flx_GCD_LIMIT),           t_Flx,10,0, speed_Flx_gcd,0.1},
-{0,  var(Flx_EXTGCD_LIMIT),        t_Flx,10,0, speed_Flx_extgcd},
+{0,   var(F2x_MUL_KARATSUBA_LIMIT),t_F2x,3,0, speed_F2x_mul,0,0,&F2x_MUL_MULII_LIMIT},
+{0,   var(F2x_MUL_MULII_LIMIT),    t_F2x,20,20000, speed_F2x_mul},
+{0,   var(Flx_MUL_KARATSUBA_LIMIT),t_Flx,5,0, speed_Flx_mul,0,0,&Flx_MUL_MULII_LIMIT},
+{0,   var(Flx_SQR_KARATSUBA_LIMIT),t_Flx,5,0, speed_Flx_sqr,0,0,&Flx_SQR_SQRI_LIMIT},
+{0,   var(Flx_MUL2_KARATSUBA_LIMIT),t_Fl2x,5,0, speed_Flx_mul,0,0,&Flx_MUL2_MULII_LIMIT},
+{0,   var(Flx_SQR2_KARATSUBA_LIMIT),t_Fl2x,5,0, speed_Flx_sqr,0,0,&Flx_SQR2_SQRI_LIMIT},
+{0,   var(Flx_MUL_MULII_LIMIT),    t_Flx,5,0, speed_Flx_mul},
+{0,   var(Flx_SQR_SQRI_LIMIT),     t_Flx,5,0, speed_Flx_sqr},
+{0,   var(Flx_MUL2_MULII_LIMIT),   t_Fl2x,5,20000, speed_Flx_mul,0.05},
+{0,   var(Flx_SQR2_SQRI_LIMIT),    t_Fl2x,5,20000, speed_Flx_sqr,0.05},
+{0,  var(Flx_INVBARRETT_LIMIT), t_NFlx, 5,0, speed_Flx_inv},
+{0,  var(Flx_INVBARRETT2_LIMIT),t_NFl2x,5,0, speed_Flx_inv},
+{0,  var(Flx_DIVREM_BARRETT_LIMIT) ,t_NFlx,10,0, speed_Flx_divrem,0.05},
+{0,  var(Flx_DIVREM2_BARRETT_LIMIT),t_NFl2x,10,0, speed_Flx_divrem,0.05},
+{0,  var(Flx_REM_BARRETT_LIMIT),   t_NFlx,10,0, speed_Flx_rem,0.05},
+{0,  var(Flx_REM2_BARRETT_LIMIT),  t_NFl2x,10,0, speed_Flx_rem,0.05},
+{0,  var(Flx_BARRETT_LIMIT), t_NFlx, 5,0, speed_Flxq_red},
+{0,  var(Flx_BARRETT2_LIMIT),t_NFl2x,5,0, speed_Flxq_red},
+{0,  var(Flx_HALFGCD_LIMIT), t_Flx, 10,0, speed_Flx_halfgcd},
+{0,  var(Flx_HALFGCD2_LIMIT),t_Fl2x,10,0, speed_Flx_halfgcd},
+{0,  var(Flx_GCD_LIMIT),            t_Flx,10,0, speed_Flx_gcd,0.1},
+{0,  var(Flx_GCD2_LIMIT),           t_Fl2x,10,0, speed_Flx_gcd,0.1},
+{0,  var(Flx_EXTGCD_LIMIT),         t_Flx,10,0, speed_Flx_extgcd},
+{0,  var(Flx_EXTGCD2_LIMIT),        t_Fl2x,10,0, speed_Flx_extgcd},
 {0,  var(F2xqX_INVBARRETT_LIMIT),t_NF2xqX,10,0, speed_F2xqX_inv,0.05},
 {0,  var(F2xqX_BARRETT_LIMIT),   t_NF2xqX,10,0, speed_F2xqXQ_red,0.05},
 {0,  var(F2xqX_DIVREM_BARRETT_LIMIT), t_NF2xqX,10,0, speed_F2xqX_divrem,0.05},
 {0,  var(F2xqX_REM_BARRETT_LIMIT), t_NF2xqX,10,0, speed_F2xqX_rem,0.05},
+{0,  var(F2xqX_HALFGCD_LIMIT),    t_F2xqX,10,0, speed_F2xqX_halfgcd,0.05},
+{0,  var(F2xqX_GCD_LIMIT),        t_F2xqX,10,0, speed_F2xqX_gcd,0.1},
+{0,  var(F2xqX_EXTGCD_LIMIT),     t_F2xqX,10,0, speed_F2xqX_extgcd,0.05},
 {0,  var(FlxqX_INVBARRETT_LIMIT),t_NFlxqX,10,0, speed_FlxqX_inv,0.05},
 {0,  var(FlxqX_BARRETT_LIMIT),   t_NFlxqX,10,0, speed_FlxqXQ_red,0.05},
 {0,  var(FlxqX_DIVREM_BARRETT_LIMIT), t_NFlxqX,10,0, speed_FlxqX_divrem,0.05},
@@ -646,7 +642,7 @@ time_fun(speed_function_t fun, speed_param *s, long enabled)
     {
       qsort (t, i+1, sizeof(t[0]), (QSCOMP)double_cmp_ptr);
       for (j = e-1; j < i; j++)
-        if (t[j] <= t[j-e+1] * TOLERANCE) { avma = av; return t[j-e+1]; }
+        if (t[j] <= t[j-e+1] * TOLERANCE) return gc_double(av, t[j-e+1]);
     }
   }
   pari_err(e_MISC,"couldn't measure time");
@@ -667,7 +663,7 @@ add_dat(long size, double d)
   if (ndat == allocdat)
   {
     allocdat += maxss(allocdat, 100);
-    dat = (struct dat_t*) pari_realloc((void*)dat, allocdat * sizeof(dat[0]));
+    pari_realloc_ip((void**)&dat, allocdat * sizeof(dat[0]));
   }
   dat[ndat].size = size;
   dat[ndat].d    = d; ndat++;
@@ -767,7 +763,7 @@ Test(tune_param *param, long linear)
     s.y = rand_g(&s);
     t1 = time_fun(param->fun, &s, 0);
     t2 = time_fun(param->fun, &s, 1);
-    avma = av;
+    set_avma(av);
     if (t2 >= t1) d = (t2 - t1) / t2;
     else          d = (t2 - t1) / t1;
 
@@ -825,13 +821,12 @@ Test(tune_param *param, long linear)
         if (new_size==s.size) linear = 1;
         s.size = new_size;
       }
-      else if (good-bad < 20*param->step_factor*bad)
+      else if (good-bad <= maxss(1,(long)(20*param->step_factor*bad)))
       {
         linear = 1;
         new_thresh = s.size = bad + 1;
       }
       else s.size = (good+bad)/2;
-      err_printf("bad= %ld good = %ld thresh = %ld linear = %ld\n",bad, good, thresh, linear);
     }
     if (s.size >= param->max_size)
     {
@@ -847,7 +842,7 @@ Test(tune_param *param, long linear)
   *(param->var) = thresh; /* set to optimal value for next tests */
   if (param->var_disable) *(param->var_disable) = save_var_disable;
   if (param->var_enable) *(param->var_enable) = s.var_enable_min;
-  avma = av;
+  set_avma(av);
 }
 
 void error(char **argv) {
@@ -857,8 +852,10 @@ void error(char **argv) {
   diag("  -t:     verbose output\n");
   diag("  -tt:    very verbose output\n");
   diag("  -ttt:   output everything\n");
-  diag("  -s xxx: set step factor between successive sizes to xxx (default 0.01)\n");
+  diag("  -l:     use linear search (slower)\n");
+  diag("  -d xxx: set finite field degree to xxx (default 10)\n");
   diag("  -p xxx: set Flx modulus to xxx (default 27449)\n");
+  diag("  -s xxx: set step factor between successive sizes to xxx (default 0.01)\n");
   diag("  -u xxx: set speed_unittime to xxx (default 1e-4s)\n");
   diag("Tunable variables (omitting variable indices tunes everybody):\n");
   for (i = 0; i < (long)numberof(param); i++)
@@ -870,20 +867,13 @@ int
 main(int argc, char **argv)
 {
   int i, r, n = 0;
-  int linear = 1;
+  int linear = 0;
   GEN v;
-  pari_init(8000000, 2);
-  DFLT_mod = 27449;
+  pari_init(160000000, 2);
   LARGE_mod=subis(powuu(3,128),62);
-#ifdef LONG_IS_64BIT
-  DFLT_qmod = 3;
-  DFLT_hmod = 257;
-  DFLT_mod2 = 281474976710677UL;
-#else
-  DFLT_qmod = 3;
-  DFLT_hmod = 3;
-  DFLT_mod1 = 1031UL;
-#endif
+  DFLT_mod  = unextprime((1UL<<((BITS_IN_LONG-2)>>1))+1);
+  DFLT_mod2 = unextprime((1UL<<(BITS_IN_LONG-1))+1);
+  DFLT_deg  = 10;
   v = new_chunk(argc);
   for (i = 1; i < argc; i++)
   {
@@ -893,8 +883,15 @@ main(int argc, char **argv)
         case 't': option_trace++;
           while (*++s == 't') option_trace++;
           break;
-        case 'd': linear = 1-linear;
+        case 'l': linear = 1-linear;
           break;
+        case 'd':
+          if (!*++s)
+          {
+            if (++i == argc) error(argv);
+            s = argv[i];
+          }
+          DFLT_deg = itou(gp_read_str(s)); break;
         case 'p':
           if (!*++s)
           {
